/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// gl_vidnt.c -- NT GL vid component

#include "gl_state.h"
#include "winquake.h"
#include "resource.h"
#include <commctrl.h>

#define MAX_MODE_LIST	60
#define VID_ROW_SIZE	3
#define WARP_WIDTH		320
#define WARP_HEIGHT		200
#define MAXWIDTH		10000
#define MAXHEIGHT		10000
#define BASEWIDTH		320
#define BASEHEIGHT		200

#define MODE_WINDOWED			0
#define NO_MODE					(MODE_WINDOWED - 1)
#define MODE_FULLSCREEN_DEFAULT	(MODE_WINDOWED + 1)

typedef struct {
	modestate_t	type;
	int			width;
	int			height;
	int			modenum;
	int			dib;
	int			fullscreen;
	int			bpp;
	int			halfscreen;
	char		modedesc[24];
} vmode_t;

typedef struct {
	int			width;
	int			height;
} lmode_t;

lmode_t	lowresmodes[] = {
	{320, 200},
	{320, 240},
	{400, 300},
	{512, 384},
};

qboolean		DDActive;
qboolean		scr_skipupdate;

static vmode_t	modelist[MAX_MODE_LIST];
static int		nummodes;
static vmode_t	*pcurrentmode;
static vmode_t	badmode;

static DEVMODE	gdevmode;
static qboolean	vid_initialized = false;
static qboolean	windowed, leavecurrentmode;
static qboolean vid_canalttab = false;
static qboolean vid_wassuspended = false;
static int		windowed_mouse;
extern qboolean	mouseactive;  // from in_win.c
static HICON	hIcon;

int			DIBWidth, DIBHeight;
RECT		WindowRect;
DWORD		WindowStyle, ExWindowStyle;

HWND	mainwindow, dibwindow;

int			vid_modenum = NO_MODE;
int			vid_realmode;
int			vid_default = MODE_WINDOWED;
static int	windowed_default;
unsigned char	vid_curpal[256*3];
static qboolean fullsbardraw = false;

static float vid_gamma = 1.0;

HGLRC	baseRC;
HDC		maindc;

glvert_t glv;

HWND WINAPI InitializeWindow (HINSTANCE hInstance, int nCmdShow);

viddef_t	vid;				// global video state

unsigned short	d_8to16table[256];
unsigned	d_8to24table[256];

float		gldepthmin, gldepthmax;

modestate_t	modestate = MS_UNINIT;

void VID_MenuDraw (void);
void VID_MenuKey (int key);

LONG WINAPI MainWndProc (HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
void AppActivate(BOOL fActive, BOOL minimize);
char *VID_GetModeDescription (int mode);
void ClearAllStates (void);
void VID_UpdateWindowStatus (void);
void GL_Init (void);


qboolean is8bit = false;

//====================================

cvar_t		vid_mode = {"vid_mode","0", false};
// Note that 0 is MODE_WINDOWED
cvar_t		_vid_default_mode = {"_vid_default_mode","0", true};
// Note that 3 is MODE_FULLSCREEN_DEFAULT
cvar_t		_vid_default_mode_win = {"_vid_default_mode_win","3", true};
cvar_t		vid_wait = {"vid_wait","0"};
cvar_t		vid_nopageflip = {"vid_nopageflip","0", true};
cvar_t		_vid_wait_override = {"_vid_wait_override", "0", true};
cvar_t		vid_config_x = {"vid_config_x","800", true};
cvar_t		vid_config_y = {"vid_config_y","600", true};
cvar_t		vid_stretch_by_2 = {"vid_stretch_by_2","1", true};
cvar_t		_windowed_mouse = {"_windowed_mouse","1", true};

int			window_center_x, window_center_y, window_x, window_y, window_width, window_height;
RECT		window_rect;



void CenterWindow(HWND hWndCenter, int width, int height, BOOL lefttopjustify)
{
    RECT    rect;
    int     CenterX, CenterY;

	CenterX = (GetSystemMetrics(SM_CXSCREEN) - width) / 2;
	CenterY = (GetSystemMetrics(SM_CYSCREEN) - height) / 2;
	if (CenterX > CenterY*2)
		CenterX >>= 1;	// dual screens
	CenterX = (CenterX < 0) ? 0: CenterX;
	CenterY = (CenterY < 0) ? 0: CenterY;
	SetWindowPos (hWndCenter, NULL, CenterX, CenterY, 0, 0,
			SWP_NOSIZE | SWP_NOZORDER | SWP_SHOWWINDOW | SWP_DRAWFRAME);
}

qboolean VID_SetWindowedMode (int modenum)
{
	HDC				hdc;
	int				lastmodestate, width, height;
	RECT			rect;

	lastmodestate = modestate;

	WindowRect.top = WindowRect.left = 0;

	WindowRect.right = modelist[modenum].width;
	WindowRect.bottom = modelist[modenum].height;

	DIBWidth = modelist[modenum].width;
	DIBHeight = modelist[modenum].height;

	WindowStyle = WS_OVERLAPPED | WS_BORDER | WS_CAPTION | WS_SYSMENU |
				  WS_MINIMIZEBOX;
	ExWindowStyle = 0;

	rect = WindowRect;
	AdjustWindowRectEx(&rect, WindowStyle, FALSE, 0);

	width = rect.right - rect.left;
	height = rect.bottom - rect.top;

	// Create the DIB window
	dibwindow = CreateWindowEx (
		 ExWindowStyle,
		 "MHQuake",
		 "MHQuake",
		 WindowStyle,
		 rect.left, rect.top,
		 width,
		 height,
		 NULL,
		 NULL,
		 global_hInstance,
		 NULL);

	if (!dibwindow)
		Sys_Error ("Couldn't create DIB window");

	// Center and show the DIB window
	CenterWindow(dibwindow, WindowRect.right - WindowRect.left,
				 WindowRect.bottom - WindowRect.top, false);

	ShowWindow (dibwindow, SW_SHOWDEFAULT);
	UpdateWindow (dibwindow);

	modestate = MS_WINDOWED;

// because we have set the background brush for the window to NULL
// (to avoid flickering when re-sizing the window on the desktop),
// we clear the window to black when created, otherwise it will be
// empty while Quake starts up.
	hdc = GetDC(dibwindow);
	PatBlt(hdc,0,0,WindowRect.right,WindowRect.bottom,BLACKNESS);
	ReleaseDC(dibwindow, hdc);

	if (vid.conheight > modelist[modenum].height)
		vid.conheight = modelist[modenum].height;
	if (vid.conwidth > modelist[modenum].width)
		vid.conwidth = modelist[modenum].width;
	vid.width = vid.conwidth;
	vid.height = vid.conheight;

	vid.numpages = 2;

	mainwindow = dibwindow;

	SendMessage (mainwindow, WM_SETICON, (WPARAM)TRUE, (LPARAM)hIcon);
	SendMessage (mainwindow, WM_SETICON, (WPARAM)FALSE, (LPARAM)hIcon);

	return true;
}


qboolean VID_SetFullDIBMode (int modenum)
{
	HDC				hdc;
	int				lastmodestate, width, height;
	RECT			rect;

	if (!leavecurrentmode)
	{
		gdevmode.dmFields = DM_BITSPERPEL | DM_PELSWIDTH | DM_PELSHEIGHT;
		gdevmode.dmBitsPerPel = modelist[modenum].bpp;
		gdevmode.dmPelsWidth = modelist[modenum].width <<
							   modelist[modenum].halfscreen;
		gdevmode.dmPelsHeight = modelist[modenum].height;
		gdevmode.dmSize = sizeof (gdevmode);

		if (ChangeDisplaySettings (&gdevmode, CDS_FULLSCREEN) != DISP_CHANGE_SUCCESSFUL)
			Sys_Error ("Couldn't set fullscreen DIB mode");
	}

	lastmodestate = modestate;
	modestate = MS_FULLDIB;

	WindowRect.top = WindowRect.left = 0;

	WindowRect.right = modelist[modenum].width;
	WindowRect.bottom = modelist[modenum].height;

	DIBWidth = modelist[modenum].width;
	DIBHeight = modelist[modenum].height;

	WindowStyle = WS_POPUP;
	ExWindowStyle = 0;

	rect = WindowRect;
	AdjustWindowRectEx(&rect, WindowStyle, FALSE, 0);

	width = rect.right - rect.left;
	height = rect.bottom - rect.top;

	// Create the DIB window
	dibwindow = CreateWindowEx (
		 ExWindowStyle,
		 "MHQuake",
		 "MHQuake",
		 WindowStyle,
		 rect.left, rect.top,
		 width,
		 height,
		 NULL,
		 NULL,
		 global_hInstance,
		 NULL);

	if (!dibwindow)
		Sys_Error ("Couldn't create DIB window");

	ShowWindow (dibwindow, SW_SHOWDEFAULT);
	UpdateWindow (dibwindow);

	// Because we have set the background brush for the window to NULL
	// (to avoid flickering when re-sizing the window on the desktop), we
	// clear the window to black when created, otherwise it will be
	// empty while Quake starts up.
	hdc = GetDC(dibwindow);
	PatBlt(hdc,0,0,WindowRect.right,WindowRect.bottom,BLACKNESS);
	ReleaseDC(dibwindow, hdc);

	if (vid.conheight > modelist[modenum].height)
		vid.conheight = modelist[modenum].height;
	if (vid.conwidth > modelist[modenum].width)
		vid.conwidth = modelist[modenum].width;
	vid.width = vid.conwidth;
	vid.height = vid.conheight;

	vid.numpages = 2;

// needed because we're not getting WM_MOVE messages fullscreen on NT
	window_x = 0;
	window_y = 0;

	mainwindow = dibwindow;

	SendMessage (mainwindow, WM_SETICON, (WPARAM)TRUE, (LPARAM)hIcon);
	SendMessage (mainwindow, WM_SETICON, (WPARAM)FALSE, (LPARAM)hIcon);

	return true;
}


int VID_SetMode (int modenum, unsigned char *palette)
{
	int				original_mode, temp;
	qboolean		stat;
    MSG				msg;
	HDC				hdc;

	if ((windowed && (modenum != 0)) ||
		(!windowed && (modenum < 1)) ||
		(!windowed && (modenum >= nummodes)))
	{
		Sys_Error ("Bad video mode\n");
	}

// so Con_Printfs don't mess us up by forcing vid and snd updates
	temp = scr_disabled_for_loading;
	scr_disabled_for_loading = true;

	CDAudio_Pause ();

	if (vid_modenum == NO_MODE)
		original_mode = windowed_default;
	else
		original_mode = vid_modenum;

	// Set either the fullscreen or windowed mode
	if (modelist[modenum].type == MS_WINDOWED)
	{
		if (_windowed_mouse.value && key_dest == key_game)
		{
			stat = VID_SetWindowedMode(modenum);
			IN_ActivateMouse ();
			IN_HideMouse ();
		}
		else
		{
			IN_DeactivateMouse ();
			IN_ShowMouse ();
			stat = VID_SetWindowedMode(modenum);
		}
	}
	else if (modelist[modenum].type == MS_FULLDIB)
	{
		stat = VID_SetFullDIBMode(modenum);
		IN_ActivateMouse ();
		IN_HideMouse ();
	}
	else
	{
		Sys_Error ("VID_SetMode: Bad mode type in modelist");
	}

	window_width = DIBWidth;
	window_height = DIBHeight;
	VID_UpdateWindowStatus ();

	CDAudio_Resume ();
	scr_disabled_for_loading = temp;

	if (!stat)
	{
		Sys_Error ("Couldn't set video mode");
	}

// now we try to make sure we get the focus on the mode switch, because
// sometimes in some systems we don't.  We grab the foreground, then
// finish setting up, pump all our messages, and sleep for a little while
// to let messages finish bouncing around the system, then we put
// ourselves at the top of the z order, then grab the foreground again,
// Who knows if it helps, but it probably doesn't hurt
	SetForegroundWindow (mainwindow);
	VID_SetPalette (palette);
	vid_modenum = modenum;
	Cvar_SetValue ("vid_mode", (float)vid_modenum);

	while (PeekMessage (&msg, NULL, 0, 0, PM_REMOVE))
	{
      	TranslateMessage (&msg);
      	DispatchMessage (&msg);
	}

	Sleep (100);

	SetWindowPos (mainwindow, HWND_TOP, 0, 0, 0, 0,
				  SWP_DRAWFRAME | SWP_NOMOVE | SWP_NOSIZE | SWP_SHOWWINDOW |
				  SWP_NOCOPYBITS);

	SetForegroundWindow (mainwindow);

// fix the leftover Alt from any Alt-Tab or the like that switched us away
	ClearAllStates ();

	if (!msg_suppress_1)
		Con_SafePrintf ("Video mode %s initialized.\n", VID_GetModeDescription (vid_modenum));

	VID_SetPalette (palette);

	// force initial r_refdef to set
	vid.recalc_refdef = 1;

	return true;
}


/*
================
VID_UpdateWindowStatus
================
*/
void VID_UpdateWindowStatus (void)
{

	window_rect.left = window_x;
	window_rect.top = window_y;
	window_rect.right = window_x + window_width;
	window_rect.bottom = window_y + window_height;
	window_center_x = (window_rect.left + window_rect.right) / 2;
	window_center_y = (window_rect.top + window_rect.bottom) / 2;

	IN_UpdateClipCursor ();
}


//====================================


//int		texture_mode = GL_NEAREST;
//int		texture_mode = GL_NEAREST_MIPMAP_NEAREST;
//int		texture_mode = GL_NEAREST_MIPMAP_LINEAR;
int		texture_mode = GL_LINEAR;
//int		texture_mode = GL_LINEAR_MIPMAP_NEAREST;
//int		texture_mode = GL_LINEAR_MIPMAP_LINEAR;


/*
==================
GL EXTENSION MALARKEY

As a general rule, I don't like the idea of proprietary extensions.  A special case of this rule is that
anything which allows me to rub ATI's nose in it must necessarily be a good thing :D

This will check vendor codes for extensions too...
==================
*/

qboolean GL_CheckExtension (char *ExtensionName, qboolean crash, qboolean silent)
{
	int i;
	char *VendorCode[] = {"ARB", "EXT", "NV", "ATI", "SGI", "SGIS", "SGIX", "INTEL", "3DFX"};

	// let us specify extensions directly
	if (ExtensionName[0] == 'G' && ExtensionName[1] == 'L' && ExtensionName[2] == '_')
	{
		// allow direct access to the extension name
		if (strstr (gl_state.gl_extensions, ExtensionName))
		{
			if (!silent) Con_Printf ("Found Extension %s.\n", ExtensionName);
			return true;
		}

		if (crash) Sys_Error ("Extension %s NOT Found!!!\n\nAborting MHQuake...", ExtensionName);

		if (!silent) Con_Printf ("Extension %s NOT Found.\n", ExtensionName);
		return false;
	}

	// use the fallback mechanism
	for (i = 0; i < 9; i++)
	{
		if (strstr (gl_state.gl_extensions, va ("GL_%s_%s", VendorCode[i], ExtensionName)))
		{
			if (!silent) Con_Printf ("Found Extension GL_%s_%s.\n", VendorCode[i], ExtensionName);
			return true;
		}
	}

	if (crash) Sys_Error ("Extension GL_*_%s NOT Found!!!\n\nAborting MHQuake...", ExtensionName);

	if (!silent) Con_Printf ("Extension GL_*_%s NOT Found.\n", ExtensionName);
	return false;
}


// this isn't standard opengl, but it's nice to have...
void glScalefv (float *v)
{
	glScalef (v[0], v[1], v[2]);
}


/*
================
MHGLGetProcAddress

Implements a work-around for earlier OpenGL versions, by checking for extensions that
support the required functions.  I'm VERY DISPLEASED at having to do this (you don't
cut down the Mona Lisa to make it fit in a smaller frame, do you?), but ATI's most
recent OpenGL implementation is only at version 1.3 and even at that it doesn't support
some 1.2 stuff.  Thank God for Doom 3 - it may make them get their act together.
================
*/
static PROC MHGLGetProcAddress (char *funcname, qboolean crash)
{
	int i;
	char *VendorCode[] = {"ARB", "EXT", "NV", "ATI", "SGI", "SGIS", "SGIX", "INTEL", "3DFX"};
	static PROC thePROC;

	// see is the function supported by the core implementation
	if (thePROC = wglGetProcAddress (funcname)) return thePROC;

	for (i = 0; i < 9; i++)
	{
		thePROC = wglGetProcAddress (va ("%s%s", funcname, VendorCode[i]));

		if (thePROC)
		{
			// return silently if not crashing
			if (crash) Con_Printf ("%s Unavailable, reverting to %s%s\n", funcname, funcname, VendorCode[i]);
			return thePROC;
		}
	}

	// oh dear.
	if (crash) Sys_Error ("Core OpenGL 1.4 function %s Unavailable!!!", funcname);

	// shut up compiler warnings
	return NULL;
}


gl_caps_s GLCaps;

void GetGLCore (void)
{
	// MHQuake doesn't have a 1 TMU render path any more, so...
	glGetIntegerv (GL_MAX_TEXTURE_UNITS, &GLCaps.GL_TextureUnits);
	if (GLCaps.GL_TextureUnits < 2) Sys_Error ("Not enough TMUs");

	// query extension availability... do this as a default because some drivers may report versions > 1.4 but not
	// actually implement the required feature set (hello ATI...) - we crash if one is not found...
	// this list is pretty exhaustive of all the required stuff that's been added from 1.1 onwards.  it may include some
	// that aren't actually used by MHQuake at present.  all of these were promoted to the core a long long time ago, and
	// as such should no longer be considered "extensions" unless you have a really creaky old card...
	GL_CheckExtension ("generate_mipmap", true, true);
	GL_CheckExtension ("vertex_array", true, true);
	GL_CheckExtension ("texture", true, true);
	GL_CheckExtension ("texture_object", true, true);

	// SGIS doesn't use glActiveTexture, so if we don't have that we'll crash later on
	GL_CheckExtension ("multitexture", true, true);

	GL_CheckExtension ("texture_compression", true, true);
	GL_CheckExtension ("texture_env_add", true, true);
	GL_CheckExtension ("texture_env_combine", true, true);
	GL_CheckExtension ("texture_env_dot3", true, true);
	GL_CheckExtension ("texture_cube_map", true, true);

	// this is an NV extension, and may not be present on ATI... MHQuake does use it though, so don't crash if it's not
	// found, and be prepared for undefined program behaviour... (it's been core since 1.4 so this should only be an
	// issue if ATI's drivers are crap...)
	GL_CheckExtension ("blend_square", false, true);

	GL_CheckExtension ("fog_coord", true, true);
	GL_CheckExtension ("point_parameters", true, true);
	GL_CheckExtension ("bgra", true, true);
	GL_CheckExtension ("texture_edge_clamp", true, true);
	GL_CheckExtension ("texture_lod", true, true);

	// now get the function pointers...
	// OpenGL 1.2 - check the ones we need only, and a few others to ensure it's OK
	glBlendColor = (void *) MHGLGetProcAddress ("glBlendColor", false);
	glBlendEquation = (void *) MHGLGetProcAddress ("glBlendEquation", false);
	glDrawRangeElements = (void *) MHGLGetProcAddress ("glDrawRangeElements", true);
	glColorTable = (void *) MHGLGetProcAddress ("glColorTable", false);
	glColorTableParameterfv = (void *) MHGLGetProcAddress ("glColorTableParameterfv", false);
	glColorTableParameteriv = (void *) MHGLGetProcAddress ("glColorTableParameteriv", false);
	glCopyColorTable = (void *) MHGLGetProcAddress ("glCopyColorTable", false);
	glGetColorTable = (void *) MHGLGetProcAddress ("glGetColorTable", false);
	glGetColorTableParameterfv = (void *) MHGLGetProcAddress ("glGetColorTableParameterfv", false);
	glGetColorTableParameteriv = (void *) MHGLGetProcAddress ("glGetColorTableParameteriv", false);
	glColorSubTable = (void *) MHGLGetProcAddress ("glColorSubTable", false);
	glCopyColorSubTable = (void *) MHGLGetProcAddress ("glCopyColorSubTable", false);
	glConvolutionFilter1D = (void *) MHGLGetProcAddress ("glConvolutionFilter1D", false);
	glConvolutionFilter2D = (void *) MHGLGetProcAddress ("glConvolutionFilter2D", false);
	glConvolutionParameterf = (void *) MHGLGetProcAddress ("glConvolutionParameterf", false);
	glConvolutionParameterfv = (void *) MHGLGetProcAddress ("glConvolutionParameterfv", false);
	glConvolutionParameteri = (void *) MHGLGetProcAddress ("glConvolutionParameteri", false);
	glConvolutionParameteriv = (void *) MHGLGetProcAddress ("glConvolutionParameteriv", false);
	glCopyConvolutionFilter1D = (void *) MHGLGetProcAddress ("glCopyConvolutionFilter1D", false);
	glCopyConvolutionFilter2D = (void *) MHGLGetProcAddress ("glCopyConvolutionFilter2D", false);
	glGetConvolutionFilter = (void *) MHGLGetProcAddress ("glGetConvolutionFilter", false);
	glGetConvolutionParameterfv = (void *) MHGLGetProcAddress ("glGetConvolutionParameterfv", false);
	glGetConvolutionParameteriv = (void *) MHGLGetProcAddress ("glGetConvolutionParameteriv", false);
	glGetSeparableFilter = (void *) MHGLGetProcAddress ("glGetSeparableFilter", false);
	glSeparableFilter2D = (void *) MHGLGetProcAddress ("glSeparableFilter2D", false);
	glGetHistogram = (void *) MHGLGetProcAddress ("glGetHistogram", false);
	glGetHistogramParameterfv = (void *) MHGLGetProcAddress ("glGetHistogramParameterfv", false);
	glGetHistogramParameteriv = (void *) MHGLGetProcAddress ("glGetHistogramParameteriv", false);
	glGetMinmax = (void *) MHGLGetProcAddress ("glGetMinmax", false);
	glGetMinmaxParameterfv = (void *) MHGLGetProcAddress ("glGetMinmaxParameterfv", false);
	glGetMinmaxParameteriv = (void *) MHGLGetProcAddress ("glGetMinmaxParameteriv", false);
	glHistogram = (void *) MHGLGetProcAddress ("glHistogram", false);
	glMinmax = (void *) MHGLGetProcAddress ("glMinmax", false);
	glResetHistogram = (void *) MHGLGetProcAddress ("glResetHistogram", false);
	glResetMinmax = (void *) MHGLGetProcAddress ("glResetMinmax", false);
	glTexImage3D = (void *) MHGLGetProcAddress ("glTexImage3D", true);
	glTexSubImage3D = (void *) MHGLGetProcAddress ("glTexSubImage3D", false);
	glCopyTexSubImage3D = (void *) MHGLGetProcAddress ("glCopyTexSubImage3D", false);

	// OpenGL 1.3 - check the ones we need only, and a few others to ensure it's OK
	glActiveTexture = (void *) MHGLGetProcAddress ("glActiveTexture", true);
	glClientActiveTexture = (void *) MHGLGetProcAddress ("glClientActiveTexture", true);
	glMultiTexCoord1d = (void *) MHGLGetProcAddress ("glMultiTexCoord1d", false);
	glMultiTexCoord1dv = (void *) MHGLGetProcAddress ("glMultiTexCoord1dv", false);
	glMultiTexCoord1f = (void *) MHGLGetProcAddress ("glMultiTexCoord1f", true);
	glMultiTexCoord1fv = (void *) MHGLGetProcAddress ("glMultiTexCoord1fv", false);
	glMultiTexCoord1i = (void *) MHGLGetProcAddress ("glMultiTexCoord1i", false);
	glMultiTexCoord1iv = (void *) MHGLGetProcAddress ("glMultiTexCoord1iv", false);
	glMultiTexCoord1s = (void *) MHGLGetProcAddress ("glMultiTexCoord1s", false);
	glMultiTexCoord1sv = (void *) MHGLGetProcAddress ("glMultiTexCoord1sv", false);
	glMultiTexCoord2d = (void *) MHGLGetProcAddress ("glMultiTexCoord2d", false);
	glMultiTexCoord2dv = (void *) MHGLGetProcAddress ("glMultiTexCoord2dv", false);
	glMultiTexCoord2f = (void *) MHGLGetProcAddress ("glMultiTexCoord2f", true);
	glMultiTexCoord2fv = (void *) MHGLGetProcAddress ("glMultiTexCoord2fv", false);
	glMultiTexCoord2i = (void *) MHGLGetProcAddress ("glMultiTexCoord2i", false);
	glMultiTexCoord2iv = (void *) MHGLGetProcAddress ("glMultiTexCoord2iv", false);
	glMultiTexCoord2s = (void *) MHGLGetProcAddress ("glMultiTexCoord2s", false);
	glMultiTexCoord2sv = (void *) MHGLGetProcAddress ("glMultiTexCoord2sv", false);
	glMultiTexCoord3d = (void *) MHGLGetProcAddress ("glMultiTexCoord3d", false);
	glMultiTexCoord3dv = (void *) MHGLGetProcAddress ("glMultiTexCoord3dv", false);
	glMultiTexCoord3f = (void *) MHGLGetProcAddress ("glMultiTexCoord3f", true);
	glMultiTexCoord3fv = (void *) MHGLGetProcAddress ("glMultiTexCoord3fv", false);
	glMultiTexCoord3i = (void *) MHGLGetProcAddress ("glMultiTexCoord3i", false);
	glMultiTexCoord3iv = (void *) MHGLGetProcAddress ("glMultiTexCoord3iv", false);
	glMultiTexCoord3s = (void *) MHGLGetProcAddress ("glMultiTexCoord3s", false);
	glMultiTexCoord3sv = (void *) MHGLGetProcAddress ("glMultiTexCoord3sv", false);
	glMultiTexCoord4d = (void *) MHGLGetProcAddress ("glMultiTexCoord4d", false);
	glMultiTexCoord4dv = (void *) MHGLGetProcAddress ("glMultiTexCoord4dv", false);
	glMultiTexCoord4f = (void *) MHGLGetProcAddress ("glMultiTexCoord4f", false);
	glMultiTexCoord4fv = (void *) MHGLGetProcAddress ("glMultiTexCoord4fv", false);
	glMultiTexCoord4i = (void *) MHGLGetProcAddress ("glMultiTexCoord4i", false);
	glMultiTexCoord4iv = (void *) MHGLGetProcAddress ("glMultiTexCoord4iv", false);
	glMultiTexCoord4s = (void *) MHGLGetProcAddress ("glMultiTexCoord4s", false);
	glMultiTexCoord4sv = (void *) MHGLGetProcAddress ("glMultiTexCoord4sv", false);
	glLoadTransposeMatrixf = (void *) MHGLGetProcAddress ("glLoadTransposeMatrixf", false);
	glLoadTransposeMatrixd = (void *) MHGLGetProcAddress ("glLoadTransposeMatrixd", false);
	glMultTransposeMatrixf = (void *) MHGLGetProcAddress ("glMultTransposeMatrixf", false);
	glMultTransposeMatrixd = (void *) MHGLGetProcAddress ("glMultTransposeMatrixd", false);
	glSampleCoverage = (void *) MHGLGetProcAddress ("glSampleCoverage", false);
	glCompressedTexImage3D = (void *) MHGLGetProcAddress ("glCompressedTexImage3D", false);
	glCompressedTexImage2D = (void *) MHGLGetProcAddress ("glCompressedTexImage2D", true);
	glCompressedTexImage1D = (void *) MHGLGetProcAddress ("glCompressedTexImage1D", false);
	glCompressedTexSubImage3D = (void *) MHGLGetProcAddress ("glCompressedTexSubImage3D", false);
	glCompressedTexSubImage2D = (void *) MHGLGetProcAddress ("glCompressedTexSubImage2D", false);
	glCompressedTexSubImage1D = (void *) MHGLGetProcAddress ("glCompressedTexSubImage1D", false);
	glGetCompressedTexImage = (void *) MHGLGetProcAddress ("glGetCompressedTexImage", false);

	// OpenGL 1.4 - check the ones we need only, and a few others to ensure it's OK
	glBlendFuncSeparate = (void *) MHGLGetProcAddress ("glBlendFuncSeparate", false);
	glFogCoordf = (void *) MHGLGetProcAddress ("glFogCoordf", true);
	glFogCoordfv = (void *) MHGLGetProcAddress ("glFogCoordfv", false);
	glFogCoordd = (void *) MHGLGetProcAddress ("glFogCoordd", false);
	glFogCoorddv = (void *) MHGLGetProcAddress ("glFogCoorddv", false);
	glFogCoordPointer = (void *) MHGLGetProcAddress ("glFogCoordPointer", true);
	glMultiDrawArrays = (void *) MHGLGetProcAddress ("glMultiDrawArrays", true);
	glMultiDrawElements = (void *) MHGLGetProcAddress ("glMultiDrawElements", false);
	glPointParameterf = (void *) MHGLGetProcAddress ("glPointParameterf", true);
	glPointParameterfv = (void *) MHGLGetProcAddress ("glPointParameterfv", false);
	glPointParameteri = (void *) MHGLGetProcAddress ("glPointParameteri", false);	// this is the entry point ATI missed...
	glPointParameteriv = (void *) MHGLGetProcAddress ("glPointParameteriv", false);
	glSecondaryColor3b = (void *) MHGLGetProcAddress ("glSecondaryColor3b", false);
	glSecondaryColor3bv = (void *) MHGLGetProcAddress ("glSecondaryColor3bv", false);
	glSecondaryColor3d = (void *) MHGLGetProcAddress ("glSecondaryColor3d", false);
	glSecondaryColor3dv = (void *) MHGLGetProcAddress ("glSecondaryColor3dv", false);
	glSecondaryColor3f = (void *) MHGLGetProcAddress ("glSecondaryColor3f", true);
	glSecondaryColor3fv = (void *) MHGLGetProcAddress ("glSecondaryColor3fv", false);
	glSecondaryColor3i = (void *) MHGLGetProcAddress ("glSecondaryColor3i", false);
	glSecondaryColor3iv = (void *) MHGLGetProcAddress ("glSecondaryColor3iv", false);
	glSecondaryColor3s = (void *) MHGLGetProcAddress ("glSecondaryColor3s", false);
	glSecondaryColor3sv = (void *) MHGLGetProcAddress ("glSecondaryColor3sv", false);
	glSecondaryColor3ub = (void *) MHGLGetProcAddress ("glSecondaryColor3ub", false);
	glSecondaryColor3ubv = (void *) MHGLGetProcAddress ("glSecondaryColor3ubv", false);
	glSecondaryColor3ui = (void *) MHGLGetProcAddress ("glSecondaryColor3ui", false);
	glSecondaryColor3uiv = (void *) MHGLGetProcAddress ("glSecondaryColor3uiv", false);
	glSecondaryColor3us = (void *) MHGLGetProcAddress ("glSecondaryColor3us", false);
	glSecondaryColor3usv = (void *) MHGLGetProcAddress ("glSecondaryColor3usv", false);
	glSecondaryColorPointer = (void *) MHGLGetProcAddress ("glSecondaryColorPointer", true);
	glWindowPos2d = (void *) MHGLGetProcAddress ("glWindowPos2d", false);
	glWindowPos2dv = (void *) MHGLGetProcAddress ("glWindowPos2dv", false);
	glWindowPos2f = (void *) MHGLGetProcAddress ("glWindowPos2f", false);
	glWindowPos2fv = (void *) MHGLGetProcAddress ("glWindowPos2fv", false);
	glWindowPos2i = (void *) MHGLGetProcAddress ("glWindowPos2i", false);
	glWindowPos2iv = (void *) MHGLGetProcAddress ("glWindowPos2iv", false);
	glWindowPos2s = (void *) MHGLGetProcAddress ("glWindowPos2s", false);
	glWindowPos2sv = (void *) MHGLGetProcAddress ("glWindowPos2sv", false);
	glWindowPos3d = (void *) MHGLGetProcAddress ("glWindowPos3d", false);
	glWindowPos3dv = (void *) MHGLGetProcAddress ("glWindowPos3dv", false);
	glWindowPos3f = (void *) MHGLGetProcAddress ("glWindowPos3f", false);
	glWindowPos3fv = (void *) MHGLGetProcAddress ("glWindowPos3fv", false);
	glWindowPos3i = (void *) MHGLGetProcAddress ("glWindowPos3i", false);
	glWindowPos3iv = (void *) MHGLGetProcAddress ("glWindowPos3iv", false);
	glWindowPos3s = (void *) MHGLGetProcAddress ("glWindowPos3s", false);
	glWindowPos3sv = (void *) MHGLGetProcAddress ("glWindowPos3sv", false);

	// OpenGL 1.5 - get them all but don't crash on a fail cos they're optional
	glGenQueries = (void *) MHGLGetProcAddress ("glGenQueries", false);
	glDeleteQueries = (void *) MHGLGetProcAddress ("glDeleteQueries", false);
	glIsQuery = (void *) MHGLGetProcAddress ("glIsQuery", false);
	glEndQuery = (void *) MHGLGetProcAddress ("glEndQuery", false);
	glGetQueryiv = (void *) MHGLGetProcAddress ("glGetQueryiv", false);
	glGetQueryObjectiv = (void *) MHGLGetProcAddress ("glGetQueryObjectiv", false);
	glGetQueryObjectuiv = (void *) MHGLGetProcAddress ("glGetQueryObjectuiv", false);
	glBindBuffer = (void *) MHGLGetProcAddress ("glBindBuffer", false);
	glDeleteBuffers = (void *) MHGLGetProcAddress ("glDeleteBuffers", false);
	glGenBuffers = (void *) MHGLGetProcAddress ("glGenBuffers", false);
	glIsBuffer = (void *) MHGLGetProcAddress ("glIsBuffer", false);
	glBufferData = (void *) MHGLGetProcAddress ("glBufferData", false);
	glBufferSubData = (void *) MHGLGetProcAddress ("glBufferSubData", false);
	glGetBufferSubData = (void *) MHGLGetProcAddress ("glGetBufferSubData", false);
	glMapBuffer = (void *) MHGLGetProcAddress ("glMapBuffer", false);
	glUnmapBuffer = (void *) MHGLGetProcAddress ("glUnmapBuffer", false);
	glGetBufferParameteriv = (void *) MHGLGetProcAddress ("glGetBufferParameteriv", false);
	glGetBufferPointerv = (void *) MHGLGetProcAddress ("glGetBufferPointerv", false);

	glGetIntegerv (GL_MAX_TEXTURE_SIZE, &GLCaps.GL_MaxTextureSize);
	Con_Printf ("\nMaximum Texture Size: %i x %i\n", GLCaps.GL_MaxTextureSize, GLCaps.GL_MaxTextureSize);

	// OH, LET'S RUB IT IN!!!
	if (!glPointParameteri) Con_Printf ("\nATI have still missed out on glPointParameteri\n");

	Con_Printf ("\nRequired OpenGL Support OK\n\n...Checking Optional Features\n");

	// check OpenGL 1.5 optional features
	if (gl_state.gl_version >= 1.5)
	{
		GL_ExtensionBits |= (HAS_VBO | HAS_OCCLUSION);
		Con_Printf ("Found core feature: vertex_buffer_object.\n");
		Con_Printf ("Found core feature: occlusion_query.\n");
	}
	else
	{
		if (GL_CheckExtension ("vertex_buffer_object", false, false)) GL_ExtensionBits |= HAS_VBO;
		if (GL_CheckExtension ("occlusion_query", false, false)) GL_ExtensionBits |= HAS_OCCLUSION;
	}

	if (GL_ExtensionBits & HAS_OCCLUSION)
	{
		glGetQueryiv (GL_SAMPLES_PASSED, GL_QUERY_COUNTER_BITS, &GLCaps.GL_OcclusionBits);
		Con_Printf ("    %i Occlusion Query Bits\n", GLCaps.GL_OcclusionBits);
	}
	else GLCaps.GL_OcclusionBits = 0;

	// check OpenGL 2.0 optional features
	if (gl_state.gl_version >= 2)
	{
		GL_ExtensionBits |= (HAS_TEXTURENPOW2 | HAS_POINTSPRITE);
		Con_Printf ("Found core feature: texture_non_power_of_two.\n");
		Con_Printf ("Found core feature: point_sprite.\n");
	}
	else
	{
		if (GL_CheckExtension ("texture_non_power_of_two", false, false)) GL_ExtensionBits |= HAS_TEXTURENPOW2;
		if (GL_CheckExtension ("point_sprite", false, false)) GL_ExtensionBits |= HAS_POINTSPRITE;
	}

	// check other non-core and/or proprietary extensions
	if (GL_CheckExtension ("texture_filter_anisotropic", false, false))
	{
		GL_ExtensionBits |= HAS_ANISOTROPIC;

		glGetFloatv (GL_MAX_TEXTURE_MAX_ANISOTROPY, &GLCaps.GL_MaxAnisotropy);
		Con_Printf ("Maximum Anisotropic Filter: %i\n", (int) GLCaps.GL_MaxAnisotropy);
	}
	else GLCaps.GL_MaxAnisotropy = 1;

	if (GL_CheckExtension ("GL_EXT_compiled_vertex_array", false, true))
	{
		glLockArrays = (void *) MHGLGetProcAddress ("glLockArrays", false);
		glUnlockArrays = (void *) MHGLGetProcAddress ("glUnlockArrays", false);

		if (glLockArrays && glUnlockArrays)
		{
			GL_ExtensionBits |= HAS_CVA;
			Con_Printf ("Found Extension GL_EXT_compiled_vertex_array.\n");
		}
	}

	if (GL_CheckExtension ("GL_NV_texture_shader", false, false)) GL_ExtensionBits |= HAS_NVTEXSHADER;

	/*
	if (GL_CheckExtension ("imaging", false, false)) GL_ExtensionBits |= HAS_COLORMATRIX;
	else if (GL_CheckExtension ("color_matrix", false, false)) GL_ExtensionBits |= HAS_COLORMATRIX;

	if (GL_ExtensionBits & HAS_COLORMATRIX) Con_Printf ("Color Matrix Manipulations Available.\n");
	*/

	// get the number of TMUs available
	Con_Printf("\nFound %i TMUs support.\n", GLCaps.GL_TextureUnits);

	Con_Printf ("\nMinimal Required OpenGL Startup OK\n");
}


void GL_InitVBOs (void);


/*
===============
GL_Init
===============
*/
void GL_Init (void)
{
	int GL_ColorBitsRed;
	int GL_ColorBitsGreen;
	int GL_ColorBitsBlue;
	int GL_DepthBits;
	int GL_StencilBits;
	int i;

	// set baseline gl state
	GL_InitState ();

	gl_state.gl_vendor = glGetString (GL_VENDOR);
	Con_Printf ("\nGL_VENDOR:   %s\n", gl_state.gl_vendor);

	gl_state.gl_renderer = glGetString (GL_RENDERER);
	Con_Printf ("GL_RENDERER: %s\n", gl_state.gl_renderer);

	gl_state.gl_version = atof (glGetString (GL_VERSION));
	Con_Printf ("GL_VERSION:  %0.2f\n", gl_state.gl_version);

	gl_state.gl_extensions = glGetString (GL_EXTENSIONS);
	Con_DPrintf ("\nGL_EXTENSIONS: %s\n", gl_state.gl_extensions);

	if (gl_state.gl_version < 1.4)
		Con_Printf ("\nWARNING\nGL_VERSION %f detected - you may experience reduced functionality\n", gl_state.gl_version);

	for (i = 0; i < strlen (gl_state.gl_renderer) - 7; i++)
	{
		// test different ways of reporting this
		if (gl_state.gl_renderer[i] == 'f' || gl_state.gl_renderer[i] == 'F' &&
			gl_state.gl_renderer[i + 1] == 'x' || gl_state.gl_renderer[i + 1] == 'X')
		{
			int j;

			if (gl_state.gl_renderer[i + 2] == ' ')
				j = i + 3;
			else j = i + 2;

			if (j + 3 >= strlen (gl_state.gl_renderer)) break;

			if (gl_state.gl_renderer[j] == '5')
			{
				gl_state.NP2BigDirtyFib = true;
				Con_Printf ("\nWARNING\nNV 3x Chipset detected - you may experience reduced functionality\n");
				break;
			}
		}
	}

	// get some caps
	glGetIntegerv (GL_RED_BITS, &GL_ColorBitsRed);
	glGetIntegerv (GL_GREEN_BITS, &GL_ColorBitsGreen);
	glGetIntegerv (GL_BLUE_BITS, &GL_ColorBitsBlue);
	glGetIntegerv (GL_DEPTH_BITS, &GL_DepthBits);
	glGetIntegerv (GL_STENCIL_BITS, &GL_StencilBits);

	Con_Printf ("\nBuffers\n");
	Con_Printf ("   Color:   %2i Bits\n", GL_ColorBitsRed + GL_ColorBitsGreen + GL_ColorBitsBlue);
	Con_Printf ("   Depth:   %2i Bits\n", GL_DepthBits);
	Con_Printf ("   Stencil: %2i Bits\n", GL_StencilBits);

	GetGLCore ();

	// init the VBOs - this function will handle behaviour if VBOs are unavailable
	GL_InitVBOs ();

	// default states
	glClearColor (0.0, 0.0, 0.0, 1.0);
	// make it white so it's easy to see where stuff is missing...
	// (set this back to black for release...)
	//glClearColor (1.0, 1.0, 1.0, 1.0);

	glCullFace(GL_FRONT);
	glEnable(GL_TEXTURE_2D);

	//glEnable(GL_ALPHA_TEST);
	glAlphaFunc(GL_GREATER, 0.666);

	glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);
	glShadeModel (GL_SMOOTH);

	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

	// set up the hints for drawing
	glHint (GL_POINT_SMOOTH_HINT, GL_NICEST);
	glHint (GL_LINE_SMOOTH_HINT, GL_NICEST);
	glHint (GL_POLYGON_SMOOTH_HINT, GL_NICEST);
	glHint (GL_FOG_HINT, GL_NICEST);
	glHint (GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	glHint (GL_GENERATE_MIPMAP_HINT, GL_NICEST);

	// explicitly enable dithering
	glEnable (GL_DITHER);

	// use fragment depth fog for higher quality
	glFogi (GL_FOG_COORDINATE_SOURCE, GL_FRAGMENT_DEPTH);
}


/*
=================
GL_BeginRendering

=================
*/
void GL_BeginRendering (int *x, int *y, int *width, int *height)
{
	*x = *y = 0;

	*width = WindowRect.right - WindowRect.left;
	*height = WindowRect.bottom - WindowRect.top;
}


void GL_EndRendering (void)
{
	if (!scr_skipupdate)
		SwapBuffers(maindc);

// handle the mouse state when windowed if that's changed
	if (modestate == MS_WINDOWED)
	{
		if (!_windowed_mouse.value) {
			if (windowed_mouse)	{
				IN_DeactivateMouse ();
				IN_ShowMouse ();
				windowed_mouse = false;
			}
		} else {
			windowed_mouse = true;
			if (key_dest == key_game && !mouseactive && ActiveApp) {
				IN_ActivateMouse ();
				IN_HideMouse ();
			} else if (mouseactive && key_dest != key_game) {
				IN_DeactivateMouse ();
				IN_ShowMouse ();
			}
		}
	}
}

void VID_SetPalette (unsigned char *palette)
{
	byte	*pal;
	unsigned r,g,b;
	unsigned v;
	unsigned short i;
	unsigned	*table;

	pal = palette;
	table = d_8to24table;

	for (i=0 ; i<256 ; i++)
	{
		r = pal[0];
		g = pal[1];
		b = pal[2];
		pal += 3;
		
		v = (255 << 24) + (r << 0) + (g << 8) + (b << 16);
		*table++ = v;
	}

	d_8to24table[255] &= 0xffffff;	// 255 is transparent
}

BOOL	gammaworks;

void	VID_ShiftPalette (unsigned char *palette)
{
	extern	byte ramps[3][256];
	
//	VID_SetPalette (palette);

//	gammaworks = SetDeviceGammaRamp (maindc, ramps);
}


void VID_SetDefaultMode (void)
{
	IN_DeactivateMouse ();
}


void VID_Shutdown (void)
{
   	HGLRC hRC;
   	HDC	  hDC;

	if (vid_initialized)
	{
		vid_canalttab = false;
		hRC = wglGetCurrentContext ();
    	hDC = wglGetCurrentDC ();

    	wglMakeCurrent (NULL, NULL);

		GL_DeleteAllTextures ();

    	if (hRC) wglDeleteContext(hRC);
		if (hDC && dibwindow) ReleaseDC (dibwindow, hDC);
		if (modestate == MS_FULLDIB) ChangeDisplaySettings (NULL, 0);
		if (maindc && dibwindow) ReleaseDC (dibwindow, maindc);

		AppActivate(false, false);
	}
}


//==========================================================================


BOOL bSetupPixelFormat (HDC hDC)
{
    static PIXELFORMATDESCRIPTOR pfd =
	{
		sizeof(PIXELFORMATDESCRIPTOR),	// size of this pfd
		1,				// version number
		PFD_DRAW_TO_WINDOW 		// support window
		|  PFD_SUPPORT_OPENGL 	// support OpenGL
		|  PFD_DOUBLEBUFFER ,	// double buffered
		PFD_TYPE_RGBA,			// RGBA type
		24,				// 24-bit color depth (8 for each of r/g/b/a)
		0, 0, 0, 0, 0, 0,		// color bits ignored
		0,				// 8 bits of your finest alpha if you please, my good man (set to none cos some implementations fall back to sw!!!)
		0,				// shift bit ignored
		0,				// no accumulation buffer
		0, 0, 0, 0, 			// accum bits ignored
		24,				// 24 bit z buffer.  because we have removed gl_ztrick, we can get away with this
		8,				// 8 bit stencil buffer
		0,				// no auxiliary buffer
		PFD_MAIN_PLANE,			// main layer
		0,				// reserved
		0, 0, 0				// layer masks ignored
    };

    int pixelformat;

    if ((pixelformat = ChoosePixelFormat (hDC, &pfd)) == 0)
    {
        MessageBox (NULL, "ChoosePixelFormat failed", "Error", MB_OK);
        return FALSE;
    }

    if (SetPixelFormat (hDC, pixelformat, &pfd) == FALSE)
    {
        MessageBox (NULL, "SetPixelFormat failed", "Error", MB_OK);
        return FALSE;
    }

	// the stencil buffer should be available on everything aside from 3dfx so this is really a redundant
	// test, cos MHQuake will work on everything aside from a 3dfx...  besides, if you have a 3dfx, not
	// having a stencil buffer will be least of your problems...
	if (!pfd.cStencilBits) Sys_Error ("No Stencil Buffer!!!");

    return TRUE;
}



byte        scantokey[128] = 
					{ 
//  0           1       2       3       4       5       6       7 
//  8           9       A       B       C       D       E       F 
	0  ,    27,     '1',    '2',    '3',    '4',    '5',    '6', 
	'7',    '8',    '9',    '0',    '-',    '=',    K_BACKSPACE, 9, // 0 
	'q',    'w',    'e',    'r',    't',    'y',    'u',    'i', 
	'o',    'p',    '[',    ']',    13 ,    K_CTRL,'a',  's',      // 1 
	'd',    'f',    'g',    'h',    'j',    'k',    'l',    ';', 
	'\'' ,    '`',    K_SHIFT,'\\',  'z',    'x',    'c',    'v',      // 2 
	'b',    'n',    'm',    ',',    '.',    '/',    K_SHIFT,'*', 
	K_ALT,' ',   0  ,    K_F1, K_F2, K_F3, K_F4, K_F5,   // 3 
	K_F6, K_F7, K_F8, K_F9, K_F10, K_PAUSE  ,    0  , K_HOME, 
	K_UPARROW,K_PGUP,'-',K_LEFTARROW,'5',K_RIGHTARROW,'+',K_END, //4 
	K_DOWNARROW,K_PGDN,K_INS,K_DEL,0,0,             0,              K_F11, 
	K_F12,0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0,        // 5 
	0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0, 
	0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0,        // 6 
	0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0, 
	0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0         // 7 
					}; 

byte        shiftscantokey[128] = 
					{ 
//  0           1       2       3       4       5       6       7 
//  8           9       A       B       C       D       E       F 
	0  ,    27,     '!',    '@',    '#',    '$',    '%',    '^', 
	'&',    '*',    '(',    ')',    '_',    '+',    K_BACKSPACE, 9, // 0 
	'Q',    'W',    'E',    'R',    'T',    'Y',    'U',    'I', 
	'O',    'P',    '{',    '}',    13 ,    K_CTRL,'A',  'S',      // 1 
	'D',    'F',    'G',    'H',    'J',    'K',    'L',    ':', 
	'"' ,    '~',    K_SHIFT,'|',  'Z',    'X',    'C',    'V',      // 2 
	'B',    'N',    'M',    '<',    '>',    '?',    K_SHIFT,'*', 
	K_ALT,' ',   0  ,    K_F1, K_F2, K_F3, K_F4, K_F5,   // 3 
	K_F6, K_F7, K_F8, K_F9, K_F10, K_PAUSE  ,    0  , K_HOME, 
	K_UPARROW,K_PGUP,'_',K_LEFTARROW,'%',K_RIGHTARROW,'+',K_END, //4 
	K_DOWNARROW,K_PGDN,K_INS,K_DEL,0,0,             0,              K_F11, 
	K_F12,0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0,        // 5 
	0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0, 
	0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0,        // 6 
	0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0, 
	0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0         // 7 
					}; 


/*
=======
MapKey

Map from windows to quake keynums
=======
*/
int MapKey (int key)
{
	key = (key>>16)&255;
	if (key > 127)
		return 0;
	if (scantokey[key] == 0)
		Con_DPrintf("key 0x%02x has no translation\n", key);
	return scantokey[key];
}

/*
===================================================================

MAIN WINDOW

===================================================================
*/

/*
================
ClearAllStates
================
*/
void ClearAllStates (void)
{
	int		i;
	
// send an up event for each key, to make sure the server clears them all
	for (i=0 ; i<256 ; i++)
	{
		Key_Event (i, false);
	}

	Key_ClearStates ();
	IN_ClearStates ();
}

void AppActivate(BOOL fActive, BOOL minimize)
/****************************************************************************
*
* Function:     AppActivate
* Parameters:   fActive - True if app is activating
*
* Description:  If the application is activating, then swap the system
*               into SYSPAL_NOSTATIC mode so that our palettes will display
*               correctly.
*
****************************************************************************/
{
	MSG msg;
    HDC			hdc;
    int			i, t;
	static BOOL	sound_active;

	ActiveApp = fActive;
	Minimized = minimize;

// enable/disable sound on focus gain/loss
	if (!ActiveApp && sound_active)
	{
		S_BlockSound ();
		sound_active = false;
	}
	else if (ActiveApp && !sound_active)
	{
		S_UnblockSound ();
		sound_active = true;
	}

	if (fActive)
	{
		if (modestate == MS_FULLDIB)
		{
			IN_ActivateMouse ();
			IN_HideMouse ();
			if (vid_canalttab && vid_wassuspended) {
				vid_wassuspended = false;
				ChangeDisplaySettings (&gdevmode, CDS_FULLSCREEN);
				ShowWindow(mainwindow, SW_SHOWNORMAL);
			}
		}
		else if ((modestate == MS_WINDOWED) && _windowed_mouse.value && key_dest == key_game)
		{
			IN_ActivateMouse ();
			IN_HideMouse ();
		}
	}

	if (!fActive)
	{
		if (modestate == MS_FULLDIB)
		{
			IN_DeactivateMouse ();
			IN_ShowMouse ();
			if (vid_canalttab) { 
				ChangeDisplaySettings (NULL, 0);
				vid_wassuspended = true;
			}
		}
		else if ((modestate == MS_WINDOWED) && _windowed_mouse.value)
		{
			IN_DeactivateMouse ();
			IN_ShowMouse ();
		}
	}
}


#ifndef WM_GRAPHNOTIFY
#define WM_GRAPHNOTIFY  WM_USER + 13
#endif

/* main window procedure */
LONG WINAPI MainWndProc (
    HWND    hWnd,
    UINT    uMsg,
    WPARAM  wParam,
    LPARAM  lParam)
{
    LONG    lRet = 1;
	int		fwKeys, xPos, yPos, fActive, fMinimized, temp;
	extern unsigned int uiWheelMessage;

	if ( uMsg == uiWheelMessage )
		uMsg = WM_MOUSEWHEEL;

    switch (uMsg)
    {
		case WM_KILLFOCUS:
			if (modestate == MS_FULLDIB)
				ShowWindow(mainwindow, SW_SHOWMINNOACTIVE);
			break;

		case WM_CREATE:
			break;

		case WM_MOVE:
			window_x = (int) LOWORD(lParam);
			window_y = (int) HIWORD(lParam);
			VID_UpdateWindowStatus ();
			break;

		case WM_KEYDOWN:
		case WM_SYSKEYDOWN:
			Key_Event (MapKey(lParam), true);
			break;
			
		case WM_KEYUP:
		case WM_SYSKEYUP:
			Key_Event (MapKey(lParam), false);
			break;

		case WM_SYSCHAR:
		// keep Alt-Space from happening
			break;

	// this is complicated because Win32 seems to pack multiple mouse events into
	// one update sometimes, so we always check all states and look for events
		case WM_LBUTTONDOWN:
		case WM_LBUTTONUP:
		case WM_RBUTTONDOWN:
		case WM_RBUTTONUP:
		case WM_MBUTTONDOWN:
		case WM_MBUTTONUP:
		case WM_MOUSEMOVE:
			temp = 0;

			if (wParam & MK_LBUTTON)
				temp |= 1;

			if (wParam & MK_RBUTTON)
				temp |= 2;

			if (wParam & MK_MBUTTON)
				temp |= 4;

			IN_MouseEvent (temp);

			break;

		// JACK: This is the mouse wheel with the Intellimouse
		// Its delta is either positive or neg, and we generate the proper
		// Event.
		case WM_MOUSEWHEEL: 
			if ((short) HIWORD(wParam) > 0) {
				Key_Event(K_MWHEELUP, true);
				Key_Event(K_MWHEELUP, false);
			} else {
				Key_Event(K_MWHEELDOWN, true);
				Key_Event(K_MWHEELDOWN, false);
			}
			break;

    	case WM_SIZE:
            break;

   	    case WM_CLOSE:
			if (MessageBox (mainwindow, "Are you sure you want to quit?", "Confirm Exit",
						MB_YESNO | MB_SETFOREGROUND | MB_ICONQUESTION) == IDYES)
			{
				Sys_Quit ();
			}

	        break;

		case WM_ACTIVATE:
			fActive = LOWORD(wParam);
			fMinimized = (BOOL) HIWORD(wParam);
			AppActivate(!(fActive == WA_INACTIVE), fMinimized);

		// fix the leftover Alt from any Alt-Tab or the like that switched us away
			ClearAllStates ();

			break;

   	    case WM_DESTROY:
        {
			if (dibwindow)
				DestroyWindow (dibwindow);

            PostQuitMessage (0);
        }
        break;

		case WM_GRAPHNOTIFY:
            lRet = CDAudio_MessageHandler (hWnd, uMsg, wParam, lParam);
			break;

    	default:
            /* pass all unhandled messages to DefWindowProc */
            lRet = DefWindowProc (hWnd, uMsg, wParam, lParam);
        break;
    }

    /* return 1 if handled message, 0 if not */
    return lRet;
}


/*
=================
VID_NumModes
=================
*/
int VID_NumModes (void)
{
	return nummodes;
}

	
/*
=================
VID_GetModePtr
=================
*/
vmode_t *VID_GetModePtr (int modenum)
{

	if ((modenum >= 0) && (modenum < nummodes))
		return &modelist[modenum];
	else
		return &badmode;
}


/*
=================
VID_GetModeDescription
=================
*/
char *VID_GetModeDescription (int mode)
{
	char		*pinfo;
	vmode_t		*pv;
	static char	temp[100];

	if ((mode < 0) || (mode >= nummodes))
		return NULL;

	if (!leavecurrentmode)
	{
		pv = VID_GetModePtr (mode);
		pinfo = pv->modedesc;
	}
	else
	{
		sprintf (temp, "Desktop resolution (%d x %d x %d)",
				 modelist[MODE_FULLSCREEN_DEFAULT].width,
				 modelist[MODE_FULLSCREEN_DEFAULT].height,
				 modelist[MODE_FULLSCREEN_DEFAULT].bpp);

		pinfo = temp;
	}

	return pinfo;
}


// KJB: Added this to return the mode driver name in description for console

char *VID_GetExtModeDescription (int mode)
{
	static char	pinfo[40];
	vmode_t		*pv;

	if ((mode < 0) || (mode >= nummodes))
		return NULL;

	pv = VID_GetModePtr (mode);
	if (modelist[mode].type == MS_FULLDIB)
	{
		if (!leavecurrentmode)
		{
			sprintf(pinfo,"%s fullscreen", pv->modedesc);
		}
		else
		{
			sprintf (pinfo, "Desktop resolution (%d x %d x %d)",
					 modelist[MODE_FULLSCREEN_DEFAULT].width,
					 modelist[MODE_FULLSCREEN_DEFAULT].height,
					 modelist[MODE_FULLSCREEN_DEFAULT].bpp);
		}
	}
	else
	{
		if (modestate == MS_WINDOWED)
			sprintf(pinfo, "%s windowed", pv->modedesc);
		else
			sprintf(pinfo, "windowed");
	}

	return pinfo;
}


/*
=================
VID_DescribeCurrentMode_f
=================
*/
void VID_DescribeCurrentMode_f (void)
{
	Con_Printf ("%s\n", VID_GetExtModeDescription (vid_modenum));
}


/*
=================
VID_NumModes_f
=================
*/
void VID_NumModes_f (void)
{

	if (nummodes == 1)
		Con_Printf ("%d video mode is available\n", nummodes);
	else
		Con_Printf ("%d video modes are available\n", nummodes);
}


/*
=================
VID_DescribeMode_f
=================
*/
void VID_DescribeMode_f (void)
{
	int		t, modenum;
	
	modenum = Q_atoi (Cmd_Argv(1));

	t = leavecurrentmode;
	leavecurrentmode = 0;

	Con_Printf ("%s\n", VID_GetExtModeDescription (modenum));

	leavecurrentmode = t;
}


/*
=================
VID_DescribeModes_f
=================
*/
void VID_DescribeModes_f (void)
{
	int			i, lnummodes, t;
	char		*pinfo;
	vmode_t		*pv;

	lnummodes = VID_NumModes ();

	t = leavecurrentmode;
	leavecurrentmode = 0;

	for (i=1 ; i<lnummodes ; i++)
	{
		pv = VID_GetModePtr (i);
		pinfo = VID_GetExtModeDescription (i);
		Con_Printf ("%2d: %s\n", i, pinfo);
	}

	leavecurrentmode = t;
}


void VID_InitDIB (HINSTANCE hInstance)
{
	WNDCLASS		wc;

	// Register the window class
    wc.style         = 0;
    wc.lpfnWndProc   = (WNDPROC) MainWndProc;
    wc.cbClsExtra    = 0;
    wc.cbWndExtra    = 0;
    wc.hInstance     = hInstance;
    wc.hIcon         = 0;
    wc.hCursor       = LoadCursor (NULL, IDC_ARROW);
	wc.hbrBackground = NULL;
    wc.lpszMenuName  = 0;
    wc.lpszClassName = "MHQuake";

    if (!RegisterClass (&wc)) Sys_Error ("Couldn't register window class");

	// set up mode 0 as the windowed mode
	modelist[0].type = MS_WINDOWED;

	if (COM_CheckParm("-width"))
		modelist[0].width = Q_atoi(com_argv[COM_CheckParm("-width")+1]);
	else modelist[0].width = 640;

	if (modelist[0].width < 640) modelist[0].width = 640;

	if (COM_CheckParm("-height"))
		modelist[0].height= Q_atoi(com_argv[COM_CheckParm("-height")+1]);
	else modelist[0].height = modelist[0].width * 3 / 4;

	if (modelist[0].height < 480) modelist[0].height = 480;

	sprintf (modelist[0].modedesc, "%d x %d x %d", modelist[0].width, modelist[0].height, modelist[0].bpp);

	modelist[0].modenum = MODE_WINDOWED;
	modelist[0].dib = 1;
	modelist[0].fullscreen = 0;
	modelist[0].halfscreen = 0;
	modelist[0].bpp = 0;

	nummodes = 1;
}


/*
=================
VID_EnumFullDIB

Because VID_InitFullDIB enumerates modes twice - once for 32bpp modes only, and the second time for other modes,
I've moved this to it's own function
=================
*/
void VID_EnumFullDIB (qboolean bpp32only)
{
	DEVMODE	devmode;
	int modenum;
	int originalnummodes;
	int i;
	qboolean existingmode;

	for (modenum = 0, originalnummodes = nummodes; ; modenum++)
	{
		// too many modes!!!  no point in testing any further
		if (nummodes >= MAX_MODE_LIST) break;

		// enumerate the next available mode, break out if we fail to enumerate
		// (per ms-help://MS.MSDNQTR.2005JAN.1033/gdi/devcons_84oj.htm)
		if (!EnumDisplaySettings (NULL, modenum, &devmode)) break;

		// condtion for first run - only report 32 bpp modes (get them to the top of the list so as to ensure that high-res
		// 32 bpp modes have a better chance of being available)
		if (bpp32only && (devmode.dmBitsPerPel != 32)) continue;

		// condtion for second run - report everything except 32bpp modes (get them at the end of the list, so it
		// doesn't matter as much if some of them fall off)
		if (!bpp32only && (devmode.dmBitsPerPel == 32)) continue;

		// invalid mode conditions
		if (devmode.dmBitsPerPel < 16) continue;			// not enough bpp
		if (devmode.dmBitsPerPel > 32) continue;			// too many bpp
		if (devmode.dmPelsWidth > MAXWIDTH) continue;		// too wide
		if (devmode.dmPelsHeight > MAXHEIGHT) continue;		// too high
		if (devmode.dmPelsWidth < 640) continue;			// not wide enough
		if (devmode.dmPelsHeight < 480) continue;			// not high enough

		// set up the fields used by the CDS test
		devmode.dmFields = DM_BITSPERPEL | DM_PELSWIDTH | DM_PELSHEIGHT;

		// run the CDS in test mode
		if (ChangeDisplaySettings (&devmode, CDS_TEST | CDS_FULLSCREEN) == DISP_CHANGE_SUCCESSFUL)
		{
			// worked OK so store out the mode
			modelist[nummodes].type = MS_FULLDIB;
			modelist[nummodes].width = devmode.dmPelsWidth;
			modelist[nummodes].height = devmode.dmPelsHeight;
			modelist[nummodes].modenum = 0;
			modelist[nummodes].halfscreen = 0;
			modelist[nummodes].dib = 1;
			modelist[nummodes].fullscreen = 1;
			modelist[nummodes].bpp = devmode.dmBitsPerPel;
			sprintf (modelist[nummodes].modedesc, "%d x %d x %d", devmode.dmPelsWidth, devmode.dmPelsHeight, devmode.dmBitsPerPel);

			// now see does this mode already exist.  this is needed cos modes can differ in matters other than
			// width, height and bpp (commonly, refresh rate)
			for (i = originalnummodes, existingmode = false; i < nummodes; i++)
			{
				// invalid conditions
				if (modelist[nummodes].width != modelist[i].width) continue;	// different width
				if (modelist[nummodes].height != modelist[i].height) continue;	// different height
				if (modelist[nummodes].bpp != modelist[i].bpp) continue;		// different bpp

				// width, height and bpp all match, so it's something inconsequential like refresh rate that's different
				existingmode = true;
				break;
			}

			// increment the mode counter if it's a new mode
			if (!existingmode) nummodes++;
		}
	}
}


/*
=================
VID_InitFullDIB
=================
*/
void VID_InitFullDIB (HINSTANCE hInstance)
{
	// store the original number of modes here
	int originalnummodes = nummodes;

	VID_EnumFullDIB (true);		// find all 32 bpp modes
	VID_EnumFullDIB (false);	// find all non-32 bpp modes

	// this should never happen
	if (nummodes == originalnummodes) Con_SafePrintf ("No fullscreen DIB modes found\n");
}


qboolean VID_Is8bit()
{
	// NEVER!!!!
	return false;
}


static void Check_Gamma (unsigned char *pal)
{
	float	f, inf;
	unsigned char	palette[768];
	int		i;

	if ((i = COM_CheckParm("-gamma")) == 0)
	{
		// no need to send through the calculation
		vid_gamma = 1.0;
		return;
	}
	else
		vid_gamma = Q_atof(com_argv[i+1]);

	for (i=0 ; i<768 ; i++)
	{
		f = pow ((pal[i] + 1) / 256.0, vid_gamma);
		inf = f * 255 + 0.5;
		if (inf < 0) inf = 0;
		if (inf > 255) inf = 255;
		palette[i] = inf;
	}

	memcpy (pal, palette, sizeof(palette));
}

/*
===================
VID_Init
===================
*/
void VID_Init (unsigned char *palette)
{
	int		i;
	int		width, height, bpp;
	HDC		hdc;
	int		desktopmatch = 0;

	Cvar_RegisterVariable (&vid_mode);
	Cvar_RegisterVariable (&vid_wait);
	Cvar_RegisterVariable (&vid_nopageflip);
	Cvar_RegisterVariable (&_vid_wait_override);
	Cvar_RegisterVariable (&_vid_default_mode);
	Cvar_RegisterVariable (&_vid_default_mode_win);
	Cvar_RegisterVariable (&vid_config_x);
	Cvar_RegisterVariable (&vid_config_y);
	Cvar_RegisterVariable (&vid_stretch_by_2);
	Cvar_RegisterVariable (&_windowed_mouse);

	Cmd_AddCommand ("vid_nummodes", VID_NumModes_f);
	Cmd_AddCommand ("vid_describecurrentmode", VID_DescribeCurrentMode_f);
	Cmd_AddCommand ("vid_describemode", VID_DescribeMode_f);
	Cmd_AddCommand ("vid_describemodes", VID_DescribeModes_f);

	hIcon = LoadIcon (global_hInstance, MAKEINTRESOURCE (IDI_ICON2));

	InitCommonControls ();

	VID_InitDIB (global_hInstance);
	VID_InitFullDIB (global_hInstance);

	// destroy the dialog here so that it doesn't interfere with any errors
	DestroyWindow (hwnd_dialog);

	// get the current device context
	hdc = GetDC (NULL);

	if (COM_CheckParm ("-window"))
	{
		if (GetDeviceCaps (hdc, RASTERCAPS) & RC_PALETTE)
			Sys_Error ("Can't run in non-RGB mode");

		windowed = true;

		vid_default = MODE_WINDOWED;
	}
	else
	{
		// attempt to match a specified mode with the list of available modes
		if (nummodes == 1) Sys_Error ("No RGB fullscreen modes available");

		windowed = false;

		if (COM_CheckParm ("-mode"))
			vid_default = Q_atoi (com_argv[COM_CheckParm ("-mode") + 1]);
		else
		{
			if (COM_CheckParm ("-current"))
			{
				// take the current desktop resolution and colour depth as the mode, and write these into the
				// default fullscreen mode (mode 1 - originally 640 x 480 x 16)
				modelist[MODE_FULLSCREEN_DEFAULT].width = GetSystemMetrics (SM_CXSCREEN);
				modelist[MODE_FULLSCREEN_DEFAULT].height = GetSystemMetrics (SM_CYSCREEN);
				modelist[MODE_FULLSCREEN_DEFAULT].bpp = GetDeviceCaps (hdc, BITSPIXEL);

				vid_default = MODE_FULLSCREEN_DEFAULT;
				leavecurrentmode = 1;
			}
			else
			{
				// check for a specified width or height, default to desktop dimensions if not specified
				if (COM_CheckParm ("-width"))
				{
					width = Q_atoi (com_argv[COM_CheckParm ("-width") + 1]);

					// if height is unspecified, pick a matching height at 1.3333 aspect ratio
					if (COM_CheckParm ("-height"))
						height = Q_atoi (com_argv[COM_CheckParm ("-height") + 1]);
					else height = (width * 3) / 4;
				}
				else if (COM_CheckParm ("-height"))
				{
					// specified height
					height = Q_atoi (com_argv[COM_CheckParm ("-height") + 1]);

					// if height is unspecified, pick a matching height at 1.3333 aspect ratio
					if (COM_CheckParm ("-width"))
						width = Q_atoi (com_argv[COM_CheckParm ("-width") + 1]);
					else width = (height * 4) / 3;
				}
				else
				{
					// unspecified, so use the desktop resolution
					width = GetSystemMetrics (SM_CXSCREEN);
					height = GetSystemMetrics (SM_CYSCREEN);

					// one match to the desktop
					desktopmatch++;
				}

				// check for a specified bpp, default to desktop bpp if not specified
				if (COM_CheckParm ("-bpp"))
					bpp = Q_atoi (com_argv[COM_CheckParm ("-bpp") + 1]);
				else
				{
					// desktop bpp
					bpp = GetDeviceCaps (hdc, BITSPIXEL);

					// one match to the desktop
					desktopmatch++;
				}

				if (desktopmatch == 2)
				{
					// if both dimensions and depth match the desktop, treat it as if it was -current
					modelist[MODE_FULLSCREEN_DEFAULT].width = width;
					modelist[MODE_FULLSCREEN_DEFAULT].height = height;
					modelist[MODE_FULLSCREEN_DEFAULT].bpp = bpp;

					vid_default = MODE_FULLSCREEN_DEFAULT;
					leavecurrentmode = 1;
				}
				else
				{
					// user specified mode, so check the available modes to see if it's valid
					for (i = 1, vid_default = 0; i < nummodes; i++)
					{
						// try to match the mode
						if ((modelist[i].width == width) && (modelist[i].height == height) && (modelist[i].bpp == bpp))
						{
							// found it
							vid_default = i;
							break;
						}
					}

					// no suitable video mode found
					if (!vid_default) Sys_Error ("Specified video mode %d x %d x %d not available.", width, height, bpp);
				}
			}
		}
	}

	// release the device context
	ReleaseDC (NULL, hdc);

	vid_initialized = true;

	// lock these
	vid.conwidth = 640;
	vid.conheight = 480;

	// software quake legacy shite
	vid.colormap = host_colormap;
	vid.fullbright = 256 - LittleLong (*((int *) vid.colormap + 2048));

	Check_Gamma (palette);
	VID_SetPalette (palette);

	VID_SetMode (vid_default, palette);

    maindc = GetDC (mainwindow);
	bSetupPixelFormat (maindc);

	// attempt to bring up OpenGL
    baseRC = wglCreateContext (maindc);

	if (!baseRC) Sys_Error ("Could not initialize GL (wglCreateContext failed).\n\n",
		"Make sure you are in at least 16 bit mode, and try reducing the resolution or running -window.\n\n",
		"If it still fails, you probably need a new video card...");

    if (!wglMakeCurrent (maindc, baseRC)) Sys_Error ("wglMakeCurrent failed");

	GL_Init ();

	vid_realmode = vid_modenum;

	vid_menudrawfn = VID_MenuDraw;
	vid_menukeyfn = VID_MenuKey;

	strcpy (badmode.modedesc, "Bad mode");
	vid_canalttab = true;

	if (COM_CheckParm ("-fullsbar")) fullsbardraw = true;
}


//========================================================
// Video menu stuff
//========================================================

extern void M_Menu_Options_f (void);
extern void M_Print (int cx, int cy, char *str);
extern void M_PrintWhite (int cx, int cy, char *str);
extern void M_DrawCharacter (int cx, int line, int num);
extern void M_DrawTransPic (int x, int y, qpic_t *pic);
extern void M_DrawPic (int x, int y, qpic_t *pic);

static int	vid_line, vid_wmodes;

typedef struct
{
	int		modenum;
	char	*desc;
	int		iscur;
} modedesc_t;

#define MAX_COLUMN_SIZE		9
#define MODE_AREA_HEIGHT	(MAX_COLUMN_SIZE + 2)
#define MAX_MODEDESCS		(MAX_COLUMN_SIZE*3)

static modedesc_t	modedescs[MAX_MODEDESCS];

/*
================
VID_MenuDraw
================
*/
void VID_MenuDraw (void)
{
	qpic_t		*p;
	char		*ptr;
	int			lnummodes, i, j, k, column, row, dup, dupmode;
	char		temp[100];
	vmode_t		*pv;

	p = Draw_CachePic ("gfx/vidmodes.lmp");
	M_DrawPic ( (320-p->width)/2, 4, p);

	vid_wmodes = 0;
	lnummodes = VID_NumModes ();
	
	for (i=1 ; (i<lnummodes) && (vid_wmodes < MAX_MODEDESCS) ; i++)
	{
		ptr = VID_GetModeDescription (i);
		pv = VID_GetModePtr (i);

		k = vid_wmodes;

		modedescs[k].modenum = i;
		modedescs[k].desc = ptr;
		modedescs[k].iscur = 0;

		if (i == vid_modenum)
			modedescs[k].iscur = 1;

		vid_wmodes++;

	}

	if (vid_wmodes > 0)
	{
		M_Print (2*8, 36+0*8, "Fullscreen Modes (WIDTHxHEIGHTxBPP)");

		column = 8;
		row = 36+2*8;

		for (i=0 ; i<vid_wmodes ; i++)
		{
			if (modedescs[i].iscur)
				M_PrintWhite (column, row, modedescs[i].desc);
			else
				M_Print (column, row, modedescs[i].desc);

			column += 13*8;

			if ((i % VID_ROW_SIZE) == (VID_ROW_SIZE - 1))
			{
				column = 8;
				row += 8;
			}
		}
	}

	M_Print (3*8, 36 + MODE_AREA_HEIGHT * 8 + 8*2,
			 "Video modes must be set from the");
	M_Print (3*8, 36 + MODE_AREA_HEIGHT * 8 + 8*3,
			 "command line with -width <width>");
	M_Print (3*8, 36 + MODE_AREA_HEIGHT * 8 + 8*4,
			 "and -bpp <bits-per-pixel>");
	M_Print (3*8, 36 + MODE_AREA_HEIGHT * 8 + 8*6,
			 "Select windowed mode with -window");
}


/*
================
VID_MenuKey
================
*/
void VID_MenuKey (int key)
{
	switch (key)
	{
	case K_ESCAPE:
		S_LocalSound ("misc/menu1.wav");
		M_Menu_Options_f ();
		break;

	default:
		break;
	}
}
