/*
Copyright (C) 1997 - 2001 Id Software, Inc.

This program is free software; you can redistribute it and / or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111 - 1307, USA.

*/ 
// r_main.c
#include "gl_local.h"

void R_Clear (void);

viddef_t	vid;

refimport_t	ri;


model_t		*r_worldmodel;

float		gldepthmin, gldepthmax;

glconfig_t gl_config;
glstate_t gl_state;

image_t		*r_notexture;		// use for bad textures
image_t		*r_particletexture;	// little dot for particles

entity_t	*currententity;
model_t		*currentmodel;

cplane_t	frustum[4];

int			r_visframecount;	// bumped when going to a new PVS
int			r_framecount;		// used for dlight push checking

int			c_brush_polys, c_alias_polys;

float		v_blend[4];			// final blending color

void GL_Strings_f (void);

// 
// view origin
// 
vec3_t	vup;
vec3_t	vpn;
vec3_t	vright;
vec3_t	r_origin;

float	r_world_matrix[16];
float	r_base_world_matrix[16];

// 
// screen size info
// 
refdef_t	r_newrefdef;

int		r_viewcluster, r_viewcluster2, r_oldviewcluster, r_oldviewcluster2;

cvar_t	*r_norefresh;
cvar_t	*r_drawentities;
cvar_t	*r_drawworld;
cvar_t	*r_speeds;
cvar_t	*r_fullbright;
cvar_t	*r_novis;
cvar_t	*r_nocull;
cvar_t	*r_lerpmodels;
cvar_t	*r_lefthand;

cvar_t	*r_lightlevel;	// FIXME: This is a HACK to get the client's light level

cvar_t	*gl_nosubimage;
cvar_t	*gl_allow_software;

cvar_t	*gl_vertex_arrays;

cvar_t	*gl_particle_min_size;
cvar_t	*gl_particle_max_size;
cvar_t	*gl_particle_size;
cvar_t	*gl_particle_att_a;
cvar_t	*gl_particle_att_b;
cvar_t	*gl_particle_att_c;

cvar_t	*gl_ext_swapinterval;
cvar_t	*gl_ext_palettedtexture;
cvar_t	*gl_ext_multitexture;
cvar_t	*gl_ext_pointparameters;
cvar_t	*gl_ext_compiled_vertex_array;
cvar_t	*gl_anisotropic_filter;

cvar_t	*gl_bitdepth;
cvar_t	*gl_drawbuffer;
cvar_t *gl_driver;
cvar_t	*gl_lightmap;
cvar_t	*gl_shadows;
cvar_t	*gl_mode;
cvar_t	*gl_dynamic;
cvar_t *gl_monolightmap;
cvar_t	*gl_modulate;
cvar_t	*gl_nobind;
cvar_t	*gl_round_down;
cvar_t	*gl_picmip;
cvar_t	*gl_skymip;
cvar_t	*gl_showtris;
cvar_t	*gl_ztrick;
cvar_t	*gl_finish;
cvar_t	*gl_clear;
cvar_t	*gl_cull;
cvar_t	*gl_polyblend;
cvar_t	*gl_flashblend;
cvar_t	*gl_playermip;
cvar_t *gl_saturatelighting;
cvar_t	*gl_swapinterval;
cvar_t	*gl_texturemode;
cvar_t	*gl_texturealphamode;
cvar_t	*gl_texturesolidmode;
cvar_t	*gl_lockpvs;

cvar_t	*gl_3dlabs_broken;

cvar_t	*vid_fullscreen;
cvar_t	*vid_gamma;
cvar_t	*lm_gamma;
cvar_t	*vid_ref;

// MH - detail textures begin
cvar_t	*gl_detailtextures;
// MH - detail textures begin

cvar_t *r_hazewibble;

/*
================= 
R_CullBox

Returns true if the box is completely outside the frustom
================= 
*/ 
qboolean R_CullBox (vec3_t mins, vec3_t maxs) 
{
	int		i;

	if (r_nocull->value) 
		return false;

	for (i = 0; i<4; i++) 
		if (BOX_ON_PLANE_SIDE (mins, maxs, &frustum[i]) == 2) 
			return true;
	return false;
}


void R_RotateForEntity (entity_t *e) 
{
	glTranslatef (e->origin[0], e->origin[1], e->origin[2]);

	glRotatef (e->angles[1], 0, 0, 1);
	glRotatef (-e->angles[0], 0, 1, 0);
	glRotatef (-e->angles[2], 1, 0, 0);
}

/*
============================================================= 

SPRITE MODELS

============================================================= 
*/ 


/*
================= 
R_DrawSpriteModel

================= 
*/ 
void R_DrawSpriteModel (entity_t *e) 
{
	float alpha = 1.0F;
	vec3_t	point;
	dsprframe_t	*frame;
	float		*up, *right;
	dsprite_t		*psprite;

	// don't even bother culling, because it's just a single
	// polygon without a surface cache

	psprite = (dsprite_t *) currentmodel->extradata;

#if 0
	if (e->frame < 0 || e->frame >= psprite->numframes) 
	{
		VID_Printf (PRINT_ALL, "no such sprite frame %i\n", e->frame);
		e->frame = 0;
	}
#endif
	e->frame %= psprite->numframes;

	frame = &psprite->frames[e->frame];

#if 0
	if (psprite->type == SPR_ORIENTED) 
	{	// bullet marks on walls
		vec3_t		v_forward, v_right, v_up;

		AngleVectors (currententity->angles, v_forward, v_right, v_up);
		up = v_up;
		right = v_right;
	}
	else
#endif
	{	// normal sprite
		up = vup;
		right = vright;
	}

	if (e->flags & RF_TRANSLUCENT) 
		alpha = e->alpha;

	if (alpha != 1.0F) 
		glEnable (GL_BLEND);

	glColor4f (1, 1, 1, alpha);

	glBindTexture (GL_TEXTURE_2D, currentmodel->skins[e->frame]->texnum);

	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	if (alpha == 1.0) 
		glEnable (GL_ALPHA_TEST);
	else
		glDisable (GL_ALPHA_TEST);

	glBegin (GL_QUADS);

	glTexCoord2f (0, 1);
	VectorMA (e->origin, - frame->origin_y, up, point);
	VectorMA (point, - frame->origin_x, right, point);
	glVertex3fv (point);

	glTexCoord2f (0, 0);
	VectorMA (e->origin, frame->height - frame->origin_y, up, point);
	VectorMA (point, - frame->origin_x, right, point);
	glVertex3fv (point);

	glTexCoord2f (1, 0);
	VectorMA (e->origin, frame->height - frame->origin_y, up, point);
	VectorMA (point, frame->width - frame->origin_x, right, point);
	glVertex3fv (point);

	glTexCoord2f (1, 1);
	VectorMA (e->origin, - frame->origin_y, up, point);
	VectorMA (point, frame->width - frame->origin_x, right, point);
	glVertex3fv (point);

	glEnd ();

	glDisable (GL_ALPHA_TEST);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

	if (alpha != 1.0F) 
		glDisable (GL_BLEND);

	glColor4f (1, 1, 1, 1);
}

// ================================================================================== 

/*
============= 
R_DrawNullModel
============= 
*/ 
void R_DrawNullModel (void) 
{
	vec3_t	shadelight;
	int		i;

	if (currententity->flags & RF_FULLBRIGHT) 
		shadelight[0] = shadelight[1] = shadelight[2] = 1.0F;
	else
		R_LightPoint (currententity->origin, shadelight);

	glPushMatrix ();
	R_RotateForEntity (currententity);

	glDisable (GL_TEXTURE_2D);
	glColor3fv (shadelight);

	glBegin (GL_TRIANGLE_FAN);
	glVertex3f (0, 0, - 16);
	for (i = 0; i<= 4; i++) 
		glVertex3f (16*cos (i*M_PI / 2) , 16*sin (i*M_PI / 2) , 0);
	glEnd ();

	glBegin (GL_TRIANGLE_FAN);
	glVertex3f (0, 0, 16);
	for (i = 4; i>= 0; i--) 
		glVertex3f (16*cos (i*M_PI / 2) , 16*sin (i*M_PI / 2) , 0);
	glEnd ();

	glColor3f (1,1,1);
	glPopMatrix ();
	glEnable (GL_TEXTURE_2D);
}

/*
============= 
R_DrawEntitiesOnList
============= 
*/ 
void R_DrawEntitiesOnList (void) 
{
	int		i;

	if (!r_drawentities->value) 
		return;

	// draw non - transparent first
	for (i = 0; i<r_newrefdef.num_entities; i++) 
	{
		currententity = &r_newrefdef.entities[i];
		if (currententity->flags & RF_TRANSLUCENT) 
			continue;	// solid

		if (currententity->flags & RF_BEAM) 
		{
			R_DrawBeam (currententity);
		}
		else
		{
			currentmodel = currententity->model;
			if (!currentmodel) 
			{
				R_DrawNullModel ();
				continue;
			}
			switch (currentmodel->type) 
			{
			case mod_alias:
				R_DrawAliasModel (currententity);
				break;
			case mod_brush:
				//R_DrawBrushModel (currententity);
				break;
			case mod_sprite:
				R_DrawSpriteModel (currententity);
				break;
			default:
				VID_Error (ERR_DROP, "Bad modeltype");
				break;
			}
		}
	}

	// draw transparent entities
	// we could sort these if it ever becomes a problem...
	glDepthMask (0);		// no z writes
	for (i = 0; i<r_newrefdef.num_entities; i++) 
	{
		currententity = &r_newrefdef.entities[i];
		if (!(currententity->flags & RF_TRANSLUCENT)) 
			continue;	// solid

		if (currententity->flags & RF_BEAM) 
		{
			R_DrawBeam (currententity);
		}
		else
		{
			currentmodel = currententity->model;

			if (!currentmodel) 
			{
				R_DrawNullModel ();
				continue;
			}
			switch (currentmodel->type) 
			{
			case mod_alias:
				R_DrawAliasModel (currententity);
				break;
			case mod_brush:
				//R_DrawBrushModel (currententity);
				break;
			case mod_sprite:
				R_DrawSpriteModel (currententity);
				break;
			default:
				VID_Error (ERR_DROP, "Bad modeltype");
				break;
			}
		}
	}
	glDepthMask (1);		// back to writing

}

/*
** GL_DrawParticles
**
*/ 
void GL_DrawParticles (int num_particles, const particle_t particles[], const unsigned colortable[768]) 
{
	const particle_t *p;
	int				i;
	vec3_t			up, right;
	float			scale;
	byte			color[4];

	glBindTexture (GL_TEXTURE_2D, r_particletexture->texnum);
	glDepthMask (GL_FALSE);		// no z buffering
	glEnable (GL_BLEND);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glBegin (GL_TRIANGLES);

	VectorScale (vup, 1.5, up);
	VectorScale (vright, 1.5, right);

	for (p = particles, i = 0; i < num_particles; i++ ,p++) 
	{
		// hack a scale up to keep particles from disapearing
		scale = (p->origin[0]- r_origin[0]) * vpn[0]+ 
			(p->origin[1]- r_origin[1]) * vpn[1]+ 
			(p->origin[2]- r_origin[2]) * vpn[2];

		if (scale < 20) 
			scale = 1;
		else
			scale = 1 + scale * 0.004;

		* (int *) color = colortable[p->color];
		color[3] = p->alpha*255;

		glColor4ubv (color);

		glTexCoord2f (0.0625, 0.0625);
		glVertex3fv (p->origin);

		glTexCoord2f (1.0625, 0.0625);
		glVertex3f (p->origin[0]+ up[0]*scale, 
			p->origin[1]+ up[1]*scale, 
			p->origin[2]+ up[2]*scale);

		glTexCoord2f (0.0625, 1.0625);
		glVertex3f (p->origin[0]+ right[0]*scale, 
			p->origin[1]+ right[1]*scale, 
			p->origin[2]+ right[2]*scale);
	}

	glEnd ();
	glDisable (GL_BLEND);
	glColor4f (1,1,1,1);
	glDepthMask (1);		// back to normal Z buffering
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
}

/*
=============== 
R_DrawParticles
=============== 
*/ 
void R_DrawParticles (void) 
{
	int i;
	unsigned char color[4];
	const particle_t *p;
	qboolean leftovers = false;

	glDepthMask (GL_FALSE);
	glEnable (GL_BLEND);
	glDisable (GL_TEXTURE_2D);

	glPointSize (gl_particle_size->value);

	glBegin (GL_POINTS);
	for (i = 0, p = r_newrefdef.particles; i < r_newrefdef.num_particles; i++ , p++) 
	{
		if (p->texnum != -1)
		{
			leftovers = true;
			continue;
		}

		* (int *) color = d_8to24table[p->color];
		color[3] = p->alpha*255;

		glColor4ubv (color);

		glVertex3fv (p->origin);
	}
	glEnd ();

	glDisable (GL_BLEND);
	glColor4f (1.0F, 1.0F, 1.0F, 1.0F);
	glDepthMask (GL_TRUE);
	glEnable (GL_TEXTURE_2D);

	if (leftovers)
	{
		// VID_Printf (PRINT_ALL, "Leftover particles\n");
		GL_DrawParticles (r_newrefdef.num_particles, r_newrefdef.particles, d_8to24table);
	}
}

/*
============ 
R_PolyBlend
============ 
*/ 
void R_PolyBlend (void) 
{
	if (!gl_polyblend->value) return;
	if (!v_blend[3]) return;

	glDisable (GL_ALPHA_TEST);
	glEnable (GL_BLEND);
	glDisable (GL_DEPTH_TEST);
	glDisable (GL_TEXTURE_2D);

	//glBlendFunc (GL_DST_COLOR, GL_SRC_COLOR);

	glLoadIdentity ();

	// FIXME: get rid of these
	glRotatef ( - 90, 1, 0, 0);	// put Z going up
	glRotatef (90, 0, 0, 1);	// put Z going up

	glColor4fv (v_blend);

	glBegin (GL_QUADS);

	glVertex3f (10, 100, 100);
	glVertex3f (10, - 100, 100);
	glVertex3f (10, - 100, - 100);
	glVertex3f (10, 100, - 100);
	glEnd ();

	//glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable (GL_BLEND);
	glEnable (GL_TEXTURE_2D);
	//glEnable (GL_ALPHA_TEST);

	glColor4f (1,1,1,1);
}

// ======================================================================= 

int SignbitsForPlane (cplane_t *out) 
{
	int	bits, j;

	// for fast box on planeside test

	bits = 0;
	for (j = 0; j<3; j++) 
	{
		if (out->normal[j]< 0) 
			bits |= 1 << j;
	}
	return bits;
}


void R_SetFrustum (void) 
{
	int		i;

#if 0
	/*
	** this code is wrong, since it presume a 90 degree FOV both in the
	** horizontal and vertical plane
	*/ 
	// front side is visible
	VectorAdd (vpn, vright, frustum[0].normal);
	VectorSubtract (vpn, vright, frustum[1].normal);
	VectorAdd (vpn, vup, frustum[2].normal);
	VectorSubtract (vpn, vup, frustum[3].normal);

	// we theoretically don't need to normalize these vectors, but I do it
	// anyway so that debugging is a little easier
	VectorNormalize (frustum[0].normal);
	VectorNormalize (frustum[1].normal);
	VectorNormalize (frustum[2].normal);
	VectorNormalize (frustum[3].normal);
#else
	// rotate VPN right by FOV_X / 2 degrees
	RotatePointAroundVector (frustum[0].normal, vup, vpn, - (90 - r_newrefdef.fov_x / 2));
	// rotate VPN left by FOV_X / 2 degrees
	RotatePointAroundVector (frustum[1].normal, vup, vpn, 90 - r_newrefdef.fov_x / 2);
	// rotate VPN up by FOV_X / 2 degrees
	RotatePointAroundVector (frustum[2].normal, vright, vpn, 90 - r_newrefdef.fov_y / 2);
	// rotate VPN down by FOV_X / 2 degrees
	RotatePointAroundVector (frustum[3].normal, vright, vpn, - (90 - r_newrefdef.fov_y / 2));
#endif

	for (i = 0; i<4; i++) 
	{
		frustum[i].type = PLANE_ANYZ;
		frustum[i].dist = DotProduct (r_origin, frustum[i].normal);
		frustum[i].signbits = SignbitsForPlane (&frustum[i]);
	}
}

// ======================================================================= 

/*
=============== 
R_SetupFrame
=============== 
*/ 
void R_SetupFrame (void) 
{
	int i;
	mleaf_t	*leaf;

	r_framecount++;

	// build the transformation matrix for the given view angles
	VectorCopy (r_newrefdef.vieworg, r_origin);

	AngleVectors (r_newrefdef.viewangles, vpn, vright, vup);

	// current viewcluster
	if (!(r_newrefdef.rdflags & RDF_NOWORLDMODEL)) 
	{
		r_oldviewcluster = r_viewcluster;
		r_oldviewcluster2 = r_viewcluster2;
		leaf = Mod_PointInLeaf (r_origin, r_worldmodel);
		r_viewcluster = r_viewcluster2 = leaf->cluster;

		// check above and below so crossing solid water doesn't draw wrong
		if (!leaf->contents) 
		{	// look down a bit
			vec3_t	temp;

			VectorCopy (r_origin, temp);
			temp[2]-= 16;
			leaf = Mod_PointInLeaf (temp, r_worldmodel);
			if (!(leaf->contents & CONTENTS_SOLID) &&
				(leaf->cluster != r_viewcluster2)) 
				r_viewcluster2 = leaf->cluster;
		}
		else
		{	// look up a bit
			vec3_t	temp;

			VectorCopy (r_origin, temp);
			temp[2]+= 16;
			leaf = Mod_PointInLeaf (temp, r_worldmodel);
			if (!(leaf->contents & CONTENTS_SOLID) &&
				(leaf->cluster != r_viewcluster2)) 
				r_viewcluster2 = leaf->cluster;
		}
	}

	for (i = 0; i<4; i++) 
		v_blend[i] = r_newrefdef.blend[i];

	c_brush_polys = 0;
	c_alias_polys = 0;

	// clear out the portion of the screen that the NOWORLDMODEL defines
	if (r_newrefdef.rdflags & RDF_NOWORLDMODEL) 
	{
		glEnable (GL_SCISSOR_TEST);
		glClearColor  (0, 0, 0, 1);
		glScissor (r_newrefdef.x, vid.height - r_newrefdef.height - r_newrefdef.y, r_newrefdef.width, r_newrefdef.height);
		glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		glClearColor (0, 0, 0, 1);
		glDisable (GL_SCISSOR_TEST);
	}
}


void MYgluPerspective (GLdouble fovy, GLdouble aspect,
					   GLdouble zNear, GLdouble zFar) 
{
	GLdouble xmin, xmax, ymin, ymax;

	ymax = zNear * tan (fovy * M_PI / 360.0);
	ymin = - ymax;

	xmin = ymin * aspect;
	xmax = ymax * aspect;

	xmin += - (2 * gl_state.camera_separation) / zNear;
	xmax += - (2 * gl_state.camera_separation) / zNear;

	glFrustum (xmin, xmax, ymin, ymax, zNear, zFar);
}


/*
============= 
R_SetupGL
============= 
*/
void R_SetupViewport (unsigned int vpx, unsigned int vpy, unsigned int vpw, unsigned int vph)
{
	float screenaspect;

	glViewport (vpx, vpy, vpw, vph);

	// set up projection matrix
	screenaspect = (float) vpw / vph;

	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();

	//if (vpw == vph)
	//	MYgluPerspective (1, 1, 4, 4096);
	//else MYgluPerspective (r_newrefdef.fov_y, screenaspect, 4, 4096);
	MYgluPerspective (r_newrefdef.fov_y, (float) vid.width / vid.height, 4, 4096); //screenaspect, 4, 4096);

	glCullFace (GL_FRONT);

	glMatrixMode (GL_MODELVIEW);
	glLoadIdentity ();

	glRotatef (-90, 1, 0, 0);	// put Z going up
	glRotatef (90, 0, 0, 1);	// put Z going up
	glRotatef (-r_newrefdef.viewangles[2], 1, 0, 0);
	glRotatef (-r_newrefdef.viewangles[0], 0, 1, 0);
	glRotatef (-r_newrefdef.viewangles[1], 0, 0, 1);
	glTranslatef (-r_newrefdef.vieworg[0], -r_newrefdef.vieworg[1], -r_newrefdef.vieworg[2]);

	glGetFloatv (GL_MODELVIEW_MATRIX, r_world_matrix);

	if (gl_cull->value) 
		glEnable (GL_CULL_FACE);
	else
		glDisable (GL_CULL_FACE);

	glDisable (GL_BLEND);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable (GL_ALPHA_TEST);
	glEnable (GL_DEPTH_TEST);
}

/*
============= 
R_Clear
============= 
*/ 
void R_Clear (void) 
{
	// clear the stencil buffer to 1
	glClearStencil (1);

	// clear all buffers each frame.  z-tricking was removed as dropping the depth buffer to 24 bits will reduce precision
	// even further.  the color buffer needs to be cleared for mapshots (in the future), blooms, and the like.
	glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

	// hack for avoiding a depth buffer clear with heat haze
	// only use 18 bits of the buffer for the main render.
	gldepthmin = 0;
	gldepthmax = 0.75;
	glDepthFunc (GL_LEQUAL);

	glDepthRange (gldepthmin, gldepthmax);
}


void R_Flash (void) 
{
	R_PolyBlend ();
}

/*
================ 
R_RenderView

r_newrefdef must be set before the first call
================ 
*/
void R_RenderParticleSprites (void);

void R_RenderView (refdef_t *fd) 
{
	if (r_norefresh->value) 
		return;

	r_newrefdef = *fd;

	if (!r_worldmodel && !(r_newrefdef.rdflags & RDF_NOWORLDMODEL)) 
		VID_Error (ERR_DROP, "R_RenderView: NULL worldmodel");

	if (r_speeds->value) 
	{
		c_brush_polys = 0;
		c_alias_polys = 0;
	}

	if (gl_finish->value) glFinish ();

	R_SetupFrame ();

	R_SetFrustum ();

	//R_SetupViewport (0, 0, vid.width, vid.height);

	R_MarkLeaves ();	// done here so we know if we're in water

	R_DrawWorld ();

	R_DrawEntitiesOnList ();

	R_RenderDlights ();

	R_RenderParticleSprites ();
	R_DrawParticles ();

	R_DrawAlphaSurfaces ();

	R_Flash ();

	if (r_speeds->value) 
	{
		VID_Printf (PRINT_ALL, "%4i wpoly %4i epoly %i tex %i lmaps\n",
			c_brush_polys, 
			c_alias_polys, 
			c_visible_textures, 
			c_visible_lightmaps); 
	}
}

int Mult8 (int numin);

void R_Set2DScreenCoords (viddef_t *myviddef)
{
	myviddef->width2d = 640;
	myviddef->height2d = (640 * myviddef->height) / myviddef->width;

	if (myviddef->height2d < 480)
	{
		myviddef->height2d = 480;
		myviddef->width2d = (480 * myviddef->width) / myviddef->height;
	}

	// make multiples of 8
	myviddef->width2d = Mult8 (myviddef->width2d);
	myviddef->height2d = Mult8 (myviddef->height2d);
}


void R_SetGL2D (void)
{
	// set 2D virtual screen size
	glViewport (0,0, vid.width, vid.height);
	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();
	//glOrtho (0, vid.width, vid.height, 0, - 99999, 99999);
	R_Set2DScreenCoords (&vid);
	glOrtho (0, vid.width2d, vid.height2d, 0, - 99999, 99999);
	glMatrixMode (GL_MODELVIEW);
	glLoadIdentity ();
	glDisable (GL_DEPTH_TEST);
	glDisable (GL_CULL_FACE);
	glDisable (GL_BLEND);
	glEnable (GL_ALPHA_TEST);
	glColor4f (1,1,1,1);
}

static void GL_DrawColoredStereoLinePair (float r, float g, float b, float y) 
{
	glColor3f (r, g, b);
	glVertex2f (0, y);
	glVertex2f (vid.width, y);
	glColor3f (0, 0, 0);
	glVertex2f (0, y + 1);
	glVertex2f (vid.width, y + 1);
}

static void GL_DrawStereoPattern (void) 
{
	int i;

	if (!(gl_config.renderer & GL_RENDERER_INTERGRAPH)) 
		return;

	if (!gl_state.stereo_enabled) 
		return;

	R_SetGL2D ();

	mhglDrawBuffer (GL_BACK_LEFT);

	for (i = 0; i < 20; i++) 
	{
		glBegin (GL_LINES);
		GL_DrawColoredStereoLinePair (1, 0, 0, 0);
		GL_DrawColoredStereoLinePair (1, 0, 0, 2);
		GL_DrawColoredStereoLinePair (1, 0, 0, 4);
		GL_DrawColoredStereoLinePair (1, 0, 0, 6);
		GL_DrawColoredStereoLinePair (0, 1, 0, 8);
		GL_DrawColoredStereoLinePair (1, 1, 0, 10);
		GL_DrawColoredStereoLinePair (1, 1, 0, 12);
		GL_DrawColoredStereoLinePair (0, 1, 0, 14);
		glEnd ();

		GLimp_EndFrame ();
	}
}


/*
==================== 
R_SetLightLevel

==================== 
*/ 
void R_SetLightLevel (void) 
{
	vec3_t		shadelight;

	if (r_newrefdef.rdflags & RDF_NOWORLDMODEL) 
		return;

	// save off light value for server to look at (BIG HACK!) 

	R_LightPoint (r_newrefdef.vieworg, shadelight);

	// pick the greatest component, which should be the same
	// as the mono value returned by software
	if (shadelight[0]> shadelight[1]) 
	{
		if (shadelight[0]> shadelight[2]) 
			r_lightlevel->value = 150*shadelight[0];
		else
			r_lightlevel->value = 150*shadelight[2];
	}
	else
	{
		if (shadelight[1]> shadelight[2]) 
			r_lightlevel->value = 150*shadelight[1];
		else
			r_lightlevel->value = 150*shadelight[2];
	}

}

/*
@@@@@@@@@@@@@@@@@@@@@
R_RenderFrame

@@@@@@@@@@@@@@@@@@@@@
*/ 
void R_RenderFrame (refdef_t *fd) 
{
	R_RenderView (fd);
	R_SetLightLevel ();
	R_SetGL2D ();
}



qboolean inScreenDumpMode = false;
float dumptime = 0;

// this command is a toggle
void SCR_Dump_f (void)
{
	if (inScreenDumpMode)
	{
		inScreenDumpMode = false;
		dumptime = 0;
	}
	else
	{
		inScreenDumpMode = true;
		dumptime = 0;
	}
}


void R_Register (void) 
{
	r_lefthand = Cvar_Get ("hand", "0", CVAR_USERINFO | CVAR_ARCHIVE);
	r_norefresh = Cvar_Get ("r_norefresh", "0", 0);
	r_fullbright = Cvar_Get ("r_fullbright", "0", 0);
	r_drawentities = Cvar_Get ("r_drawentities", "1", 0);
	r_drawworld = Cvar_Get ("r_drawworld", "1", 0);
	r_novis = Cvar_Get ("r_novis", "0", 0);
	r_nocull = Cvar_Get ("r_nocull", "0", 0);
	r_lerpmodels = Cvar_Get ("r_lerpmodels", "1", 0);
	r_speeds = Cvar_Get ("r_speeds", "0", 0);

	r_lightlevel = Cvar_Get ("r_lightlevel", "0", 0);

	gl_nosubimage = Cvar_Get ("gl_nosubimage", "0", 0);
	gl_allow_software = Cvar_Get ("gl_allow_software", "0", 0);

	gl_particle_min_size = Cvar_Get ("gl_particle_min_size", "2", CVAR_ARCHIVE);
	gl_particle_max_size = Cvar_Get ("gl_particle_max_size", "40", CVAR_ARCHIVE);
	gl_particle_size = Cvar_Get ("gl_particle_size", "40", CVAR_ARCHIVE);
	gl_particle_att_a = Cvar_Get ("gl_particle_att_a", "0.01", CVAR_ARCHIVE);
	gl_particle_att_b = Cvar_Get ("gl_particle_att_b", "0.0", CVAR_ARCHIVE);
	gl_particle_att_c = Cvar_Get ("gl_particle_att_c", "0.01", CVAR_ARCHIVE);

	gl_modulate = Cvar_Get ("gl_modulate", "1", CVAR_ARCHIVE);
	gl_bitdepth = Cvar_Get ("gl_bitdepth", "0", 0);
	gl_mode = Cvar_Get ("gl_mode", "3", CVAR_ARCHIVE);
	gl_lightmap = Cvar_Get ("gl_lightmap", "0", 0);
	gl_shadows = Cvar_Get ("gl_shadows", "0", CVAR_ARCHIVE);
	gl_dynamic = Cvar_Get ("gl_dynamic", "1", 0);
	gl_nobind = Cvar_Get ("gl_nobind", "0", 0);
	gl_round_down = Cvar_Get ("gl_round_down", "1", 0);
	gl_picmip = Cvar_Get ("gl_picmip", "0", 0);
	gl_skymip = Cvar_Get ("gl_skymip", "0", 0);
	gl_showtris = Cvar_Get ("gl_showtris", "0", 0);
	gl_ztrick = Cvar_Get ("gl_ztrick", "0", 0);
	gl_finish = Cvar_Get ("gl_finish", "0", CVAR_ARCHIVE);
	gl_clear = Cvar_Get ("gl_clear", "0", 0);
	gl_cull = Cvar_Get ("gl_cull", "1", 0);
	gl_polyblend = Cvar_Get ("gl_polyblend", "1", 0);
	gl_flashblend = Cvar_Get ("gl_flashblend", "0", 0);
	gl_playermip = Cvar_Get ("gl_playermip", "0", 0);
	gl_monolightmap = Cvar_Get ("gl_monolightmap", "0", 0);
	gl_driver = Cvar_Get ("gl_driver", "opengl32", CVAR_ARCHIVE);
	gl_texturemode = Cvar_Get ("gl_texturemode", "GL_LINEAR_MIPMAP_LINEAR", CVAR_ARCHIVE);
	gl_texturealphamode = Cvar_Get ("gl_texturealphamode", "default", CVAR_ARCHIVE);
	gl_texturesolidmode = Cvar_Get ("gl_texturesolidmode", "default", CVAR_ARCHIVE);
	gl_lockpvs = Cvar_Get ("gl_lockpvs", "0", 0);

	gl_vertex_arrays = Cvar_Get ("gl_vertex_arrays", "1", CVAR_ARCHIVE);

	gl_ext_swapinterval = Cvar_Get ("gl_ext_swapinterval", "1", CVAR_ARCHIVE);
	gl_ext_palettedtexture = Cvar_Get ("gl_ext_palettedtexture", "1", CVAR_ARCHIVE);
	gl_ext_multitexture = Cvar_Get ("gl_ext_multitexture", "1", CVAR_ARCHIVE);
	gl_anisotropic_filter = Cvar_Get ("gl_anisotropic_filter", "1", CVAR_ARCHIVE);
	gl_ext_pointparameters = Cvar_Get ("gl_ext_pointparameters", "1", CVAR_ARCHIVE);
	gl_ext_compiled_vertex_array = Cvar_Get ("gl_ext_compiled_vertex_array", "1", CVAR_ARCHIVE);

	gl_drawbuffer = Cvar_Get ("gl_drawbuffer", "GL_BACK", 0);
	gl_swapinterval = Cvar_Get ("gl_swapinterval", "1", CVAR_ARCHIVE);

	gl_saturatelighting = Cvar_Get ("gl_saturatelighting", "0", 0);

	gl_3dlabs_broken = Cvar_Get ("gl_3dlabs_broken", "1", CVAR_ARCHIVE);

	vid_fullscreen = Cvar_Get ("vid_fullscreen", "0", CVAR_ARCHIVE);
	vid_gamma = Cvar_Get ("vid_gamma", "1.0", CVAR_ARCHIVE);
	lm_gamma = Cvar_Get ("lm_gamma", "1.0", CVAR_ARCHIVE);
	vid_ref = Cvar_Get ("vid_ref", "gl", 0);

	r_hazewibble = Cvar_Get ("r_hazewibble", "0.666", CVAR_ARCHIVE);

	// MH - detail textures begin
	// the amount of scaling to apply to detail textures depends on the value of this
	// cvar. set to 0 to switch off detail texturing.
	// ID recommend using non - integer values in the Q3A shading manual to avoid tiling artefacts
	gl_detailtextures = Cvar_Get ("gl_detailtextures", "0", CVAR_ARCHIVE);
	// MH - detail textures begin

	Cmd_AddCommand ("imagelist", GL_ImageList_f);
	Cmd_AddCommand ("screenshot", GL_ScreenShot_f);
	Cmd_AddCommand ("modellist", Mod_Modellist_f);
	Cmd_AddCommand ("gl_strings", GL_Strings_f);
	Cmd_AddCommand ("screendump", SCR_Dump_f);
}

/*
================== 
R_SetMode
================== 
*/ 
qboolean R_SetMode (void) 
{
	rserr_t err;
	qboolean fullscreen;

	if (vid_fullscreen->modified && !gl_config.allow_cds) 
	{
		VID_Printf (PRINT_ALL, "R_SetMode () - CDS not allowed with this driver\n");
		Cvar_SetValue ("vid_fullscreen", !vid_fullscreen->value);
		vid_fullscreen->modified = false;
	}

	fullscreen = vid_fullscreen->value;

	vid_fullscreen->modified = false;
	gl_mode->modified = false;

	if ((err = GLimp_SetMode (&vid.width, &vid.height, gl_mode->value, fullscreen)) == rserr_ok) 
	{
		gl_state.prev_mode = gl_mode->value;
	}
	else
	{
		if (err == rserr_invalid_fullscreen) 
		{
			Cvar_SetValue ("vid_fullscreen", 0);
			vid_fullscreen->modified = false;
			VID_Printf (PRINT_ALL, "ref_gl::R_SetMode () - fullscreen unavailable in this mode\n");
			if ((err = GLimp_SetMode (&vid.width, &vid.height, gl_mode->value, false)) == rserr_ok) 
				return true;
		}
		else if (err == rserr_invalid_mode) 
		{
			Cvar_SetValue ("gl_mode", gl_state.prev_mode);
			gl_mode->modified = false;
			VID_Printf (PRINT_ALL, "ref_gl::R_SetMode () - invalid mode\n");
		}

		// try setting it back to something safe
		if ((err = GLimp_SetMode (&vid.width, &vid.height, gl_state.prev_mode, false)) != rserr_ok) 
		{
			VID_Printf (PRINT_ALL, "ref_gl::R_SetMode () - could not revert to safe mode\n");
			return false;
		}
	}
	return true;
}


/*
=============== 
R_Init
=============== 
*/ 
int R_Init (void *hinstance, void *hWnd) 
{	
	char renderer_buffer[1000];
	char vendor_buffer[1000];
	int		err;
	int		j;
	extern float r_turbsin[256];

	// make sure we only get this once
	if (r_turbsin[1] > 0.18)
	{
		// hack ahoy!!!
		for (j = 0; j < 256; j++) 
		{
			r_turbsin[j] *= 0.5;
		}
	}

	VID_Printf (PRINT_ALL, "ref_gl version: "REF_VERSION"\n");

	Draw_GetPalette ();

	R_Register ();

	// initialize our QGL dynamic bindings
	if (!QGL_Init (gl_driver->string)) 
	{
		QGL_Shutdown ();
		VID_Printf (PRINT_ALL, "ref_gl::R_Init () - could not load \"%s\"\n", gl_driver->string);
		return - 1;
	}

	// initialize OS - specific parts of OpenGL
	if (!GLimp_Init (hinstance, hWnd)) 
	{
		QGL_Shutdown ();
		return - 1;
	}

	// set our "safe" modes
	gl_state.prev_mode = 3;

	// create the window and set up the context
	if (!R_SetMode ()) 
	{
		QGL_Shutdown ();
		VID_Printf (PRINT_ALL, "ref_gl::R_Init () - could not R_SetMode () \n");
		return - 1;
	}

	VID_MenuInit ();

	/*
	** get our various GL strings
	*/ 
	gl_config.vendor_string = glGetString (GL_VENDOR);
	VID_Printf (PRINT_ALL, "GL_VENDOR: %s\n", gl_config.vendor_string);
	gl_config.renderer_string = glGetString (GL_RENDERER);
	VID_Printf (PRINT_ALL, "GL_RENDERER: %s\n", gl_config.renderer_string);
	gl_config.version_string = glGetString (GL_VERSION);
	VID_Printf (PRINT_ALL, "GL_VERSION: %s\n", gl_config.version_string);
	gl_config.extensions_string = glGetString (GL_EXTENSIONS);
	VID_Printf (PRINT_ALL, "GL_EXTENSIONS: %s\n", gl_config.extensions_string);

	strcpy (renderer_buffer, gl_config.renderer_string);
	_strlwr (renderer_buffer);

	strcpy (vendor_buffer, gl_config.vendor_string);
	_strlwr (vendor_buffer);

	if (strstr (renderer_buffer, "voodoo")) 
	{
		if (!strstr (renderer_buffer, "rush")) 
			gl_config.renderer = GL_RENDERER_VOODOO;
		else
			gl_config.renderer = GL_RENDERER_VOODOO_RUSH;
	}
	else if (strstr (vendor_buffer, "sgi")) 
		gl_config.renderer = GL_RENDERER_SGI;
	else if (strstr (renderer_buffer, "permedia")) 
		gl_config.renderer = GL_RENDERER_PERMEDIA2;
	else if (strstr (renderer_buffer, "glint")) 
		gl_config.renderer = GL_RENDERER_GLINT_MX;
	else if (strstr (renderer_buffer, "glzicd")) 
		gl_config.renderer = GL_RENDERER_REALIZM;
	else if (strstr (renderer_buffer, "gdi")) 
		gl_config.renderer = GL_RENDERER_MCD;
	else if (strstr (renderer_buffer, "pcx2")) 
		gl_config.renderer = GL_RENDERER_PCX2;
	else if (strstr (renderer_buffer, "verite")) 
		gl_config.renderer = GL_RENDERER_RENDITION;
	else
		gl_config.renderer = GL_RENDERER_OTHER;

	if (toupper (gl_monolightmap->string[1]) != 'F') 
	{
		if (gl_config.renderer == GL_RENDERER_PERMEDIA2) 
		{
			Cvar_Set ("gl_monolightmap", "A");
			VID_Printf (PRINT_ALL, "...using gl_monolightmap 'a'\n");
		}
		else if (gl_config.renderer & GL_RENDERER_POWERVR) 
		{
			Cvar_Set ("gl_monolightmap", "0");
		}
		else
		{
			Cvar_Set ("gl_monolightmap", "0");
		}
	}

	// power vr can't have anything stay in the framebuffer, so
	// the screen needs to redraw the tiled background every frame
	if (gl_config.renderer & GL_RENDERER_POWERVR) 
	{
		Cvar_Set ("scr_drawall", "1");
	}
	else
	{
		Cvar_Set ("scr_drawall", "0");
	}

#ifdef __linux__
	Cvar_SetValue ("gl_finish", 1);
#endif

	// MCD has buffering issues
	if (gl_config.renderer == GL_RENDERER_MCD) 
	{
		Cvar_SetValue ("gl_finish", 1);
	}

	if (gl_config.renderer & GL_RENDERER_3DLABS) 
	{
		if (gl_3dlabs_broken->value) 
			gl_config.allow_cds = false;
		else
			gl_config.allow_cds = true;
	}
	else
	{
		gl_config.allow_cds = true;
	}

	if (gl_config.allow_cds) 
		VID_Printf (PRINT_ALL, "...allowing CDS\n");
	else
		VID_Printf (PRINT_ALL, "...disabling CDS\n");

	/*
	** grab extensions
	*/ 
	if (strstr (gl_config.extensions_string, "GL_EXT_compiled_vertex_array") || 
		strstr (gl_config.extensions_string, "GL_SGI_compiled_vertex_array")) 
	{
		VID_Printf (PRINT_ALL, "...enabling GL_EXT_compiled_vertex_array\n");
		glLockArrays = (void *) MHGLGetProcAddress ("glLockArrays");
		glUnlockArrays = (void *) MHGLGetProcAddress ("glUnlockArrays");
	}
	else
	{
		VID_Printf (PRINT_ALL, "...GL_EXT_compiled_vertex_array not found\n");
	}

#ifdef _WIN32
	if (strstr (gl_config.extensions_string, "WGL_EXT_swap_control")) 
	{
		qwglSwapInterval = (BOOL (WINAPI *) (int)) MHGLGetProcAddress ("wglSwapInterval");
		VID_Printf (PRINT_ALL, "...enabling WGL_EXT_swap_control\n");
	}
	else
	{
		VID_Printf (PRINT_ALL, "...WGL_EXT_swap_control not found\n");
	}
#endif

	if (strstr (gl_config.extensions_string, "GL_EXT_point_parameters")) 
	{
		if (gl_ext_pointparameters->value) 
		{
			VID_Printf (PRINT_ALL, "...using GL_EXT_point_parameters\n");
		}
		else
		{
			VID_Printf (PRINT_ALL, "...ignoring GL_EXT_point_parameters\n");
		}
	}
	else
	{
		VID_Printf (PRINT_ALL, "...GL_EXT_point_parameters not found\n");
	}

	// don't bother with SGIS cos it's just so much old hat...
	if (strstr (gl_config.extensions_string, "GL_ARB_multitexture")) 
	{
		if (gl_ext_multitexture->value) 
		{
			VID_Printf (PRINT_ALL, "...using GL_ARB_multitexture\n");
		}
		else
		{
			VID_Printf (PRINT_ALL, "...ignoring GL_ARB_multitexture\n");
		}
	}
	else
	{
		VID_Printf (PRINT_ALL, "...GL_ARB_multitexture not found\n");
	}

	if (strstr (gl_config.extensions_string, "GL_EXT_texture_filter_anisotropic")) 
	{
		VID_Printf (PRINT_ALL, "...using GL_EXT_texture_filter_anisotropic\n");
		glGetIntegerv (GL_MAX_TEXTURE_MAX_ANISOTROPY, &gl_config.maxanisotropy);
		VID_Printf (PRINT_ALL, "...Maximum Anisotropy %d\n", gl_config.maxanisotropy);
	}
	else
	{
		VID_Printf (PRINT_ALL, "...GL_EXT_texture_filter_anisotropic not found\n");
		gl_config.maxanisotropy = 1;
	}

	if (strstr (gl_config.extensions_string, "GL_ARB_imaging"))
	{
		VID_Printf (PRINT_ALL, "...using GL_ARB_imaging\n");

		gl_config.GL_ARB_imaging = true;
	}
	else
	{
		gl_config.GL_ARB_imaging = false;
	}

	// now get all of our core procs
	// OpenGL 1.2 with Imaging Subset
	glBlendColor = (void *) MHGLGetProcAddress ("glBlendColor");
	glBlendEquation = (void *) MHGLGetProcAddress ("glBlendEquation");
	glDrawRangeElements = (void *) MHGLGetProcAddress ("glDrawRangeElements");
	glTexImage3D = (void *) MHGLGetProcAddress ("glTexImage3D");
	glTexSubImage3D = (void *) MHGLGetProcAddress ("glTexSubImage3D");
	glCopyTexSubImage3D = (void *) MHGLGetProcAddress ("glCopyTexSubImage3D");

	if (gl_config.GL_ARB_imaging)
	{
		// don't pollute the ATI fanboys' displays with a load of unavailable entrypoints
		glColorTable = (void *) MHGLGetProcAddress ("glColorTable");
		glColorTableParameterfv = (void *) MHGLGetProcAddress ("glColorTableParameterfv");
		glColorTableParameteriv = (void *) MHGLGetProcAddress ("glColorTableParameteriv");
		glCopyColorTable = (void *) MHGLGetProcAddress ("glCopyColorTable");
		glGetColorTable = (void *) MHGLGetProcAddress ("glGetColorTable");
		glGetColorTableParameterfv = (void *) MHGLGetProcAddress ("glGetColorTableParameterfv");
		glGetColorTableParameteriv = (void *) MHGLGetProcAddress ("glGetColorTableParameteriv");
		glColorSubTable = (void *) MHGLGetProcAddress ("glColorSubTable");
		glCopyColorSubTable = (void *) MHGLGetProcAddress ("glCopyColorSubTable");
		glConvolutionFilter1D = (void *) MHGLGetProcAddress ("glConvolutionFilter1D");
		glConvolutionFilter2D = (void *) MHGLGetProcAddress ("glConvolutionFilter2D");
		glConvolutionParameterf = (void *) MHGLGetProcAddress ("glConvolutionParameterf");
		glConvolutionParameterfv = (void *) MHGLGetProcAddress ("glConvolutionParameterfv");
		glConvolutionParameteri = (void *) MHGLGetProcAddress ("glConvolutionParameteri");
		glConvolutionParameteriv = (void *) MHGLGetProcAddress ("glConvolutionParameteriv");
		glCopyConvolutionFilter1D = (void *) MHGLGetProcAddress ("glCopyConvolutionFilter1D");
		glCopyConvolutionFilter2D = (void *) MHGLGetProcAddress ("glCopyConvolutionFilter2D");
		glGetConvolutionFilter = (void *) MHGLGetProcAddress ("glGetConvolutionFilter");
		glGetConvolutionParameterfv = (void *) MHGLGetProcAddress ("glGetConvolutionParameterfv");
		glGetConvolutionParameteriv = (void *) MHGLGetProcAddress ("glGetConvolutionParameteriv");
		glGetSeparableFilter = (void *) MHGLGetProcAddress ("glGetSeparableFilter");
		glSeparableFilter2D = (void *) MHGLGetProcAddress ("glSeparableFilter2D");
		glGetHistogram = (void *) MHGLGetProcAddress ("glGetHistogram");
		glGetHistogramParameterfv = (void *) MHGLGetProcAddress ("glGetHistogramParameterfv");
		glGetHistogramParameteriv = (void *) MHGLGetProcAddress ("glGetHistogramParameteriv");
		glGetMinmax = (void *) MHGLGetProcAddress ("glGetMinmax");
		glGetMinmaxParameterfv = (void *) MHGLGetProcAddress ("glGetMinmaxParameterfv");
		glGetMinmaxParameteriv = (void *) MHGLGetProcAddress ("glGetMinmaxParameteriv");
		glHistogram = (void *) MHGLGetProcAddress ("glHistogram");
		glMinmax = (void *) MHGLGetProcAddress ("glMinmax");
		glResetHistogram = (void *) MHGLGetProcAddress ("glResetHistogram");
		glResetMinmax = (void *) MHGLGetProcAddress ("glResetMinmax");
	}
	else
	{
		glColorTable = NULL;
		glColorTableParameterfv = NULL;
		glColorTableParameteriv = NULL;
		glCopyColorTable = NULL;
		glGetColorTable = NULL;
		glGetColorTableParameterfv = NULL;
		glGetColorTableParameteriv = NULL;
		glColorSubTable = NULL;
		glCopyColorSubTable = NULL;
		glConvolutionFilter1D = NULL;
		glConvolutionFilter2D = NULL;
		glConvolutionParameterf = NULL;
		glConvolutionParameterfv = NULL;
		glConvolutionParameteri = NULL;
		glConvolutionParameteriv = NULL;
		glCopyConvolutionFilter1D = NULL;
		glCopyConvolutionFilter2D = NULL;
		glGetConvolutionFilter = NULL;
		glGetConvolutionParameterfv = NULL;
		glGetConvolutionParameteriv = NULL;
		glGetSeparableFilter = NULL;
		glSeparableFilter2D = NULL;
		glGetHistogram = NULL;
		glGetHistogramParameterfv = NULL;
		glGetHistogramParameteriv = NULL;
		glGetMinmax = NULL;
		glGetMinmaxParameterfv = NULL;
		glGetMinmaxParameteriv = NULL;
		glHistogram = NULL;
		glMinmax = NULL;
		glResetHistogram = NULL;
		glResetMinmax = NULL;
	}

	// OpenGL 1.3
	glActiveTexture = (void *) MHGLGetProcAddress ("glActiveTexture");
	glClientActiveTexture = (void *) MHGLGetProcAddress ("glClientActiveTexture");
	glMultiTexCoord1d = (void *) MHGLGetProcAddress ("glMultiTexCoord1d");
	glMultiTexCoord1dv = (void *) MHGLGetProcAddress ("glMultiTexCoord1dv");
	glMultiTexCoord1f = (void *) MHGLGetProcAddress ("glMultiTexCoord1f");
	glMultiTexCoord1fv = (void *) MHGLGetProcAddress ("glMultiTexCoord1fv");
	glMultiTexCoord1i = (void *) MHGLGetProcAddress ("glMultiTexCoord1i");
	glMultiTexCoord1iv = (void *) MHGLGetProcAddress ("glMultiTexCoord1iv");
	glMultiTexCoord1s = (void *) MHGLGetProcAddress ("glMultiTexCoord1s");
	glMultiTexCoord1sv = (void *) MHGLGetProcAddress ("glMultiTexCoord1sv");
	glMultiTexCoord2d = (void *) MHGLGetProcAddress ("glMultiTexCoord2d");
	glMultiTexCoord2dv = (void *) MHGLGetProcAddress ("glMultiTexCoord2dv");
	glMultiTexCoord2f = (void *) MHGLGetProcAddress ("glMultiTexCoord2f");
	glMultiTexCoord2fv = (void *) MHGLGetProcAddress ("glMultiTexCoord2fv");
	glMultiTexCoord2i = (void *) MHGLGetProcAddress ("glMultiTexCoord2i");
	glMultiTexCoord2iv = (void *) MHGLGetProcAddress ("glMultiTexCoord2iv");
	glMultiTexCoord2s = (void *) MHGLGetProcAddress ("glMultiTexCoord2s");
	glMultiTexCoord2sv = (void *) MHGLGetProcAddress ("glMultiTexCoord2sv");
	glMultiTexCoord3d = (void *) MHGLGetProcAddress ("glMultiTexCoord3d");
	glMultiTexCoord3dv = (void *) MHGLGetProcAddress ("glMultiTexCoord3dv");
	glMultiTexCoord3f = (void *) MHGLGetProcAddress ("glMultiTexCoord3f");
	glMultiTexCoord3fv = (void *) MHGLGetProcAddress ("glMultiTexCoord3fv");
	glMultiTexCoord3i = (void *) MHGLGetProcAddress ("glMultiTexCoord3i");
	glMultiTexCoord3iv = (void *) MHGLGetProcAddress ("glMultiTexCoord3iv");
	glMultiTexCoord3s = (void *) MHGLGetProcAddress ("glMultiTexCoord3s");
	glMultiTexCoord3sv = (void *) MHGLGetProcAddress ("glMultiTexCoord3sv");
	glMultiTexCoord4d = (void *) MHGLGetProcAddress ("glMultiTexCoord4d");
	glMultiTexCoord4dv = (void *) MHGLGetProcAddress ("glMultiTexCoord4dv");
	glMultiTexCoord4f = (void *) MHGLGetProcAddress ("glMultiTexCoord4f");
	glMultiTexCoord4fv = (void *) MHGLGetProcAddress ("glMultiTexCoord4fv");
	glMultiTexCoord4i = (void *) MHGLGetProcAddress ("glMultiTexCoord4i");
	glMultiTexCoord4iv = (void *) MHGLGetProcAddress ("glMultiTexCoord4iv");
	glMultiTexCoord4s = (void *) MHGLGetProcAddress ("glMultiTexCoord4s");
	glMultiTexCoord4sv = (void *) MHGLGetProcAddress ("glMultiTexCoord4sv");
	glLoadTransposeMatrixf = (void *) MHGLGetProcAddress ("glLoadTransposeMatrixf");
	glLoadTransposeMatrixd = (void *) MHGLGetProcAddress ("glLoadTransposeMatrixd");
	glMultTransposeMatrixf = (void *) MHGLGetProcAddress ("glMultTransposeMatrixf");
	glMultTransposeMatrixd = (void *) MHGLGetProcAddress ("glMultTransposeMatrixd");
	glSampleCoverage = (void *) MHGLGetProcAddress ("glSampleCoverage");
	glCompressedTexImage3D = (void *) MHGLGetProcAddress ("glCompressedTexImage3D");
	glCompressedTexImage2D = (void *) MHGLGetProcAddress ("glCompressedTexImage2D");
	glCompressedTexImage1D = (void *) MHGLGetProcAddress ("glCompressedTexImage1D");
	glCompressedTexSubImage3D = (void *) MHGLGetProcAddress ("glCompressedTexSubImage3D");
	glCompressedTexSubImage2D = (void *) MHGLGetProcAddress ("glCompressedTexSubImage2D");
	glCompressedTexSubImage1D = (void *) MHGLGetProcAddress ("glCompressedTexSubImage1D");
	glGetCompressedTexImage = (void *) MHGLGetProcAddress ("glGetCompressedTexImage");

	// OpenGL 1.4
	glBlendFuncSeparate = (void *) MHGLGetProcAddress ("glBlendFuncSeparate");
	glFogCoordf = (void *) MHGLGetProcAddress ("glFogCoordf");
	glFogCoordfv = (void *) MHGLGetProcAddress ("glFogCoordfv");
	glFogCoordd = (void *) MHGLGetProcAddress ("glFogCoordd");
	glFogCoorddv = (void *) MHGLGetProcAddress ("glFogCoorddv");
	glFogCoordPointer = (void *) MHGLGetProcAddress ("glFogCoordPointer");
	glMultiDrawArrays = (void *) MHGLGetProcAddress ("glMultiDrawArrays");
	glMultiDrawElements = (void *) MHGLGetProcAddress ("glMultiDrawElements");
	glPointParameterf = (void *) MHGLGetProcAddress ("glPointParameterf");
	glPointParameterfv = (void *) MHGLGetProcAddress ("glPointParameterfv");
	glPointParameteri = (void *) MHGLGetProcAddress ("glPointParameteri");	// this is the entry point ATI missed...
	glPointParameteriv = (void *) MHGLGetProcAddress ("glPointParameteriv");
	glSecondaryColor3b = (void *) MHGLGetProcAddress ("glSecondaryColor3b");
	glSecondaryColor3bv = (void *) MHGLGetProcAddress ("glSecondaryColor3bv");
	glSecondaryColor3d = (void *) MHGLGetProcAddress ("glSecondaryColor3d");
	glSecondaryColor3dv = (void *) MHGLGetProcAddress ("glSecondaryColor3dv");
	glSecondaryColor3f = (void *) MHGLGetProcAddress ("glSecondaryColor3f");
	glSecondaryColor3fv = (void *) MHGLGetProcAddress ("glSecondaryColor3fv");
	glSecondaryColor3i = (void *) MHGLGetProcAddress ("glSecondaryColor3i");
	glSecondaryColor3iv = (void *) MHGLGetProcAddress ("glSecondaryColor3iv");
	glSecondaryColor3s = (void *) MHGLGetProcAddress ("glSecondaryColor3s");
	glSecondaryColor3sv = (void *) MHGLGetProcAddress ("glSecondaryColor3sv");
	glSecondaryColor3ub = (void *) MHGLGetProcAddress ("glSecondaryColor3ub");
	glSecondaryColor3ubv = (void *) MHGLGetProcAddress ("glSecondaryColor3ubv");
	glSecondaryColor3ui = (void *) MHGLGetProcAddress ("glSecondaryColor3ui");
	glSecondaryColor3uiv = (void *) MHGLGetProcAddress ("glSecondaryColor3uiv");
	glSecondaryColor3us = (void *) MHGLGetProcAddress ("glSecondaryColor3us");
	glSecondaryColor3usv = (void *) MHGLGetProcAddress ("glSecondaryColor3usv");
	glSecondaryColorPointer = (void *) MHGLGetProcAddress ("glSecondaryColorPointer");
	glWindowPos2d = (void *) MHGLGetProcAddress ("glWindowPos2d");
	glWindowPos2dv = (void *) MHGLGetProcAddress ("glWindowPos2dv");
	glWindowPos2f = (void *) MHGLGetProcAddress ("glWindowPos2f");
	glWindowPos2fv = (void *) MHGLGetProcAddress ("glWindowPos2fv");
	glWindowPos2i = (void *) MHGLGetProcAddress ("glWindowPos2i");
	glWindowPos2iv = (void *) MHGLGetProcAddress ("glWindowPos2iv");
	glWindowPos2s = (void *) MHGLGetProcAddress ("glWindowPos2s");
	glWindowPos2sv = (void *) MHGLGetProcAddress ("glWindowPos2sv");
	glWindowPos3d = (void *) MHGLGetProcAddress ("glWindowPos3d");
	glWindowPos3dv = (void *) MHGLGetProcAddress ("glWindowPos3dv");
	glWindowPos3f = (void *) MHGLGetProcAddress ("glWindowPos3f");
	glWindowPos3fv = (void *) MHGLGetProcAddress ("glWindowPos3fv");
	glWindowPos3i = (void *) MHGLGetProcAddress ("glWindowPos3i");
	glWindowPos3iv = (void *) MHGLGetProcAddress ("glWindowPos3iv");
	glWindowPos3s = (void *) MHGLGetProcAddress ("glWindowPos3s");
	glWindowPos3sv = (void *) MHGLGetProcAddress ("glWindowPos3sv");

	// OpenGL 1.5
	glGenQueries = (void *) MHGLGetProcAddress ("glGenQueries");
	glDeleteQueries = (void *) MHGLGetProcAddress ("glDeleteQueries");
	glIsQuery = (void *) MHGLGetProcAddress ("glIsQuery");
	glEndQuery = (void *) MHGLGetProcAddress ("glEndQuery");
	glGetQueryiv = (void *) MHGLGetProcAddress ("glGetQueryiv");
	glGetQueryObjectiv = (void *) MHGLGetProcAddress ("glGetQueryObjectiv");
	glGetQueryObjectuiv = (void *) MHGLGetProcAddress ("glGetQueryObjectuiv");
	glBindBuffer = (void *) MHGLGetProcAddress ("glBindBuffer");
	glDeleteBuffers = (void *) MHGLGetProcAddress ("glDeleteBuffers");
	glGenBuffers = (void *) MHGLGetProcAddress ("glGenBuffers");
	glIsBuffer = (void *) MHGLGetProcAddress ("glIsBuffer");
	glBufferData = (void *) MHGLGetProcAddress ("glBufferData");
	glBufferSubData = (void *) MHGLGetProcAddress ("glBufferSubData");
	glGetBufferSubData = (void *) MHGLGetProcAddress ("glGetBufferSubData");
	glMapBuffer = (void *) MHGLGetProcAddress ("glMapBuffer");
	glUnmapBuffer = (void *) MHGLGetProcAddress ("glUnmapBuffer");
	glGetBufferParameteriv = (void *) MHGLGetProcAddress ("glGetBufferParameteriv");
	glGetBufferPointerv = (void *) MHGLGetProcAddress ("glGetBufferPointerv");

	// max texture size
	glGetIntegerv (GL_MAX_TEXTURE_SIZE, &gl_config.maxtexsize);

	// clamp it at the next power of 2 equal to or below the lowest of width or height
	{
		int targ = (vid.width > vid.height) ? vid.height : vid.width;
		for (j = 1; j < targ; j *= 2);

		if (j > targ) j /= 2;

		if (gl_config.maxtexsize > j) gl_config.maxtexsize = j;

		VID_Printf (PRINT_ALL, "\n...Maximum Texture Size for Quake II %d x %d\n", gl_config.maxtexsize, gl_config.maxtexsize);
	}

	GL_SetDefaultState ();

	/*
	** draw our stereo patterns
	*/ 
#if 0// commented out until H3D pays us the money they owe us
	GL_DrawStereoPattern ();
#endif

	GL_InitImages ();
	Mod_Init ();
	R_InitParticleTexture ();
	Draw_InitLocal ();

	err = glGetError ();
	if (err != GL_NO_ERROR) 
		VID_Printf (PRINT_DEVELOPER, "glGetError () = 0x%x\n", err);

	return 0;
}

/*
=============== 
R_Shutdown
=============== 
*/ 
void R_Shutdown (void) 
{	
	Cmd_RemoveCommand ("modellist");
	Cmd_RemoveCommand ("screenshot");
	Cmd_RemoveCommand ("imagelist");
	Cmd_RemoveCommand ("gl_strings");

	Mod_FreeAll ();

	GL_ShutdownImages ();

	/*
	** shut down OS specific OpenGL stuff like contexts, etc.
	*/ 
	GLimp_Shutdown ();

	/*
	** shutdown our QGL subsystem
	*/ 
	QGL_Shutdown ();
}



/*
@@@@@@@@@@@@@@@@@@@@@
R_BeginFrame
@@@@@@@@@@@@@@@@@@@@@
*/ 
void R_BeginFrame (float camera_separation) 
{

	gl_state.camera_separation = camera_separation;

	/*
	** change modes if necessary
	*/ 
	if (gl_mode->modified || vid_fullscreen->modified) 
	{
		// FIXME: only restart if CDS is required
		cvar_t	*ref;

		ref = Cvar_Get ("vid_ref", "gl", 0);
		ref->modified = true;
	}

	/*
	** update 3Dfx gamma -- it is expected that a user will do a vid_restart
	** after tweaking this value
	*/ 
	// mh - is this a crap idea?
	if (vid_gamma->modified) 
		vid_gamma->modified = false;

	if (lm_gamma->modified)
		lm_gamma->modified = false;

	GLimp_BeginFrame (camera_separation);

	/*
	** go into 2D mode
	*/ 
	glViewport (0,0, vid.width, vid.height);
	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();
	//glOrtho (0, vid.width, vid.height, 0, - 99999, 99999);
	R_Set2DScreenCoords (&vid);
	glOrtho (0, vid.width2d, vid.height2d, 0, - 99999, 99999);
	glMatrixMode (GL_MODELVIEW);
	glLoadIdentity ();
	glDisable (GL_DEPTH_TEST);
	glDisable (GL_CULL_FACE);
	glDisable (GL_BLEND);
	glEnable (GL_ALPHA_TEST);
	glColor4f (1,1,1,1);

	/*
	** draw buffer stuff
	*/ 
	if (gl_drawbuffer->modified) 
	{
		gl_drawbuffer->modified = false;

		if (gl_state.camera_separation == 0 || !gl_state.stereo_enabled) 
		{
			if (Q_stricmp (gl_drawbuffer->string, "GL_FRONT") == 0) 
				mhglDrawBuffer (GL_FRONT);
			else
				mhglDrawBuffer (GL_BACK);
		}
	}

	/*
	** texturemode stuff
	*/ 
	if (gl_texturemode->modified) 
	{
		GL_TextureMode (gl_texturemode->string);
		gl_texturemode->modified = false;
	}

	if (gl_texturealphamode->modified) 
	{
		GL_TextureAlphaMode (gl_texturealphamode->string);
		gl_texturealphamode->modified = false;
	}

	if (gl_texturesolidmode->modified) 
	{
		GL_TextureSolidMode (gl_texturesolidmode->string);
		gl_texturesolidmode->modified = false;
	}

	if (gl_anisotropic_filter->modified) 
	{
		GL_AnisotropicFilter ();
		gl_anisotropic_filter->modified = false;
	}

	/*
	** swapinterval stuff
	*/ 
	GL_UpdateSwapInterval ();

	if (inScreenDumpMode)
	{
		dumptime += r_newrefdef.frametime;

		// dump screenshots at 8fps to match windows movie maker minimum
		// note - depending on your pc and video card this may or may not be fast or slow
		// i recommend running at 640 x 480 resolution if you're going to use it...!
		if (dumptime >= 0.125)
		{
			// take a screenshot
			GL_ScreenShot_f ();
			dumptime = 0;
		}
	}

	// clear screen if desired
	R_Clear ();
}

/*
============= 
R_SetPalette
============= 
*/ 
unsigned r_rawpalette[256];

void R_SetPalette (const unsigned char *palette) 
{
	int		i;

	byte *rp = (byte *) r_rawpalette;

	if (palette) 
	{
		for (i = 0; i < 256; i++) 
		{
			rp[i*4 + 0] = palette[i*3 + 0];
			rp[i*4 + 1] = palette[i*3 + 1];
			rp[i*4 + 2] = palette[i*3 + 2];
			rp[i*4 + 3] = 0xff;
		}
	}
	else
	{
		for (i = 0; i < 256; i++) 
		{
			rp[i*4 + 0] = d_8to24table[i]& 0xff;
			rp[i*4 + 1] = (d_8to24table[i]>> 8) & 0xff;
			rp[i*4 + 2] = (d_8to24table[i]>> 16) & 0xff;
			rp[i*4 + 3] = 0xff;
		}
	}
	GL_SetTexturePalette (r_rawpalette);

	glClearColor (0,0,0,0);
	glClear (GL_COLOR_BUFFER_BIT);
	glClearColor (0, 0, 0, 1);
}

/*
** R_DrawBeam
*/ 
void R_DrawBeam (entity_t *e) 
{
#define NUM_BEAM_SEGS 6

	int	i;
	float r, g, b;

	vec3_t perpvec;
	vec3_t direction, normalized_direction;
	vec3_t	start_points[NUM_BEAM_SEGS], end_points[NUM_BEAM_SEGS];
	vec3_t oldorigin, origin;

	oldorigin[0] = e->oldorigin[0];
	oldorigin[1] = e->oldorigin[1];
	oldorigin[2] = e->oldorigin[2];

	origin[0] = e->origin[0];
	origin[1] = e->origin[1];
	origin[2] = e->origin[2];

	normalized_direction[0] = direction[0] = oldorigin[0]- origin[0];
	normalized_direction[1] = direction[1] = oldorigin[1]- origin[1];
	normalized_direction[2] = direction[2] = oldorigin[2]- origin[2];

	if (VectorNormalize (normalized_direction) == 0) 
		return;

	PerpendicularVector (perpvec, normalized_direction);
	VectorScale (perpvec, e->frame / 2, perpvec);

	for (i = 0; i < 6; i++) 
	{
		RotatePointAroundVector (start_points[i], normalized_direction, perpvec, (360.0 / NUM_BEAM_SEGS) *i);
		VectorAdd (start_points[i], origin, start_points[i]);
		VectorAdd (start_points[i], direction, end_points[i]);
	}

	glDisable (GL_TEXTURE_2D);
	glEnable (GL_BLEND);
	glDepthMask (GL_FALSE);

	r = (d_8to24table[e->skinnum & 0xFF]) & 0xFF;
	g = (d_8to24table[e->skinnum & 0xFF]>> 8) & 0xFF;
	b = (d_8to24table[e->skinnum & 0xFF]>> 16) & 0xFF;

	r *= 1 / 255.0F;
	g *= 1 / 255.0F;
	b *= 1 / 255.0F;

	glColor4f (r, g, b, e->alpha);

	glBegin (GL_TRIANGLE_STRIP);
	for (i = 0; i < NUM_BEAM_SEGS; i++) 
	{
		glVertex3fv (start_points[i]);
		glVertex3fv (end_points[i]);
		glVertex3fv (start_points[(i + 1) %NUM_BEAM_SEGS]);
		glVertex3fv (end_points[(i + 1) %NUM_BEAM_SEGS]);
	}
	glEnd ();

	glEnable (GL_TEXTURE_2D);
	glDisable (GL_BLEND);
	glDepthMask (GL_TRUE);
}

// =================================================================== 


void	R_BeginRegistration (char *map);
struct model_s	*R_RegisterModel (char *name);
struct image_s	*R_RegisterSkin (char *name);
void R_SetSky (char *name, float rotate, vec3_t axis);
void	R_EndRegistration (void);

void	R_RenderFrame (refdef_t *fd);

struct image_s	*Draw_FindPic (char *name);

void	Draw_Pic (int x, int y, char *name);
void	Draw_Char (int x, int y, int c);
void	Draw_TileClear (int x, int y, int w, int h, char *name);
void	Draw_Fill (int x, int y, int w, int h, int c);
void	Draw_FadeScreen (void);

/*
@@@@@@@@@@@@@@@@@@@@@
GetRefAPI

@@@@@@@@@@@@@@@@@@@@@
*/ 
void R_InitSwap (void) 
{
	Swap_Init ();
}

