/*
Copyright (C) 2007, MH

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


#include "quakedef.h"


/*
=================================================================================================================================

		GL_SHADER.C

	Q3A style shader system.  Currently these shaders are implemented directly within the engine, but there's no reason why
	they can't be extended to support external shader files also.  I won't be doing it, but anyone else should feel free.

	All rendering should ultimately go through here.

=================================================================================================================================
*/


/*
================
VIEWPORT SUPPORT
================
*/

typedef struct r_viewport_s
{
	// location of viewport
	int x;
	int y;

	// size of viewport
	int width;
	int height;

	// modelview and projection matrices for direct loading up
	GLfloat modelview_matrix[16];
	GLfloat projection_matrix[16];

	// display list number of default OpenGL State for the viewport
	int GL_State;

	// next viewport in list
	struct r_viewport_s *r_next_viewport;
} r_viewport_t;


// pointer to the previous viewport in use
r_viewport_t *r_previous_viewport = NULL;

// pointer to the current viewport in use
r_viewport_t *r_current_viewport = NULL;

// linked list of already available viewports
r_viewport_t *r_viewport_list = NULL;


void R_SetViewPortObject (r_viewport_t *vp)
{
	// only create a viewport if we need to
	if (vp->x != r_current_viewport->x || vp->y != r_current_viewport->y ||
		vp->width != r_current_viewport->width || vp->height != r_current_viewport->height)
	{
		glViewport (vp->x, vp->y, vp->width, vp->height);
	}
}


void R_SetViewPort (int x, int y, int width, int height, qboolean vp3D)
{
	r_viewport_t *vp;

	// see does it match the current viewport
	if (1)
	{
		// no need to change
		return;
	}

	for (vp = r_viewport_list; vp; vp = vp->r_next_viewport)
	{
		// see does it match the new settings
		if (vp->width != width) continue;

		// match, so load up this viewport and get out
		R_SetViewPortObject (vp);
		return;
	}

	// create a new viewport
	vp = (r_viewport_t *) malloc (sizeof (r_viewport_t));

	// link it in
	vp->r_next_viewport = r_viewport_list;

	// set it's properties
	vp->x = x;
	vp->y = y;
	vp->width = width;
	vp->height = height;

	// load it up
	R_SetViewPortObject (vp);
}

