/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// gl_warp.c -- sky and water polygons

#include "quakedef.h"

extern	model_t	*loadmodel;

int		skytexturenum;

int		solidskytexture;
int		alphaskytexture;
float	speedscale;		// for top sky and bottom sky

msurface_t	*warpface;

float GL_SUBDIVIDE_SIZE;

typedef struct warpvert_s
{
	float v[3];
	float st[2];
} warpvert_t;

warpvert_t *warpverts = NULL;

int maxwarpverts;

void BoundPoly (int numverts, float *verts, vec3_t mins, vec3_t maxs)
{
	int		i, j;
	float	*v;

	mins[0] = mins[1] = mins[2] = 9999;
	maxs[0] = maxs[1] = maxs[2] = -9999;
	v = verts;
	for (i=0 ; i<numverts ; i++)
	{
		for (j=0 ; j<3 ; j++, v++)
		{
			if (*v < mins[j])
				mins[j] = *v;
			if (*v > maxs[j])
				maxs[j] = *v;
		}
	}
}


void SubdividePolygon (int numverts, float *verts)
{
	int		i, j, k;
	vec3_t	mins, maxs;
	float	m;
	float	*v;
	vec3_t	front[64], back[64];
	int		f, b;
	float	dist[64];
	float	frac;
	glpoly_t	*poly;
	float	s, t;

	if (numverts > 60)
		Sys_Error ("numverts = %i", numverts);

	BoundPoly (numverts, verts, mins, maxs);

	for (i=0 ; i<3 ; i++)
	{
		m = (mins[i] + maxs[i]) * 0.5;
		m = GL_SUBDIVIDE_SIZE * floor (m/GL_SUBDIVIDE_SIZE + 0.5);
		if (maxs[i] - m < 8)
			continue;
		if (m - mins[i] < 8)
			continue;

		// cut it
		v = verts + i;
		for (j=0 ; j<numverts ; j++, v+= 3)
			dist[j] = *v - m;

		// wrap cases
		dist[j] = dist[0];
		v-=i;
		VectorCopy (verts, v);

		f = b = 0;
		v = verts;
		for (j=0 ; j<numverts ; j++, v+= 3)
		{
			if (dist[j] >= 0)
			{
				VectorCopy (v, front[f]);
				f++;
			}

			if (dist[j] <= 0)
			{
				VectorCopy (v, back[b]);
				b++;
			}

			if (dist[j] == 0 || dist[j+1] == 0)
				continue;

			if ((dist[j] > 0) != (dist[j+1] > 0))
			{
				// clip point
				frac = dist[j] / (dist[j] - dist[j+1]);
				for (k=0 ; k<3 ; k++)
					front[f][k] = back[b][k] = v[k] + frac*(v[3+k] - v[k]);
				f++;
				b++;
			}
		}

		SubdividePolygon (f, front[0]);
		SubdividePolygon (b, back[0]);
		return;
	}

	poly = Hunk_Alloc (sizeof (glpoly_t));
	poly->verts = Hunk_Alloc (sizeof (glvertex_t) * (numverts + 2));
	poly->next = warpface->polys;
	warpface->polys = poly;

	poly->numverts = numverts + 2;

	poly->midpoint[0] = poly->midpoint[1] = poly->midpoint[2] = 0;

	for (i = 0 ; i < numverts ; i++, verts += 3)
	{
		VectorCopy (verts, poly->verts[i + 1].v);

		s = DotProduct (verts, warpface->texinfo->vecs[0]);
		t = DotProduct (verts, warpface->texinfo->vecs[1]);
		poly->verts[i + 1].st[0] = s;
		poly->verts[i + 1].st[1] = t;

		poly->midpoint[0] += poly->verts[i + 1].v[0];
		poly->midpoint[1] += poly->verts[i + 1].v[1];
		poly->midpoint[2] += poly->verts[i + 1].v[2];
	}

	poly->midpoint[0] /= (float) numverts;
	poly->midpoint[1] /= (float) numverts;
	poly->midpoint[2] /= (float) numverts;

	// copy first to last
	memcpy (&poly->verts[poly->numverts - 1], &poly->verts[1], sizeof (glvertex_t));

	// fill in the new first
	VectorCopy (poly->midpoint, poly->verts[0].v);

	s = DotProduct (poly->midpoint, warpface->texinfo->vecs[0]);
	t = DotProduct (poly->midpoint, warpface->texinfo->vecs[1]);
	poly->verts[0].st[0] = s;
	poly->verts[0].st[1] = t;

	poly->fxofs = (rand () % 64) - 64;

	// allocate vertexes
	if (!warpverts)
	{
		// take at least 9 to begin with and grow as required
		maxwarpverts = poly->numverts > 9 ? poly->numverts : 9;
		warpverts = (warpvert_t *) malloc (sizeof (warpvert_t) * maxwarpverts);
	}
	else if (poly->numverts > maxwarpverts)
	{
		// go to the new max
		maxwarpverts = poly->numverts;

		// reallocate
		free (warpverts);
		warpverts = (warpvert_t *) malloc (sizeof (warpvert_t) * maxwarpverts);
	}

	// copy v to tv
	for (i = 0; i < poly->numverts; i++)
	{
		poly->verts[i].tv[0] = poly->verts[i].v[0];
		poly->verts[i].tv[1] = poly->verts[i].v[1];
		poly->verts[i].tv[2] = poly->verts[i].v[2];
	}
}

/*
================
GL_SubdivideSurface

Breaks a polygon up along axial 64 unit
boundaries so that turbulent and sky warps
can be done reasonably.
================
*/
void GL_DontSubdivideSky (msurface_t *surf)
{
	int i;
	int lindex;
	float *vec;
	glpoly_t *poly;

	// just build a normal surf, but without s and t and all of the other malarkey
	// dramatic speedups on skybox polygon clipping will result!
	poly = Hunk_Alloc (sizeof (glpoly_t));
	poly->verts = Hunk_Alloc (sizeof (glvertex_t) * surf->numedges);
	poly->next = NULL;
	poly->flags = surf->flags;
	surf->polys = poly;
	poly->numverts = surf->numedges;

	for (i = 0; i < surf->numedges; i++)
	{
		lindex = loadmodel->surfedges[surf->firstedge + i];

		if (lindex > 0)
			vec = loadmodel->vertexes[loadmodel->edges[lindex].v[0]].position;
		else
			vec = loadmodel->vertexes[loadmodel->edges[-lindex].v[1]].position;

		VectorCopy (vec, poly->verts[i].v);
	}
}


void GL_SubdivideSurface (msurface_t *surf)
{
	vec3_t		verts[64];
	int			numverts;
	int			i;
	int			lindex;
	float		*vec;
	char		*name = surf->texinfo->texture->name;

	// change the subdivision algorithmaccording to the type of surface
	if (!strncmp (name, "sky", 3))
	{
		GL_DontSubdivideSky (surf);
		return;
		//GL_SUBDIVIDE_SIZE = 256.0f;
	}
	else if (!strncmp (&name[1], "lava", 4))
		GL_SUBDIVIDE_SIZE = 128.0f;
	else GL_SUBDIVIDE_SIZE = 64.0f;

	warpface = surf;

	// convert edges back to a normal polygon
	numverts = 0;

	for (i=0 ; i<surf->numedges ; i++)
	{
		lindex = loadmodel->surfedges[surf->firstedge + i];

		if (lindex > 0)
			vec = loadmodel->vertexes[loadmodel->edges[lindex].v[0]].position;
		else
			vec = loadmodel->vertexes[loadmodel->edges[-lindex].v[1]].position;
		VectorCopy (vec, verts[numverts]);
		numverts++;
	}

	SubdividePolygon (numverts, verts[0]);
}

//=========================================================



// speed up sin calculations - Ed
float	turbsin[] =
{
	#include "gl_warp_sin.h"
};
#define TURBSCALE (256.0 / (2 * M_PI))


/*
=============
EmitWaterPolys

Does a water warp on the pre-fragmented glpoly_t chain.  only the midpoint warps.
=============
*/
void EmitWaterPolys (msurface_t *surf, qboolean warpwave)
{
	glpoly_t	*p;
	glvertex_t	*v;
	int			i;
	int			vindex[2];

	// get the verts to use
	switch (surf->plane->type)
	{
	case PLANE_Y:
		// vertical in line with the world grid
		vindex[0] = 0;
		vindex[1] = 2;
		break;

	case PLANE_X:
		// vertical in line with the world grid
		vindex[0] = 2;
		vindex[1] = 1;
		break;

	case PLANE_ANYX:
	case PLANE_ANYY:
		// vertical at an angle to the world grid
		vindex[0] = 2;
		vindex[1] = 0;
		break;

	case PLANE_Z:
		// flat
	default:
		// same as flat
		vindex[0] = 0;
		vindex[1] = 1;
		break;
	}

	for (p = surf->polys; p; p = p->next)
	{
		float	warpscale = 4.0f;

		v = p->verts;

		warpverts[0].st[0] = (v[0].v[vindex[0]] + 4 * sin(v[0].v[vindex[1]] * 0.025 + realtime)) * 0.015625f; // v[0].st[0] * 0.015625f;
		warpverts[0].st[1] = (v[0].v[vindex[1]] + 4 * sin(v[0].v[vindex[0]] * 0.025 + realtime)) * 0.015625f; // v[0].st[1] * 0.015625f;

		if (warpwave)
		{
			// water/slime/lava have a wave effect added
			warpverts->v[0] = v->v[0] + warpscale * sin (v->v[1] * 0.05f + realtime) * sin (v->v[2] * 0.05f + realtime);
			warpverts->v[1] = v->v[1] + warpscale * sin (v->v[0] * 0.05f + realtime) * sin (v->v[2] * 0.05f + realtime);
			warpverts->v[2] = v->v[2] + warpscale * sin (v->v[0] * 0.05f + realtime) * sin (v->v[1] * 0.05f + realtime);
		}
		else
		{
			// teleports don't...
			warpverts->v[0] = v->v[0];
			warpverts->v[1] = v->v[1];
			warpverts->v[2] = v->v[2];
		}

		// now do the remainder normally
		for (i = 1; i < p->numverts; i++)
		{
			warpverts[i].st[0] = (v[i].v[vindex[0]] + 4 * sin(v[i].v[vindex[1]] * 0.025 + realtime)) * 0.015625f; // v[0].st[0] * 0.015625f;
			warpverts[i].st[1] = (v[i].v[vindex[1]] + 4 * sin(v[i].v[vindex[0]] * 0.025 + realtime)) * 0.015625f; // v[0].st[1] * 0.015625f;

			warpverts[i].v[0] = v[i].v[0];
			warpverts[i].v[1] = v[i].v[1];
			warpverts[i].v[2] = v[i].v[2];
		}

		glDrawArrays (GL_TRIANGLE_FAN, 0, p->numverts);
	}
}


void EmitWaterBloomPolys (msurface_t *surf)
{
	glpoly_t	*p;
	glvertex_t	*v;
	int			i;

	for (p = surf->polys; p; p = p->next)
	{
		v = p->verts;

		for (i = 0; i < p->numverts; i++)
		{
			warpverts[i].st[0] = v[i].st[0] * 0.015625f;
			warpverts[i].st[1] = v[i].st[1] * 0.015625f;

			warpverts[i].v[0] = v[i].v[0];
			warpverts[i].v[1] = v[i].v[1];
			warpverts[i].v[2] = v[i].v[2];
		}

		glDrawArrays (GL_TRIANGLE_FAN, 0, p->numverts);
	}
}


/*
================
R_DrawWaterSurfaces
================
*/
extern texture_t *texturelist;

void R_DrawWaterSurfaces (qboolean Bloom)
{
	msurface_t	*s;
	texture_t	*t;
	extern qboolean liquidpvs;
	qboolean warpwave;

	// go back to the world matrix
    glLoadMatrixf (r_world_matrix);

	// no liquid in the PVS
	if (!liquidpvs) return;

	// bring up vertex pointer
	glEnableClientState (GL_VERTEX_ARRAY);
	glVertexPointer (3, GL_FLOAT, sizeof (warpvert_t), warpverts[0].v);

	// bring up image texture array
	glEnableClientState (GL_TEXTURE_COORD_ARRAY);
	glTexCoordPointer (2, GL_FLOAT, sizeof (warpvert_t), warpverts[0].st);

	if (Bloom)
	{
		for (t = texturelist; t; t = t->texturelist)
		{
			if (!(s = t->texturechain)) continue;
			if (!((s->flags & SURF_DRAWLAVA) || (s->flags & SURF_DRAWSLIME))) continue;

			glBindTexture (GL_TEXTURE_2D, t->gltexture->texnum);

			for (; s; s = s->texturechain)
				EmitWaterBloomPolys (s);
		}
	}
	else
	{
		glEnable (GL_BLEND);
		glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

		for (t = texturelist; t; t = t->texturelist)
		{
			if (!(s = t->texturechain)) continue;
			if (!(s->flags & SURF_DRAWTURB)) continue;

			glBindTexture (GL_TEXTURE_2D, t->gltexture->texnum);

			warpwave = true;

			// the amount of translucency depends on the type of texture
			if (s->flags & SURF_DRAWLAVA)
				glColor4f (1, 1, 1, 1);
			else if (s->flags & SURF_DRAWTELE)
			{
				// don't vertex warp these
				warpwave = false;
				glColor4f (1, 1, 1, 1);
			}
			else if (s->flags & SURF_DRAWSLIME)
				glColor4f (1, 1, 1, 0.75);
			else glColor4f (1, 1, 1, 0.6);

			for (; s; s = s->texturechain)
				EmitWaterPolys (s, warpwave);
		}

		glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
		glColor4f (1, 1, 1, 1);
		glDisable (GL_BLEND);
	}

	glDisableClientState (GL_TEXTURE_COORD_ARRAY);
	glDisableClientState (GL_VERTEX_ARRAY);
}


//===============================================================

gltexture_t *skyboxtexes[6] = {NULL, NULL, NULL, NULL, NULL, NULL};


/*
=================
R_DrawSkyChain
=================
*/
float r_farclip = 4096;


vec3_t	skyclip[6] =
{
	{1, 1, 0},
	{1, -1, 0},
	{0, -1, 1},
	{0, 1, 1},
	{1, 0, 1},
	{-1, 0, 1} 
};

// 1 = s, 2 = t, 3 = 2048
int	st_to_vec[6][3] = 
{
	{3, -1, 2},
	{-3, 1, 2},

	{1, 3, 2},
	{-1, -3, 2},

	{-2, -1, 3},		// 0 degrees yaw, look straight up
	{2, -1, -3}		// look straight down
};

// s =[0]/[2], t =[1]/[2]
int	vec_to_st[6][3] = 
{
	{-2, 3, 1},
	{2, 3, -1},

	{1, 3, 2},
	{-1, 3, -2},

	{-2, -1, 3},
	{-2, 1, -3}
};

float	skymins[2][6], skymaxs[2][6];
float	sky_min, sky_max;

void DrawSkyPolygon (int nump, vec3_t vecs) 
{
	int		i,j;
	vec3_t	v, av;
	float	s, t, dv;
	int		axis;
	float	*vp;

	// decide which face it maps to
	VectorCopy (vec3_origin, v);

	for (i = 0, vp = vecs; i < nump; i++, vp += 3) 
		VectorAdd (vp, v, v);

	av[0] = fabs (v[0]);
	av[1] = fabs (v[1]);
	av[2] = fabs (v[2]);

	if (av[0] > av[1] && av[0] > av[2]) 
	{
		if (v[0] < 0) 
			axis = 1;
		else axis = 0;
	}
	else if (av[1] > av[2] && av[1] > av[0]) 
	{
		if (v[1] < 0) 
			axis = 3;
		else axis = 2;
	}
	else
	{
		if (v[2] < 0) 
			axis = 5;
		else axis = 4;
	}

	// project new texture coords
	for (i = 0; i < nump; i++, vecs += 3) 
	{
		j = vec_to_st[axis][2];

		if (j > 0) 
			dv = vecs[j - 1];
		else
			dv = - vecs[-j - 1];

		// don't divide by zero
		if (dv < 0.001) continue;

		j = vec_to_st[axis][0];

		if (j < 0) 
			s = -vecs[-j - 1] / dv;
		else s = vecs[j - 1] / dv;

		j = vec_to_st[axis][1];

		if (j < 0) 
			t = -vecs[-j - 1] / dv;
		else t = vecs[j - 1] / dv;

		if (s < skymins[0][axis]) skymins[0][axis] = s;
		if (t < skymins[1][axis]) skymins[1][axis] = t;
		if (s > skymaxs[0][axis]) skymaxs[0][axis] = s;
		if (t > skymaxs[1][axis]) skymaxs[1][axis] = t;
	}
}


#define	ON_EPSILON		0.1			// point on plane side epsilon
#define	MAX_CLIP_VERTS	64

void ClipSkyPolygon (int nump, vec3_t vecs, int stage) 
{
	float	*norm;
	float	*v;
	qboolean	front, back;
	float	d, e;
	float	dists[MAX_CLIP_VERTS];
	int		sides[MAX_CLIP_VERTS];
	vec3_t	newv[2][MAX_CLIP_VERTS];
	int		newc[2];
	int		i, j;

	if (nump > MAX_CLIP_VERTS - 2) Sys_Error ("ClipSkyPolygon: MAX_CLIP_VERTS");

	if (stage == 6) 
	{
		// fully clipped, so draw it
		DrawSkyPolygon (nump, vecs);
		return;
	}

	front = back = false;
	norm = skyclip[stage];

	for (i = 0, v = vecs; i < nump; i++, v += 3) 
	{
		d = DotProduct (v, norm);

		if (d > ON_EPSILON) 
		{
			front = true;
			sides[i] = SIDE_FRONT;
		}
		else if (d < -ON_EPSILON) 
		{
			back = true;
			sides[i] = SIDE_BACK;
		}
		else sides[i] = SIDE_ON;

		dists[i] = d;
	}

	if (!front || !back) 
	{
		// not clipped
		ClipSkyPolygon (nump, vecs, stage + 1);
		return;
	}

	// clip it
	sides[i] = sides[0];
	dists[i] = dists[0];

	VectorCopy (vecs, (vecs + (i * 3)));
	newc[0] = newc[1] = 0;

	for (i = 0, v = vecs; i < nump; i++, v += 3) 
	{
		switch (sides[i]) 
		{
		case SIDE_FRONT:
			VectorCopy (v, newv[0][newc[0]]);
			newc[0]++;
			break;

		case SIDE_BACK:
			VectorCopy (v, newv[1][newc[1]]);
			newc[1]++;
			break;

		case SIDE_ON:
			VectorCopy (v, newv[0][newc[0]]);
			newc[0]++;
			VectorCopy (v, newv[1][newc[1]]);
			newc[1]++;
			break;
		}

		if (sides[i] == SIDE_ON || sides[i + 1] == SIDE_ON || sides[i + 1] == sides[i]) 
			continue;

		d = dists[i] / (dists[i] - dists[i + 1]);

		for (j = 0; j < 3; j++)
		{
			e = v[j] + d * (v[j + 3] - v[j]);

			newv[0][newc[0]][j] = e;
			newv[1][newc[1]][j] = e;
		}

		newc[0]++;
		newc[1]++;
	}

	// continue
	ClipSkyPolygon (newc[0], newv[0][0], stage + 1);
	ClipSkyPolygon (newc[1], newv[1][0], stage + 1);
}


/*
 ================= 
R_AddSkySurface
 ================= 
*/
void R_AddSkySurface (msurface_t *surf) 
{
	int			i;
	vec3_t		verts[MAX_CLIP_VERTS];
	glpoly_t	*poly;

	// calculate vertex values for sky box
	for (poly = surf->polys; poly; poly = poly->next) 
	{
		for (i = 0; i < poly->numverts; i++)
		{
			verts[i][0] = poly->verts[i].v[0] - r_origin[0];
			verts[i][1] = poly->verts[i].v[1] - r_origin[1];
			verts[i][2] = poly->verts[i].v[2] - r_origin[2];
		}

		ClipSkyPolygon (poly->numverts, verts[0], 0);
	}
}


/*
 ============== 
R_ClearSkyBox
 ============== 
*/ 
void R_ClearSkyBox (void)
{
	int		i;

	for (i = 0; i < 6; i++) 
	{
		skymins[0][i] = skymins[1][i] = 99999999;
		skymaxs[0][i] = skymaxs[1][i] = -99999999;
	}
}


int	skytexorder[6] = {0, 2, 1, 3, 4, 5};

typedef struct sbverts_s
{
	float v[3];
	float st[2];
} sbverts_t;

sbverts_t sbverts[4];

void MakeSkyVec (float s, float t, int axis, int vnum) 
{
	vec3_t		b;
	int			j, k;
	float		w, h;

	b[0] = s * r_farclip * 0.57735f;
	b[1] = t * r_farclip * 0.57735f;
	b[2] = r_farclip * 0.57735f;

	for (j = 0; j < 3; j++)
	{
		k = st_to_vec[axis][j];

		if (k < 0)
			sbverts[vnum].v[j] = -b[-k - 1];
		else
			sbverts[vnum].v[j] = b[k - 1];

		sbverts[vnum].v[j] += r_origin[j];
	}

	// convert from range [-1,1] to [0,1]
	s = (s + 1.0f) * 0.5f;
	t = (t + 1.0f) * 0.5f;

	// avoid bilerp seam
	// fixme - this still isn't perfect.  it's there if you know what to look for.  ideally, the texture
	// should have it's first and last row and column repeated (fixed - or use clamp to edge)
	w = skyboxtexes[skytexorder[axis]]->width;
	h = skyboxtexes[skytexorder[axis]]->height;

	s = s * (w - 1) / w + 0.5f / w;
	t = t * (h - 1) / h + 0.5f / h;

	sbverts[vnum].st[0] = s;
	sbverts[vnum].st[1] = 1.0f - t;
}


/*
 ============== 
R_DrawSkyBox
 ============== 
*/ 
void R_DrawSkyBox (void)
{
	int		i;
	qboolean init = false;

	for (i = 0; i < 6; i++) 
	{
		if (skymins[0][i] >= skymaxs[0][i] || skymins[1][i] >= skymaxs[1][i]) 
			continue;

		if (!init)
		{
			// bring up vertex pointer
			glEnableClientState (GL_VERTEX_ARRAY);
			glVertexPointer (3, GL_FLOAT, sizeof (sbverts_t), sbverts[0].v);

			// bring up image texture array
			glEnableClientState (GL_TEXTURE_COORD_ARRAY);
			glTexCoordPointer (2, GL_FLOAT, sizeof (sbverts_t), sbverts[0].st);

			init = true;
		}

		glBindTexture (GL_TEXTURE_2D, skyboxtexes[skytexorder[i]]->texnum);

		MakeSkyVec (skymins[0][i], skymins[1][i], i, 0);
		MakeSkyVec (skymins[0][i], skymaxs[1][i], i, 1);
		MakeSkyVec (skymaxs[0][i], skymaxs[1][i], i, 2);
		MakeSkyVec (skymaxs[0][i], skymins[1][i], i, 3);

		glDrawArrays (GL_QUADS, 0, 4);
	}

	if (init)
	{
		glDisableClientState (GL_TEXTURE_COORD_ARRAY);
		glDisableClientState (GL_VERTEX_ARRAY);
	}
}


cvar_t r_skyspeed = {"r_skyspeed", "2.5", true};

float rotateBack = 0;
float rotateFore = 0;

float skytexes[440];
float skyverts[660];

void R_DrawSkyArrays (void)
{
	int i;
	int j;

	int numverts = 0;
	int start;

	for (i = 0; i < 10; i++)
	{
		start = numverts;

		for (j = 0; j <= 10; j++)
			numverts += 2;

		glDrawArrays (GL_TRIANGLE_STRIP, start, (numverts - start));
	}
}


void R_DrawSkyChain (qboolean skydraw)
{
	// sky scaling
	float fullscale;
	float halfscale;
	float reducedfull;
	float reducedhalf;
	int i;

	// do these calcs even if we're not drawing
	// sky rotation
	// bound rotation speed 0 to 100
	if (r_skyspeed.value < 0) r_skyspeed.value = 0;
	if (r_skyspeed.value > 100) r_skyspeed.value = 100;

	// always rotate even if we're paused!!!
	rotateBack = anglemod (realtime * (2.5 * r_skyspeed.value));
	rotateFore = anglemod (realtime * (4.0 * r_skyspeed.value));

	if (!skydraw) return;

	// check for a skybox - all 6must be OK, so break on the first failure
	for (i = 0; i < 6; i++)
	{
		// no skybox texture!!!
		// (this pointer is only valid after the first call to R_LoadSky, so it will be NULL if we haven't called that yet,
		// which is quite likely as at present it must be called manually)
		if (!skyboxtexes[i]) break;

		// no texture loaded
		if (!skyboxtexes[i]->cache) break;
	}

	if (i == 6)
	{
		// draw the skybox instead and get out.
		// note: the rotation check above should be done even if a skybox is drawn, so that it's up to date
		// if we subsequently switch the skybox off
		R_DrawSkyBox ();
		return;
	}

	// calculate the scales
	fullscale = r_farclip / 4096;
	halfscale = r_farclip / 8192;
	reducedfull = (r_farclip / 4096) * 0.9;
	reducedhalf = (r_farclip / 8192) * 0.9;

	// sky texture scaling
	// note that in terms of size this is *NOT* visually the same as the "classic" quake sky - drawing it on a 
	// proper sphere generates an "obviously a tiled texture" look if we shrink it more, and looks very wrong, 
	// so we go for somewhat larger instead.  increase to 8, 4, 4 if that really bothers you.
	glMatrixMode (GL_TEXTURE);
	glLoadIdentity ();
	glScalef (4, 2, 2);
	glMatrixMode (GL_MODELVIEW);

	glEnableClientState (GL_VERTEX_ARRAY);
	glVertexPointer (3, GL_FLOAT, 3 * sizeof (float), skyverts);

	glEnableClientState (GL_TEXTURE_COORD_ARRAY);
	glTexCoordPointer (2, GL_FLOAT, 2 * sizeof (float), skytexes);

	// background
	// ==========
	// go to a new matrix
	glPushMatrix ();

	// center it on the players position
	glTranslatef (r_origin[0], r_origin[1], r_origin[2]);

	// flatten the sphere
	glScalef (fullscale, fullscale, halfscale);

	// orient it so that the poles are unobtrusive
	glRotatef (-90, 1, 0, 0);

	// make it not always at right angles to the player
	glRotatef (-22, 0 ,1, 0);

	// rotate it around the poles
	glRotatef (-rotateBack, 0, 0, 1);

	// solid sky texture
	glBindTexture (GL_TEXTURE_2D, solidskytexture);

	// draw the sphere
	R_DrawSkyArrays ();

	// restore the previous matrix
	glPopMatrix ();

	// foreground
	// ==========
	// go to a new matrix
	glPushMatrix ();

	// center it on the players position
	glTranslatef (r_origin[0], r_origin[1], r_origin[2]);

	// flatten the sphere and shrink it a little - the reduced scale prevents artefacts appearing when
	// corners on the skysphere may potentially overlap
	glScalef (reducedfull, reducedfull, reducedhalf);

	// orient it so that the poles are unobtrusive
	glRotatef (-90, 1, 0, 0);

	// make it not always at right angles to the player
	glRotatef (-22, 0 ,1, 0);

	// rotate it around the poles
	glRotatef (-rotateFore, 0, 0, 1);

	// blend mode
	glEnable (GL_BLEND);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	// alpha sky texture
	glBindTexture (GL_TEXTURE_2D, alphaskytexture);

	// draw the sphere
	R_DrawSkyArrays ();

	glDisableClientState (GL_TEXTURE_COORD_ARRAY);
	glDisableClientState (GL_VERTEX_ARRAY);

	// back to normal mode
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable (GL_BLEND);

	// restore the previous matrix
	glPopMatrix ();

	glMatrixMode (GL_TEXTURE);
	glLoadIdentity ();
	glMatrixMode (GL_MODELVIEW);
}


void R_InitSkySphere (void)
{
	float drho = 0.3141592653589;
	float dtheta = 0.6283185307180;

	float ds = 0.1;
	float dt = 0.1;

	float t = 1.0f;	
	float s = 0.0f;

	int i;
	int j;

	int vertspos = 0;
	int texespos = 0;

	for (i = 0; i < 10; i++)
	{
		float rho = (float) i * drho;
		float srho = (float) (sin (rho));
		float crho = (float) (cos (rho));
		float srhodrho = (float) (sin (rho + drho));
		float crhodrho = (float) (cos (rho + drho));

		s = 0.0f;

		for (j = 0; j <= 10; j++)
		{
			float theta = (j == 10) ? 0.0f : j * dtheta;
			float stheta = (float) (-sin (theta));
			float ctheta = (float) (cos (theta));

			skytexes[texespos++] = s * 2;
			skytexes[texespos++] = t * 2;

			skyverts[vertspos++] = stheta * srho * 4096.0;
			skyverts[vertspos++] = ctheta * srho * 4096.0;
			skyverts[vertspos++] = crho * 4096.0;

			skytexes[texespos++] = s * 2;
			skytexes[texespos++] = (t - dt) * 2;

			skyverts[vertspos++] = stheta * srhodrho * 4096.0;
			skyverts[vertspos++] = ctheta * srhodrho * 4096.0;
			skyverts[vertspos++] = crhodrho * 4096.0;

			s += ds;
		}

		t -= dt;
	}
}


/*
=============
R_InitSky

A sky texture is 256*128, with the right side being a masked overlay
==============
*/
void R_LoadSky (char *skyname, qboolean verbose);

void R_InitSky (miptex_t *mt)
{
	int			i, j, p;
	byte		*src;
	unsigned	trans[128*128];
	unsigned	transpix;
	int			r, g, b;
	unsigned	*rgba;
	extern int	skytexturenum;

	// attempt a load of a skybox using this name
	R_LoadSky (mt->name, false);

	// always send through the regular loader to ensure that the texture objects are created OK
	src = (byte *) mt + mt->offsets[0];

	// make an average value for the back to avoid
	// a fringe on the top level
	r = g = b = 0;

	for (i = 0; i < 128; i++)
	{
		for (j = 0; j < 128; j++)
		{
			p = src[i * 256 + j + 128];
			rgba = &d_8to24table[p];
			trans[(i * 128) + j] = *rgba;

			r += ((byte *) rgba)[0];
			g += ((byte *) rgba)[1];
			b += ((byte *) rgba)[2];
		}
	}

	((byte *) &transpix)[0] = r / (128 * 128);
	((byte *) &transpix)[1] = g / (128 * 128);
	((byte *) &transpix)[2] = b / (128 * 128);
	((byte *) &transpix)[3] = 0;


	if (!solidskytexture) GL_AllocTexture (&solidskytexture);

	glBindTexture (GL_TEXTURE_2D, solidskytexture);
	glTexImage2D (GL_TEXTURE_2D, 0, gl_solid_format, 128, 128, 0, GL_RGBA, GL_UNSIGNED_BYTE, trans);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	for (i = 0; i < 128; i++)
	{
		for (j = 0; j < 128; j++)
		{
			p = src[i * 256 + j];

			if (p == 0)
				trans[(i * 128) + j] = transpix;
			else
			{
				trans[(i * 128) + j] = d_8to24table[p];

				// set alpha to the greyscaled average of the components
				r = (int) (((byte *) &trans[(i * 128) + j])[0]) * 30;
				g = (int) (((byte *) &trans[(i * 128) + j])[1]) * 59;
				b = (int) (((byte *) &trans[(i * 128) + j])[2]) * 11;

				// to be mathematically correct, we would divide this by 100 to derive the alpha.  however, this
				// gives a result too low for the purple sky, so we use a lesser number.
				p = (r + g + b) / 50;

				if (p > 255) p = 255;

				((byte *) &trans[(i * 128) + j])[3] = p;
			}
		}
	}

	if (!alphaskytexture) GL_AllocTexture (&alphaskytexture);

	glBindTexture (GL_TEXTURE_2D, alphaskytexture);
	glTexImage2D (GL_TEXTURE_2D, 0, gl_alpha_format, 128, 128, 0, GL_RGBA, GL_UNSIGNED_BYTE, trans);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
}


qboolean LoadExternalTexture (char *basename, qboolean luma, qboolean mipmap);
extern qboolean IntelIntegrated;
extern int upload_width;
extern int upload_height;

void R_LoadSky (char *skyname, qboolean verbose)
{
	int i;
	qboolean oldIntelIntegrated;

	// some people like using _ before the suffix, so for now we just store the base suffix, and will check it twice
	// the specific order below is required for the render, so don't change it
	char *suffixes[] = {"rt", "bk", "lf", "ft", "up", "dn"};

	char componentname[MAX_OSPATH];

	// before we do anything, check if we need to init the texture objects
	for (i = 0; i < 6; i++)
	{
		if (!skyboxtexes[i])
		{
			// these aren't kept in the regular cache...
			skyboxtexes[i] = (gltexture_t *) malloc (sizeof (gltexture_t));
			memset (skyboxtexes[i], 0, sizeof (gltexture_t));
			GL_AllocTexture (&skyboxtexes[i]->texnum);
		}

		// hack - use the "cache" member to indicate if a texture has been loaded
		// these aren't stored in the regular cache, so it's ok to do this
		skyboxtexes[i]->cache = false;
	}

	// disable intel integrated check in LoadExternalTexture
	oldIntelIntegrated = IntelIntegrated;
	IntelIntegrated = false;

	// these won't load from under textures/ if we constrain to this, so we make two passes...
	BuildTextureSearchPaths ("gfx/", "env/", "gfx/env/", NULL);

	// load each component in turn
	for (i = 0; i < 6; i++)
	{
		// already loaded
		if (skyboxtexes[i]->cache) continue;

		// bind the correct texture
		glBindTexture (GL_TEXTURE_2D, skyboxtexes[i]->texnum);

		// here we need two passes again... one with an _ prefixing the suffix, and one without
		// (some peoplel like to use an _ here... first we run without it.
		sprintf (componentname, "%s%s", skyname, suffixes[i]);

		if (LoadExternalTexture (componentname, false, false))
		{
			// mark as loaded OK
			skyboxtexes[i]->cache = true;
			skyboxtexes[i]->width = upload_width;
			skyboxtexes[i]->height = upload_height;

			// no need for any more
			continue;
		}

		// now run the second pass (with the _)
		sprintf (componentname, "%s_%s", skyname, suffixes[i]);

		// attempt a load and mark if OK
		if (LoadExternalTexture (componentname, false, false))
		{
			skyboxtexes[i]->cache = true;
			skyboxtexes[i]->width = upload_width;
			skyboxtexes[i]->height = upload_height;
		}
	}

	// now run a *second* full pass, including "textures" in the search path
	BuildTextureSearchPaths ("textures/gfx/", "textures/env/", "textures/gfx/env/", NULL);

	// load each component in turn
	for (i = 0; i < 6; i++)
	{
		// already loaded
		if (skyboxtexes[i]->cache) continue;

		// bind the correct texture
		glBindTexture (GL_TEXTURE_2D, skyboxtexes[i]->texnum);

		// here we need two passes again... one with an _ prefixing the suffix, and one without
		// (some peoplel like to use an _ here... first we run without it.
		sprintf (componentname, "%s%s", skyname, suffixes[i]);

		if (LoadExternalTexture (componentname, false, false))
		{
			// mark as loaded OK
			skyboxtexes[i]->cache = true;
			skyboxtexes[i]->width = upload_width;
			skyboxtexes[i]->height = upload_height;

			// no need for any more
			continue;
		}

		// now run the second pass (with the _)
		sprintf (componentname, "%s_%s", skyname, suffixes[i]);

		// attempt a load and mark if OK
		if (LoadExternalTexture (componentname, false, false))
		{
			skyboxtexes[i]->cache = true;
			skyboxtexes[i]->width = upload_width;
			skyboxtexes[i]->height = upload_height;
		}
	}

	// re-enable intel integrated check
	// this must be done before the check, to ensure it's actually done at all!
	IntelIntegrated = oldIntelIntegrated;

	// now check that we have all 6 components
	for (i = 0; i < 6; i++)
	{
		// any one not present indicates a failure
		if (!skyboxtexes[i]->cache)
		{
			if (verbose) Con_Printf ("Failed to load component %s for skybox %s\n", suffixes[i], skyname);
			return;
		}
	}

	// success!
	// set clamp env mode to help avoid bilerp seam
	for (i = 0; i < 6; i++)
	{
		glBindTexture (GL_TEXTURE_2D, skyboxtexes[i]->texnum);

		glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	}

	if (verbose) Con_Printf ("Loaded %s\n", skyname);
}


void R_LoadSky_f (void)
{
	if (Cmd_Argc () != 2)
	{
		Con_Printf ("loadsky <skyname> : loads the specified skybox\n");
		Con_Printf ("                    call with invalid name to clear skybox\n");
		return;
	}

	R_LoadSky (Cmd_Argv (1), true);
}
