/*
Code taken from bsp2prt type thing.

Is this properly GPL or what???
*/

#include "quakedef.h"

byte *model_base;

int numnodes;
int numplanes;
int numbrushplanes;

typedef struct
{
	int	numpoints;
	vec3_t	points[8];			// variable sized
} winding_t;

typedef struct portal_s
{
	int		planenum;
	struct node_s		*nodes[2];	// [0] = front side of planenum
	struct portal_s	*next[2];
	winding_t	*winding;
} portal_t;

typedef struct node_s
{
	vec3_t			mins,maxs;	// bounding volume, not just points inside

	// information for decision nodes
	int			planenum;	// -1 = leaf node
	int			outputplanenum;	// only valid after WriteNodePlanes
	int			firstface;	// decision node only
	int			numfaces;	// decision node only
	struct node_s		*children[2];	// only valid for decision nodes
	int			*faces;		// decision nodes only, list for both sides

	// information for leafs
	int			contents;	// leaf nodes (0 for decision nodes)
	int			**markfaces;	// leaf nodes only, point to node faces
	struct portal_s		*portals;
	int			visleafnum;	// -1 = solid
	int			valid;		// for flood filling
	int			occupied;	// light number in leaf for outside filling
} node_t;


mplane_t *planes;
node_t *nodes;
node_t *node_p;

dnode_t *dnodes;
dleaf_t *dleafs;

typedef struct brush_s
{
	struct brush_s	*next;
	vec3_t		mins, maxs;
	int		*faces;
	int		contents;
} brush_t;

typedef struct
{
	vec3_t		mins, maxs;
	brush_t		*brushes;		// NULL terminated list
} brushset_t;

brushset_t *brushset;

#define SIDESPACE 24
node_t outside_node;

#define	DISTEPSILON	0.01
#define	ANGLEEPSILON	0.00001

int c_activeportals, c_peakportals;

double VectorLength(vec3_t v)
{
	int		i;
	double	length;
	
	length = 0;
	for (i=0 ; i< 3 ; i++)
		length += v[i]*v[i];
	length = sqrt (length);		// FIXME

	return length;
}


portal_t *AllocPortal (void)
{
	portal_t	*p;

	c_activeportals++;

	if (c_activeportals > c_peakportals)
		c_peakportals = c_activeportals;

	p = (portal_t *) malloc (sizeof (portal_t));
	memset (p, 0, sizeof (portal_t));

	return p;
}


void NormalizePlane (mplane_t *dp)
{
	vec_t	ax, ay, az;

	if (dp->normal[0] == -1.0)
	{
		dp->normal[0] = 1.0;
		dp->dist = -dp->dist;
	}
	if (dp->normal[1] == -1.0)
	{
		dp->normal[1] = 1.0;
		dp->dist = -dp->dist;
	}
	if (dp->normal[2] == -1.0)
	{
		dp->normal[2] = 1.0;
		dp->dist = -dp->dist;
	}

	if (dp->normal[0] == 1.0)
	{
		dp->type = PLANE_X;
		return;
	}
	if (dp->normal[1] == 1.0)
	{
		dp->type = PLANE_Y;
		return;
	}
	if (dp->normal[2] == 1.0)
	{
		dp->type = PLANE_Z;
		return;
	}

	ax = fabs(dp->normal[0]);
	ay = fabs(dp->normal[1]);
	az = fabs(dp->normal[2]);

	if (ax >= ay && ax >= az)
		dp->type = PLANE_ANYX;
	else if (ay >= ax && ay >= az)
		dp->type = PLANE_ANYY;
	else
		dp->type = PLANE_ANYZ;
	if (dp->normal[dp->type-PLANE_ANYX] < 0)
	{
		VectorSubtract (vec3_origin, dp->normal, dp->normal);
		dp->dist = -dp->dist;
	}

}

/*
===============
FindPlane

Returns a global plane number and the side that will be the front
===============
*/
int	FindPlane (mplane_t *dplane, int *side)
{
	int			i;
	mplane_t	*dp, pl;
	vec_t		dot;

	dot = VectorLength (dplane->normal);

	if (dot < 1.0 - ANGLEEPSILON || dot > 1.0 + ANGLEEPSILON)
		Sys_Error ("FindPlane: Normalization error");

	pl = *dplane;
	NormalizePlane (&pl);

	if (DotProduct(pl.normal, dplane->normal) > 0)
		*side = 0;
	else
		*side = 1;

	dp = planes;

	for (i=0 ; i<numbrushplanes;i++, dp++)
	{
		dot = DotProduct (dp->normal, pl.normal);

		if (dot > 1.0 - ANGLEEPSILON && fabs(dp->dist - pl.dist) < DISTEPSILON )
		{
			// regular match
			return i;
		}
	}

	if (numbrushplanes > numplanes + 6)
		Sys_Error ("FindPlane: numbrushplanes exceeds max (%d)", numplanes + 6);

	planes[numbrushplanes] = pl;

	numbrushplanes++;

	return numbrushplanes-1;
}


/*
================
MakeHeadnodePortals

The created portals will face the global outside_node
================
*/
/*
void MakeHeadnodePortals (node_t *node)
{
	vec3_t		bounds[2];
	int			i, j, n;
	portal_t	*p, *portals[6];
	mplane_t	bplanes[6], *pl;
	int			side;

	// pad with some space so there will never be null volume leafs
	for (i = 0; i < 3; i++)
	{
		bounds[0][i] = brushset->mins[i] - SIDESPACE;
		bounds[1][i] = brushset->maxs[i] + SIDESPACE;
	}

	outside_node.contents = CONTENTS_SOLID;
	outside_node.portals = NULL;

	for (i = 0; i < 3; i++)
	{
		for (j = 0; j < 2; j++)
		{
			n = j * 3 + i;

			p = AllocPortal ();
			portals[n] = p;

			pl = &bplanes[n];
			memset (pl, 0, sizeof (*pl));

			if (j)
			{
				pl->normal[i] = -1;
				pl->dist = -bounds[j][i];
			}
			else
			{
				pl->normal[i] = 1;
				pl->dist = bounds[j][i];
			}

			p->planenum = FindPlane (pl, &side);

			p->winding = BaseWindingForPlane (pl);

			if (side)
				AddPortalToNodes (p, &outside_node, node);
			else
				AddPortalToNodes (p, node, &outside_node);
		}
	}

	// clip the basewindings by all the other planes
	for (i = 0; i < 6; i++)
	{
		for (j = 0; j < 6; j++)
		{
			if (j == i)
				continue;

			portals[i]->winding = ClipWinding (portals[i]->winding, &bplanes[j], true);
		}
	}
}
*/


void PortalizeWorld (node_t *headnode)
{
	//MakeHeadnodePortals (headnode);
	//CutNodePortals_r (headnode);
}


qboolean CheckWorld (lump_t *l)
{
	int i;
	char *ents;

	if (!l->filelen) return 0;

	ents = (char *) (model_base + l->fileofs);

	// look through the entities for certain ones to confirm if it's the world
	// the -30 isn't based on anything, it's just big enough to prevent an overshoot
	for (i = 0; i < l->filelen - 30; i++)
	{
		if (!strncmp ("info_player_start", &ents[i], 17)) return true;
		if (!strncmp ("info_player_coop", &ents[i], 16)) return true;
		if (!strncmp ("info_player_deathmatch", &ents[i], 22)) return true;
	}

	return false;
}


void MakeNode (int nodenum)
{
	node_t	*n;
	dnode_t *dn;
	int	i;

	n = node_p++;

	if (nodenum < 0)
	{
		n->contents = dleafs[-nodenum - 1].contents;
		return;
	}

	dn = dnodes + nodenum;

	n->planenum = dn->planenum;

	for (i = 0; i < 2; ++i)
	{
		n->children[i] = node_p;
		MakeNode (dn->children[i]);
	}
}


qboolean CheckTransVis (model_t *mdl)
{
	int i;
	int j;
	int k;
	byte *vis;
	byte *vis2;
	mleaf_t *leaf;
	mleaf_t *leaf2;
	mleaf_t *vleaf;

	for (i = 1; i < mdl->numleafs; i++)
	{
		leaf = &mdl->leafs[i];

		// only run with empty leafs
		if (leaf->contents != CONTENTS_EMPTY)
			continue;

		// only interested in leafs that contains water
		if (!leaf->haswater)
			continue;

		// get a PVS for this leaf
		vis = Mod_LeafPVS (leaf, mdl);

		// check all leafs
		for (j = 0; j < mdl->numleafs; j++)
		{
			// visible?
			if (vis[j >> 3] & (1 << (j & 7)))
			{
				vleaf = &mdl->leafs[j + 1];

				// only if it has water
				if (!vleaf->haswater) continue;

				// check contents
				if (vleaf->contents == CONTENTS_WATER || vleaf->contents == CONTENTS_LAVA || vleaf->contents == CONTENTS_SLIME)
				{
					// now do a reciprocal check
					vis2 = Mod_LeafPVS (vleaf, mdl);

					// check all again
					for (k = 0; k < mdl->numleafs; k++)
					{
						// visible?
						if (vis2[k >> 3] & (1 << (k & 7)))
						{
							leaf2 = &mdl->leafs[k + 1];

							// match!
							if (leaf2 == leaf) return true;
						}
					}
				}
			}
		}
	}

	return false;
}


void Mod_DoPrt (model_t *mdl, byte *base, dheader_t *header)
{
	int i;
	int j;
	brushset_t bs;

	model_base = base;

	// do a quick check of the entities lump for info_player stuff
	if (!CheckWorld (&header->lumps[LUMP_ENTITIES]))
		return;

	// see have we got a map already vised for translucent water, abort if we do
	if (CheckTransVis (mdl))
	{
		Con_DPrintf ("Map %s is vissed for translucent water\n", mdl->name);
		return;
	}

	Con_Printf ("Map %s is NOT vissed for translucent water\n", mdl->name);

	// get node and leaf pointers
	dnodes = (dnode_t *) (base + header->lumps[LUMP_NODES].fileofs);
	dleafs = (dleaf_t *) (base + header->lumps[LUMP_LEAFS].fileofs);

	// load up nodes into temp storage
	numnodes = header->lumps[LUMP_NODES].filelen / sizeof (dnode_t);

	// memory nodes can max be 3 * numnodes, usually less than 2 * numnodes
	node_p = nodes = (node_t *) malloc (sizeof (node_t) * numnodes * 3);
	memset (nodes, 0, sizeof (node_t) * numnodes * 3);

	// load up planes into temp storage.
	numplanes = header->lumps[LUMP_PLANES].filelen / sizeof (dplane_t);

	planes = malloc (sizeof (mplane_t) * (numplanes + 6));

	for (i = 0; i < numplanes; i++)
	{
		dplane_t *dp = (dplane_t *) (base + header->lumps[LUMP_PLANES].fileofs);
		dp += i;

		for (j = 0; j < 3; j++)
			planes[i].normal[j] = dp->normal[j];

		planes[i].dist = dp->dist;
		planes[i].type = dp->type;
	}

	numbrushplanes = numplanes;

	// build nodes tree
	Con_Printf ("Building node tree...\n");
	MakeNode (0);

	// initialize brushset boundaries
	brushset = &bs;
	memset (brushset, 0, sizeof (brushset_t));

	for (i = 0; i < 3; ++i)
	{
		brushset->mins[i] = dnodes[0].mins[i];
		brushset->maxs[i] = dnodes[0].maxs[i];
	}

	// make the real portals for vis tracing
	Con_Printf ("Portalize World...\n");
	PortalizeWorld (nodes);

	free (planes);
	free (nodes);
}

