/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_misc.c

#include "quakedef.h"
#include "cvar.h"
#include "render.h"
#include "client.h"
#include "gl_model.h"
#include "glquake.h"
#include "glext.h"

// 2 new particle textures
int bloodtexture = 0;
int smoketexture = 0;

/*
==================
R_InitTextures
==================
*/
void GL_Upload8 (byte *data, int width, int height, qboolean mipmap, qboolean alpha);

texture_t *r_notexture_mip;


void R_InitTextures (void)
{
	int x, y, m;
	byte *dest;

	// create a simple checkerboard texture for the default
	r_notexture_mip = Hunk_AllocName (sizeof (texture_t) + 16 * 16 + 8 * 8 + 4 * 4 + 2 * 2, "notexture");

	r_notexture_mip->width = r_notexture_mip->height = 16;
	r_notexture_mip->offsets[0] = sizeof (texture_t);
	r_notexture_mip->offsets[1] = r_notexture_mip->offsets[0] + 16 * 16;
	r_notexture_mip->offsets[2] = r_notexture_mip->offsets[1] + 8 * 8;
	r_notexture_mip->offsets[3] = r_notexture_mip->offsets[2] + 4 * 4;

	for (m = 0; m < 4; m++)
	{
		dest = (byte *) r_notexture_mip + r_notexture_mip->offsets[m];

		for (y = 0; y < (16 >> m); y++)
		{
			for (x = 0; x < (16 >> m); x++)
			{
				if ((y < (8 >> m)) ^ (x < (8 >> m)))
					*dest++ = 0;
				else
					*dest++ = 0xff;
			}
		}
	}

	// MH - need to upload this as a proper texture
	// done - see Draw_Init ()
}


unsigned char blood_texture[32][32] =
{
	// the "blood texture" is a font containing 9 seperate textures - each blood particle is
	// assigned one of these 9 (randomly chosen) to keep blood from looking all samey
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, },
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, },
	{0, 0, 24, 99, 21, 0, 0, 0, 0, 0, 0, 0, 0, 18, 43, 36, 6, 124, 33, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, },
	{0, 0, 15, 60, 7, 0, 0, 0, 0, 0, 0, 0, 0, 116, 210, 70, 48, 106, 2, 0, 0, 0, 0, 0, 0, 32, 84, 169, 24, 0, 0, 0, },
	{0, 0, 0, 0, 0, 0, 107, 0, 0, 0, 0, 0, 0, 55, 124, 28, 11, 12, 2, 0, 0, 0, 0, 0, 0, 24, 177, 218, 74, 0, 0, 0, },
	{0, 0, 0, 24, 32, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 23, 232, 102, 0, 0, 0, 0, 0, 0, 0, 75, 94, 39, 0, 0, 0, },
	{0, 0, 75, 177, 84, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 17, 213, 143, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, },
	{0, 0, 94, 218, 169, 0, 0, 9, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, },
	{0, 0, 39, 74, 24, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, },
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, },
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, },
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 14, 54, 18, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, },
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 19, 133, 146, 15, 0, 0, 0, 0, 0, 1, 20, 4, 0, 0, 0, 0, 0, },
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3, 64, 223, 82, 0, 0, 0, 0, 0, 26, 155, 20, 0, 0, 0, 0, 0, },
	{0, 0, 0, 0, 0, 64, 38, 0, 0, 0, 0, 0, 0, 0, 0, 1, 13, 193, 225, 0, 0, 0, 0, 0, 67, 206, 105, 0, 0, 0, 0, 0, },
	{0, 0, 0, 0, 0, 39, 120, 36, 0, 0, 0, 0, 0, 0, 1, 2, 17, 200, 175, 0, 0, 0, 0, 0, 11, 96, 7, 0, 0, 0, 0, 0, },
	{0, 0, 0, 107, 0, 11, 93, 32, 0, 0, 0, 0, 0, 2, 3, 16, 72, 118, 20, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, },
	{0, 0, 0, 23, 0, 0, 0, 0, 0, 0, 0, 0, 0, 8, 27, 49, 45, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6, 7, 0, },
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 56, 141, 36, 22, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 24, 99, 21, 0, },
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 28, 113, 190, 91, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 15, 60, 7, 0, },
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 26, 82, 6, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, },
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, },
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3, 1, 0, 0, 0, 0, 0, 0, },
	{0, 7, 96, 11, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 1, 11, 39, 64, 0, 0, 0, },
	{0, 105, 206, 67, 0, 0, 0, 0, 0, 0, 0, 0, 0, 107, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 8, 2, 93, 120, 38, 0, 0, 0, },
	{0, 20, 155, 26, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 27, 3, 32, 36, 0, 0, 0, 0, },
	{0, 4, 20, 1, 0, 23, 107, 0, 0, 0, 0, 0, 0, 11, 39, 64, 0, 0, 0, 0, 0, 0, 0, 0, 49, 16, 0, 0, 0, 19, 14, 0, },
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 93, 120, 38, 0, 0, 24, 32, 0, 0, 0, 0, 45, 72, 17, 13, 64, 133, 54, 0, },
	{0, 0, 0, 0, 0, 0, 11, 39, 64, 0, 0, 0, 0, 32, 36, 0, 0, 75, 177, 84, 0, 0, 0, 0, 14, 118, 200, 193, 223, 146, 18, 0, },
	{0, 0, 0, 0, 0, 0, 93, 120, 38, 0, 0, 0, 0, 0, 0, 0, 0, 94, 218, 169, 0, 0, 0, 0, 0, 20, 175, 225, 82, 15, 0, 0, },
	{0, 0, 0, 0, 0, 0, 32, 36, 0, 0, 0, 0, 0, 0, 0, 0, 0, 39, 74, 24, 0, 0, 0, 0, 0, 0, 3, 7, 0, 0, 0, 0, },
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, },
};


unsigned char smoke_texture[32][32] = 
{
	// blatantly stolen from TomazQuake, by way of Q3A
	{  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  5, 28, 32, 34, 29, 26,  4,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0},
	{  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  7, 27, 46, 41, 40, 30, 29, 16,  6,  0,  0,  6, 18,  6,  0,  0,  0,  0,  0,  0,  0,  0},
	{  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 18, 62, 65, 62, 53, 35, 51, 57, 15,  0, 12, 25, 48, 25,  2,  0,  0,  0,  0,  0,  0,  0},
	{  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 14, 63, 98,100, 67, 69, 86, 72, 24, 11, 31, 39, 41, 30, 11,  0,  0,  0,  0,  0,  0,  0},
	{  0,  0,  0,  0,  0,  0,  0,  0,  7, 17, 25, 62, 97,102, 93,101,102, 89, 62, 43, 36, 49, 48, 23,  2,  0,  0,  0,  0,  0,  0,  0},
	{  0,  0,  0,  0,  0,  0,  0,  0, 19, 69, 61, 74,113,111,100,110,137,139,132, 97, 65, 78, 54, 35, 13,  0,  0,  0,  0,  0,  0,  0},
	{  0,  0,  0,  0,  7, 20, 18, 13, 35,104, 74, 88,130,139,142,151,162,162,159,146,127,106, 64, 21,  5,  0,  0,  0,  0,  0,  0,  0},
	{  0,  0,  0,  0, 19, 62, 55, 43, 56, 93, 65, 83,132,161,181,193,187,175,165,166,164,123, 69, 24,  6,  1,  0,  0,  0,  0,  0,  0},
	{  0,  0,  0,  0,  4, 26, 54, 68, 69, 64, 67, 85,142,175,195,203,198,188,177,176,177,146, 98, 63, 39, 12,  0,  0,  0,  0,  0,  0},
	{  0,  0,  0,  6, 17, 48,119,121, 96, 74, 80,105,150,183,196,201,197,194,188,184,179,162,133,101, 69, 15,  9, 12,  1,  0,  0,  0},
	{  0,  0,  6, 36, 75,142,158,127,111,101,106,127,160,187,195,200,197,196,196,187,169,147,136,129, 92, 33, 36, 34, 14,  0,  0,  0},
	{  0,  1, 27, 72,105,161,144,124,128,127,118,146,174,191,196,200,200,201,195,181,154,126,124,132,126, 90, 77, 54, 20,  0,  0,  0},
	{  0,  9, 38, 94,137,156,138,119,135,142,123,155,182,192,196,197,201,200,190,180,163,121, 94,115,131,110, 97, 90, 34,  6,  0,  0},
	{  0,  7, 32, 86,134,150,139,123,135,140,131,149,172,185,191,192,194,193,183,174,161,124, 88, 99,126,111, 94,110, 73, 14,  0,  0},
	{  0, 16, 52, 87,121,128,140,139,133,138,156,161,169,177,182,182,184,184,172,163,146,114, 95,110,115, 98,102,101, 86, 34,  5,  0},
	{  6, 19, 51, 77,107,126,146,141,138,143,161,167,169,173,179,176,174,172,161,151,131,101, 94,101, 92, 73, 94,101, 95, 80, 27,  0},
	{ 20, 31, 81, 92,112,123,133,128,141,157,162,165,166,172,176,171,170,166,154,136,113, 90, 70, 71, 81, 80, 93, 97, 96, 89, 46, 12},
	{ 24, 70,104, 80, 88, 76,102,113,136,149,143,142,156,170,168,160,166,154,142,120, 91, 64, 45, 52, 76, 78, 81, 84, 84, 80, 58, 29},
	{ 42, 91,100, 55, 57, 41, 66, 91, 97,118,122,120,135,151,153,148,148,144,131,104, 88, 50, 34, 47, 58, 49, 54, 62, 68, 75, 64, 37},
	{ 43, 82, 84, 70, 60, 41, 39, 65, 50, 85,106,110,104,115,131,142,139,135,110, 83, 80, 60, 42, 46, 27, 13, 21, 38, 46, 60, 58, 35},
	{ 15, 29, 46, 82, 94, 56, 21, 34, 31, 57, 94,109, 88, 88,113,128,136,126, 94, 64, 53, 55, 55, 47, 36, 11,  0, 21, 35, 40, 42, 22},
	{  0,  0, 25, 50, 83, 69, 15, 30, 42, 54, 84,101, 93, 80, 87,111,127,119, 93, 62, 52, 62, 49, 37, 38, 24, 12, 33, 45, 36, 22,  5},
	{  0,  0,  6, 25, 69, 63, 23, 39, 47, 65, 87, 96, 92, 75, 87,114,121,106, 87, 65, 60, 62, 34, 27, 15,  6, 12, 35, 49, 45, 14,  0},
	{  0,  0,  0, 15, 60, 61, 39, 55, 59, 71, 86, 97, 87, 72, 90,110,109, 96, 73, 67, 70, 60, 26, 11,  0,  0, 12, 31, 39, 41, 16,  0},
	{  0,  0,  0, 16, 54, 61, 51, 62, 61, 62, 81, 87, 69, 58, 79, 82, 85, 84, 46, 57, 74, 63, 39, 12,  0,  0, 12, 33, 42, 33,  7,  0},
	{  0,  0,  0, 14, 47, 59, 50, 52, 57, 53, 53, 51, 46, 57, 65, 72, 78, 68, 36, 42, 59, 57, 26,  6,  0,  0, 15, 41, 42, 21,  0,  0},
	{  0,  0,  0, 11, 35, 42, 31, 25, 42, 40, 34, 36, 47, 56, 49, 60, 68, 55, 48, 35, 27, 44, 15,  0,  0,  0, 14, 33, 17,  5,  0,  0},
	{  0,  0,  0,  7, 29, 21,  5, 12, 39, 41, 44, 46, 51, 50, 45, 51, 45, 39, 48, 35, 24, 21,  7,  0,  0,  0,  2, 11,  0,  0,  0,  0},
	{  0,  0,  1, 22, 39, 13,  0, 12, 29, 36, 42, 32, 18, 34, 36, 37, 38, 32, 34, 32, 14,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0},
	{  0,  0, 12, 34, 36, 14,  0,  1,  2, 12, 15,  6,  0, 11, 13, 13, 15, 13, 13, 12,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0},
	{  0,  0,  1, 14, 29, 13,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0},
	{  0,  0,  0,  0,  6,  2,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0},
};

void R_InitParticleTexture (void)
{
	int		x,y;
	byte	data[32][32][4];

	// blood texture;
	if (!bloodtexture) bloodtexture = texture_extension_number++;
    glBindTexture (GL_TEXTURE_2D, bloodtexture);

	for (x = 0; x < 32; x++)
	{
		for (y = 0; y < 32; y++)
		{
			data[y][x][0] = 255;
			data[y][x][1] = 255;
			data[y][x][2] = 255;
			data[y][x][3] = blood_texture[x][y];
		}
	}

	glTexImage2D (GL_TEXTURE_2D, 0, GL_RGBA, 32, 32, 0, GL_RGBA, GL_UNSIGNED_BYTE, data);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	// smoke texture;
	if (!smoketexture) smoketexture = texture_extension_number++;
    glBindTexture (GL_TEXTURE_2D, smoketexture);

	for (x = 0; x < 32; x++)
	{
		// MH - this texture is upside-down!!!
		for (y = 31; y >= 0; y--)
		{
			data[y][x][0] = 255;
			data[y][x][1] = 255;
			data[y][x][2] = 255;

			// it's too bright so take it down
			data[y][x][3] = (smoke_texture[x][y] / 2);
		}
	}

	glTexImage2D (GL_TEXTURE_2D, 0, GL_RGBA, 32, 32, 0, GL_RGBA, GL_UNSIGNED_BYTE, data);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
}

/*
===============
R_Envmap_f

Grab six views for environment mapping tests
===============
*/
void R_Envmap_f (void)
{
}


void R_MapData_f (void)
{
	int i;
	int leafsurfs;

	if (!cl.worldmodel) return;

	Con_Printf ("Map: %i\n", cl.worldmodel->name);
	Con_Printf ("- Textures: %i\n", cl.worldmodel->numtextures);
	Con_Printf ("- Surfaces: %i\n", cl.worldmodel->numsurfaces);
	Con_Printf ("- Leafs: %i\n", cl.worldmodel->numleafs);
	Con_Printf ("- Nodes: %i\n", cl.worldmodel->numnodes);
	Con_Printf ("- MarkSurfaces: %i\n", cl.worldmodel->nummarksurfaces);

	leafsurfs = 0;

	for (i = 0; i < cl.worldmodel->numleafs + 1; i++)
	{
		leafsurfs += cl.worldmodel->leafs[i].nummarksurfaces;
	}

	Con_Printf ("- LeafSurfaces: %i\n", leafsurfs);
}


void R_DumpIndividualSurface (msurface_t *surf, FILE *f)
{
	glpoly_t	*p;
	float		*v;
	int			i;
	int			poly;
	int			check;

	/*
	fprintf (f, "SURFACE %5i     Styles [%i] [%i] [%i] [%i]   Cached [%i] [%i] [%i] [%i]\n",
		d_lightstylevalue[surf->styles[0]], 
		d_lightstylevalue[surf->styles[1]], 
		d_lightstylevalue[surf->styles[2]], 
		d_lightstylevalue[surf->styles[3]], 
		surf->cached_light[0],
		surf->cached_light[1],
		surf->cached_light[2],
		surf->cached_light[3]);
	*/

	/*
	fprintf (f, "SURFACE %5i     [Style][Value] [%i][%i] - [%i][%i] - [%i][%i] - [%i][%i]   Cached [%i] [%i] [%i] [%i]\n",
		surf->surfNum,
		surf->styles[0], 
		d_lightstylevalue[surf->styles[0]], 
		surf->styles[1], 
		d_lightstylevalue[surf->styles[1]], 
		surf->styles[2], 
		d_lightstylevalue[surf->styles[2]], 
		surf->styles[3], 
		d_lightstylevalue[surf->styles[3]], 
		surf->cached_light[0],
		surf->cached_light[1],
		surf->cached_light[2],
		surf->cached_light[3]);
	*/

	fprintf (f, "SURFACE %5i   [style][value][cached]", surf->surfNum);

	for (i = 0; i < 4; i++)
	{
		if (surf->styles[i] == 255) break;

		if (d_lightstylevalue[surf->styles[i]] != surf->cached_light[i])
		{
			fprintf (f, "   *[%3i][%3i][%3i]*", surf->styles[i], d_lightstylevalue[surf->styles[i]], surf->cached_light[i]);
		}
		else
		{
			fprintf (f, "    [%3i][%3i][%3i] ", surf->styles[i], d_lightstylevalue[surf->styles[i]], surf->cached_light[i]);
		}
	}

	fprintf (f, "\n");
}


void R_DumpSolidSurf (msurface_t *surf, FILE *f)
{
}


void R_SurfDump_f (void)
{
	int i;
	FILE *f;
	msurface_t *surf;

	if (!cl.worldmodel) return;

	f = fopen ("surfdump.txt", "w");

	for (i = 0, surf = cl.worldmodel->surfaces; i < cl.worldmodel->numsurfaces; i++, surf++)
		R_DumpIndividualSurface (surf, f);

	fclose (f);
}


void R_LMDump_f (void)
{
	int i;
	int j;
	int scount;

	FILE *f;
	msurface_t *surf;

	if (!cl.worldmodel) return;

	f = fopen ("lmdump.txt", "w");

	for (j = 0; ; j++)
	{
		scount = 0;

		for (i = 0, surf = cl.worldmodel->surfaces; i < cl.worldmodel->numsurfaces; i++, surf++)
			if (surf->lightmaptexturenum == j) scount++;

		if (!scount) break;

		fprintf (f, "LIGHTMAP %02i Used by %i Surfs\n", j, scount);
	}

	fprintf (f, "\n%i Lightmaps Used\n", j);

	fclose (f);
}


void R_LeafDump_f (void)
{
	int i;
	int j;
	FILE *f;

	if (!cl.worldmodel) return;

	f = fopen ("leafdump.txt", "w");

	for (i = 0; i < cl.worldmodel->numleafs + 1; i++)
	{
		switch (cl.worldmodel->leafs[i].contents)
		{
		case CONTENTS_SOLID:
			fprintf (f, "Leaf %4i Has CONTENTS_SOLID\n", cl.worldmodel->leafs[i].leafnum);
			break;

		case CONTENTS_EMPTY:
			fprintf (f, "Leaf %4i Has CONTENTS_EMPTY\n", cl.worldmodel->leafs[i].leafnum);
			break;

		case CONTENTS_WATER:
			fprintf (f, "Leaf %4i Has CONTENTS_WATER\n", cl.worldmodel->leafs[i].leafnum);
			break;

		case CONTENTS_LAVA:
			fprintf (f, "Leaf %4i Has CONTENTS_LAVA\n", cl.worldmodel->leafs[i].leafnum);
			break;

		case CONTENTS_SLIME:
			fprintf (f, "Leaf %4i Has CONTENTS_SLIME\n", cl.worldmodel->leafs[i].leafnum);
			break;

		case CONTENTS_SKY:
			fprintf (f, "Leaf %4i Has CONTENTS_SKY\n", cl.worldmodel->leafs[i].leafnum);
			break;

		case CONTENTS_ORIGIN:
			fprintf (f, "Leaf %4i Has CONTENTS_ORIGIN\n", cl.worldmodel->leafs[i].leafnum);
			break;

		case CONTENTS_CLIP:
			fprintf (f, "Leaf %4i Has CONTENTS_CLIP\n", cl.worldmodel->leafs[i].leafnum);
			break;

		default:
			fprintf (f, "Leaf %4i Has UNKNOWN CONTENTS!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n", cl.worldmodel->leafs[i].leafnum);
			break;
		}
	}

	fclose (f);
}


void R_VolDump_f (void)
{
	int i;
	FILE *f;
	mleaf_t *leaf;

	if (!cl.worldmodel) return;

	f = fopen ("voldump.txt", "w");

	for (i = 0; i < cl.worldmodel->numVolumes; i++)
	{
		fprintf (f, "VOLUME %i  Contents %i", i, cl.worldmodel->volumes[i].contents);

		leaf = cl.worldmodel->volumes[i].chain;

		if (cl.worldmodel->volumes[i].visframe != -2)
			fprintf (f, "      Drawing volume");

		fprintf (f, "\n\n");
	}

	fclose (f);
}


// exports visdata to a .VIS file
void R_Visexport_f (void)
{
}


/*
===============
R_Init
===============
*/
cvar_t gl_fogenable = {"gl_fogenable", "1", true};
cvar_t gl_fogstart = {"gl_fogstart", "100", true};
cvar_t gl_fogend = {"gl_fogend", "4096", true};
cvar_t gl_fogdensity = {"gl_fogdensity", "0.2", true};
cvar_t gl_fogr = {"gl_fogr", "0.1", true};
cvar_t gl_fogg = {"gl_fogg", "0.1", true};
cvar_t gl_fogb = {"gl_fogb", "0.1", true};
cvar_t gl_lockpvs = {"gl_lockpvs", "0", false};

cvar_t gl_foguser = {"gl_foguser", "0", true};

cvar_t gl_fogvol = {"gl_fogvol", "0", true};

extern cvar_t r_skyspeed;

int mapshottexture;
int currentmapshottexture[21];


void R_Mapshots_f (void)
{
	int i;

	for (i = 0; i < 21; i++)
	{
		Con_Printf ("Mapshot %02i   Texture Number: %i\n", i, currentmapshottexture[i]);
	}
}


void R_GL_Extensions_f (void)
{
	FILE *f;
	char *exts = (char *) glGetString (GL_EXTENSIONS);
	int i;

	for (i = 0; ; i++)
	{
		if (!exts[i]) break;

		if (exts[i] == ' ') exts[i] = '\n';
	}

	f = fopen ("glextensions.txt", "w");

	fprintf (f, "%s\n", exts);

	Con_Printf ("%s\n", exts);

	fclose (f);
}


void GL_InitLightmapTextures (void);
void GL_RSurfInit (void);

void R_Init (void)
{
	int i;
	extern byte *hunk_base;
	extern cvar_t gl_finish;

	Cmd_AddCommand ("timerefresh", R_TimeRefresh_f);	
	Cmd_AddCommand ("envmap", R_Envmap_f);	
	Cmd_AddCommand ("pointfile", R_ReadPointFile_f);	
	Cmd_AddCommand ("mapdata", R_MapData_f);
	Cmd_AddCommand ("surfdump", R_SurfDump_f);
	Cmd_AddCommand ("leafdump", R_LeafDump_f);
	Cmd_AddCommand ("voldump", R_VolDump_f);
	Cmd_AddCommand ("mapshots", R_Mapshots_f);
	Cmd_AddCommand ("lmdump", R_LMDump_f);

	Cmd_AddCommand ("visexport", R_Visexport_f);
	Cmd_AddCommand ("gl_extensions", R_GL_Extensions_f);

	Cvar_RegisterVariable (&gl_blend_src);
	Cvar_RegisterVariable (&gl_blend_dst);

	Cvar_RegisterVariable (&gl_lockpvs);
	Cvar_RegisterVariable (&r_norefresh);
	Cvar_RegisterVariable (&r_lightmap);
	Cvar_RegisterVariable (&r_drawentities);
	Cvar_RegisterVariable (&r_drawviewmodel);
	Cvar_RegisterVariable (&r_shadows);
	Cvar_RegisterVariable (&r_wateralpha);
	Cvar_RegisterVariable (&r_dynamic);
	Cvar_RegisterVariable (&r_novis);
	Cvar_RegisterVariable (&r_speeds);

	Cvar_RegisterVariable (&gl_finish);

	Cvar_RegisterVariable (&gl_fogenable);

	Cvar_RegisterVariable (&gl_fogstart);
	Cvar_RegisterVariable (&gl_fogend);
	Cvar_RegisterVariable (&gl_fogdensity);
	Cvar_RegisterVariable (&gl_fogr);
	Cvar_RegisterVariable (&gl_fogg);
	Cvar_RegisterVariable (&gl_fogb);

	Cvar_RegisterVariable (&gl_foguser);

	Cvar_RegisterVariable (&gl_clear);

	Cvar_RegisterVariable (&gl_cull);
	Cvar_RegisterVariable (&gl_smoothmodels);
	Cvar_RegisterVariable (&gl_affinemodels);
	Cvar_RegisterVariable (&gl_polyblend);
	Cvar_RegisterVariable (&gl_playermip);
	Cvar_RegisterVariable (&gl_nocolors);

	Cvar_RegisterVariable (&gl_keeptjunctions);
	Cvar_RegisterVariable (&gl_reporttjunctions);

	Cvar_RegisterVariable (&gl_doubleeyes);

	R_InitParticles ();
	R_InitParticleTexture ();
	GL_InitLightmapTextures ();

	playertextures = texture_extension_number;
	texture_extension_number += 16;

	mapshottexture = texture_extension_number;
	texture_extension_number += 21;

	for (i = 0; i < 21; i++)
		currentmapshottexture[i] = -1;

	Cvar_RegisterVariable (&r_skyspeed);
	GL_RSurfInit ();
}

/*
===============
R_TranslatePlayerSkin

Translates a skin texture by the per-player color lookup
===============
*/
void R_TranslatePlayerSkin (int playernum)
{
	int		top, bottom;
	byte	translate[256];
	unsigned	translate32[256];
	int		i, j, s;
	model_t	*model;
	struct aliashdr_s *paliashdr;
	byte	*original;
	unsigned	pixels[512*256], *out;
	unsigned	scaled_width, scaled_height;
	int			inwidth, inheight;
	byte		*inrow;
	unsigned	frac, fracstep;
	extern	byte		**player_8bit_texels_tbl;
	byte	translated[320*200];

	top = cl.scores[playernum].colors & 0xf0;
	bottom = (cl.scores[playernum].colors &15)<<4;

	for (i=0 ; i<256 ; i++)
		translate[i] = i;

	for (i=0 ; i<16 ; i++)
	{
		if (top < 128)	// the artists made some backwards ranges.  sigh.
			translate[TOP_RANGE+i] = top+i;
		else
			translate[TOP_RANGE+i] = top+15-i;
				
		if (bottom < 128)
			translate[BOTTOM_RANGE+i] = bottom+i;
		else
			translate[BOTTOM_RANGE+i] = bottom+15-i;
	}

	//
	// locate the original skin pixels
	//
	currententity = &cl_entities[1+playernum];
	model = currententity->model;
	if (!model)
		return;		// player doesn't have a model yet
	if (model->type != mod_alias)
		return; // only translate skins on alias models

	paliashdr = (struct aliashdr_s *)Mod_Extradata (model);
	s = paliashdr->skinwidth * paliashdr->skinheight;

	if (currententity->skinnum < 0 || currententity->skinnum >= paliashdr->numskins)
	{
		Con_Printf("(%d): Invalid player skin #%d\n", playernum, currententity->skinnum);
		original = (byte *)paliashdr + paliashdr->texels[0];
	}
	else
		original = (byte *)paliashdr + paliashdr->texels[currententity->skinnum];

	if (s & 3)
		Sys_Error ("R_TranslateSkin: s&3");

	inwidth = paliashdr->skinwidth;
	inheight = paliashdr->skinheight;

	// because this happens during gameplay, do it fast
	// instead of sending it through gl_upload 8
    glBindTexture (GL_TEXTURE_2D, playertextures + playernum);

	for (i=0 ; i<s ; i+=4)
	{
		translated[i] = translate[original[i]];
		translated[i+1] = translate[original[i+1]];
		translated[i+2] = translate[original[i+2]];
		translated[i+3] = translate[original[i+3]];
	}

	// don't mipmap these, because it takes too long
	GL_Upload8 (translated, paliashdr->skinwidth, paliashdr->skinheight, true, false);
}


extern qboolean WorldModelLoaded;


extern int worldtype;
int Mod_FindWorldType (model_t *mod);
extern qboolean SbarPreload;

/*
====================
R_IdentifyUnderwater

reliable determination of an * model's underwater status
it's a dirty hack but - hey! - it works.
====================
*/
void R_IdentifyUnderwater (void)
{
	int x;
	model_t *mod;
	msurface_t *psurf;
	vec3_t org;
	float *v;
	glpoly_t *p;
	int numVerts;
	int i, j;

	for (x = 1; x < MAX_MODELS; x++)
	{
		mod = cl.model_precache[x];

		if (!mod) break;
		mod->modelleaf = NULL;

		if (mod->name[0] != '*') continue;

		psurf = &mod->surfaces[mod->firstmodelsurface];
		numVerts = org[0] = org[1] = org[2] = 0;

		// get the midpoint of the bmodel based on the verts in each surf
		for (i = 0; i < mod->nummodelsurfaces; i++, psurf++)
		{
			// some bmodels have these
			if (psurf->flags & (SURF_DRAWSKY | SURF_DRAWTURB)) continue;

			p = psurf->polys;
			v = p->verts[0];

			for (j = 0; j < p->numverts; j++, v += VERTEXSIZE)
			{
				org[0] += v[0];
				org[1] += v[1];
				org[2] += v[2];

				numVerts++;
			}
		}

		if (!numVerts) continue;

		org[0] /= (float) numVerts;
		org[1] /= (float) numVerts;
		org[2] /= (float) numVerts;

		mod->modelleaf = Mod_PointInLeaf (org, cl.worldmodel);
		mod->origin[0] = org[0];
		mod->origin[1] = org[1];
		mod->origin[2] = org[2];

		// now that we have the info we need, mark the surfs in the model
		psurf = &mod->surfaces[mod->firstmodelsurface];

		for (i = 0; i < mod->nummodelsurfaces; i++, psurf++)
		{
			psurf->caustic = mod->modelleaf->caustic;

			if (mod->modelleaf->contents == CONTENTS_WATER ||
				mod->modelleaf->contents == CONTENTS_SLIME ||
				mod->modelleaf->contents == CONTENTS_LAVA)
			{
				psurf->flags |= SURF_UNDERWATER;
			}
		}
	}
}


// bout 23K... not bad...
// there's never gonna be more than 128 teleports in a map...
#define MAX_MH_PORTALS 128

MHTeleports_t PortalData[MAX_MH_PORTALS];


void Mod_SaveTeleportData (char *entdata)
{
	int i;
	byte targName[1024];
	byte modelName[1024];
	byte key[1024];
	byte value[1024];
	int count;

	memset (targName, 0, sizeof (targName));
	memset (modelName, 0, sizeof (modelName));

	for (i = 0;;)
	{
		memset (key, 0, sizeof (key));
		memset (value, 0, sizeof (value));

		// read to the character after the first "
		for (;; i++)
			if (entdata[i - 1] == '\"') break;

		// read in the key
		for (count = 0;; i++, count++)
		{
			key[count] = entdata[i];
			if (entdata[i + 1] == '\"') break;
		}

		// skip to after the quote
		i += 3;

		// skip to the start of the value
		for (;; i++)
			if (entdata[i - 1] == '\"') break;

		// read in the value
		for (count = 0;; i++, count++)
		{
			value[count] = entdata[i];
			if (entdata[i + 1] == '\"') break;
		}

		// and advance
		i += 3;

		// process the data
		if (!strcmp (key, "target")) strcpy (targName, value);
		if (!strcmp (key, "model")) strcpy (modelName, value);

		// see have we got it all
		if (modelName[0] != '\0' && targName[0] != '\0')
		{
			// save it and get out
			for (count = 0; count < MAX_MH_PORTALS; count++)
			{
				if (!PortalData[count].active)
				{
					// save it
					PortalData[count].active = true;
					strcpy (PortalData[count].targ, targName);
					strcpy (PortalData[count].model, modelName);
					break;
				}
			}

			return;
		}

		// emergency termination condition
		if (entdata[i + 5] == '\0') break;
	}
}


void Mod_SaveDestinationData (char *entdata)
{
	int i;
	byte targName[1024];
	byte angle[1024];
	byte origin[1024];
	byte key[1024];
	byte value[1024];
	int count;
	int		j;
	char	string[128];
	char	*v, *w;

	memset (targName, 0, sizeof (targName));
	memset (angle, 0, sizeof (targName));
	memset (origin, 0, sizeof (targName));

	for (i = 0;;)
	{
		memset (key, 0, sizeof (key));
		memset (value, 0, sizeof (value));

		// read to the character after the first "
		for (;; i++)
			if (entdata[i - 1] == '\"') break;

		// read in the key
		for (count = 0;; i++, count++)
		{
			key[count] = entdata[i];
			if (entdata[i + 1] == '\"') break;
		}

		// skip to after the quote
		i += 3;

		// skip to the start of the value
		for (;; i++)
			if (entdata[i - 1] == '\"') break;

		// read in the value
		for (count = 0;; i++, count++)
		{
			value[count] = entdata[i];
			if (entdata[i + 1] == '\"') break;
		}

		// and advance
		i += 3;

		// process the data
		if (!strcmp (key, "targetname")) strcpy (targName, value);
		if (!strcmp (key, "angle")) strcpy (angle, value);
		if (!strcmp (key, "origin")) strcpy (origin, value);

		// see have we got it all - angle is not obligatory, it defaults to 0 if not present
		if (origin[0] != '\0' && targName[0] != '\0')
		{
			// save it and get out
			for (count = 0; count < MAX_MH_PORTALS; count++)
			{
				if (!PortalData[count].active)
					break;

				if (!strcmp (targName, PortalData[count].targ))
				{
					// save the angle and origin (convert to floats first
					PortalData[count].destangles[0] = 0;
					PortalData[count].destangles[1] = 0;
					PortalData[count].destangles[2] = 0;

					// angle is easy
					if (angle[0] != '\0') PortalData[count].destangles[1] = atof (angle);

					// origin...
					strcpy (string, origin);
					v = string;
					w = string;

					for (j = 0; j < 3; j++)
					{
						while (*v && *v != ' ')
							v++;

						*v = 0;
						PortalData[count].destorigin[j] = atof (w);
						w = v = v + 1;
					}

					break;
				}
			}

			return;
		}

		// emergency termination condition
		if (entdata[i + 5] == '\0') break;
	}
}


void Mod_LinkToStarMods (void)
{
	int x;
	model_t *mod;
	int i, j;

	for (x = 1; x < MAX_MODELS; x++)
	{
		mod = cl.model_precache[x];

		if (!mod) break;
		if (mod->name[0] != '*') continue;

		for (i = 0; i < MAX_MH_PORTALS; i++)
		{
			if (!PortalData[i].active) break;

			if (!strcmp (mod->name, PortalData[i].model))
				mod->portaldata = &PortalData[i];
		}
	}
}


// could be done at load time but the data is only valid for the client so it's done here
void Mod_LinkTeleports (model_t *mod)
{
	int i;
	byte buffer[4096];	// unsafe???
	int count;
	byte *buf;

	if (!mod->entLumpSize) return;

	memset (PortalData, 0, sizeof (PortalData));

	// PASS 1 - identify teleports
	// no-one ever said it was gonna be pretty...
	for (i = 0;;)
	{
		// read to an opening brace
		for (;; i++)
			if (mod->entities[i] == '{') break;

		// init the buffer
		memset (buffer, 0, sizeof (buffer));

		// load into the buffer until we hit a closing brace
		for (count = 0; ; i++, count++)
		{
			buffer[count] = mod->entities[i];
			if (buffer[count] == '}') break;
		}

		// see have we got a trigger teleport in the buffer
		buf = buffer;

		for (;;)
		{
			if (!buf[0]) break;

			if (!strncmp (buf, "trigger_teleport", 16))
			{
				// save out the data - we need the target and the model
				Mod_SaveTeleportData (buffer);

				break;
			}

			buf++;
		}

		// emergency termination condition - if i is getting near to the max size, bomb out
		// 40 is the minimum number of bytes required for a meaningful entry
		if (i > (mod->entLumpSize - 39)) break;
	}

	// PASS 2 - identify teleport targets
	for (i = 0;;)
	{
		// read to an opening brace
		for (;; i++)
			if (mod->entities[i] == '{') break;

		// init the buffer
		memset (buffer, 0, sizeof (buffer));

		// load into the buffer until we hit a closing brace
		for (count = 0; ; i++, count++)
		{
			buffer[count] = mod->entities[i];
			if (buffer[count] == '}') break;
		}

		// see have we got a trigger teleport in the buffer
		buf = buffer;

		for (;;)
		{
			if (!buf[0]) break;

			if (!strncmp (buf, "info_teleport_destination", 25))
			{
				// save out the data - we need the target, the angle and the origin
				Mod_SaveDestinationData (buffer);

				break;
			}

			buf++;
		}

		// emergency termination condition - if i is getting near to the max size, bomb out
		// 40 is the minimum number of bytes required for a meaningful entry
		if (i > (mod->entLumpSize - 39)) break;
	}

	Mod_LinkToStarMods ();
}


/*
===============
R_NewMap
===============
*/
void GL_BuildAllLightmaps (qboolean staticmap);
void GL_TexManMapLoadFinish (void);
extern int uploads;
extern int addSurfs;

void R_NewMap (void)
{
	int i;
	int j;
	dlight_t *dl = cl_dlights;

	memset (cl_dlights, 0, sizeof (cl_dlights));

	// set all dlights to die
	for (i = 0; i < MAX_DLIGHTS; i++, dl++)
	{
		dl->die = -1;
	}

	// make sure lightmaps aren't rebuilt the first time they're seen!!!
	for (i = 0; i < 256; i++)
		d_lightstylevalue[i] = 264;

	uploads = 0;
	addSurfs = 0;

	memset (&r_worldentity, 0, sizeof(r_worldentity));
	r_worldentity.model = cl.worldmodel;

	// clear out efrags in case the level hasn't been reloaded
	// FIXME: is this one short?
	// MH - it is...
	for (i = 0; i < cl.worldmodel->numleafs + 1; i++)
		cl.worldmodel->leafs[i].efrags = NULL;

	r_viewleaf = NULL;
	R_ClearParticles ();

	// this has to be set *before* building the lightmaps or the dlightframe for the
	// surfs will be set incorrectly, resulting in rogue dlights...!
	r_framecount = 1;

	// build static lightstyles first
	GL_BuildAllLightmaps (true);

	// identify underwater * models
	R_IdentifyUnderwater ();

	// put all chains in a known initial state
	for (i = 0; i < cl.worldmodel->numtextures; i++)
	{
		if (!cl.worldmodel->textures[i])
			continue;

		cl.worldmodel->textures[i]->fbChain = NULL;
		cl.worldmodel->textures[i]->lqChain = NULL;
		cl.worldmodel->textures[i]->dtChain = NULL;

		// in theory we *could* use a seperate texture chain for each volume,
		// but volume merging and some astutely placed Con_Printfs indicates
		// that no more than 4 drawing volumes (i.e. volumes after merging) are
		// used in ID1.  I've upped it to 16 to be utterly safe.
		for (j = 0; j < 16; j++)
		{
			cl.worldmodel->textures[i]->texturechain[j] = NULL;
		}
	}

	// set the world fog based on worldtype
	// also, load a generic type skybox if we don't have a specific one for the map
	worldtype = -1;
	worldtype = Mod_FindWorldType (cl.worldmodel);

	switch (worldtype)
	{
	case 1:		// runic red
		world_fog[0] = 0.6;
		world_fog[1] = 0.1;
		world_fog[2] = 0.0;

		break;

	case 2:		// base grey
		world_fog[0] = 0.4;
		world_fog[1] = 0.4;
		world_fog[2] = 0.4;

		break;

	case 0:		// medieval mud
	default:	// y' never know
		world_fog[0] = 0.5;
		world_fog[1] = 0.4;
		world_fog[2] = 0.3;

		break;
	}

	// do some texture management housekeeping
	GL_TexManMapLoadFinish ();

	WorldModelLoaded = false;
	SbarPreload = false;
}


/*
====================
R_TimeRefresh_f

For program optimization
====================
*/
void R_TimeRefresh_f (void)
{
	int			i;
	float		start, stop, time;
	int			startangle;
	vrect_t		vr;

	glDrawBuffer  (GL_FRONT);
	glFinish ();

	start = Sys_FloatTime ();

	for (i=0 ; i<128 ; i++)
	{
		r_refdef.viewangles[1] = i/128.0*360.0;
		R_RenderView ();
	}

	glFinish ();
	stop = Sys_FloatTime ();
	time = stop-start;
	Con_Printf ("%f seconds (%f fps)\n", time, 128/time);

	glDrawBuffer  (GL_BACK);
	GL_EndRendering ();
}

