/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

/*
MHQuake 3D revolving HUD

This whole file is one big fixme - everything is hard-coded!!!
*/

#include "gl_local.h"

typedef struct hudmodel_s
{
	model_t *mod;
	float scale;
	float z;
} hudmodel_t;

hudmodel_t HUD_Health;
hudmodel_t HUD_Armour[4];
hudmodel_t HUD_Ammo;

// have a temp entity for the drawing
entity_t HUD_TempEnt;


float bobjrotate;
float clienttime;

void R_LoadHUDModel (hudmodel_t *hm, char *path)
{
	mmdl_t *paliashdr;
	dtrivertx_t	*verts;
	daliasframe_t *frame;
	int count;
	int *order;
	float v;
	float minz = 99999;
	float maxz = -99999;

	hm->mod = R_RegisterModel (path);
	hm->z = 0;

	if (!hm->mod) return;

	// now we get the average z of the mdl so we can offset it by the midpoint when drawing
	// get the header
	paliashdr = (mmdl_t *) hm->mod->extradata;
	order = (int *) ((byte *) paliashdr + paliashdr->ofs_glcmds);

	// get the frame and verts data
	frame = (daliasframe_t *) ((byte *) paliashdr + paliashdr->ofs_frames);
	verts = frame->verts;

	while (1) 
	{
		// get the vertex count and primitive type
		count = *order++;

		// done
		if (!count) break;

		if (count < 0) 
			count = -count;

		do
		{
			v = ((float) verts[order[2]].v[2] * frame->scale[2] + frame->translate[2]);

			if (v > maxz) maxz = v;
			if (v < minz) minz = v;

			order += 3;
		} while (--count);
	}

	// center vertically
	hm->z = (minz + maxz) / 2;

	VID_Printf (PRINT_DEVELOPER, "MDL %s with z offset %f loaded OK\n", path, hm->z);
}


/*
====================
R_RegisterHUDModels

Register the models we need for the HUD
====================
*/
void R_RegisterHUDModels (void)
{
	R_LoadHUDModel (&HUD_Health, "models/items/adrenal/tris.md2");
	R_LoadHUDModel (&HUD_Armour[0], "models/items/armor/shard/tris.md2");
	R_LoadHUDModel (&HUD_Armour[1], "models/items/armor/jacket/tris.md2");
	R_LoadHUDModel (&HUD_Armour[2], "models/items/armor/body/tris.md2");
	R_LoadHUDModel (&HUD_Armour[3], "models/items/armor/combat/tris.md2");
	R_LoadHUDModel (&HUD_Ammo, "models/items/band/tris.md2");

	// rescaling factors for mdls
	HUD_Health.scale = 1.5;
	HUD_Armour[0].scale = 1.5;
	HUD_Armour[1].scale = 1.25;
	HUD_Armour[2].scale = 1.25;
	HUD_Armour[3].scale = 1.25;
	HUD_Ammo.scale = 1.25;
}


void R_RenderHUDModel (entity_t *e, hudmodel_t *hm)
{
	mmdl_t *paliashdr;
	dtrivertx_t	*verts;
	daliasframe_t *frame;

	if (!hm) return;
	if (!hm->mod) return;
	if (!e) return;

	// get the header
	paliashdr = (mmdl_t *) hm->mod->extradata;

	// get the frame and verts data
	frame = (daliasframe_t *) ((byte *) paliashdr + paliashdr->ofs_frames + e->frame * paliashdr->framesize);
	verts = frame->verts;

    glPushMatrix ();

	// offset by the mdl's z factor so that the vertical centerpoint of the mdl is at the specified z origin
	e->origin[2] -= (hm->z * hm->scale);

	// turn it upside down
	e->origin[2] = vid.height2d - e->origin[2];

	// modelview transformations
	glTranslatef (e->origin[0], -e->origin[1], e->origin[2]);
	glRotatef (e->angles[1],  0, 0, 1);
	glRotatef (-e->angles[0],  0, 1, 0);
	glRotatef (e->angles[2],  1, 0, 0);

	// scale negatively in z cos the orthographic projection will render these mdls upside-down
	glScalef (hm->scale, hm->scale, -hm->scale);

	glActiveTexture (GL_TEXTURE0);
	glBindTexture (GL_TEXTURE_2D, hm->mod->skins[e->skinnum]->texnum);

	glActiveTexture (GL_TEXTURE1);
	glBindTexture (GL_TEXTURE_2D, hm->mod->skins[e->skinnum]->lumatex);

	R_AddHUDMeshToVArray (frame, verts, (int *) ((byte *) paliashdr + paliashdr->ofs_glcmds));

	glPopMatrix ();

	// increase rotation angle so they don't look crap all lined up
	bobjrotate += 63;
}


void R_RenderPlayerMenuModel (void *m, char *skin)
{
	hudmodel_t MDL_Player;

	bobjrotate = anglemod (clienttime * 0.1);

	memset (&HUD_TempEnt, 0, sizeof (HUD_TempEnt));

	HUD_TempEnt.origin[1] = (vid.width2d / 2) + 75;
	HUD_TempEnt.origin[2] = (vid.height2d / 2);

	HUD_TempEnt.angles[1] = bobjrotate;

	// set up hudmodel variables
	MDL_Player.mod = (model_t *) m;
	MDL_Player.scale = 3.5;
	MDL_Player.z = 7;
	MDL_Player.mod->skins[0] = HUD_TempEnt.skin = R_RegisterSkin (skin);

	glPushMatrix ();

	// put Z going up and X going deep to match Quakes main render
	// Y remains at right to left, Z becomes top to bottom and X becomes near to far
	glRotatef (-90, 1, 0, 0);
	glRotatef (90, 0, 0, 1);

	// hack the depth range to prevent sbar models being clipped by world geometry
	glEnable (GL_DEPTH_TEST);
	glDepthRange (gldepthmin, gldepthmin + 0.1 * (gldepthmax - gldepthmin));

	glColor3f (0.8, 0.8, 0.8);
	glDisable (GL_ALPHA_TEST);

	glActiveTexture (GL_TEXTURE0);
	glEnable (GL_TEXTURE_2D);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	glActiveTexture (GL_TEXTURE1);
	glEnable (GL_TEXTURE_2D);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_ADD);

	R_InitVArrays (VERT_MULTI_TEXTURED);
	R_RenderHUDModel (&HUD_TempEnt, &MDL_Player);
	R_KillVArrays ();

	glActiveTexture (GL_TEXTURE1);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glDisable (GL_TEXTURE_2D);

	glActiveTexture (GL_TEXTURE0);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

	glPopMatrix ();

	glEnable (GL_ALPHA_TEST);

	glGetError ();

	// restore the previous depth range
	glDepthRange (gldepthmin, gldepthmax);
	glDisable (GL_DEPTH_TEST);

	glColor3f (1, 1, 1);
}


void R_RenderHUDHealth (void)
{
	memset (&HUD_TempEnt, 0, sizeof (HUD_TempEnt));

	HUD_TempEnt.origin[1] = (vid.width2d / 2) - 180;
	HUD_TempEnt.origin[2] = 25;

	HUD_TempEnt.angles[1] = bobjrotate;

	R_RenderHUDModel (&HUD_TempEnt, &HUD_Health);
}


void R_RenderHUDArmour (int armour)
{
	memset (&HUD_TempEnt, 0, sizeof (HUD_TempEnt));

	HUD_TempEnt.origin[1] = (vid.width2d / 2) - 30;
	HUD_TempEnt.origin[2] = 25;

	HUD_TempEnt.angles[1] = bobjrotate;

	// MH - this is definitely not the right thing to do!!!
	// we should read the armour icon and adjust accordingly
	// trouble is, we don't have the gitem_t struct and itemlist array available to the engine...
	// this modular construction thing will never catch on
	if (armour < 25)
		R_RenderHUDModel (&HUD_TempEnt, &HUD_Armour[0]);
	else if (armour < 50)
		R_RenderHUDModel (&HUD_TempEnt, &HUD_Armour[1]);
	else if (armour < 100)
		R_RenderHUDModel (&HUD_TempEnt, &HUD_Armour[2]);
	else R_RenderHUDModel (&HUD_TempEnt, &HUD_Armour[3]);
}


void R_RenderHUDAmmo (void)
{
	memset (&HUD_TempEnt, 0, sizeof (HUD_TempEnt));

	HUD_TempEnt.origin[1] = (vid.width2d / 2) + 120;
	HUD_TempEnt.origin[2] = 25;

	HUD_TempEnt.angles[1] = bobjrotate;

	R_RenderHUDModel (&HUD_TempEnt, &HUD_Ammo);
}


void R_RenderHUDModels (short *stats)
{
	unsigned int err;

	glGetError ();

	bobjrotate = anglemod (clienttime * 0.1);

	glPushMatrix ();

	// put Z going up and X going deep to match Quakes main render
	// Y remains at right to left, Z becomes top to bottom and X becomes near to far
	glRotatef (-90, 1, 0, 0);
	glRotatef (90, 0, 0, 1);

	// hack the depth range to prevent sbar models being clipped by world geometry
	glEnable (GL_DEPTH_TEST);
	glDepthRange (gldepthmin, gldepthmin + 0.1 * (gldepthmax - gldepthmin));

	glColor3f (0.8, 0.8, 0.8);
	glDisable (GL_ALPHA_TEST);

	glActiveTexture (GL_TEXTURE0);
	glEnable (GL_TEXTURE_2D);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	glActiveTexture (GL_TEXTURE1);
	glEnable (GL_TEXTURE_2D);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_ADD);

	R_InitVArrays (VERT_MULTI_TEXTURED);

	if (stats[STAT_HEALTH]) R_RenderHUDHealth ();
	if (stats[STAT_ARMOR]) R_RenderHUDArmour (stats[STAT_ARMOR]);
	if (stats[STAT_AMMO]) R_RenderHUDAmmo ();

	R_KillVArrays ();

	glActiveTexture (GL_TEXTURE1);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glDisable (GL_TEXTURE_2D);

	glActiveTexture (GL_TEXTURE0);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

	glPopMatrix ();

	glEnable (GL_ALPHA_TEST);

	err = glGetError ();

	// clear down any errors
	if (err) VID_Printf (PRINT_DEVELOPER, "glGetError: %i\n", err);

	// restore the previous depth range
	glDepthRange (gldepthmin, gldepthmax);
	glDisable (GL_DEPTH_TEST);

	glColor3f (1, 1, 1);
}
