/*
Copyright (C) 1997 - 2001 Id Software, Inc.

This program is free software; you can redistribute it and / or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111 - 1307, USA.

*/ 


/*
MH - this whole file is one big "fixme".

it was crudely hacked to remove the software options from visibility in the menus, but much of
the backend support code for them is still here.
*/

#include "client.h"
#include "qmenu.h"

#define REF_OPENGL 1

extern cvar_t *vid_ref;
extern cvar_t *vid_fullscreen;
extern cvar_t *vid_gamma;
extern cvar_t *lm_gamma;
extern cvar_t *scr_viewsize;

static cvar_t *gl_mode;
static cvar_t *gl_driver;
static cvar_t *gl_picmip;
static cvar_t *gl_ext_palettedtexture;
static cvar_t *gl_finish;

static cvar_t *gl_dot3bump;

static cvar_t *gl_lumamode = NULL;

static cvar_t *sw_mode;
static cvar_t *sw_stipplealpha;

extern void M_ForceMenuOff (void);

/*
 ==================================================================== 

MENU INTERACTION

 ==================================================================== 
*/ 
#define OPENGL_MENU 1

static menuframework_s	s_opengl_menu;
static menuframework_s *s_current_menu;
static int				s_current_menu_index;

static menulist_s		s_mode_list[2];
static menulist_s		s_ref_list[2];
//static menuslider_s		s_tq_slider;
static menuslider_s		s_screensize_slider[2];
static menuslider_s		s_brightness_slider[2];
static menuslider_s		s_lmbrightness_slider[2];
static menulist_s		s_bump_slider;
static menulist_s		s_lumamode_slider;
static menulist_s 		s_fs_box[2];
static menulist_s 		s_stipple_box;
static menulist_s 		s_finish_box;
static menuaction_s		s_cancel_action[2];
static menuaction_s		s_defaults_action[2];

static void DriverCallback (void *unused) 
{
	s_current_menu = &s_opengl_menu;
	s_current_menu_index = 1;
}

static void ScreenSizeCallback (void *s) 
{
	menuslider_s *slider = (menuslider_s *) s;

	Cvar_SetValue ("viewsize", slider->curvalue * 10);
}

static void BrightnessCallback (void *s) 
{
	menuslider_s *slider = (menuslider_s *) s;

	s_brightness_slider[0].curvalue = s_brightness_slider[1].curvalue;

	if (_stricmp (vid_ref->string, "soft") == 0) 
	{
		float gamma = (0.8 - (slider->curvalue / 10.0 - 0.5)) + 0.5;

		Cvar_SetValue ("vid_gamma", gamma);
	}
}


static void LMBrightnessCallback (void *s)
{
	menuslider_s *slider = (menuslider_s *) s;

	s_lmbrightness_slider[0].curvalue = s_lmbrightness_slider[1].curvalue;

	if (_stricmp (vid_ref->string, "soft") == 0) 
	{
		float gamma = (0.8 - (slider->curvalue / 10.0 - 0.5)) + 0.5;

		Cvar_SetValue ("lm_gamma", gamma);
	}
}


static void ResetDefaults (void *unused) 
{
	VID_MenuInit ();
}

static void ApplyChanges (void *unused) 
{
	float gamma;

	/*
	** make values consistent
	*/ 
	s_fs_box[!s_current_menu_index].curvalue = s_fs_box[s_current_menu_index].curvalue;
	s_brightness_slider[!s_current_menu_index].curvalue = s_brightness_slider[s_current_menu_index].curvalue;
	s_lmbrightness_slider[!s_current_menu_index].curvalue = s_lmbrightness_slider[s_current_menu_index].curvalue;
	s_ref_list[!s_current_menu_index].curvalue = s_ref_list[s_current_menu_index].curvalue;

	/*
	** invert sense so greater = brighter, and scale to a range of 0.5 to 1.3
	*/ 
	gamma = (0.8 - (s_brightness_slider[s_current_menu_index].curvalue / 10.0 - 0.5)) + 0.5;

	Cvar_SetValue ("vid_gamma", gamma);

	gamma = (0.8 - (s_lmbrightness_slider[s_current_menu_index].curvalue / 10.0 - 0.5)) + 0.5;

	Cvar_SetValue ("lm_gamma", gamma);

	Cvar_SetValue ("sw_stipplealpha", s_stipple_box.curvalue);
	//Cvar_SetValue ("gl_picmip", 3 - s_tq_slider.curvalue);
	Cvar_SetValue ("vid_fullscreen", s_fs_box[s_current_menu_index].curvalue);
	Cvar_SetValue ("gl_finish", s_finish_box.curvalue);
	Cvar_SetValue ("gl_mode", s_mode_list[OPENGL_MENU].curvalue);
	Cvar_SetValue ("gl_lumamode", s_lumamode_slider.curvalue);
	Cvar_SetValue ("gl_dot3bump", s_bump_slider.curvalue);

	Cvar_Set ("vid_ref", "gl");
	Cvar_Set ("gl_driver", "opengl32");

	// initially unmodified
	vid_ref->modified = false;
	gl_driver->modified = false;

	/*
	** update appropriate stuff if we're running OpenGL and gamma
	** has been modified
	*/ 
	if (_stricmp (vid_ref->string, "gl") == 0) 
	{
		if (vid_gamma->modified) vid_ref->modified = true;

		if (lm_gamma->modified) vid_ref->modified = true;

		// this can never happen now...
		// if (gl_driver->modified) vid_ref->modified = true;
	}

	M_ForceMenuOff ();
}

static void CancelChanges (void *unused) 
{
	extern void M_PopMenu (void);

	M_PopMenu ();
}

/*
** VID_MenuInit
*/ 
void VID_MenuInit (void) 
{
	extern char *resolutionlist[];

	static const char *yesno_names[] = 
	{
		"no",
		"yes",
		0
	};

	static const char *lumamode_names[] =
	{
		"diffuse only",
		"luminescense",
		"specular",
		"light blooms",
		0
	};

	static const char *bump_names[] =
	{
		"off",
		"on",
		"test mode",
		0
	};

	int i;

	if (!gl_driver) gl_driver = Cvar_Get ("gl_driver", "opengl32", 0);
	if (!gl_picmip) gl_picmip = Cvar_Get ("gl_picmip", "0", 0);
	if (!gl_mode) gl_mode = Cvar_Get ("gl_mode", "3", 0);
	if (!sw_mode) sw_mode = Cvar_Get ("sw_mode", "0", 0);
	if (!gl_ext_palettedtexture) gl_ext_palettedtexture = Cvar_Get ("gl_ext_palettedtexture", "1", CVAR_ARCHIVE);
	if (!gl_finish) gl_finish = Cvar_Get ("gl_finish", "0", CVAR_ARCHIVE);
	if (!sw_stipplealpha) sw_stipplealpha = Cvar_Get ("sw_stipplealpha", "0", CVAR_ARCHIVE);
	if (!gl_lumamode) gl_lumamode = Cvar_Get ("gl_lumamode", "0", CVAR_ARCHIVE);
	if (!gl_dot3bump) gl_dot3bump = Cvar_Get ("gl_dot3bump", "0", CVAR_ARCHIVE);

	s_mode_list[OPENGL_MENU].curvalue = gl_mode->value;

	if (!scr_viewsize) scr_viewsize = Cvar_Get ("viewsize", "100", CVAR_ARCHIVE);

	s_screensize_slider[OPENGL_MENU].curvalue = scr_viewsize->value / 10;

	s_current_menu_index = OPENGL_MENU;

	s_ref_list[s_current_menu_index].curvalue = REF_OPENGL;

	s_opengl_menu.x = viddef.width2d * 0.50;
	s_opengl_menu.nitems = 0;

	for (i = 0; i < 2; i++) 
	{
		s_mode_list[i].generic.type = MTYPE_SPINCONTROL;
		s_mode_list[i].generic.name = "video mode";
		s_mode_list[i].generic.x = 0;
		s_mode_list[i].generic.y = 0;
		s_mode_list[i].itemnames = resolutionlist; //resolutions;

		s_screensize_slider[i].generic.type	 = MTYPE_SLIDER;
		s_screensize_slider[i].generic.x		 = 0;
		s_screensize_slider[i].generic.y		 = 10;
		s_screensize_slider[i].generic.name	 = "screen size";
		s_screensize_slider[i].minvalue = 3;
		s_screensize_slider[i].maxvalue = 12;
		s_screensize_slider[i].generic.callback = ScreenSizeCallback;

		s_brightness_slider[i].generic.type	 = MTYPE_SLIDER;
		s_brightness_slider[i].generic.x	 = 0;
		s_brightness_slider[i].generic.y	 = 20;
		s_brightness_slider[i].generic.name	 = "texture gamma";
		s_brightness_slider[i].generic.callback = BrightnessCallback;
		s_brightness_slider[i].minvalue = 5;
		s_brightness_slider[i].maxvalue = 13;
		s_brightness_slider[i].curvalue = (1.3 - vid_gamma->value + 0.5) * 10;

		s_lmbrightness_slider[i].generic.type	 = MTYPE_SLIDER;
		s_lmbrightness_slider[i].generic.x	 = 0;
		s_lmbrightness_slider[i].generic.y	 = 30;
		s_lmbrightness_slider[i].generic.name	 = "lightmap gamma";
		s_lmbrightness_slider[i].generic.callback = LMBrightnessCallback;
		s_lmbrightness_slider[i].minvalue = 5;
		s_lmbrightness_slider[i].maxvalue = 13;
		s_lmbrightness_slider[i].curvalue = (1.3 - lm_gamma->value + 0.5) * 10;

		s_fs_box[i].generic.type = MTYPE_SPINCONTROL;
		s_fs_box[i].generic.x	 = 0;
		s_fs_box[i].generic.y	 = 40;
		s_fs_box[i].generic.name	 = "fullscreen";
		s_fs_box[i].itemnames = yesno_names;
		s_fs_box[i].curvalue = vid_fullscreen->value;

		s_defaults_action[i].generic.type = MTYPE_ACTION;
		s_defaults_action[i].generic.name = "reset to defaults";
		s_defaults_action[i].generic.x = 0;
		s_defaults_action[i].generic.y = 80;
		s_defaults_action[i].generic.callback = ResetDefaults;

		s_cancel_action[i].generic.type = MTYPE_ACTION;
		s_cancel_action[i].generic.name = "cancel";
		s_cancel_action[i].generic.x = 0;
		s_cancel_action[i].generic.y = 90;
		s_cancel_action[i].generic.callback = CancelChanges;
	}

	s_bump_slider.generic.type = MTYPE_SPINCONTROL;
	s_bump_slider.generic.x	 = 0;
	s_bump_slider.generic.y	 = 50;
	s_bump_slider.generic.name	 = "bumpmapping";
	s_bump_slider.itemnames = bump_names;
	s_bump_slider.curvalue = gl_dot3bump->value;

	s_lumamode_slider.generic.type = MTYPE_SPINCONTROL;
	s_lumamode_slider.generic.x	 = 0;
	s_lumamode_slider.generic.y	 = 60;
	s_lumamode_slider.generic.name	 = "luma mode";
	s_lumamode_slider.itemnames = lumamode_names;
	s_lumamode_slider.curvalue = gl_lumamode->value;

	s_finish_box.generic.type = MTYPE_SPINCONTROL;
	s_finish_box.generic.x	 = 0;
	s_finish_box.generic.y	 = 70;
	s_finish_box.generic.name	 = "sync every frame";
	s_finish_box.curvalue = gl_finish->value;
	s_finish_box.itemnames = yesno_names;

	// Menu_AddItem (&s_opengl_menu, (void *) &s_ref_list[OPENGL_MENU]);
	Menu_AddItem (&s_opengl_menu, (void *) &s_mode_list[OPENGL_MENU]);
	Menu_AddItem (&s_opengl_menu, (void *) &s_screensize_slider[OPENGL_MENU]);
	Menu_AddItem (&s_opengl_menu, (void *) &s_brightness_slider[OPENGL_MENU]);
	Menu_AddItem (&s_opengl_menu, (void *) &s_lmbrightness_slider[OPENGL_MENU]);
	Menu_AddItem (&s_opengl_menu, (void *) &s_fs_box[OPENGL_MENU]);
	Menu_AddItem (&s_opengl_menu, (void *) &s_bump_slider);
	Menu_AddItem (&s_opengl_menu, (void *) &s_lumamode_slider);
	Menu_AddItem (&s_opengl_menu, (void *) &s_finish_box);

	Menu_AddItem (&s_opengl_menu, (void *) &s_defaults_action[OPENGL_MENU]);
	Menu_AddItem (&s_opengl_menu, (void *) &s_cancel_action[OPENGL_MENU]);

	Menu_Center (&s_opengl_menu);
	s_opengl_menu.x -= 8;
}

/*
 ================ 
VID_MenuDraw
 ================ 
*/ 
void VID_MenuDraw (void) 
{
	int w, h;

	s_current_menu = &s_opengl_menu;

	/*
	** draw the banner
	*/ 
	Draw_GetPicSize (&w, &h, "m_banner_video");
	Draw_Pic (viddef.width2d / 2 - w / 2, viddef.height2d / 2 - 110, "m_banner_video");

	/*
	** move cursor to a reasonable starting position
	*/ 
	Menu_AdjustCursor (s_current_menu, 1);

	/*
	** draw the menu
	*/ 
	Menu_Draw (s_current_menu);
}

/*
 ================ 
VID_MenuKey
 ================ 
*/ 
const char *VID_MenuKey (int key) 
{
	menuframework_s *m = s_current_menu;
	static const char *sound = "misc/menu1.wav";

	switch (key) 
	{
	case K_ESCAPE:
		ApplyChanges (0);
		return NULL;
	case K_KP_UPARROW:
	case K_UPARROW:
		m->cursor--;
		Menu_AdjustCursor (m, - 1);
		break;
	case K_KP_DOWNARROW:
	case K_DOWNARROW:
		m->cursor++;
		Menu_AdjustCursor (m, 1);
		break;
	case K_KP_LEFTARROW:
	case K_LEFTARROW:
		Menu_SlideItem (m, - 1);
		break;
	case K_KP_RIGHTARROW:
	case K_RIGHTARROW:
		Menu_SlideItem (m, 1);
		break;
	case K_KP_ENTER:
	case K_ENTER:
		if (!Menu_SelectItem (m)) 
			ApplyChanges (NULL);
		break;
	}

	return sound;
}


