/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// gl_warp.c -- sky and water polygons

#include "gl_state.h"

// needed for sky
#include "resource.h"

extern	model_t	*loadmodel;

int		skytexturenum;

gltexture_t solidskytexture;
gltexture_t alphaskytexture;

msurface_t	*warpface;

#define GL_SUBDIVIDE_SIZE 64

int spherelist;
int spherelist2;
float sky_fog[4];


// gl_rsurf.c imports
extern float sintable[];
extern int STBitMask;
extern float STMult;
extern qboolean waterthisframe;
extern polyverts_t vertlist[];
extern gltexture_t *TextureList;


void BoundPoly (int numverts, float *verts, vec3_t mins, vec3_t maxs)
{
	int		i, j;
	float	*v;

	mins[0] = mins[1] = mins[2] = 9999;
	maxs[0] = maxs[1] = maxs[2] = -9999;
	v = verts;
	for (i=0 ; i<numverts ; i++)
		for (j=0 ; j<3 ; j++, v++)
		{
			if (*v < mins[j])
				mins[j] = *v;
			if (*v > maxs[j])
				maxs[j] = *v;
		}
}


static vec3_t surfmins;
static vec3_t surfmaxs;

void SubdividePolygon (int numverts, float *verts)
{
	int		i, j, k;
	vec3_t	mins, maxs;
	float	m;
	float	*v;
	vec3_t	front[64], back[64];
	int		f, b;
	float	dist[64];
	float	frac;
	glpoly_t	*poly;
	float	s, t;

	if (numverts > 60)
		Sys_Error ("numverts = %i", numverts);

	BoundPoly (numverts, verts, mins, maxs);

	for (i=0 ; i<3 ; i++)
	{
		m = (mins[i] + maxs[i]) * 0.5;
		m = GL_SUBDIVIDE_SIZE * floor (m/GL_SUBDIVIDE_SIZE + 0.5);
		if (maxs[i] - m < 8)
			continue;
		if (m - mins[i] < 8)
			continue;

		// cut it
		v = verts + i;
		for (j=0 ; j<numverts ; j++, v+= 3)
			dist[j] = *v - m;

		// wrap cases
		dist[j] = dist[0];
		v-=i;
		VectorCopy (verts, v);

		f = b = 0;
		v = verts;
		for (j=0 ; j<numverts ; j++, v+= 3)
		{
			if (dist[j] >= 0)
			{
				VectorCopy (v, front[f]);
				f++;
			}
			if (dist[j] <= 0)
			{
				VectorCopy (v, back[b]);
				b++;
			}
			if (dist[j] == 0 || dist[j+1] == 0)
				continue;
			if ( (dist[j] > 0) != (dist[j+1] > 0) )
			{
				// clip point
				frac = dist[j] / (dist[j] - dist[j+1]);
				for (k=0 ; k<3 ; k++)
					front[f][k] = back[b][k] = v[k] + frac*(v[3+k] - v[k]);
				f++;
				b++;
			}
		}

		SubdividePolygon (f, front[0]);
		SubdividePolygon (b, back[0]);
		return;
	}

	poly = Hunk_Alloc (sizeof (glpoly_t));
	poly->next = warpface->polys;
	warpface->polys = poly;
	poly->numverts = numverts;

	// set the vertex pointer
	poly->verts = &vertlist[allocverts];
	poly->vertindex = allocverts;

	// increase the vertex count
	allocverts += numverts;

	// set the initial index of the water verts.  there's nothing for it to
	// point to yet, but we'll fix that up in a bit...
	poly->waterindex = watervertcount;

	// increase the count of water verts too
	watervertcount += numverts;

	poly->midpoint[0] = poly->midpoint[1] = poly->midpoint[2] = 0;

	for (i = 0; i < numverts; i++, verts += 3)
	{
		poly->midpoint[0] += verts[0];
		poly->midpoint[1] += verts[1];
		poly->midpoint[2] += verts[2];

		// set bbox for the surf
		// useful to have for testing and stuff...
		if (verts[0] > surfmaxs[0]) surfmaxs[0] = verts[0];
		if (verts[1] > surfmaxs[1]) surfmaxs[1] = verts[1];
		if (verts[2] > surfmaxs[2]) surfmaxs[2] = verts[2];

		if (verts[0] < surfmins[0]) surfmins[0] = verts[0];
		if (verts[1] < surfmins[1]) surfmins[1] = verts[1];
		if (verts[2] < surfmins[2]) surfmins[2] = verts[2];

		VectorCopy (verts, poly->verts[i].v);
		s = DotProduct (verts, warpface->texinfo->vecs[0]);
		t = DotProduct (verts, warpface->texinfo->vecs[1]);
		poly->verts[i].st[0] = s;
		poly->verts[i].st[1] = t;
	}

	poly->midpoint[0] /= (float) numverts;
	poly->midpoint[1] /= (float) numverts;
	poly->midpoint[2] /= (float) numverts;
}

/*
================
GL_SubdivideSurface

Breaks a polygon up along axial 64 unit
boundaries so that turbulent and sky warps
can be done reasonably.
================
*/
void GL_SubdivideSurface (model_t *mod, msurface_t *surf)
{
	vec3_t		verts[64];
	int			numverts;
	int			i;
	int			lindex;
	float		*vec;
	texture_t	*t;

	warpface = surf;

	// convert edges back to a normal polygon
	numverts = 0;

	surfmins[0] = 99999999;
	surfmins[1] = 99999999;
	surfmins[2] = 99999999;

	surfmaxs[0] = -99999999;
	surfmaxs[1] = -99999999;
	surfmaxs[2] = -99999999;

	for (i = 0; i < surf->numedges; i++)
	{
		lindex = mod->surfedges[surf->firstedge + i];

		if (lindex > 0)
			vec = mod->vertexes[mod->edges[lindex].v[0]].position;
		else vec = mod->vertexes[mod->edges[-lindex].v[1]].position;

		VectorCopy (vec, verts[numverts]);
		numverts++;
	}

	SubdividePolygon (numverts, verts[0]);

	// increase the maxs for lava surf bboxes
	// (only place they're used for liquids)
	surfmaxs[2] += 256;

	// store out the completed bbox
	VectorCopy (surfmins, surf->mins);
	VectorCopy (surfmaxs, surf->maxs);
}

//=========================================================


/*
================
R_DrawWaterSurfaces
================
*/
void R_DrawWaterSurfaces (void)
{
	gltexture_t *tex;
	msurface_t *surf;
	glpoly_t *p;
	int i;
	float sadd, tadd;
	float *v;

	// drawing parameters
	vector_t warp1scale;
	vector_t warp2scale;
	float wateralpha;
	float warp2sign;

	GLenum WARP1_TEXENV;
	GLenum WARP2_TEXENV;

	// only teleports visible
	qboolean teleonly = true;

	int Warp1Texture;
	int Warp2Texture;

	// the waterverts are an array of structs of floats, so this operation is valid
	float *waterarray = (float *) waterverts;

	if (!waterthisframe) return;

	// back to the world matrix
	glLoadMatrixf (cl_entities[0].MViewMatrix);

	// enable blending
	glEnable (GL_BLEND);
	glDepthMask (GL_FALSE);

	glActiveTexture (GL_TEXTURE1);
	glEnable (GL_TEXTURE_2D);

	// enable vertex pointer - index 0 is the regular verts
	glEnableClientState (GL_VERTEX_ARRAY);
	glVertexPointer (3, GL_FLOAT, sizeof (waterverts_t), &waterarray[0]);

	// enable texcoord pointer for textures - index 3 is the st coords for warp 1
	glClientActiveTexture (GL_TEXTURE0);
	glEnableClientState (GL_TEXTURE_COORD_ARRAY);
	glTexCoordPointer (2, GL_FLOAT, sizeof (waterverts_t), &waterarray[3]);

	// enable texcoord pointer for textures - index 5 is the st coords for warp 2
	glClientActiveTexture (GL_TEXTURE1);
	glEnableClientState (GL_TEXTURE_COORD_ARRAY);
	glTexCoordPointer (2, GL_FLOAT, sizeof (waterverts_t), &waterarray[5]);

	// go through the texture list
	for (tex = TextureList; tex; tex = tex->TextureList)
	{
		// nothing chained
		if (!(surf = tex->texturechain)) continue;

		// it ain't liquid
		if (!(surf->flags & SURF_DRAWTURB)) continue;

		// setup rendering parameters
		// these are actually absolutely excellent
		if (surf->flags & SURF_DRAWLAVA)
		{
			glDisable (GL_BLEND);
			glDepthMask (GL_TRUE);

			// because this layer isn't visible, give it a massive size to make rendering it less expensive
			// (actually this doesn't matter cos the texture matrix is ignored here...)
			warp1scale.vec.x = 1.0f / 1024.0f;
			warp1scale.vec.y = 1.0f / 1024.0f;
			warp1scale.vec.z = 1.0f;

			wateralpha = 1.0;

			warp2scale.vec.x = 1.0f / 128.0f;
			warp2scale.vec.y = 1.0f / 128.0f;
			warp2scale.vec.z = 1.0f;

			warp2sign = 1;

			// special - instruct the renderer to ignore fancy stuff here
			WARP1_TEXENV = GL_ZERO;

			// make the 2nd warp a replace texenv also to avoid dual layer madness with lava
			WARP2_TEXENV = GL_REPLACE;

			// we could use any old crap for warp1 here..
			Warp1Texture = surf->texinfo->texture->gl_texture->gltexnum;
			Warp2Texture = surf->texinfo->texture->gl_texture->gltexnum;

			teleonly = false;
		}
		else if (surf->flags & SURF_DRAWTELE)
		{
			glDisable (GL_BLEND);
			glDepthMask (GL_TRUE);

			warp1scale.vec.x = 1.0f / 64.0f;
			warp1scale.vec.y = 1.0f / 64.0f;
			warp1scale.vec.z = 1.0f;

			wateralpha = 1.0;

			warp2scale.vec.x = 1.0f / 24.0f;
			warp2scale.vec.y = 1.0f / 24.0f;
			warp2scale.vec.z = 1.0f;

			warp2sign = 1;

			WARP1_TEXENV = GL_REPLACE;
			WARP2_TEXENV = GL_ADD;

			Warp1Texture = surf->texinfo->texture->gl_texture->gltexnum;
			Warp2Texture = surf->texinfo->texture->gl_texture->gltexnum;
		}
		else if (surf->flags & SURF_DRAWSLIME)
		{
			glEnable (GL_BLEND);
			glDepthMask (GL_FALSE);

			warp1scale.vec.x = 1.0f / 128.0f;
			warp1scale.vec.y = 1.0f / 128.0f;
			warp1scale.vec.z = 1.0f;

			wateralpha = 0.75;

			warp2scale.vec.x = 1.0f / 256.0f;
			warp2scale.vec.y = 1.0f / 256.0f;
			warp2scale.vec.z = 1.0f;

			warp2sign = -1;

			WARP1_TEXENV = GL_MODULATE;
			WARP2_TEXENV = GL_ADD;

			Warp1Texture = surf->texinfo->texture->gl_texture->gltexnum;
			Warp2Texture = surf->texinfo->texture->gl_texture->gltexnum;

			teleonly = false;
		}
		else
		{
			glEnable (GL_BLEND);
			glDepthMask (GL_FALSE);

			warp1scale.vec.x = 1.0f / 128.0f;
			warp1scale.vec.y = 1.0f / 128.0f;
			warp1scale.vec.z = 1.0f;

			wateralpha = 0.5;

			warp2scale.vec.x = 1.0f / 256.0f;
			warp2scale.vec.y = 1.0f / 256.0f;
			warp2scale.vec.z = 1.0f;

			warp2sign = 1;

			WARP1_TEXENV = GL_MODULATE;
			WARP2_TEXENV = GL_ADD;

			Warp1Texture = surf->texinfo->texture->gl_texture->gltexnum;
			Warp2Texture = surf->texinfo->texture->gl_texture->gltexnum;

			teleonly = false;
		}

		// apply parameters to tmu 0
		glActiveTexture (GL_TEXTURE0);
		if (WARP1_TEXENV != GL_ZERO)
		{
			// only do this if we're doing dual layer, otherwise it's cheaper to do a very basic render than it
			// would be to change TMU states
			glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, WARP1_TEXENV);
			glMatrixMode (GL_TEXTURE);
			glLoadIdentity ();
			glScalefv (warp1scale.vecs);
			glMatrixMode (GL_MODELVIEW);
		}
		glBindTexture (GL_TEXTURE_2D, Warp1Texture);

		// apply parameters to tmu 1
		glActiveTexture (GL_TEXTURE1);
		glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, WARP2_TEXENV);
		glMatrixMode (GL_TEXTURE);
		glLoadIdentity ();
		glScalefv (warp2scale.vecs);
		glMatrixMode (GL_MODELVIEW);
		glBindTexture (GL_TEXTURE_2D, Warp2Texture);

		// set the colour
		glColor4f (1, 1, 1, wateralpha);

		// go through the surfs in the chain
		for (; surf; surf = surf->texturechain)
		{
			// water surfs have a chained list of polys rather than just 1
			for (p = surf->polys; p; p = p->next)
			{
				// need to go through each vert in the list to set the warped texcoords
				for (i = 0; i < p->numverts; i++)
				{
					// get a pointer to the verts to keep the code more readable
					v = p->water[i].v;

					// calculate the amount to add for the warp
					sadd = sintable[(int) ((v[surf->warpindex[1]] * 0.025 + realtime) * STMult) & STBitMask];
					tadd = sintable[(int) ((v[surf->warpindex[0]] * 0.025 + realtime) * STMult) & STBitMask];

					// set the warp verts
					// fix water warping - derive the texcoords from the verts!!!  mad, but it works!!!
					// warpindex is precalculated in gl_model.c - use the x2 sintable for 8 unit warps
					p->water[i].warp1[0] = v[surf->warpindex[0]] + sadd;
					p->water[i].warp1[1] = v[surf->warpindex[1]] + tadd;

					// warp 2
					p->water[i].warp2[0] = (v[surf->warpindex[0]] * warp2sign) + sadd;
					p->water[i].warp2[1] = (v[surf->warpindex[1]] * warp2sign) + tadd;
				}

				// draw the vertex array
				glDrawArrays (GL_TRIANGLE_FAN, p->waterindex, p->numverts);
			}
		}
	}

	// restore state
	glActiveTexture (GL_TEXTURE1);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glMatrixMode (GL_TEXTURE);
	glLoadIdentity ();
	glMatrixMode (GL_MODELVIEW);
	glDisable (GL_TEXTURE_2D);

	glActiveTexture (GL_TEXTURE0);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glMatrixMode (GL_TEXTURE);
	glLoadIdentity ();
	glMatrixMode (GL_MODELVIEW);

	// kill vertex arrays
	glClientActiveTexture (GL_TEXTURE1);
	glDisableClientState (GL_TEXTURE_COORD_ARRAY);
	glClientActiveTexture (GL_TEXTURE0);
	glDisableClientState (GL_TEXTURE_COORD_ARRAY);

	glColor4f (1, 1, 1, 1);

	// draw fog if we're underwater
	if (r_viewleaf->contents != CONTENTS_EMPTY && !teleonly)
	{
		float fogdist;
		float fogalpha;

		// if teleports or lava are drawn last they'll leave blending disabled coming in here, so
		// make a potentially redundant state change.  it's only once per frame, so live with it.
		glEnable (GL_BLEND);
		glDepthMask (GL_FALSE);

		glEnableClientState (GL_COLOR_ARRAY);
		glColorPointer (4, GL_FLOAT, sizeof (waterverts_t), &waterarray[7]);

		// turn off texturing and ensure we have smooth shading
		glDisable (GL_TEXTURE_2D);
		glShadeModel (GL_SMOOTH);

		// default blendfunc for this
		glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

		// go through the texture list
		for (tex = TextureList; tex; tex = tex->TextureList)
		{
			// nothing chained
			if (!(surf = tex->texturechain)) continue;

			// it ain't liquid
			if (!(surf->flags & SURF_DRAWTURB)) continue;

			// don't put fog on teleports
			if (surf->flags & SURF_DRAWTELE) continue;

			// go through the surfs in the chain
			for (; surf; surf = surf->texturechain)
			{
				// water surfs have a chained list of polys rather than just 1
				for (p = surf->polys; p; p = p->next)
				{
					// need to go through each vert in the list to set the colours
					for (i = 0; i < p->numverts; i++)
					{
						// get a pointer to the verts to keep the code more readable
						v = p->water[i].v;

						p->water[i].rgba[0] = surf->texinfo->texture->gl_texture->rgb[0];
						p->water[i].rgba[1] = surf->texinfo->texture->gl_texture->rgb[1];
						p->water[i].rgba[2] = surf->texinfo->texture->gl_texture->rgb[2];

						if (surf->flags & SURF_DRAWLAVA)
						{
							// constant fog
							fogalpha = FOG_ALPHA_LAVA_CONST;
						}
						else if (surf->flags & SURF_DRAWSLIME)
						{
							// constant fog
							fogalpha = FOG_ALPHA_SLIME_CONST;
						}
						else
						{
							// distance fog
							// use the transformed verts to get the dist
							fogdist = (v[0] - r_origin[0]) * vpn[0] + (v[1] - r_origin[1]) * vpn[1] + (v[2] - r_origin[2]) * vpn[2];

							if (fogdist < 0)
								fogdist = 0;
							else if (fogdist > FOG_RANGE)
								fogdist = FOG_RANGE;

							fogalpha = 0.3 + (fogdist / FOG_RANGE * FOG_ALPHA_MAX) / 2;

							// boost alpha a little cos the translucent water can make it hard to see
							fogalpha *= 1.25;
						}

						p->water[i].rgba[3] = fogalpha;
					}

					// draw the vertex array
					glDrawArrays (GL_TRIANGLE_FAN, p->waterindex, p->numverts);
				}
			}
		}

		// bring texturing back
		glEnable (GL_TEXTURE_2D);

		// paranoia!!!
		glColor4f (1, 1, 1, 1);

		// now bring vertex arrays back to the way they were.
		glDisableClientState (GL_COLOR_ARRAY);
	}

	glDisableClientState (GL_VERTEX_ARRAY);

	// potentially redundant if teleports or lava were drawn last and we're not underwater.
	// it's only once per frame so live with it.
	glDisable (GL_BLEND);
	glDepthMask (GL_TRUE);
}


//===============================================================


cvar_t r_farclip = {"r_farclip", "16384", true};
cvar_t r_skyspeed = {"r_skyspeed", "10", true};

float rotateBack = 0;
float rotateFore = 0;

void R_DrawSkyPickup (qboolean skydraw)
{
	// sky scaling
	float fullscale;
	float halfscale;
	float reducedfull;
	float reducedhalf;

	// do these calcs even if we're not drawing
	// sky rotation
	// bound rotation speed 0 to 100
	if (r_skyspeed.value < 0) r_skyspeed.value = 0;
	if (r_skyspeed.value > 100) r_skyspeed.value = 100;

	// always rotate even if we're paused!!!
	rotateBack = anglemod (realtime * (2.5 * r_skyspeed.value));
	rotateFore = anglemod (realtime * (4.0 * r_skyspeed.value));

	if (!skydraw) return;

	// calculate the scales
	fullscale = r_farclip.value / 4096;
	halfscale = r_farclip.value / 8192;
	reducedfull = (r_farclip.value / 4096) * 0.9;
	reducedhalf = (r_farclip.value / 8192) * 0.9;

	// turn on fogging.  fog looks good on the skies - it gives them a more 
	// "airy" far-away look, and has the knock-on effect of preventing the 
	// old "texture distortion at the poles" problem.
	glFogi (GL_FOG_MODE, GL_LINEAR);
	glFogfv (GL_FOG_COLOR, sky_fog);
	glFogf (GL_FOG_START, -r_farclip.value);

	// must tweak the fog end too!!!
	glFogf (GL_FOG_END, r_farclip.value);
	glEnable (GL_FOG);

	// sky texture scaling
	// previous releases made a tiled version of the sky texture.  here i just shrink it using the
	// texture matrix, for much the same effect
	glMatrixMode (GL_TEXTURE);
	glLoadIdentity ();
	glScalef (2, 1, 1);
	glMatrixMode (GL_MODELVIEW);

	// background
	// ==========
	// go to a new matrix
	glPushMatrix ();

	// center it on the players position
	glTranslatef (r_origin[0], r_origin[1], r_origin[2]);

	// flatten the sphere
	glScalef (fullscale, fullscale, halfscale);

	// orient it so that the poles are unobtrusive
	glRotatef (-90, 1, 0, 0);

	// make it not always at right angles to the player
	glRotatef (-22, 0 ,1, 0);

	// rotate it around the poles
	glRotatef (-rotateBack, 0, 0, 1);

	// solid sky texture
	glBindTexture (GL_TEXTURE_2D, solidskytexture.gltexnum);

	// draw the sphere
	glCallList (spherelist);

	// restore the previous matrix
	glPopMatrix ();

	// foreground
	// ==========
	// go to a new matrix
	glPushMatrix ();

	// center it on the players position
	glTranslatef (r_origin[0], r_origin[1], r_origin[2]);

	// flatten the sphere and shrink it a little - the reduced scale prevents artefacts appearing when
	// corners on the skysphere may potentially overlap
	glScalef (reducedfull, reducedfull, reducedhalf);

	// orient it so that the poles are unobtrusive
	glRotatef (-90, 1, 0, 0);

	// make it not always at right angles to the player
	glRotatef (-22, 0 ,1, 0);

	// rotate it around the poles
	glRotatef (-rotateFore, 0, 0, 1);

	// blend mode
	glEnable (GL_BLEND);
	glBlendFunc (GL_SRC_COLOR, GL_SRC_ALPHA);

	// alpha sky texture
	glBindTexture (GL_TEXTURE_2D, alphaskytexture.gltexnum);

	// draw the sphere
	glCallList (spherelist);

	// back to normal mode
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable (GL_BLEND);

	// restore the previous matrix
	glPopMatrix ();

	// turn off fog
	glDisable (GL_FOG);

	glMatrixMode (GL_TEXTURE);
	glLoadIdentity ();
	glMatrixMode (GL_MODELVIEW);
}


void R_InitSkySphere (void)
{
	float skytexes[440];
	float skyverts[660];

	float drho = 0.3141592653589;
	float dtheta = 0.6283185307180;

	float ds = 0.1;
	float dt = 0.1;

	float t = 1.0f;	
	float s = 0.0f;

	int i;
	int j;

	int vertspos = 0;
	int texespos = 0;

	for (i = 0; i < 10; i++)
	{
		float rho = (float) i * drho;
		float srho = (float) (sin (rho));
		float crho = (float) (cos (rho));
		float srhodrho = (float) (sin (rho + drho));
		float crhodrho = (float) (cos (rho + drho));

		s = 0.0f;

		for (j = 0; j <= 10; j++)
		{
			float theta = (j == 10) ? 0.0f : j * dtheta;
			float stheta = (float) (-sin (theta));
			float ctheta = (float) (cos (theta));

			skytexes[texespos++] = s * 2;
			skytexes[texespos++] = t * 2;

			skyverts[vertspos++] = stheta * srho * 4096.0;
			skyverts[vertspos++] = ctheta * srho * 4096.0;
			skyverts[vertspos++] = crho * 4096.0;

			skytexes[texespos++] = s * 2;
			skytexes[texespos++] = (t - dt) * 2;

			skyverts[vertspos++] = stheta * srhodrho * 4096.0;
			skyverts[vertspos++] = ctheta * srhodrho * 4096.0;
			skyverts[vertspos++] = crhodrho * 4096.0;

			s += ds;
		}

		t -= dt;
	}

	vertspos = 0;
	texespos = 0;

	// build the sphere display list
	spherelist = GL_MakeDisplayListObjects (1);

	// fire it up in compile mode
	glNewList (spherelist, GL_COMPILE);

	// draw in immediate mode to properly store the commands
	for (i = 0; i < 10; i++)
	{
		glBegin (GL_TRIANGLE_STRIP);

		for (j = 0; j <= 10; j++)
		{
			glTexCoord2fv (&skytexes[texespos]);
			glVertex3fv (&skyverts[vertspos]);

			texespos += 2;
			vertspos += 3;

			glTexCoord2fv (&skytexes[texespos]);
			glVertex3fv (&skyverts[vertspos]);

			texespos += 2;
			vertspos += 3;
		}

		glEnd ();
	}

	// done
	glEndList ();

	vertspos = 0;

	spherelist2 = GL_MakeDisplayListObjects (1);

	// display list for heat haze stencil things
	glNewList (spherelist2, GL_COMPILE);

	// draw in immediate mode to properly store the commands
	for (i = 0; i < 10; i++)
	{
		float rho = (float) i * drho;
		float srho = (float) (sin (rho));
		float crho = (float) (cos (rho));
		float srhodrho = (float) (sin (rho + drho));
		float crhodrho = (float) (cos (rho + drho));

		glBegin (GL_TRIANGLE_STRIP);

		for (j = 0; j <= 10; j++)
		{
			float theta = (j == 10) ? 0.0f : j * dtheta;
			float stheta = (float) (-sin (theta));
			float ctheta = (float) (cos (theta));

			skyverts[vertspos++] = stheta * srho * (4096.0 - (rand () % 8192));
			skyverts[vertspos++] = ctheta * srho * (4096.0 - (rand () % 8192));
			skyverts[vertspos++] = crho * (4096.0 - (rand () % 8192));

			// no texes here
			glVertex3fv (&skyverts[vertspos - 3]);

			skyverts[vertspos++] = stheta * srhodrho * (4096.0 - (rand () % 8192));
			skyverts[vertspos++] = ctheta * srhodrho * (4096.0 - (rand () % 8192));
			skyverts[vertspos++] = crhodrho * (4096.0 - (rand () % 8192));

			glVertex3fv (&skyverts[vertspos - 3]);
		}

		glEnd ();
	}

	glEndList ();

	// load it from the resource version
	GL_LoadBitmapResourceTexture (&alphaskytexture, IDB_BITMAP6, TEX_TEXTURE);
	GL_LoadBitmapResourceTexture (&solidskytexture, IDB_BITMAP7, TEX_TEXTURE);
}


/*
=============
R_InitSky

A sky texture is 256*128, with the right side being a masked overlay
this is only used to determine the correct fogging colours
==============
*/
void R_InitSky (texture_t *mt)
{
	int			i, j, p;
	byte		*src;
	unsigned	transpix;
	int			r, g, b;
	unsigned	*rgba;
	unsigned	topalpha;
	int			div;
	unsigned	*trans;
	gltexture_t *skytex;

	// initialize our pointers
	trans = malloc (128 * 128 * 4);

	// -----------------
	// solid sky texture
	// -----------------
	src = (byte *) (mt + 1);

	// make an average value for the back to avoid
	// a fringe on the top level
	r = g = b = 0;

	for (i = 0; i < 128; i++)
	{
		for (j = 0; j < 128; j++)
		{
			p = src[i * 256 + j + 128];
			rgba = &d_8to24table[p];
			trans[(i * 128) + j] = *rgba;

			r += ((byte *) rgba)[0];
			g += ((byte *) rgba)[1];
			b += ((byte *) rgba)[2];
		}
	}

	((byte *) &transpix)[0] = r / (128 * 128);
	((byte *) &transpix)[1] = g / (128 * 128);
	((byte *) &transpix)[2] = b / (128 * 128);
	((byte *) &transpix)[3] = 0;

	// -----------------
	// alpha sky texture
	// -----------------
	// get another average cos the bottom layer average can be too dark
	div = r = g = b = 0;

	for (i = 0; i < 128; i++)
	{
		for (j = 0; j < 128; j++)
		{
			p = src[i * 256 + j];

			if (p == 0)
				topalpha = transpix;
			else
			{
				rgba = &d_8to24table[p];

				g += ((byte *) rgba)[1];
				r += ((byte *) rgba)[0];
				b += ((byte *) rgba)[2];

				div++;

				topalpha = d_8to24table[p];
			}

			((byte *) &topalpha)[3] = ((byte *) &topalpha)[3] / 2;

			trans[(i * 128) + j] = topalpha;
		}
	}

	// get fog colours for the sky - we halve these
	sky_fog[0] = ((float) r / (float) div) / 512.0;
	sky_fog[1] = ((float) g / (float) div) / 512.0;
	sky_fog[2] = ((float) b / (float) div) / 512.0;
	sky_fog[3] = 0.1;

	// free the used memory
	free (trans);
}

