/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// gl_vidnt.c -- NT GL vid component

#include "quakedef.h"
#include "winquake.h"
#include "resource.h"
#include <commctrl.h>

#define MAX_MODE_LIST	99
#define VID_ROW_SIZE	3
#define WARP_WIDTH		320
#define WARP_HEIGHT		200
#define MAXWIDTH		10000
#define MAXHEIGHT		10000
#define BASEWIDTH		320
#define BASEHEIGHT		200

#define MODE_WINDOWED			0
#define NO_MODE					(MODE_WINDOWED - 1)
#define MODE_FULLSCREEN_DEFAULT	(MODE_WINDOWED + 1)

LONG CDAudio_MessageHandler(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);

typedef struct {
	modestate_t	type;
	int			width;
	int			height;
	int			modenum;
	int			dib;
	int			fullscreen;
	int			bpp;
	int			halfscreen;
	char		modedesc[17];
} vmode_t;

typedef struct {
	int			width;
	int			height;
} lmode_t;

lmode_t	lowresmodes[] = {
	{320, 200},
	{320, 240},
	{400, 300},
	{512, 384},
};

const char *gl_vendor;
const char *gl_renderer;
const char *gl_version;
const char *gl_extensions;

qboolean		DDActive;
qboolean		scr_skipupdate;

static vmode_t	modelist[MAX_MODE_LIST];
static int		nummodes;
static vmode_t	*pcurrentmode;
static vmode_t	badmode;

static DEVMODE	gdevmode;
static qboolean	vid_initialized = false;
static qboolean	windowed, leavecurrentmode;
static qboolean vid_canalttab = false;
static qboolean vid_wassuspended = false;
static int		windowed_mouse;
extern qboolean	mouseactive;  // from in_win.c
static HICON	hIcon;

int			DIBWidth, DIBHeight;
RECT		WindowRect;
DWORD		WindowStyle, ExWindowStyle;

HWND	mainwindow, dibwindow;

int			vid_modenum = NO_MODE;
int			vid_realmode;
int			vid_default = MODE_WINDOWED;
static int	windowed_default;
unsigned char	vid_curpal[256*3];

static float vid_gamma = 1.0;

HGLRC	baseRC;
HDC		maindc;

glvert_t glv;

HWND WINAPI InitializeWindow (HINSTANCE hInstance, int nCmdShow);

viddef_t	vid;				// global video state

unsigned	d_8to24tableRGBA[256];
unsigned	d_8to24tableBGRA[256];

modestate_t	modestate = MS_UNINIT;

void VID_MenuDraw (void);
void VID_MenuKey (int key);

LONG WINAPI MainWndProc (HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
void AppActivate(BOOL fActive, BOOL minimize);
char *VID_GetModeDescription (int mode);
void ClearAllStates (void);
void VID_UpdateWindowStatus (void);
void GL_Init (void);

PROC glArrayElementEXT;
PROC glColorPointerEXT;
PROC glTexCoordPointerEXT;
PROC glVertexPointerEXT;

//====================================

cvar_t		vid_mode = {"vid_mode","0", false};
// Note that 0 is MODE_WINDOWED
cvar_t		_vid_default_mode = {"_vid_default_mode","0", true};
// Note that 3 is MODE_FULLSCREEN_DEFAULT
cvar_t		_vid_default_mode_win = {"_vid_default_mode_win","3", true};
cvar_t		vid_wait = {"vid_wait","0"};
cvar_t		vid_nopageflip = {"vid_nopageflip","0", true};
cvar_t		_vid_wait_override = {"_vid_wait_override", "0", true};
cvar_t		vid_config_x = {"vid_config_x","800", true};
cvar_t		vid_config_y = {"vid_config_y","600", true};
cvar_t		vid_stretch_by_2 = {"vid_stretch_by_2","1", true};
cvar_t		_windowed_mouse = {"_windowed_mouse","1", true};

int			window_center_x, window_center_y, window_x, window_y, window_width, window_height;
RECT		window_rect;

// direct draw software compatability stuff

void VID_HandlePause (qboolean pause)
{
}

void VID_ForceLockState (int lk)
{
}

void VID_LockBuffer (void)
{
}

void VID_UnlockBuffer (void)
{
}

int VID_ForceUnlockedAndReturnState (void)
{
	return 0;
}

void D_BeginDirectRect (int x, int y, byte *pbitmap, int width, int height)
{
}

void D_EndDirectRect (int x, int y, int width, int height)
{
}


void CenterWindow(HWND hWndCenter, int width, int height, BOOL lefttopjustify)
{
    int     CenterX, CenterY;

	CenterX = (GetSystemMetrics(SM_CXSCREEN) - width) / 2;
	CenterY = (GetSystemMetrics(SM_CYSCREEN) - height) / 2;
	if (CenterX > CenterY*2)
		CenterX >>= 1;	// dual screens
	CenterX = (CenterX < 0) ? 0: CenterX;
	CenterY = (CenterY < 0) ? 0: CenterY;
	SetWindowPos (hWndCenter, NULL, CenterX, CenterY, 0, 0,
			SWP_NOSIZE | SWP_NOZORDER | SWP_SHOWWINDOW | SWP_DRAWFRAME);
}

qboolean VID_SetWindowedMode (int modenum)
{
	HDC				hdc;
	int				lastmodestate, width, height;
	RECT			rect;

	lastmodestate = modestate;

	WindowRect.top = WindowRect.left = 0;

	WindowRect.right = modelist[modenum].width;
	WindowRect.bottom = modelist[modenum].height;

	DIBWidth = modelist[modenum].width;
	DIBHeight = modelist[modenum].height;

	WindowStyle = WS_OVERLAPPED | WS_BORDER | WS_CAPTION | WS_SYSMENU |
				  WS_MINIMIZEBOX;
	ExWindowStyle = 0;

	rect = WindowRect;
	AdjustWindowRectEx(&rect, WindowStyle, FALSE, 0);

	width = rect.right - rect.left;
	height = rect.bottom - rect.top;

	// Create the DIB window
	dibwindow = CreateWindowEx (
		 ExWindowStyle,
		 "WinQuake",
		 "GLQuake",
		 WindowStyle,
		 rect.left, rect.top,
		 width,
		 height,
		 NULL,
		 NULL,
		 global_hInstance,
		 NULL);

	if (!dibwindow)
		Sys_Error ("Couldn't create DIB window");

	// Center and show the DIB window
	CenterWindow(dibwindow, WindowRect.right - WindowRect.left,
				 WindowRect.bottom - WindowRect.top, false);

	ShowWindow (dibwindow, SW_SHOWDEFAULT);
	UpdateWindow (dibwindow);

	modestate = MS_WINDOWED;

// because we have set the background brush for the window to NULL
// (to avoid flickering when re-sizing the window on the desktop),
// we clear the window to black when created, otherwise it will be
// empty while Quake starts up.
	hdc = GetDC(dibwindow);
	PatBlt(hdc,0,0,WindowRect.right,WindowRect.bottom,BLACKNESS);
	ReleaseDC(dibwindow, hdc);

	// don't let console dimensions exceed video resolution
	if (vid.conheight > modelist[modenum].height) vid.conheight = modelist[modenum].height;
	if (vid.conwidth > modelist[modenum].width) vid.conwidth = modelist[modenum].width;

	vid.width = vid.conwidth;
	vid.height = vid.conheight;

	vid.numpages = 2;

	mainwindow = dibwindow;

	SendMessage (mainwindow, WM_SETICON, (WPARAM)TRUE, (LPARAM)hIcon);
	SendMessage (mainwindow, WM_SETICON, (WPARAM)FALSE, (LPARAM)hIcon);

	return true;
}


qboolean VID_SetFullDIBMode (int modenum)
{
	HDC				hdc;
	int				lastmodestate, width, height;
	RECT			rect;

	if (!leavecurrentmode)
	{
		gdevmode.dmFields = DM_BITSPERPEL | DM_PELSWIDTH | DM_PELSHEIGHT;
		gdevmode.dmBitsPerPel = modelist[modenum].bpp;
		gdevmode.dmPelsWidth = modelist[modenum].width <<
							   modelist[modenum].halfscreen;
		gdevmode.dmPelsHeight = modelist[modenum].height;
		gdevmode.dmSize = sizeof (gdevmode);

		if (ChangeDisplaySettings (&gdevmode, CDS_FULLSCREEN) != DISP_CHANGE_SUCCESSFUL)
			Sys_Error ("Couldn't set fullscreen DIB mode");
	}

	lastmodestate = modestate;
	modestate = MS_FULLDIB;

	WindowRect.top = WindowRect.left = 0;

	WindowRect.right = modelist[modenum].width;
	WindowRect.bottom = modelist[modenum].height;

	DIBWidth = modelist[modenum].width;
	DIBHeight = modelist[modenum].height;

	WindowStyle = WS_POPUP;
	ExWindowStyle = 0;

	rect = WindowRect;
	AdjustWindowRectEx(&rect, WindowStyle, FALSE, 0);

	width = rect.right - rect.left;
	height = rect.bottom - rect.top;

	// Create the DIB window
	dibwindow = CreateWindowEx (
		 ExWindowStyle,
		 "WinQuake",
		 "GLQuake",
		 WindowStyle,
		 rect.left, rect.top,
		 width,
		 height,
		 NULL,
		 NULL,
		 global_hInstance,
		 NULL);

	if (!dibwindow)
		Sys_Error ("Couldn't create DIB window");

	ShowWindow (dibwindow, SW_SHOWDEFAULT);
	UpdateWindow (dibwindow);

	// Because we have set the background brush for the window to NULL
	// (to avoid flickering when re-sizing the window on the desktop), we
	// clear the window to black when created, otherwise it will be
	// empty while Quake starts up.
	hdc = GetDC(dibwindow);
	PatBlt(hdc,0,0,WindowRect.right,WindowRect.bottom,BLACKNESS);
	ReleaseDC(dibwindow, hdc);

	if (vid.conheight > modelist[modenum].height) vid.conheight = modelist[modenum].height;
	if (vid.conwidth > modelist[modenum].width) vid.conwidth = modelist[modenum].width;

	vid.width = vid.conwidth;
	vid.height = vid.conheight;

	vid.numpages = 2;

	// needed because we're not getting WM_MOVE messages fullscreen on NT
	window_x = 0;
	window_y = 0;

	mainwindow = dibwindow;

	SendMessage (mainwindow, WM_SETICON, (WPARAM)TRUE, (LPARAM)hIcon);
	SendMessage (mainwindow, WM_SETICON, (WPARAM)FALSE, (LPARAM)hIcon);

	return true;
}


int VID_SetMode (int modenum, unsigned char *palette)
{
	int				original_mode, temp;
	qboolean		stat = false;
    MSG				msg;

	if ((windowed && (modenum != 0)) ||
		(!windowed && (modenum < 1)) ||
		(!windowed && (modenum >= nummodes)))
	{
		Sys_Error ("Bad video mode\n");
	}

// so Con_Printfs don't mess us up by forcing vid and snd updates
	temp = scr_disabled_for_loading;
	scr_disabled_for_loading = true;

	CDAudio_Pause ();

	if (vid_modenum == NO_MODE)
		original_mode = windowed_default;
	else
		original_mode = vid_modenum;

	// Set either the fullscreen or windowed mode
	if (modelist[modenum].type == MS_WINDOWED)
	{
		if (_windowed_mouse.value && key_dest == key_game)
		{
			stat = VID_SetWindowedMode(modenum);
			IN_ActivateMouse ();
			IN_HideMouse ();
		}
		else
		{
			IN_DeactivateMouse ();
			IN_ShowMouse ();
			stat = VID_SetWindowedMode(modenum);
		}
	}
	else if (modelist[modenum].type == MS_FULLDIB)
	{
		stat = VID_SetFullDIBMode(modenum);
		IN_ActivateMouse ();
		IN_HideMouse ();
	}
	else
	{
		Sys_Error ("VID_SetMode: Bad mode type in modelist");
	}

	window_width = DIBWidth;
	window_height = DIBHeight;
	VID_UpdateWindowStatus ();

	CDAudio_Resume ();
	scr_disabled_for_loading = temp;

	if (!stat) Sys_Error ("Couldn't set video mode");

// now we try to make sure we get the focus on the mode switch, because
// sometimes in some systems we don't.  We grab the foreground, then
// finish setting up, pump all our messages, and sleep for a little while
// to let messages finish bouncing around the system, then we put
// ourselves at the top of the z order, then grab the foreground again,
// Who knows if it helps, but it probably doesn't hurt
	SetForegroundWindow (mainwindow);
	VID_SetPalette (palette);
	vid_modenum = modenum;
	Cvar_SetValue ("vid_mode", (float)vid_modenum);

	while (PeekMessage (&msg, NULL, 0, 0, PM_REMOVE))
	{
      	TranslateMessage (&msg);
      	DispatchMessage (&msg);
	}

	Sleep (100);

	SetWindowPos (mainwindow, HWND_TOP, 0, 0, 0, 0,
				  SWP_DRAWFRAME | SWP_NOMOVE | SWP_NOSIZE | SWP_SHOWWINDOW |
				  SWP_NOCOPYBITS);

	SetForegroundWindow (mainwindow);

// fix the leftover Alt from any Alt-Tab or the like that switched us away
	ClearAllStates ();

	if (!msg_suppress_1)
		Con_SafePrintf ("Video mode %s initialized.\n", VID_GetModeDescription (vid_modenum));

	VID_SetPalette (palette);

	vid.recalc_refdef = 1;

	return true;
}


/*
================
VID_UpdateWindowStatus
================
*/
void VID_UpdateWindowStatus (void)
{

	window_rect.left = window_x;
	window_rect.top = window_y;
	window_rect.right = window_x + window_width;
	window_rect.bottom = window_y + window_height;
	window_center_x = (window_rect.left + window_rect.right) / 2;
	window_center_y = (window_rect.top + window_rect.bottom) / 2;

	IN_UpdateClipCursor ();
}


//====================================

char *VendorCode[] = {"ARB", "EXT", "NV", "ATI", "SGI", "SGIS", "SGIX", "INTEL", "3DFX", NULL};

/*
================
MHGLGetProcAddress

Implements a work-around for earlier OpenGL versions, by checking for extensions that
support the required functions.  I'm VERY DISPLEASED at having to do this (you don't
cut down the Mona Lisa to make it fit in a smaller frame, do you?), but ATI's most
recent OpenGL implementation is only at version 1.3 and even at that it doesn't support
some 1.2 stuff.  Thank God for Doom 3 - it may make them get their act together.
================
*/
PROC MHGLGetProcAddress (char *funcname, qboolean crash, qboolean *ver)
{
	int i;
	static PROC thePROC;

	// see is the function supported by the core implementation
	if (thePROC = wglGetProcAddress (funcname))
	{
		*ver = (*ver && true);
		return thePROC;
	}

	for (i = 0; ; i++)
	{
		if (!VendorCode[i]) break;

		thePROC = wglGetProcAddress (va ("%s%s", funcname, VendorCode[i]));

		if (thePROC)
		{
			Con_Printf ("%s Unavailable, reverting to %s%s\n", funcname, funcname, VendorCode[i]);
			*ver = (*ver && true);
			return thePROC;
		}
	}

	// oh dear.
	if (crash) Sys_Error ("Core OpenGL function %s Unavailable!!!", funcname);

	*ver = (*ver && false);
	return NULL;
}


// gl capabilities struct
gl_caps_t GL_Caps;


void GL_GetFunctionPointers (void)
{
	// initially all true
	GL_Caps.OpenGL12 = true;
	GL_Caps.OpenGL12Img = true;
	GL_Caps.OpenGL13 = true;
	GL_Caps.OpenGL14 = true;

	// opengl 1.2 and imaging subset prototypes
	glBlendColor = MHGLGetProcAddress ("glBlendColor", true, &GL_Caps.OpenGL12);
	glBlendEquation = MHGLGetProcAddress ("glBlendEquation", true, &GL_Caps.OpenGL12);
	glDrawRangeElements = MHGLGetProcAddress ("glDrawRangeElements", true, &GL_Caps.OpenGL12);
	glColorTable = MHGLGetProcAddress ("glColorTable", false, &GL_Caps.OpenGL12Img);
	glColorTableParameterfv = MHGLGetProcAddress ("glColorTableParameterfv", false, &GL_Caps.OpenGL12Img);
	glColorTableParameteriv = MHGLGetProcAddress ("glColorTableParameteriv", false, &GL_Caps.OpenGL12Img);
	glCopyColorTable = MHGLGetProcAddress ("glCopyColorTable", false, &GL_Caps.OpenGL12Img);
	glGetColorTable = MHGLGetProcAddress ("glGetColorTable", false, &GL_Caps.OpenGL12Img);
	glGetColorTableParameterfv = MHGLGetProcAddress ("glGetColorTableParameterfv", false, &GL_Caps.OpenGL12Img);
	glGetColorTableParameteriv = MHGLGetProcAddress ("glGetColorTableParameteriv", false, &GL_Caps.OpenGL12Img);
	glColorSubTable = MHGLGetProcAddress ("glColorSubTable", false, &GL_Caps.OpenGL12Img);
	glCopyColorSubTable = MHGLGetProcAddress ("glCopyColorSubTable", false, &GL_Caps.OpenGL12Img);
	glConvolutionFilter1D = MHGLGetProcAddress ("glConvolutionFilter1D", false, &GL_Caps.OpenGL12Img);
	glConvolutionFilter2D = MHGLGetProcAddress ("glConvolutionFilter2D", false, &GL_Caps.OpenGL12Img);
	glConvolutionParameterf = MHGLGetProcAddress ("glConvolutionParameterf", false, &GL_Caps.OpenGL12Img);
	glConvolutionParameterfv = MHGLGetProcAddress ("glConvolutionParameterfv", false, &GL_Caps.OpenGL12Img);
	glConvolutionParameteri = MHGLGetProcAddress ("glConvolutionParameteri", false, &GL_Caps.OpenGL12Img);
	glConvolutionParameteriv = MHGLGetProcAddress ("glConvolutionParameteriv", false, &GL_Caps.OpenGL12Img);
	glCopyConvolutionFilter1D = MHGLGetProcAddress ("glCopyConvolutionFilter1D", false, &GL_Caps.OpenGL12Img);
	glCopyConvolutionFilter2D = MHGLGetProcAddress ("glCopyConvolutionFilter2D", false, &GL_Caps.OpenGL12Img);
	glGetConvolutionFilter = MHGLGetProcAddress ("glGetConvolutionFilter", false, &GL_Caps.OpenGL12Img);
	glGetConvolutionParameterfv = MHGLGetProcAddress ("glGetConvolutionParameterfv", false, &GL_Caps.OpenGL12Img);
	glGetConvolutionParameteriv = MHGLGetProcAddress ("glGetConvolutionParameteriv", false, &GL_Caps.OpenGL12Img);
	glGetSeparableFilter = MHGLGetProcAddress ("glGetSeparableFilter", false, &GL_Caps.OpenGL12Img);
	glSeparableFilter2D = MHGLGetProcAddress ("glSeparableFilter2D", false, &GL_Caps.OpenGL12Img);
	glGetHistogram = MHGLGetProcAddress ("glGetHistogram", false, &GL_Caps.OpenGL12Img);
	glGetHistogramParameterfv = MHGLGetProcAddress ("glGetHistogramParameterfv", false, &GL_Caps.OpenGL12Img);
	glGetHistogramParameteriv = MHGLGetProcAddress ("glGetHistogramParameteriv", false, &GL_Caps.OpenGL12Img);
	glGetMinmax = MHGLGetProcAddress ("glGetMinmax", false, &GL_Caps.OpenGL12Img);
	glGetMinmaxParameterfv = MHGLGetProcAddress ("glGetMinmaxParameterfv", false, &GL_Caps.OpenGL12Img);
	glGetMinmaxParameteriv = MHGLGetProcAddress ("glGetMinmaxParameteriv", false, &GL_Caps.OpenGL12Img);
	glHistogram = MHGLGetProcAddress ("glHistogram", false, &GL_Caps.OpenGL12Img);
	glMinmax = MHGLGetProcAddress ("glMinmax", false, &GL_Caps.OpenGL12Img);
	glResetHistogram = MHGLGetProcAddress ("glResetHistogram", false, &GL_Caps.OpenGL12Img);
	glResetMinmax = MHGLGetProcAddress ("glResetMinmax", false, &GL_Caps.OpenGL12Img);
	glTexImage3D = MHGLGetProcAddress ("glTexImage3D", true, &GL_Caps.OpenGL12);
	glTexSubImage3D = MHGLGetProcAddress ("glTexSubImage3D", true, &GL_Caps.OpenGL12);
	glCopyTexSubImage3D = MHGLGetProcAddress ("glCopyTexSubImage3D", true, &GL_Caps.OpenGL12);

	// OpenGL 1.3 Prototypes
	glActiveTexture = MHGLGetProcAddress ("glActiveTexture", true, &GL_Caps.OpenGL13);
	glClientActiveTexture = MHGLGetProcAddress ("glClientActiveTexture", true, &GL_Caps.OpenGL13);
	glMultiTexCoord1d = MHGLGetProcAddress ("glMultiTexCoord1d", true, &GL_Caps.OpenGL13);
	glMultiTexCoord1dv = MHGLGetProcAddress ("glMultiTexCoord1dv", true, &GL_Caps.OpenGL13);
	glMultiTexCoord1f = MHGLGetProcAddress ("glMultiTexCoord1f", true, &GL_Caps.OpenGL13);
	glMultiTexCoord1fv = MHGLGetProcAddress ("glMultiTexCoord1fv", true, &GL_Caps.OpenGL13);
	glMultiTexCoord1i = MHGLGetProcAddress ("glMultiTexCoord1i", true, &GL_Caps.OpenGL13);
	glMultiTexCoord1iv = MHGLGetProcAddress ("glMultiTexCoord1iv", true, &GL_Caps.OpenGL13);
	glMultiTexCoord1s = MHGLGetProcAddress ("glMultiTexCoord1s", true, &GL_Caps.OpenGL13);
	glMultiTexCoord1sv = MHGLGetProcAddress ("glMultiTexCoord1sv", true, &GL_Caps.OpenGL13);
	glMultiTexCoord2d = MHGLGetProcAddress ("glMultiTexCoord2d", true, &GL_Caps.OpenGL13);
	glMultiTexCoord2dv = MHGLGetProcAddress ("glMultiTexCoord2dv", true, &GL_Caps.OpenGL13);
	glMultiTexCoord2f = MHGLGetProcAddress ("glMultiTexCoord2f", true, &GL_Caps.OpenGL13);
	glMultiTexCoord2fv = MHGLGetProcAddress ("glMultiTexCoord2fv", true, &GL_Caps.OpenGL13);
	glMultiTexCoord2i = MHGLGetProcAddress ("glMultiTexCoord2i", true, &GL_Caps.OpenGL13);
	glMultiTexCoord2iv = MHGLGetProcAddress ("glMultiTexCoord2iv", true, &GL_Caps.OpenGL13);
	glMultiTexCoord2s = MHGLGetProcAddress ("glMultiTexCoord2s", true, &GL_Caps.OpenGL13);
	glMultiTexCoord2sv = MHGLGetProcAddress ("glMultiTexCoord2sv", true, &GL_Caps.OpenGL13);
	glMultiTexCoord3d = MHGLGetProcAddress ("glMultiTexCoord3d", true, &GL_Caps.OpenGL13);
	glMultiTexCoord3dv = MHGLGetProcAddress ("glMultiTexCoord3dv", true, &GL_Caps.OpenGL13);
	glMultiTexCoord3f = MHGLGetProcAddress ("glMultiTexCoord3f", true, &GL_Caps.OpenGL13);
	glMultiTexCoord3fv = MHGLGetProcAddress ("glMultiTexCoord3fv", true, &GL_Caps.OpenGL13);
	glMultiTexCoord3i = MHGLGetProcAddress ("glMultiTexCoord3i", true, &GL_Caps.OpenGL13);
	glMultiTexCoord3iv = MHGLGetProcAddress ("glMultiTexCoord3iv", true, &GL_Caps.OpenGL13);
	glMultiTexCoord3s = MHGLGetProcAddress ("glMultiTexCoord3s", true, &GL_Caps.OpenGL13);
	glMultiTexCoord3sv = MHGLGetProcAddress ("glMultiTexCoord3sv", true, &GL_Caps.OpenGL13);
	glMultiTexCoord4d = MHGLGetProcAddress ("glMultiTexCoord4d", true, &GL_Caps.OpenGL13);
	glMultiTexCoord4dv = MHGLGetProcAddress ("glMultiTexCoord4dv", true, &GL_Caps.OpenGL13);
	glMultiTexCoord4f = MHGLGetProcAddress ("glMultiTexCoord4f", true, &GL_Caps.OpenGL13);
	glMultiTexCoord4fv = MHGLGetProcAddress ("glMultiTexCoord4fv", true, &GL_Caps.OpenGL13);
	glMultiTexCoord4i = MHGLGetProcAddress ("glMultiTexCoord4i", true, &GL_Caps.OpenGL13);
	glMultiTexCoord4iv = MHGLGetProcAddress ("glMultiTexCoord4iv", true, &GL_Caps.OpenGL13);
	glMultiTexCoord4s = MHGLGetProcAddress ("glMultiTexCoord4s", true, &GL_Caps.OpenGL13);
	glMultiTexCoord4sv = MHGLGetProcAddress ("glMultiTexCoord4sv", true, &GL_Caps.OpenGL13);
	glLoadTransposeMatrixf = MHGLGetProcAddress ("glLoadTransposeMatrixf", true, &GL_Caps.OpenGL13);
	glLoadTransposeMatrixd = MHGLGetProcAddress ("glLoadTransposeMatrixd", true, &GL_Caps.OpenGL13);
	glMultTransposeMatrixf = MHGLGetProcAddress ("glMultTransposeMatrixf", true, &GL_Caps.OpenGL13);
	glMultTransposeMatrixd = MHGLGetProcAddress ("glMultTransposeMatrixd", true, &GL_Caps.OpenGL13);
	glSampleCoverage = MHGLGetProcAddress ("glSampleCoverage", true, &GL_Caps.OpenGL13);
	glCompressedTexImage3D = MHGLGetProcAddress ("glCompressedTexImage3D", true, &GL_Caps.OpenGL13);
	glCompressedTexImage2D = MHGLGetProcAddress ("glCompressedTexImage2D", true, &GL_Caps.OpenGL13);
	glCompressedTexImage1D = MHGLGetProcAddress ("glCompressedTexImage1D", true, &GL_Caps.OpenGL13);
	glCompressedTexSubImage3D = MHGLGetProcAddress ("glCompressedTexSubImage3D", true, &GL_Caps.OpenGL13);
	glCompressedTexSubImage2D = MHGLGetProcAddress ("glCompressedTexSubImage2D", true, &GL_Caps.OpenGL13);
	glCompressedTexSubImage1D = MHGLGetProcAddress ("glCompressedTexSubImage1D", true, &GL_Caps.OpenGL13);
	glGetCompressedTexImage = MHGLGetProcAddress ("glGetCompressedTexImage", true, &GL_Caps.OpenGL13);

	// OpenGL 1.4 Prototypes
	glBlendFuncSeparate = MHGLGetProcAddress ("glBlendFuncSeparate", true, &GL_Caps.OpenGL14);
	glFogCoordf = MHGLGetProcAddress ("glFogCoordf", true, &GL_Caps.OpenGL14);
	glFogCoordfv = MHGLGetProcAddress ("glFogCoordfv", true, &GL_Caps.OpenGL14);
	glFogCoordd = MHGLGetProcAddress ("glFogCoordd", true, &GL_Caps.OpenGL14);
	glFogCoorddv = MHGLGetProcAddress ("glFogCoorddv", true, &GL_Caps.OpenGL14);
	glFogCoordPointer = MHGLGetProcAddress ("glFogCoordPointer", true, &GL_Caps.OpenGL14);
	glMultiDrawArrays = MHGLGetProcAddress ("glMultiDrawArrays", true, &GL_Caps.OpenGL14);
	glMultiDrawElements = MHGLGetProcAddress ("glMultiDrawElements", true, &GL_Caps.OpenGL14);
	glPointParameterf = MHGLGetProcAddress ("glPointParameterf", true, &GL_Caps.OpenGL14);
	glPointParameterfv = MHGLGetProcAddress ("glPointParameterfv", true, &GL_Caps.OpenGL14);
	glSecondaryColor3b = MHGLGetProcAddress ("glSecondaryColor3b", true, &GL_Caps.OpenGL14);
	glSecondaryColor3bv = MHGLGetProcAddress ("glSecondaryColor3bv", true, &GL_Caps.OpenGL14);
	glSecondaryColor3d = MHGLGetProcAddress ("glSecondaryColor3d", true, &GL_Caps.OpenGL14);
	glSecondaryColor3dv = MHGLGetProcAddress ("glSecondaryColor3dv", true, &GL_Caps.OpenGL14);
	glSecondaryColor3f = MHGLGetProcAddress ("glSecondaryColor3f", true, &GL_Caps.OpenGL14);
	glSecondaryColor3fv = MHGLGetProcAddress ("glSecondaryColor3fv", true, &GL_Caps.OpenGL14);
	glSecondaryColor3i = MHGLGetProcAddress ("glSecondaryColor3i", true, &GL_Caps.OpenGL14);
	glSecondaryColor3iv = MHGLGetProcAddress ("glSecondaryColor3iv", true, &GL_Caps.OpenGL14);
	glSecondaryColor3s = MHGLGetProcAddress ("glSecondaryColor3s", true, &GL_Caps.OpenGL14);
	glSecondaryColor3sv = MHGLGetProcAddress ("glSecondaryColor3sv", true, &GL_Caps.OpenGL14);
	glSecondaryColor3ub = MHGLGetProcAddress ("glSecondaryColor3ub", true, &GL_Caps.OpenGL14);
	glSecondaryColor3ubv = MHGLGetProcAddress ("glSecondaryColor3ubv", true, &GL_Caps.OpenGL14);
	glSecondaryColor3ui = MHGLGetProcAddress ("glSecondaryColor3ui", true, &GL_Caps.OpenGL14);
	glSecondaryColor3uiv = MHGLGetProcAddress ("glSecondaryColor3uiv", true, &GL_Caps.OpenGL14);
	glSecondaryColor3us = MHGLGetProcAddress ("glSecondaryColor3us", true, &GL_Caps.OpenGL14);
	glSecondaryColor3usv = MHGLGetProcAddress ("glSecondaryColor3usv", true, &GL_Caps.OpenGL14);
	glSecondaryColorPointer = MHGLGetProcAddress ("glSecondaryColorPointer", true, &GL_Caps.OpenGL14);
	glWindowPos2d = MHGLGetProcAddress ("glWindowPos2d", true, &GL_Caps.OpenGL14);
	glWindowPos2dv = MHGLGetProcAddress ("glWindowPos2dv", true, &GL_Caps.OpenGL14);
	glWindowPos2f = MHGLGetProcAddress ("glWindowPos2f", true, &GL_Caps.OpenGL14);
	glWindowPos2fv = MHGLGetProcAddress ("glWindowPos2fv", true, &GL_Caps.OpenGL14);
	glWindowPos2i = MHGLGetProcAddress ("glWindowPos2i", true, &GL_Caps.OpenGL14);
	glWindowPos2iv = MHGLGetProcAddress ("glWindowPos2iv", true, &GL_Caps.OpenGL14);
	glWindowPos2s = MHGLGetProcAddress ("glWindowPos2s", true, &GL_Caps.OpenGL14);
	glWindowPos2sv = MHGLGetProcAddress ("glWindowPos2sv", true, &GL_Caps.OpenGL14);
	glWindowPos3d = MHGLGetProcAddress ("glWindowPos3d", true, &GL_Caps.OpenGL14);
	glWindowPos3dv = MHGLGetProcAddress ("glWindowPos3dv", true, &GL_Caps.OpenGL14);
	glWindowPos3f = MHGLGetProcAddress ("glWindowPos3f", true, &GL_Caps.OpenGL14);
	glWindowPos3fv = MHGLGetProcAddress ("glWindowPos3fv", true, &GL_Caps.OpenGL14);
	glWindowPos3i = MHGLGetProcAddress ("glWindowPos3i", true, &GL_Caps.OpenGL14);
	glWindowPos3iv = MHGLGetProcAddress ("glWindowPos3iv", true, &GL_Caps.OpenGL14);
	glWindowPos3s = MHGLGetProcAddress ("glWindowPos3s", true, &GL_Caps.OpenGL14);
	glWindowPos3sv = MHGLGetProcAddress ("glWindowPos3sv", true, &GL_Caps.OpenGL14);

	if (GL_Caps.OpenGL12) Con_Printf ("OpenGL 1.2 OK\n"); else Con_Printf ("OpenGL 1.2 Not fully supported\n");
	if (GL_Caps.OpenGL12Img) Con_Printf ("OpenGL Imaging OK\n"); else Con_Printf ("OpenGL Imaging Not fully supported\n");
	if (GL_Caps.OpenGL13) Con_Printf ("OpenGL 1.3 OK\n"); else Con_Printf ("OpenGL 1.3 Not fully supported\n");
	if (GL_Caps.OpenGL14) Con_Printf ("OpenGL 1.4 OK\n"); else Con_Printf ("OpenGL 1.4 Not fully supported\n");
}


int		texture_mode = GL_LINEAR;

/*
===============
GL_Init
===============
*/
void GL_Init (void)
{
	gl_vendor = (char *) glGetString (GL_VENDOR);
	Con_Printf ("GL_VENDOR: %s\n", gl_vendor);
	gl_renderer = (char *) glGetString (GL_RENDERER);
	Con_Printf ("GL_RENDERER: %s\n", gl_renderer);

	gl_version = (char *) glGetString (GL_VERSION);
	Con_Printf ("GL_VERSION: %s\n", gl_version);

	gl_extensions = (char *) glGetString (GL_EXTENSIONS);

	glGetIntegerv (GL_MAX_TEXTURE_SIZE, &gl_max_size);

	// take one quarter of the maximum size we're going to use, but never less than
	// 512, unless the absolute maximum is 256 or less...
	if (gl_max_size > 256)
	{
		gl_max_size /= 4;

		if (gl_max_size < 512) gl_max_size = 512;
	}

	Con_Printf ("Maximum Usable Texture Size: %i x %i\n", gl_max_size, gl_max_size);

	// Quake doesn't have a 1 TMU render path any more, so...
	glGetIntegerv (GL_MAX_TEXTURE_UNITS, &gl_num_tmus);
	Con_Printf ("Found %i Texture Memory Units\n\n", gl_num_tmus);

	// mhquake requires 4 tmus.  if an intel integrated can have 8, you really have no excuse
	if (gl_num_tmus < 4) Sys_Error ("Not enough TMUs");

	GL_GetFunctionPointers ();

	glClearColor (0, 0, 0, 0);
	glClearColor (0, 0, 0, 0);
	glCullFace(GL_FRONT);
	glEnable(GL_TEXTURE_2D);

	glEnable(GL_ALPHA_TEST);
	glAlphaFunc(GL_GREATER, 0.666);

	glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);
	glShadeModel (GL_FLAT);

	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

//	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
}

/*
=================
GL_BeginRendering

=================
*/
void GL_BeginRendering (int *x, int *y, int *width, int *height)
{
	*x = *y = 0;
	*width = WindowRect.right - WindowRect.left;
	*height = WindowRect.bottom - WindowRect.top;

//    if (!wglMakeCurrent( maindc, baseRC ))
//		Sys_Error ("wglMakeCurrent failed");

//	glViewport (*x, *y, *width, *height);
}


void GL_EndRendering (void)
{
	if (!scr_skipupdate || block_drawing)
		SwapBuffers(maindc);

// handle the mouse state when windowed if that's changed
	if (modestate == MS_WINDOWED)
	{
		if (!_windowed_mouse.value) {
			if (windowed_mouse)	{
				IN_DeactivateMouse ();
				IN_ShowMouse ();
				windowed_mouse = false;
			}
		} else {
			windowed_mouse = true;
			if (key_dest == key_game && !mouseactive && ActiveApp) {
				IN_ActivateMouse ();
				IN_HideMouse ();
			} else if (mouseactive && key_dest != key_game) {
				IN_DeactivateMouse ();
				IN_ShowMouse ();
			}
		}
	}

	Sbar_Changed();
}

void VID_SetPalette (unsigned char *palette)
{
	int i;
	byte *pal;
	unsigned int r, g, b;
	unsigned int *tableRGBA;
	unsigned int *tableBGRA;

	// 8 8 8 encoding
	pal = palette;
	tableRGBA = d_8to24tableRGBA;
	tableBGRA = d_8to24tableBGRA;

	for (i = 0; i < 256; i++)
	{
		r = pal[0];
		g = pal[1];
		b = pal[2];
		pal += 3;

		*tableRGBA++ = ((255 << 24) + (r << 0) + (g << 8) + (b << 16));
		*tableBGRA++ = ((255 << 24) + (b << 0) + (g << 8) + (r << 16));
	}

	// set proper alpha colour
	d_8to24tableRGBA[255] = 0;
	d_8to24tableBGRA[255] = 0;
}


BOOL	gammaworks;

void	VID_ShiftPalette (unsigned char *palette)
{
	extern	byte ramps[3][256];
	
//	VID_SetPalette (palette);

//	gammaworks = SetDeviceGammaRamp (maindc, ramps);
}


void VID_SetDefaultMode (void)
{
	IN_DeactivateMouse ();
}


void	VID_Shutdown (void)
{
   	HGLRC hRC;
   	HDC	  hDC;

	if (vid_initialized)
	{
		vid_canalttab = false;
		hRC = wglGetCurrentContext();
    	hDC = wglGetCurrentDC();

    	wglMakeCurrent(NULL, NULL);

    	if (hRC)
    	    wglDeleteContext(hRC);

		if (hDC && dibwindow)
			ReleaseDC(dibwindow, hDC);

		if (modestate == MS_FULLDIB)
			ChangeDisplaySettings (NULL, 0);

		if (maindc && dibwindow)
			ReleaseDC (dibwindow, maindc);

		AppActivate(false, false);
	}
}


//==========================================================================


BOOL bSetupPixelFormat(HDC hDC)
{
    static PIXELFORMATDESCRIPTOR pfd =
	{
		sizeof (PIXELFORMATDESCRIPTOR),	// size of this pfd
		1,								// version number
		PFD_DRAW_TO_WINDOW 				// support window
		|  PFD_SUPPORT_OPENGL 			// support OpenGL
		|  PFD_DOUBLEBUFFER ,			// double buffered
		PFD_TYPE_RGBA,					// RGBA type
		24,								// 24-bit color depth
		0, 0, 0, 0, 0, 0,				// color bits ignored
		0,								// no alpha buffer
		0,								// shift bit ignored
		0,								// no accumulation buffer
		0, 0, 0, 0, 					// accum bits ignored
		32,								// 32-bit z-buffer	
		0,								// 8 bit stencil buffer
		0,								// no auxiliary buffer
		PFD_MAIN_PLANE,					// main layer
		0,								// reserved
		0, 0, 0							// layer masks ignored
    };

    int pixelformat;

    if ( (pixelformat = ChoosePixelFormat(hDC, &pfd)) == 0 )
    {
        MessageBox(NULL, "ChoosePixelFormat failed", "Error", MB_OK);
        return FALSE;
    }

    if (SetPixelFormat(hDC, pixelformat, &pfd) == FALSE)
    {
        MessageBox(NULL, "SetPixelFormat failed", "Error", MB_OK);
        return FALSE;
    }

    return TRUE;
}


byte        scantokey[128] = 
					{ 
//  0           1       2       3       4       5       6       7 
//  8           9       A       B       C       D       E       F 
	0  ,    27,     '1',    '2',    '3',    '4',    '5',    '6', 
	'7',    '8',    '9',    '0',    '-',    '=',    K_BACKSPACE, 9, // 0 
	'q',    'w',    'e',    'r',    't',    'y',    'u',    'i', 
	'o',    'p',    '[',    ']',    13 ,    K_CTRL,'a',  's',      // 1 
	'd',    'f',    'g',    'h',    'j',    'k',    'l',    ';', 
	'\'' ,    '`',    K_SHIFT,'\\',  'z',    'x',    'c',    'v',      // 2 
	'b',    'n',    'm',    ',',    '.',    '/',    K_SHIFT,'*', 
	K_ALT,' ',   0  ,    K_F1, K_F2, K_F3, K_F4, K_F5,   // 3 
	K_F6, K_F7, K_F8, K_F9, K_F10, K_PAUSE  ,    0  , K_HOME, 
	K_UPARROW,K_PGUP,'-',K_LEFTARROW,'5',K_RIGHTARROW,'+',K_END, //4 
	K_DOWNARROW,K_PGDN,K_INS,K_DEL,0,0,             0,              K_F11, 
	K_F12,0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0,        // 5 
	0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0, 
	0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0,        // 6 
	0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0, 
	0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0         // 7 
					}; 

byte        shiftscantokey[128] = 
					{ 
//  0           1       2       3       4       5       6       7 
//  8           9       A       B       C       D       E       F 
	0  ,    27,     '!',    '@',    '#',    '$',    '%',    '^', 
	'&',    '*',    '(',    ')',    '_',    '+',    K_BACKSPACE, 9, // 0 
	'Q',    'W',    'E',    'R',    'T',    'Y',    'U',    'I', 
	'O',    'P',    '{',    '}',    13 ,    K_CTRL,'A',  'S',      // 1 
	'D',    'F',    'G',    'H',    'J',    'K',    'L',    ':', 
	'"' ,    '~',    K_SHIFT,'|',  'Z',    'X',    'C',    'V',      // 2 
	'B',    'N',    'M',    '<',    '>',    '?',    K_SHIFT,'*', 
	K_ALT,' ',   0  ,    K_F1, K_F2, K_F3, K_F4, K_F5,   // 3 
	K_F6, K_F7, K_F8, K_F9, K_F10, K_PAUSE  ,    0  , K_HOME, 
	K_UPARROW,K_PGUP,'_',K_LEFTARROW,'%',K_RIGHTARROW,'+',K_END, //4 
	K_DOWNARROW,K_PGDN,K_INS,K_DEL,0,0,             0,              K_F11, 
	K_F12,0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0,        // 5 
	0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0, 
	0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0,        // 6 
	0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0, 
	0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0         // 7 
					}; 


/*
=======
MapKey

Map from windows to quake keynums
=======
*/
int MapKey (int key)
{
	key = (key>>16)&255;
	if (key > 127)
		return 0;
	if (scantokey[key] == 0)
		Con_DPrintf("key 0x%02x has no translation\n", key);
	return scantokey[key];
}

/*
===================================================================

MAIN WINDOW

===================================================================
*/

/*
================
ClearAllStates
================
*/
void ClearAllStates (void)
{
	int		i;
	
// send an up event for each key, to make sure the server clears them all
	for (i=0 ; i<256 ; i++)
	{
		Key_Event (i, false);
	}

	Key_ClearStates ();
	IN_ClearStates ();
}

void AppActivate(BOOL fActive, BOOL minimize)
/****************************************************************************
*
* Function:     AppActivate
* Parameters:   fActive - True if app is activating
*
* Description:  If the application is activating, then swap the system
*               into SYSPAL_NOSTATIC mode so that our palettes will display
*               correctly.
*
****************************************************************************/
{
	static BOOL	sound_active;

	ActiveApp = fActive;
	Minimized = minimize;

// enable/disable sound on focus gain/loss
	if (!ActiveApp && sound_active)
	{
		S_BlockSound ();
		sound_active = false;
	}
	else if (ActiveApp && !sound_active)
	{
		S_UnblockSound ();
		sound_active = true;
	}

	if (fActive)
	{
		if (modestate == MS_FULLDIB)
		{
			IN_ActivateMouse ();
			IN_HideMouse ();
			if (vid_canalttab && vid_wassuspended) {
				vid_wassuspended = false;
				ChangeDisplaySettings (&gdevmode, CDS_FULLSCREEN);
				ShowWindow(mainwindow, SW_SHOWNORMAL);
			}
		}
		else if ((modestate == MS_WINDOWED) && _windowed_mouse.value && key_dest == key_game)
		{
			IN_ActivateMouse ();
			IN_HideMouse ();
		}
	}

	if (!fActive)
	{
		if (modestate == MS_FULLDIB)
		{
			IN_DeactivateMouse ();
			IN_ShowMouse ();
			if (vid_canalttab) { 
				ChangeDisplaySettings (NULL, 0);
				vid_wassuspended = true;
			}
		}
		else if ((modestate == MS_WINDOWED) && _windowed_mouse.value)
		{
			IN_DeactivateMouse ();
			IN_ShowMouse ();
		}
	}
}

void GDI_KillGDI (void);

/* main window procedure */
LONG WINAPI MainWndProc (
    HWND    hWnd,
    UINT    uMsg,
    WPARAM  wParam,
    LPARAM  lParam)
{
    LONG    lRet = 1;
	int		fActive, fMinimized, temp;
	extern unsigned int uiWheelMessage;

	if ( uMsg == uiWheelMessage )
		uMsg = WM_MOUSEWHEEL;

    switch (uMsg)
    {
		case WM_KILLFOCUS:
			if (modestate == MS_FULLDIB)
				ShowWindow(mainwindow, SW_SHOWMINNOACTIVE);
			break;

		case WM_CREATE:
			break;

		case WM_MOVE:
			window_x = (int) LOWORD(lParam);
			window_y = (int) HIWORD(lParam);
			VID_UpdateWindowStatus ();
			break;

		case WM_KEYDOWN:
		case WM_SYSKEYDOWN:
			Key_Event (MapKey(lParam), true);
			break;
			
		case WM_KEYUP:
		case WM_SYSKEYUP:
			Key_Event (MapKey(lParam), false);
			break;

		case WM_SYSCHAR:
		// keep Alt-Space from happening
			break;

	// this is complicated because Win32 seems to pack multiple mouse events into
	// one update sometimes, so we always check all states and look for events
		case WM_LBUTTONDOWN:
		case WM_LBUTTONUP:
		case WM_RBUTTONDOWN:
		case WM_RBUTTONUP:
		case WM_MBUTTONDOWN:
		case WM_MBUTTONUP:
		case WM_MOUSEMOVE:
			temp = 0;

			if (wParam & MK_LBUTTON)
				temp |= 1;

			if (wParam & MK_RBUTTON)
				temp |= 2;

			if (wParam & MK_MBUTTON)
				temp |= 4;

			IN_MouseEvent (temp);

			break;

		// JACK: This is the mouse wheel with the Intellimouse
		// Its delta is either positive or neg, and we generate the proper
		// Event.
		case WM_MOUSEWHEEL: 
			if ((short) HIWORD(wParam) > 0) {
				Key_Event(K_MWHEELUP, true);
				Key_Event(K_MWHEELUP, false);
			} else {
				Key_Event(K_MWHEELDOWN, true);
				Key_Event(K_MWHEELDOWN, false);
			}
			break;

    	case WM_SIZE:
            break;

   	    case WM_CLOSE:
			if (MessageBox (mainwindow, "Are you sure you want to quit?", "Confirm Exit",
						MB_YESNO | MB_SETFOREGROUND | MB_ICONQUESTION) == IDYES)
			{
				Sys_Quit ();
			}

	        break;

		case WM_ACTIVATE:
			fActive = LOWORD(wParam);
			fMinimized = (BOOL) HIWORD(wParam);
			AppActivate(!(fActive == WA_INACTIVE), fMinimized);

		// fix the leftover Alt from any Alt-Tab or the like that switched us away
			ClearAllStates ();

			break;

   	    case WM_DESTROY:
        {
			GDI_KillGDI ();

			if (dibwindow)
				DestroyWindow (dibwindow);

            PostQuitMessage (0);
        }
        break;

		case MM_MCINOTIFY:
            lRet = CDAudio_MessageHandler (hWnd, uMsg, wParam, lParam);
			break;

    	default:
            /* pass all unhandled messages to DefWindowProc */
            lRet = DefWindowProc (hWnd, uMsg, wParam, lParam);
        break;
    }

    /* return 1 if handled message, 0 if not */
    return lRet;
}


/*
=================
VID_NumModes
=================
*/
int VID_NumModes (void)
{
	return nummodes;
}

	
/*
=================
VID_GetModePtr
=================
*/
vmode_t *VID_GetModePtr (int modenum)
{

	if ((modenum >= 0) && (modenum < nummodes))
		return &modelist[modenum];
	else
		return &badmode;
}


/*
=================
VID_GetModeDescription
=================
*/
char *VID_GetModeDescription (int mode)
{
	char		*pinfo;
	vmode_t		*pv;
	static char	temp[100];

	if ((mode < 0) || (mode >= nummodes))
		return NULL;

//	if (!leavecurrentmode)
//	{
		pv = VID_GetModePtr (mode);
		pinfo = pv->modedesc;
//	}
//	else
//	{
//		sprintf (temp, "Desktop resolution (%dx%d)",
//				 modelist[MODE_FULLSCREEN_DEFAULT].width,
//				 modelist[MODE_FULLSCREEN_DEFAULT].height);
//		pinfo = temp;
//	}

	return pinfo;
}


// KJB: Added this to return the mode driver name in description for console

char *VID_GetExtModeDescription (int mode)
{
	static char	pinfo[40];
	vmode_t		*pv;

	if ((mode < 0) || (mode >= nummodes))
		return NULL;

	pv = VID_GetModePtr (mode);
	if (modelist[mode].type == MS_FULLDIB)
	{
//		if (!leavecurrentmode)
//		{
			sprintf(pinfo,"%s fullscreen", pv->modedesc);
//		}
//		else
//		{
//			sprintf (pinfo, "Desktop resolution (%dx%d)",
//					 modelist[MODE_FULLSCREEN_DEFAULT].width,
//					 modelist[MODE_FULLSCREEN_DEFAULT].height);
//		}
	}
	else
	{
		if (modestate == MS_WINDOWED)
			sprintf(pinfo, "%s windowed", pv->modedesc);
		else
			sprintf(pinfo, "windowed");
	}

	return pinfo;
}


/*
=================
VID_DescribeCurrentMode_f
=================
*/
void VID_DescribeCurrentMode_f (void)
{
	Con_Printf ("%s\n", VID_GetExtModeDescription (vid_modenum));
}


/*
=================
VID_NumModes_f
=================
*/
void VID_NumModes_f (void)
{

	if (nummodes == 1)
		Con_Printf ("%d video mode is available\n", nummodes);
	else
		Con_Printf ("%d video modes are available\n", nummodes);
}


/*
=================
VID_DescribeMode_f
=================
*/
void VID_DescribeMode_f (void)
{
	int		t, modenum;
	
	modenum = Q_atoi (Cmd_Argv(1));

	t = leavecurrentmode;
	leavecurrentmode = 0;

	Con_Printf ("%s\n", VID_GetExtModeDescription (modenum));

	leavecurrentmode = t;
}


/*
=================
VID_DescribeModes_f
=================
*/
void VID_DescribeModes_f (void)
{
	int			i, lnummodes, t;
	char		*pinfo;
	vmode_t		*pv;

	lnummodes = VID_NumModes ();

	t = leavecurrentmode;
	leavecurrentmode = 0;

	for (i=1 ; i<lnummodes ; i++)
	{
		pv = VID_GetModePtr (i);
		pinfo = VID_GetExtModeDescription (i);
		Con_Printf ("%2d: %s\n", i, pinfo);
	}

	leavecurrentmode = t;
}


void VID_InitDIB (HINSTANCE hInstance)
{
	WNDCLASS		wc;

	/* Register the frame class */
    wc.style         = 0;
    wc.lpfnWndProc   = (WNDPROC)MainWndProc;
    wc.cbClsExtra    = 0;
    wc.cbWndExtra    = 0;
    wc.hInstance     = hInstance;
    wc.hIcon         = 0;
    wc.hCursor       = LoadCursor (NULL,IDC_ARROW);
	wc.hbrBackground = NULL;
    wc.lpszMenuName  = 0;
    wc.lpszClassName = "WinQuake";

    if (!RegisterClass (&wc) )
		Sys_Error ("Couldn't register window class");

	modelist[0].type = MS_WINDOWED;

	if (COM_CheckParm("-width"))
		modelist[0].width = Q_atoi(com_argv[COM_CheckParm("-width")+1]);
	else
		modelist[0].width = 640;

	if (modelist[0].width < 320)
		modelist[0].width = 320;

	if (COM_CheckParm("-height"))
		modelist[0].height= Q_atoi(com_argv[COM_CheckParm("-height")+1]);
	else
		modelist[0].height = modelist[0].width * 240/320;

	if (modelist[0].height < 240)
		modelist[0].height = 240;

	sprintf (modelist[0].modedesc, "%dx%d",
			 modelist[0].width, modelist[0].height);

	modelist[0].modenum = MODE_WINDOWED;
	modelist[0].dib = 1;
	modelist[0].fullscreen = 0;
	modelist[0].halfscreen = 0;
	modelist[0].bpp = 0;

	nummodes = 1;
}


/*
=================
VID_InitFullDIB
=================
*/
void VID_InitFullDIB (HINSTANCE hInstance)
{
	DEVMODE	devmode;
	int		i, modenum, originalnummodes, existingmode;
	BOOL	stat;
	int		DesiredBPP;

	// enumerate > 8 bpp modes
	originalnummodes = nummodes;

	// take 32 bpp modes first, then 16
	for (DesiredBPP = 32; DesiredBPP > 0; DesiredBPP -= 16)
	{
		// start at mode -1 because we're going to increment it as soon as we enter the next loop
		modenum = -1;

		// enumerate modes
		do
		{
			// increment it here rather than at the end because we do a few continues later on
			modenum++;

			// get the current mode
			stat = EnumDisplaySettings (NULL, modenum, &devmode);

			// make sure that we got a mode we can use before testing anything!!!
			// EnumDisplaySettings returns 0 when it runs out of modes, so don't enumerate any more
			if (!stat) break;

			// check it's properties.  bpp must be what we're looking for, and we ignore modes below 640 x 480
			if (devmode.dmBitsPerPel != DesiredBPP) continue;
			if (devmode.dmPelsWidth < 640) continue;
			if (devmode.dmPelsHeight < 480) continue;
			if (devmode.dmPelsWidth > MAXWIDTH) continue;
			if (devmode.dmPelsHeight > MAXHEIGHT) continue;

			// no room left for any more modes!!!
			if (nummodes >= MAX_MODE_LIST) break;

			// so far so good - now see if it's already in the list because EnumDisplaySettings
			// reports modes that have the same width/height/bpp but different refresh rates
			// do this before the CDS test
			for (i = originalnummodes, existingmode = 0; i < nummodes; i++)
			{
				if ((devmode.dmPelsWidth == modelist[i].width) &&
					(devmode.dmPelsHeight == modelist[i].height) &&
					(devmode.dmBitsPerPel == modelist[i].bpp))
				{
					// found a duplicate
					existingmode = 1;
					break;
				}
			}

			// don't add duplicates
			if (existingmode) continue;

			// set the fields we want to test
			devmode.dmFields = DM_BITSPERPEL | DM_PELSWIDTH | DM_PELSHEIGHT;

			// test the mode
			if (ChangeDisplaySettings (&devmode, CDS_TEST | CDS_FULLSCREEN) != DISP_CHANGE_SUCCESSFUL) continue;

			// add this mode to the list
			modelist[nummodes].type = MS_FULLDIB;
			modelist[nummodes].width = devmode.dmPelsWidth;
			modelist[nummodes].height = devmode.dmPelsHeight;
			modelist[nummodes].modenum = 0;
			modelist[nummodes].dib = 1;
			modelist[nummodes].fullscreen = 1;
			modelist[nummodes].bpp = devmode.dmBitsPerPel;
			sprintf (modelist[nummodes].modedesc, "%d x %d x %d", devmode.dmPelsWidth, devmode.dmPelsHeight, devmode.dmBitsPerPel);

			// go to the next mode slot
			nummodes++;
		} while (stat);
	}

	if (nummodes == originalnummodes) Con_SafePrintf ("No fullscreen DIB modes found\n");
}


static void Check_Gamma (unsigned char *pal)
{
	float	f, inf;
	unsigned char	palette[768];
	int		i;

	if ((i = COM_CheckParm("-gamma")) == 0)
		vid_gamma = 1;
	else
		vid_gamma = Q_atof(com_argv[i+1]);

	for (i=0 ; i<768 ; i++)
	{
		f = pow ( (pal[i]+1)/256.0 , vid_gamma );
		inf = f*255 + 0.5;
		if (inf < 0)
			inf = 0;
		if (inf > 255)
			inf = 255;
		palette[i] = inf;
	}

	memcpy (pal, palette, sizeof(palette));
}

/*
===================
VID_Init
===================
*/
void	VID_Init (unsigned char *palette)
{
	int		i;
	int		basenummodes, width, height, bpp;
	HDC		hdc;
	DEVMODE	devmode;

	memset(&devmode, 0, sizeof(devmode));

	Cvar_RegisterVariable (&vid_mode);
	Cvar_RegisterVariable (&vid_wait);
	Cvar_RegisterVariable (&vid_nopageflip);
	Cvar_RegisterVariable (&_vid_wait_override);
	Cvar_RegisterVariable (&_vid_default_mode);
	Cvar_RegisterVariable (&_vid_default_mode_win);
	Cvar_RegisterVariable (&vid_config_x);
	Cvar_RegisterVariable (&vid_config_y);
	Cvar_RegisterVariable (&vid_stretch_by_2);
	Cvar_RegisterVariable (&_windowed_mouse);

	Cmd_AddCommand ("vid_nummodes", VID_NumModes_f);
	Cmd_AddCommand ("vid_describecurrentmode", VID_DescribeCurrentMode_f);
	Cmd_AddCommand ("vid_describemode", VID_DescribeMode_f);
	Cmd_AddCommand ("vid_describemodes", VID_DescribeModes_f);

	hIcon = LoadIcon (global_hInstance, MAKEINTRESOURCE (IDI_ICON2));

	InitCommonControls();

	VID_InitDIB (global_hInstance);
	basenummodes = nummodes = 1;

	VID_InitFullDIB (global_hInstance);

	// don't CDS on window creation
	leavecurrentmode = true;

	// if any of these are set we need a CDS
	if (COM_CheckParm ("-width")) leavecurrentmode = false;
	if (COM_CheckParm ("-height")) leavecurrentmode = false;
	if (COM_CheckParm ("-bpp")) leavecurrentmode = false;

	if (COM_CheckParm("-window"))
	{
		hdc = GetDC (NULL);

		if (GetDeviceCaps(hdc, RASTERCAPS) & RC_PALETTE)
		{
			Sys_Error ("Can't run in non-RGB mode");
		}

		ReleaseDC (NULL, hdc);

		windowed = true;

		vid_default = MODE_WINDOWED;
	}
	else
	{
		if (nummodes == 1) Sys_Error ("No RGB fullscreen modes available");

		windowed = false;

		// get the desktop device context
		hdc = GetDC (NULL);

		// check for command-line parameters, set to desktop resolution and colour depth if not specified
		if (COM_CheckParm ("-width"))
			width = Q_atoi (com_argv[COM_CheckParm ("-width") + 1]);
		else width = GetDeviceCaps (hdc, HORZRES);

		if (COM_CheckParm ("-height"))
			height = Q_atoi (com_argv[COM_CheckParm ("-height") + 1]);
		else height = GetDeviceCaps (hdc, VERTRES);

		if (COM_CheckParm ("-bpp"))
			bpp = Q_atoi (com_argv[COM_CheckParm ("-bpp") + 1]);
		else bpp = GetDeviceCaps (hdc, BITSPIXEL);

		// release the device context
		ReleaseDC (NULL, hdc);

		// find a mode that matches the selection
		for (i = 1, vid_default = 0; i < nummodes; i++)
		{
			// does it match?
			if ((modelist[i].width == width) && (modelist[i].height == height) && (modelist[i].bpp == bpp))
			{
				// yes!!!
				vid_default = i;
				break;
			}
		}

		// no matching mode found
		if (!vid_default) Sys_Error ("Specified video mode not available");
	}

	vid_initialized = true;

	vid.conwidth = 512; //modelist[vid_default].width;

	vid.conwidth &= 0xfff8; // make it a multiple of eight

	if (vid.conwidth < 320)
		vid.conwidth = 320;

	// pick a conheight that matches with correct aspect
	vid.conheight = vid.conwidth * modelist[vid_default].height / modelist[vid_default].width;

	vid.maxwarpwidth = WARP_WIDTH;
	vid.maxwarpheight = WARP_HEIGHT;
	vid.colormap = host_colormap;
	vid.fullbright = 256 - LittleLong (*((int *)vid.colormap + 2048));

	DestroyWindow (hwnd_dialog);

	Check_Gamma(palette);
	VID_SetPalette (palette);

	VID_SetMode (vid_default, palette);

    maindc = GetDC(mainwindow);
	bSetupPixelFormat(maindc);

    baseRC = wglCreateContext( maindc );
	if (!baseRC)
		Sys_Error ("Could not initialize GL (wglCreateContext failed).\n\nMake sure you in are 65535 color mode, and try running -window.");
    if (!wglMakeCurrent( maindc, baseRC ))
		Sys_Error ("wglMakeCurrent failed");

	GL_Init ();

	vid_realmode = vid_modenum;

	// removed
	vid_menudrawfn = NULL; //VID_MenuDraw;
	vid_menukeyfn = VID_MenuKey;

	strcpy (badmode.modedesc, "Bad mode");
	vid_canalttab = true;
}


//========================================================
// Video menu stuff
//========================================================

extern void M_Menu_Options_f (void);
extern void M_Print (int cx, int cy, char *str);
extern void M_PrintWhite (int cx, int cy, char *str);
extern void M_DrawCharacter (int cx, int line, int num);
extern void M_DrawTransPic (int x, int y, qpic_t *pic);
extern void M_DrawPic (int x, int y, qpic_t *pic);

static int	vid_line, vid_wmodes;

typedef struct
{
	int		modenum;
	char	*desc;
	int		iscur;
} modedesc_t;

#define MAX_COLUMN_SIZE		9
#define MODE_AREA_HEIGHT	(MAX_COLUMN_SIZE + 2)
#define MAX_MODEDESCS		(MAX_COLUMN_SIZE*3)

static modedesc_t	modedescs[MAX_MODEDESCS];

/*
================
VID_MenuDraw
================
*/
void VID_MenuDraw (void)
{
	qpic_t		*p;
	char		*ptr;
	int			lnummodes, i, k, column, row;
	vmode_t		*pv;

	p = Draw_CachePic ("gfx/vidmodes.lmp");
	M_DrawPic ( (320-p->width)/2, 4, p);

	vid_wmodes = 0;
	lnummodes = VID_NumModes ();
	
	for (i=1 ; (i<lnummodes) && (vid_wmodes < MAX_MODEDESCS) ; i++)
	{
		ptr = VID_GetModeDescription (i);
		pv = VID_GetModePtr (i);

		k = vid_wmodes;

		modedescs[k].modenum = i;
		modedescs[k].desc = ptr;
		modedescs[k].iscur = 0;

		if (i == vid_modenum)
			modedescs[k].iscur = 1;

		vid_wmodes++;

	}

	if (vid_wmodes > 0)
	{
		M_Print (2*8, 36+0*8, "Fullscreen Modes (WIDTHxHEIGHTxBPP)");

		column = 8;
		row = 36+2*8;

		for (i=0 ; i<vid_wmodes ; i++)
		{
			if (modedescs[i].iscur)
				M_PrintWhite (column, row, modedescs[i].desc);
			else
				M_Print (column, row, modedescs[i].desc);

			column += 13*8;

			if ((i % VID_ROW_SIZE) == (VID_ROW_SIZE - 1))
			{
				column = 8;
				row += 8;
			}
		}
	}

	M_Print (3*8, 36 + MODE_AREA_HEIGHT * 8 + 8*2,
			 "Video modes must be set from the");
	M_Print (3*8, 36 + MODE_AREA_HEIGHT * 8 + 8*3,
			 "command line with -width <width>");
	M_Print (3*8, 36 + MODE_AREA_HEIGHT * 8 + 8*4,
			 "and -bpp <bits-per-pixel>");
	M_Print (3*8, 36 + MODE_AREA_HEIGHT * 8 + 8*6,
			 "Select windowed mode with -window");
}


/*
================
VID_MenuKey
================
*/
void VID_MenuKey (int key)
{
	switch (key)
	{
	case K_ESCAPE:
		S_LocalSound ("misc/menu1.wav");
		M_Menu_Options_f ();
		break;

	default:
		break;
	}
}
