/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// gl_warp.c -- sky and water polygons

#include "quakedef.h"

extern	model_t	*loadmodel;

int		skytexturenum;

int		solidskytexture;
int		alphaskytexture;
float	speedscale;		// for top sky and bottom sky

msurface_t	*warpface;

extern cvar_t gl_subdivide_size;

// surface vertex array

void BoundPoly (int numverts, float *verts, vec3_t mins, vec3_t maxs)
{
	int		i, j;
	float	*v;

	mins[0] = mins[1] = mins[2] = 9999;
	maxs[0] = maxs[1] = maxs[2] = -9999;

	v = verts;

	for (i = 0; i < numverts; i++)
	{
		for (j = 0; j < 3; j++, v++)
		{
			if (*v < mins[j]) mins[j] = *v;
			if (*v > maxs[j]) maxs[j] = *v;
		}
	}
}

extern int vertcount;

void SubdividePolygon (int numverts, float *verts)
{
	int		i, j, k;
	vec3_t	mins, maxs;
	float	m;
	float	*v;
	vec3_t	front[64], back[64];
	int		f, b;
	float	dist[64];
	float	frac;
	glpoly_t	*poly;
	float	s, t;

	if (numverts > 60) Sys_Error ("numverts = %i", numverts);

	BoundPoly (numverts, verts, mins, maxs);

	for (i = 0; i < 3; i++)
	{
		m = (mins[i] + maxs[i]) * 0.5;

		m = gl_subdivide_size.value * floor (m / gl_subdivide_size.value + 0.5);

		if (maxs[i] - m < 8) continue;
		if (m - mins[i] < 8) continue;

		// cut it
		v = verts + i;

		for (j = 0; j < numverts; j++, v += 3)
			dist[j] = *v - m;

		// wrap cases
		dist[j] = dist[0];
		v -= i;
		VectorCopy (verts, v);

		f = b = 0;
		v = verts;

		for (j = 0; j < numverts; j++, v += 3)
		{
			if (dist[j] >= 0)
			{
				VectorCopy (v, front[f]);
				f++;
			}

			if (dist[j] <= 0)
			{
				VectorCopy (v, back[b]);
				b++;
			}

			if (dist[j] == 0 || dist[j + 1] == 0) continue;

			if ((dist[j] > 0) != (dist[j + 1] > 0))
			{
				// clip point
				frac = dist[j] / (dist[j] - dist[j + 1]);

				for (k = 0; k < 3; k++)
					front[f][k] = back[b][k] = v[k] + frac * (v[3 + k] - v[k]);

				f++;
				b++;
			}
		}

		SubdividePolygon (f, front[0]);
		SubdividePolygon (b, back[0]);
		return;
	}

	poly = Hunk_Alloc (sizeof (glpoly_t));
	poly->verts = Hunk_Alloc (numverts * sizeof (glvertex_t));

	poly->next = warpface->polys;
	warpface->polys = poly;
	poly->numverts = numverts;

	for (i = 0; i < numverts; i++, verts += 3)
	{
		VectorCopy (verts, poly->verts[i].v);

		s = DotProduct (verts, warpface->texinfo->vecs[0]);
		t = DotProduct (verts, warpface->texinfo->vecs[1]);

		poly->verts[i].st[0] = s;
		poly->verts[i].st[1] = t;
	}

	// if the max number of verts increases, reallocate storage for the surface verts
	if (poly->numverts > maxverts)
	{
		maxverts = poly->numverts;
		if (maxverts < 4) maxverts = 4;

		if (surfverts) free (surfverts);

		surfverts = (glvertex_t *) malloc (maxverts * sizeof (glvertex_t));
	}

	// set the primitive type
	poly->PRIMITIVE_TYPE = GL_TRIANGLE_FAN;

	vertcount += poly->numverts;
}


/*
================
GL_SubdivideSurface

Breaks a polygon up along axial 64 unit
boundaries so that turbulent and sky warps
can be done reasonably.
================
*/
void GL_SubdivideSurface (msurface_t *surf)
{
	vec3_t		verts[64];
	int			numverts;
	int			i;
	int			lindex;
	float		*vec;

	warpface = surf;

	// convert edges back to a normal polygon
	numverts = 0;

	for (i = 0; i < surf->numedges; i++)
	{
		lindex = loadmodel->surfedges[surf->firstedge + i];

		if (lindex > 0)
			vec = loadmodel->vertexes[loadmodel->edges[lindex].v[0]].position;
		else
			vec = loadmodel->vertexes[loadmodel->edges[-lindex].v[1]].position;

		VectorCopy (vec, verts[numverts]);
		numverts++;
	}

	SubdividePolygon (numverts, verts[0]);

	// untransformed initially
	surf->passframe = -1;
}

//=========================================================



/*
=============
EmitSkyPolys
=============
*/
void EmitSkyPolys (msurface_t *surf)
{
}

/*
=================
R_DrawSkyChain
=================
*/
float r_farclip;
cvar_t r_skyspeed = {"r_skyspeed", "5", true};

float rotateBack = 0;
float rotateFore = 0;

float skytexes[440];
float skyverts[660];

void R_DrawSkyArrays (void)
{
	int i;
	int j;

	int numverts = 0;
	int start;

	for (i = 0; i < 10; i++)
	{
		start = numverts;

		for (j = 0; j <= 10; j++)
			numverts += 2;

		glDrawArrays (GL_TRIANGLE_STRIP, start, (numverts - start));
	}
}


void R_DrawSkyChain (qboolean skydraw)
{
	// sky scaling
	float fullscale;
	float halfscale;
	float reducedfull;
	float reducedhalf;

	// do these calcs even if we're not drawing
	// sky rotation
	// bound rotation speed 0 to 100
	if (r_skyspeed.value < 0) r_skyspeed.value = 0;
	if (r_skyspeed.value > 100) r_skyspeed.value = 100;

	// always rotate even if we're paused!!!
	rotateBack = anglemod (realtime * (2.5 * r_skyspeed.value));
	rotateFore = anglemod (realtime * (4.0 * r_skyspeed.value));

	if (!skydraw) return;

	// calculate the scales
	fullscale = r_farclip / 4096;
	halfscale = r_farclip / 8192;
	reducedfull = (r_farclip / 4096) * 0.9;
	reducedhalf = (r_farclip / 8192) * 0.9;

	// sky texture scaling
	// previous releases made a tiled version of the sky texture.  here i just shrink it using the
	// texture matrix, for much the same effect
	glMatrixMode (GL_TEXTURE);
	glLoadIdentity ();
	glScalef (4, 2, 2);
	glMatrixMode (GL_MODELVIEW);

	glEnableClientState (GL_VERTEX_ARRAY);
	glVertexPointer (3, GL_FLOAT, 3 * sizeof (float), skyverts);

	glClientActiveTexture (GL_TEXTURE0);
	glEnableClientState (GL_TEXTURE_COORD_ARRAY);
	glTexCoordPointer (2, GL_FLOAT, 2 * sizeof (float), skytexes);

	// background
	// ==========
	// go to a new matrix
	glPushMatrix ();

	// center it on the players position
	glTranslatef (r_origin[0], r_origin[1], r_origin[2]);

	// flatten the sphere
	glScalef (fullscale, fullscale, halfscale);

	// orient it so that the poles are unobtrusive
	glRotatef (-90, 1, 0, 0);

	// make it not always at right angles to the player
	glRotatef (-22, 0 ,1, 0);

	// rotate it around the poles
	glRotatef (-rotateBack, 0, 0, 1);

	// solid sky texture
	GL_BindTexture (solidskytexture);

	// draw the sphere
	R_DrawSkyArrays ();

	// restore the previous matrix
	glPopMatrix ();

	// foreground
	// ==========
	// go to a new matrix
	glPushMatrix ();

	// center it on the players position
	glTranslatef (r_origin[0], r_origin[1], r_origin[2]);

	// flatten the sphere and shrink it a little - the reduced scale prevents artefacts appearing when
	// corners on the skysphere may potentially overlap
	glScalef (reducedfull, reducedfull, reducedhalf);

	// orient it so that the poles are unobtrusive
	glRotatef (-90, 1, 0, 0);

	// make it not always at right angles to the player
	glRotatef (-22, 0 ,1, 0);

	// rotate it around the poles
	glRotatef (-rotateFore, 0, 0, 1);

	// blend mode
	glEnable (GL_BLEND);

	// alpha sky texture
	GL_BindTexture (alphaskytexture);

	// draw the sphere
	R_DrawSkyArrays ();

	glDisableClientState (GL_TEXTURE_COORD_ARRAY);
	glDisableClientState (GL_VERTEX_ARRAY);

	// back to normal mode
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable (GL_BLEND);

	// restore the previous matrix
	glPopMatrix ();

	glMatrixMode (GL_TEXTURE);
	glLoadIdentity ();
	glMatrixMode (GL_MODELVIEW);
}


//===============================================================


void R_InitSkySphere (void)
{
	float drho = 0.3141592653589;
	float dtheta = 0.6283185307180;

	float ds = 0.1;
	float dt = 0.1;

	float t = 1.0f;	
	float s = 0.0f;

	int i;
	int j;

	int vertspos = 0;
	int texespos = 0;

	for (i = 0; i < 10; i++)
	{
		float rho = (float) i * drho;
		float srho = (float) (sin (rho));
		float crho = (float) (cos (rho));
		float srhodrho = (float) (sin (rho + drho));
		float crhodrho = (float) (cos (rho + drho));

		s = 0.0f;

		for (j = 0; j <= 10; j++)
		{
			float theta = (j == 10) ? 0.0f : j * dtheta;
			float stheta = (float) (-sin (theta));
			float ctheta = (float) (cos (theta));

			skytexes[texespos++] = s * 2;
			skytexes[texespos++] = t * 2;

			skyverts[vertspos++] = stheta * srho * 4096.0;
			skyverts[vertspos++] = ctheta * srho * 4096.0;
			skyverts[vertspos++] = crho * 4096.0;

			skytexes[texespos++] = s * 2;
			skytexes[texespos++] = (t - dt) * 2;

			skyverts[vertspos++] = stheta * srhodrho * 4096.0;
			skyverts[vertspos++] = ctheta * srhodrho * 4096.0;
			skyverts[vertspos++] = crhodrho * 4096.0;

			s += ds;
		}

		t -= dt;
	}
}


/*
=============
R_InitSky

A sky texture is 256*128, with the right side being a masked overlay
==============
*/
void R_InitSky (texture_t *mt)
{
	int			i, j, p;
	byte		*src;
	unsigned	trans[128 * 128];
	byte		*rgba;
	extern int	skytexturenum;
	int			topalpha;
	int			average[3];
	int			gs;
	gltexture_t glt;

	glt.width = 128;
	glt.height = 128;
	glt.flags = TEX_32BPP;
	glt.data = (byte *) trans;

	src = (byte *) mt + mt->offsets[0];

	average[0] = average[1] = average[2] = 0;

	for (i = 0; i < 128; i++)
	{
		for (j = 0; j < 128; j++)
		{
			p = src[i * 256 + j + 128];

			trans[(i * 128) + j] = d_8to24tableBGRA[p];

			// accumulate the average at RGBA
			average[0] += ((byte *) &d_8to24tableRGBA[p]) [0];
			average[1] += ((byte *) &d_8to24tableRGBA[p]) [1];
			average[2] += ((byte *) &d_8to24tableRGBA[p]) [2];
		}
	}

	average[0] /= (128 * 128);
	average[1] /= (128 * 128);
	average[2] /= (128 * 128);

	if (!solidskytexture) GL_AllocTexture (&solidskytexture);

	sprintf (glt.identifier, "%s_solid", mt->name);
	glt.texnum = solidskytexture;

	GL_UploadTexture (&glt);

	for (i = 0; i < 128; i++)
	{
		for (j = 0; j < 128; j++)
		{
			p = src[i * 256 + j];

			if (p == 0)
			{
				// set to average (swapped to bgra)
				trans[(i * 128) + j] = ((0 << 24) + (average[0] << 16) + (average[1] << 8) + (average[2] << 0));
				continue;
			}

			rgba = (byte *) &d_8to24tableRGBA[p];

			gs = (int) rgba[0] * 30;
			gs += (int) rgba[1] * 59;
			gs += (int) rgba[2] * 11;

			topalpha = gs / 50;

			if (topalpha > 255) topalpha = 255;

			// swap to bgra
			trans[(i * 128) + j] = ((topalpha << 24) + (rgba[0] << 16) + (rgba[1] << 8) + (rgba[2] << 0));
		}
	}

	if (!alphaskytexture) GL_AllocTexture (&alphaskytexture);

	sprintf (glt.identifier, "%s_alpha", mt->name);
	glt.texnum = alphaskytexture;

	GL_UploadTexture (&glt);
}

