/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_misc.c

#include "gl_state.h"

// needed for resource texture loading
#include "resource.h"

// we have no easy way of getting at the worldtype from client side, so let's do a ghastly vile hack
int worldtype;

// chained list of all textures in the world and all bmodels
gltexture_t *TextureList;

void R_InitSurfStuff (void);
void R_InitBloomTextures(void);
void R_ClearParticles (void);
void GL_BuildLightmaps (void);
void GL_TextureAging (void);

/*
==========================================================================================
testing of blend funcs
----------------------
this is a good way to experiment with different blending modes to see what looks
best - just replace a call to glBlendFunc with this.  be careful to only test one
blend at a time (e.g. water/light/caustic/etc) or you may get confused!!!
==========================================================================================
*/
GLenum bFactors[11] =
{
	GL_ZERO,					// 0
	GL_ONE,						// 1
	GL_SRC_COLOR,				// 2
	GL_ONE_MINUS_SRC_COLOR,		// 3
	GL_DST_COLOR,				// 4
	GL_ONE_MINUS_DST_COLOR,		// 5
	GL_SRC_ALPHA,				// 6
	GL_ONE_MINUS_SRC_ALPHA,		// 7
	GL_DST_ALPHA,				// 8
	GL_ONE_MINUS_DST_ALPHA,		// 9
	GL_SRC_ALPHA_SATURATE		// 10
};

int gl_blend_src;
int gl_blend_dst;

void GL_BlendTest (void)
{
	if (gl_blend_src < 0) gl_blend_src = 0;
	if (gl_blend_src > 10) gl_blend_src = 10;

	if (gl_blend_dst < 0) gl_blend_dst = 0;
	if (gl_blend_dst > 10) gl_blend_dst = 10;

	glBlendFunc (bFactors[gl_blend_src], bFactors[gl_blend_dst]);
}


void GL_Blend_f (void)
{
	if (Cmd_Argc () == 1)
	{
		gl_blend_src = 0;
		gl_blend_dst = 0;
	}
	else if (Cmd_Argc () == 2)
	{
		gl_blend_src = atoi (Cmd_Argv (1));
		gl_blend_dst = 0;
	}
	else
	{
		gl_blend_src = atoi (Cmd_Argv (1));
		gl_blend_dst = atoi (Cmd_Argv (2));
	}
}


// generic bounding box drawing.  might be useful if we ever do occlusion queries
// (although Quake's testing really renders them redundant unless we ever do something
// sexy with particles) and also handy to have for testing
void R_DrawBBox (vec3_t mins, vec3_t maxs)
{
	glDisable (GL_CULL_FACE);
	glDisable (GL_TEXTURE_2D);

	glColor3f (1, 0, 0);

	glBegin (GL_QUADS);

	glVertex3f (mins[0], mins[1], mins[2]);
	glVertex3f (mins[0], maxs[1], mins[2]);
	glVertex3f (mins[0], maxs[1], maxs[2]);
	glVertex3f (mins[0], mins[1], maxs[2]);

	glVertex3f (maxs[0], mins[1], mins[2]);
	glVertex3f (maxs[0], maxs[1], mins[2]);
	glVertex3f (maxs[0], maxs[1], maxs[2]);
	glVertex3f (maxs[0], mins[1], maxs[2]);

	glVertex3f (mins[0], mins[1], maxs[2]);
	glVertex3f (mins[0], maxs[1], maxs[2]);
	glVertex3f (maxs[0], maxs[1], maxs[2]);
	glVertex3f (maxs[0], mins[1], maxs[2]);

	glVertex3f (mins[0], mins[1], mins[2]);
	glVertex3f (mins[0], maxs[1], mins[2]);
	glVertex3f (maxs[0], maxs[1], mins[2]);
	glVertex3f (maxs[0], mins[1], mins[2]);

	glVertex3f (mins[0], mins[1], mins[2]);
	glVertex3f (mins[0], mins[1], maxs[2]);
	glVertex3f (maxs[0], mins[1], maxs[2]);
	glVertex3f (maxs[0], mins[1], mins[2]);

	glVertex3f (mins[0], maxs[1], mins[2]);
	glVertex3f (mins[0], maxs[1], maxs[2]);
	glVertex3f (maxs[0], maxs[1], maxs[2]);
	glVertex3f (maxs[0], maxs[1], mins[2]);

	glEnd ();

	glColor3f (1, 1, 1);

	glEnable (GL_CULL_FACE);
	glEnable (GL_TEXTURE_2D);
}


/*
==================
R_InitTextures
==================
*/
void R_InitTextures (void)
{
	// create a simple checkerboard texture for the default
	r_notexture_mip = Hunk_AllocName (sizeof(texture_t) + 16*16+8*8+4*4+2*2, "notexture");
	
	r_notexture_mip->width = r_notexture_mip->height = 16;
}


void R_InitParticleTexture (void)
{
	// try it from the resource
	if (GL_LoadBitmapResourceTexture (&particletexture, IDB_BITMAP2, (TEX_MIPMAP | TEX_ALPHAMASK))) return;

	// crash 'n' burn, baybeeee
	// this is ok to do cos if it ain't in the resource then the exe has gone wrong and we have bigger problems
	Sys_Error ("Could not init particle texture");
}

extern float XHairS[];
extern float XHairT[];
extern float XHairSAdd;
extern float XHairTAdd;

void R_InitXHairTexture (void)
{
	// try it from the resource
	if (GL_LoadBitmapResourceTexture (&CrosshairTexture, IDB_BITMAP3, TEX_ALPHAMASK)) return;

	// crash 'n' burn, baybeeee
	// this is ok to do cos if it ain't in the resource then the exe has gone wrong and we have bigger problems
	Sys_Error ("Could not init crosshair texture");
}


void fractalnoise (byte *noise, int size, int startgrid)
{
	int x;
	int y;
	int g;
	int g2;
	int amplitude;
	int min;
	int max;
	int size1 = size - 1;
	int sizepower;
	int gridpower;
	int *noisebuf;

#define n(x,y) noisebuf[((y) & size1) * size + ((x) & size1)]

	for (sizepower = 0; (1 << sizepower) < size; sizepower++);

	if (size != (1 << sizepower)) Sys_Error ("fractalnoise: size must be power of 2\n");

	for (gridpower = 0; (1 << gridpower) < startgrid; gridpower++);

	if (startgrid != (1 << gridpower)) Sys_Error ("fractalnoise: grid must be power of 2\n");

	if (startgrid < 0) startgrid = 0;
	if (startgrid > size) startgrid = size;

	amplitude = 0xFFFF; // this gets halved before use
	noisebuf = malloc (size * size * sizeof (int));
	memset (noisebuf, 0, size * size * sizeof (int));

	for (g2 = startgrid; g2; g2 >>= 1)
	{
		// brownian motion (at every smaller level there is random behavior)
		amplitude >>= 1;

		for (y = 0; y < size; y += g2)
		{
			for (x = 0; x < size; x += g2)
			{
				// keep as a rand cos it doesn't happen during gameplay
				n (x,y) += (rand() & amplitude);
			}
		}

		g = g2 >> 1;

		if (g)
		{
			// subdivide, diamond-square algorithm (really this has little to do with squares)
			// diamond
			for (y = 0;y < size;y += g2)
			{
				for (x = 0;x < size;x += g2)
				{
					n (x + g, y + g) = (n (x, y) + n (x + g2, y) + n (x, y + g2) + n (x + g2, y + g2)) >> 2;
				}
			}

			// square
			for (y = 0; y < size; y += g2)
			{
				for (x = 0; x < size; x += g2)
				{
					n (x + g, y) = (n (x, y) + n (x + g2, y) + n (x + g, y - g) + n (x + g, y + g)) >> 2;
					n (x, y + g) = (n (x, y) + n (x, y + g2) + n (x - g, y + g) + n (x + g, y + g)) >> 2;
				}
			}
		}
	}

	// find range of noise values
	min = max = 0;

	for (y = 0; y < size; y++)
	{
		for (x = 0; x < size; x++)
		{
			if (n (x, y) < min) min = n (x, y);
			if (n (x, y) > max) max = n (x, y);
		}
	}

	max -= min;
	max++;

	// normalize noise and copy to output
	for (y = 0; y < size; y++)
	{
		for (x = 0; x < size; x++)
		{
			*noise++ = (byte) (((n (x, y) - min) * 256) / max);
		}
	}
	free (noisebuf);
#undef n
}


void R_InitDetailTexture (void)
{
	int x, y, light;
	float vc[3], vx[3], vy[3], vn[3], lightdir[3];

// increase this if you want
#define DETAILRESOLUTION 256

	byte data[DETAILRESOLUTION][DETAILRESOLUTION][4], noise[DETAILRESOLUTION][DETAILRESOLUTION];

	// fall back on internal
	lightdir[0] = 0.5;
	lightdir[1] = 1;
	lightdir[2] = -0.25;
	VectorNormalize (lightdir);

	fractalnoise (&noise[0][0], DETAILRESOLUTION, DETAILRESOLUTION >> 4);

	for (y = 0; y < DETAILRESOLUTION; y++)
	{
		for (x = 0; x < DETAILRESOLUTION; x++)
		{
			vc[0] = x;
			vc[1] = y;
			vc[2] = noise[y][x] * (1.0f / 32.0f);

			vx[0] = x + 1;
			vx[1] = y;
			vx[2] = noise[y][(x + 1) % DETAILRESOLUTION] * (1.0f / 32.0f);

			vy[0] = x;
			vy[1] = y + 1;
			vy[2] = noise[(y + 1) % DETAILRESOLUTION][x] * (1.0f / 32.0f);

			VectorSubtract (vx, vc, vx);
			VectorSubtract (vy, vc, vy);
			CrossProduct (vx, vy, vn);
			VectorNormalize (vn);

			light = 128 - DotProduct (vn, lightdir) * 128;
			if (light < 0) light = 0;
			if (light > 255) light = 255;

			// rescale light to maintain the same colour balance
			light = (light * 5) / 6;

			// store it
			data[y][x][0] = data[y][x][1] = data[y][x][2] = light;
			data[y][x][3] = 255;
		}
	}

	// and upload it
	// help m'boab!!!  we get a *major* dose of sparklies here with nearest filtering...!
	r_detailtexture.gltexnum = GL_UploadTextureToOpenGL (data, DETAILRESOLUTION, DETAILRESOLUTION, TEX_MIPMAP);

	// now we hack the brightness again to maintain colour balance for the 3 tmu vesion.  (in 
	// an ideal world opengl would allow multitexturing to be able to reproduce the same effects
	// as multiple passes, but this isn't an ideal world, and to be honest, the way the ARB seem
	// to get bogged down in squabbling over minor points of principle, it's never likely to 
	// happen.  what we really need is a closed API under the control of a single company.  oh...
	// wait a minute...  translation: i'm pissed off that i have to do this).
	for (y = 0; y < DETAILRESOLUTION; y++)
	{
		for (x = 0; x < DETAILRESOLUTION; x++)
		{
			// grab the pixel.  as these are greyscale, any component will do.  lets see..... red!
			light = data[y][x][0];

			// rescale again to maintain the same colour balance
			// (trial and error can be a very good friend sometimes).
			light = (light * 41) / 44;

			// store it
			data[y][x][0] = data[y][x][1] = data[y][x][2] = light;
			data[y][x][3] = 255;
		}
	}

	// and upload it in the second texture object (just reuse the luma object here)
	// help m'boab!!!  we get a *major* dose of sparklies here with nearest filtering...!
	r_detailtexture.lumatexnum = GL_UploadTextureToOpenGL (data, DETAILRESOLUTION, DETAILRESOLUTION, TEX_MIPMAP);
#undef DETAILRESOLUTION
}

void R_InitCausticTexture (void)
{
	// try it from the resource version
	if (GL_LoadBitmapResourceTexture (&r_caustictexture, IDB_BITMAP4, TEX_MIPMAP)) return;

	// crash 'n' burn, baybeeee
	// this is ok to do cos if it ain't in the resource then the exe has gone wrong and we have bigger problems
	Sys_Error ("Could not init caustic texture");
}

void R_InitColourTexture (gltexture_t *ColourTexture, byte r, byte g, byte b)
{
	int x, y;
	byte data[8][8][4];

	// make the texture
	for (x = 0; x < 8; x++)
	{
		for (y = 0; y < 8; y++)
		{
			data[y][x][0] = r;
			data[y][x][1] = g;
			data[y][x][2] = b;
			data[y][x][3] = 255;
		}
	}

	// upload to opengl - these are small and one of them gets updated at runtime, so no compression, please
	ColourTexture->gltexnum = GL_UploadTextureToOpenGL (data, 8, 8, (TEX_MIPMAP | TEX_NOCOMPRESS));
}

void R_InitHCTextures (void)
{
	R_InitParticleTexture ();
	R_InitXHairTexture ();
	R_InitDetailTexture ();
	R_InitCausticTexture ();

	// initialize colour textures
	R_InitColourTexture (&r_white, 255, 255, 255);
	R_InitColourTexture (&r_black, 0, 0, 0);
	R_InitColourTexture (&r_red, 255, 0, 0);
	R_InitColourTexture (&r_green, 0, 255, 0);
	R_InitColourTexture (&r_blue, 0, 0, 255);
	R_InitColourTexture (&r_gamma, 64, 64, 64);
}

void R_LavaSplash_f (void)
{
	R_LavaSplash (r_origin);
}

/*
===============
R_Init
===============
*/
void R_Init (void)
{	
	extern byte *hunk_base;
	extern cvar_t gl_finish;

	Cmd_AddCommand ("timerefresh", R_TimeRefresh_f);	
	Cmd_AddCommand ("pointfile", R_ReadPointFile_f);	
	Cmd_AddCommand ("lavasplash", R_LavaSplash_f);	

	Cvar_RegisterVariable (&r_farclip);
	Cvar_RegisterVariable (&r_skyspeed);
	Cvar_RegisterVariable (&r_norefresh);
	Cvar_RegisterVariable (&r_lightmap);
	Cvar_RegisterVariable (&r_drawentities);
	Cvar_RegisterVariable (&r_drawviewmodel);
	Cvar_RegisterVariable (&r_shadows);
	Cvar_RegisterVariable (&r_wateralpha);
	Cvar_RegisterVariable (&r_dynamic);
	Cvar_RegisterVariable (&r_novis);
	Cvar_RegisterVariable (&r_speeds);

	Cvar_RegisterVariable(&r_bloom);
	Cvar_RegisterVariable (&r_bloom_alpha);
	Cvar_RegisterVariable (&r_bloom_diamond_size);
	Cvar_RegisterVariable (&r_bloom_intensity);
	Cvar_RegisterVariable (&r_bloom_darken);
	Cvar_RegisterVariable (&r_bloom_sample_size);
	Cvar_RegisterVariable (&r_bloom_fast_sample);

	Cvar_RegisterVariable (&gl_finish);
	Cvar_RegisterVariable (&gl_clear);

	Cvar_RegisterVariable (&gl_cull);
	Cvar_RegisterVariable (&gl_smoothmodels);
	Cvar_RegisterVariable (&gl_affinemodels);
	Cvar_RegisterVariable (&gl_polyblend);
	Cvar_RegisterVariable (&gl_playermip);
	Cvar_RegisterVariable (&gl_nocolors);

	Cvar_RegisterVariable (&gl_keeptjunctions);
	Cvar_RegisterVariable (&gl_reporttjunctions);
	Cvar_RegisterVariable (&gl_doubleeyes);
	Cvar_RegisterVariable (&gl_lockpvs);

	Cmd_AddCommand ("gl_blend", GL_Blend_f);

	R_InitParticles ();
	R_InitHCTextures ();
	R_InitSurfStuff ();
	R_InitBloomTextures();

	GL_MakeTextureObjects (16, playertextures);
}

/*
===============
R_TranslatePlayerSkin

Translates a skin texture by the per-player color lookup
===============
*/
void R_TranslatePlayerSkin (int playernum)
{
	int				top, bottom;
	byte			   translate[256];
	int				i, s;
	model_t			*model;
	aliashdr_t		*paliashdr;
	byte			   *original;
	int				inwidth, inheight;
	extern	byte	**player_8bit_texels_tbl;
	byte			   translated[320*200];
	gltexture_t		glt;

	top = cl.scores[playernum].colors & 0xf0;
	bottom = (cl.scores[playernum].colors &15)<<4;

	for (i=0 ; i<256 ; i++) translate[i] = i;

	for (i=0 ; i<16 ; i++)
	{
		if (top < 128)	// the artists made some backwards ranges.  sigh.
			translate[TOP_RANGE+i] = top+i;
		else
			translate[TOP_RANGE+i] = top+15-i;
				
		if (bottom < 128)
			translate[BOTTOM_RANGE+i] = bottom+i;
		else
			translate[BOTTOM_RANGE+i] = bottom+15-i;
	}

	//
	// locate the original skin pixels
	//
	currententity = &cl_entities[1+playernum];
	model = currententity->model;

	if (!model)	return;						// player doesn't have a model yet
	if (model->type != mod_alias) return;	// only translate skins on alias models

	paliashdr = (aliashdr_t *)Mod_Extradata (model);
	s = paliashdr->skinwidth * paliashdr->skinheight;

	if (currententity->skinnum < 0 || currententity->skinnum >= paliashdr->numskins)
	{
		Con_Printf("(%d): Invalid player skin #%d\n", playernum, currententity->skinnum);
		original = (byte *)paliashdr + paliashdr->texels[0];
	}
	else 
	{
		original = (byte *)paliashdr + paliashdr->texels[currententity->skinnum];
	}

	if (s & 3) 
	{
		Sys_Error ("R_TranslateSkin: s&3");
	}
	inwidth = paliashdr->skinwidth;
	inheight = paliashdr->skinheight;

	for (i=0 ; i<s ; i+=4)
	{
		translated[i] = translate[original[i]];
		translated[i+1] = translate[original[i+1]];
		translated[i+2] = translate[original[i+2]];
		translated[i+3] = translate[original[i+3]];
	}

	// we already have a texture object for this, so just bind it
    glBindTexture (GL_TEXTURE_2D, playertextures[playernum]);

	// and load it into a dummy glt - don;t make a texture object cos we already have one from the mdl loading!!!
	GL_LoadQuakeTexture (&glt, "***blah***", translated, paliashdr->skinwidth, paliashdr->skinheight, (TEX_NOGEN | TEX_MIPMAP));
}


/*
================================================================================================
TEXTURE LIST BUILDING

These routines build a single chained texture list of all textures used by all bmodels in the
current map, including BSP bmodels, so that all brush-based geometry stuff can be drawn in a
single pass.
================================================================================================
*/
void R_AddTexturesToListForModel (model_t *mod)
{
	int         i;
	texture_t   *tex;
	gltexture_t *texcheck;

	Con_DPrintf ("Adding Textures to list for %s\n", mod->name);

	for (i = 0; i < mod->numtextures; i++)
	{
		tex = mod->textures[i];

		// this only happens in e2m3
		if (!tex) continue;

		// lets be extra paranoid here
		if (!tex->gl_texture) continue;
		if (!tex->gl_texture->gltexnum) continue;

		// see is it already in the texture list
		for (texcheck = TextureList; texcheck; texcheck = texcheck->TextureList)
		{
			// already there
			if (texcheck == tex->gl_texture) break;
		}

		// texcheck will be NULL if it's not already present
		if (texcheck) continue;

		// now add it to the chain
		tex->gl_texture->TextureList = TextureList;
		TextureList = tex->gl_texture;

		// null the initial texturechain
		tex->gl_texture->texturechain = NULL;
	}
}

void R_BuildTextureList (void)
{
	int      x;
	model_t  *mod;

	// initially NULL
	TextureList = NULL;

	// the valid model list starts at 1 for some reason
	// i'm sure i knew once what model 0 was but can't for the life of me remember now
	for (x = 1; x < MAX_MODELS; x++)
	{
		// get the model
		mod = cl.model_precache[x];

		// no more models
		if (!mod) break;

		// we're only interested in brush models
		if (mod->type == mod_brush) 
		{
			R_AddTexturesToListForModel (mod);
		}
	}
}

char *R_FindWorldSpawnEntry (model_t *mod, char *KeyToFind)
{
	char		*entlump = mod->entities;
	char		WSQuotedKey[128];
	static char WSValue[128];
	int			seekOffset;
	int			i;

	// build the key we wanna find and get the seekoffset
	// we add the quotes around it to ensure that we get an exact match, i.e. so that light for example
	// (not that you'd have light in worldspawn, but you know what i mean) isn't reported as light_blah
	sprintf (WSQuotedKey, "\"%s\"", KeyToFind);

	seekOffset = strlen (WSQuotedKey) + 1;

	// clear to NULLs
	memset (WSValue, 0, sizeof (WSValue));

	while (1)
	{
		// end of the worldspawn entity
		if (*entlump == '}') return WSValue;

		if (!strncmp (entlump, WSQuotedKey, seekOffset - 1))
		{
			goto Eat_Your_Heart_Out_Bono;
		}
		entlump++;

		continue;

Eat_Your_Heart_Out_Bono:;
		// I've just found what I'm looking for!!!!
		// advance by the seekOffset
		entlump += seekOffset;

		// skip whitespace
		while (1)
		{
			if (*entlump == '\"') break;
			entlump++;
		}

		// skip the quote we just found
		entlump++;

		// copy to WSValue
		for (i = 0; i < 127; i++)
		{
			if (*entlump == '\"') break;

			WSValue[i] = *entlump;
			entlump++;
		}
		return WSValue;
	}

	// never gets here, just keep the compiler happy
	return WSValue;
}


/*
===============
R_NewMap
===============
*/
void R_NewMap (void)
{
	int		i;
	
	for (i=0 ; i<256 ; i++)
	{
		d_lightstylevalue[i] = 264;		// normal light value
	}
	memset (&r_worldentity, 0, sizeof(r_worldentity));
	r_worldentity.model = cl.worldmodel;

	// clear out efrags in case the level hasn't been reloaded
	// FIXME: is this one short?  MH - yes
	for (i = 0; i < cl.worldmodel->numleafs + 1; i++)
	{
		cl.worldmodel->leafs[i].efrags = NULL;
	}
	r_viewleaf = NULL;
	R_ClearParticles ();

	// build the texture list before the lightmaps
	R_BuildTextureList ();

	GL_BuildLightmaps ();

	// identify sky texture
	skytexturenum = -1;

	for (i=0 ; i<cl.worldmodel->numtextures ; i++)
	{
		if (!cl.worldmodel->textures[i]) continue;

		if (!Q_strncmp(cl.worldmodel->textures[i]->name, "sky",3)) 
		{
			skytexturenum = i;
		}

		if (!cl.worldmodel->textures[i]->gl_texture) continue;
 		cl.worldmodel->textures[i]->gl_texture->texturechain = NULL;
	}

	// set the world fog based on worldtype
	// also, load a generic type skybox if we don't have a specific one for the map
	worldtype = -1;

	// hack - fix for wrong worldtype in dm2
	if (!strcmp (cl.levelname, "Claustrophobopolis"))
	{
		worldtype = 1;
	}
	else
	{
		// hmmm - what happens if we send a NULL into atoi here...
		// FIXED - modded R_FindWorldSpawnEntry so it never returns NULL
		worldtype = atoi (R_FindWorldSpawnEntry (cl.worldmodel, "worldtype"));
	}

	// implement texture aging
	GL_TextureAging ();

	// mh - auto water trans begin
	if (cl.worldmodel->watertrans)
	{
		Con_Printf ("Map vis'ed for translucent water\n");
	}
	else
	{
		Con_Printf ("Map NOT vis'ed for translucent water\n");
	}
}


/*
====================
R_TimeRefresh_f

For program optimization
====================
*/
void R_TimeRefresh_f (void)
{
	int			i;
	float		start, stop, time;

	glDrawBuffer  (GL_FRONT);
	glFinish ();

	start = Sys_FloatTime ();

	for (i=0 ; i<128 ; i++)
	{
		r_refdef.viewangles[1] = i/128.0*360.0;
		R_RenderView ();
	}
	glFinish ();
	stop = Sys_FloatTime ();
	time = stop-start;
	Con_Printf ("%f seconds (%f fps)\n", time, 128/time);

	glDrawBuffer  (GL_BACK);
	GL_EndRendering ();
}

void R_FlushCaches (void)
{
}


