/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_surf.c: surface-related refresh code

#include "gl_state.h"
#include "gl_swmatrix.h"

int		skytexturenum;

cvar_t gl_detailtexture = {"gl_detailtexture", "0", true};
cvar_t gl_dot3bump = {"gl_dot3bump", "1", true};
cvar_t gl_doublelight = {"gl_doublelight", "0", true};
cvar_t r_fullbright = {"r_fullbright", "0"};
cvar_t gl_picmip = {"gl_picmip", "0"};
cvar_t gl_anisotropicfilter = {"gl_anisotropicfilter", "1", true};

gltexture_t r_detailtexture;
gltexture_t r_caustictexture;

// gl_rlight.c imports
void R_StoreEfrags (efrag_t **ppefrag);
void R_AddDynamicLights (msurface_t *surf);
void InitBLTable (void);
void R_BuildLightMap (msurface_t *surf, int destindex, int stride);
void R_UpdateLightmapTextures (void);
void R_AnimateSurfaceLight (msurface_t *surf);
void R_ClearLightmaps (void);
void GL_SubdivideSurface (model_t *mod, msurface_t *surf);
void GL_CreateSurfaceLightmap (msurface_t *surf);
void GL_RemakeSurfaceLightmaps (void);
void R_UploadSingleLightmap (lightmapinfo_t *lm);
void R_PushDlights (mnode_t *headnode);
void R_PostBumpMapChange (qboolean scale);
void R_StencilHazeParticles (void);
qboolean R_CullBox (vec3_t mins, vec3_t maxs);

extern lightmapinfo_t *lightmapinfo[];

// glwarp.c imports
void R_InitSkySphere (void);
void R_DrawSkyPickup (qboolean skydraw);


#define SINTABLE_SIZE 1024
float	sintable[SINTABLE_SIZE];
int		STBitMask;
float	STMult;

// use a seperate vertex type for water verts to avoid polluting the standard vertex array
// with a load of extra data that only applies to a small number of polys
waterverts_t	*waterverts = NULL;
int				watervertcount;

qboolean skythisframe;
qboolean waterthisframe;
qboolean onlynontransthisframe;
qboolean lavathisframe;

// linked list of textures
extern gltexture_t *TextureList;

// caches for surfs and polys
msurface_t	*surfcache;
glpoly_t	*polycache;

int			allocverts;
int			tempallocverts;
int			cachepos;

gltexture_t r_normcubemap;
gltexture_t r_dst_texture;

// the vertlist is an array of structs of floats, so this operation is valid.
// made this a global as it's needed in many funcs now
// we could put this on the hunk, but lets just get lazy instead
// this is almost 5 mb!!!
polyverts_t vertlist[MAX_MAP_VERTS * 2];
float		*surfarray = (float *)vertlist;

// keep a master drawlist too...
msurface_t *drawlist;
model_t		*currentmodel;

mleaf_t *R_LeafForEntity (vec3_t mins, vec3_t maxs)
{
	vec3_t originhack;
	mleaf_t *EntLeaf;
	int i;

	// test a few quick and easy conditions first.
	if (r_viewleaf->contents == CONTENTS_SOLID)
	{
		// oh well...
		return r_viewleaf;
	}
	else if (onlynontransthisframe)
	{
		// lava isn't translucent so just use the viewleaf
		return r_viewleaf;
	}
	else if (!waterthisframe)
	{
		// no water visible so the viewleaf is valid
		return r_viewleaf;
	}

	// translucent water is visible, so here we go...
	// get the centerpoint for x and y
	originhack[0] = (mins[0] + maxs[0]) / 2.0;
	originhack[1] = (mins[1] + maxs[1]) / 2.0;

	// get the z point - if we're above water it's the max, if we're underwater it's the min.
	// the water line is more likely to be below the player if in empty space, and above if underwater.
	if (r_viewleaf->contents == CONTENTS_EMPTY)
		originhack[2] = maxs[2];
	else
		originhack[2] = mins[2];

	// get the model leaf
	EntLeaf = Mod_PointInLeaf (originhack, cl.worldmodel);

	// avoid contents solid
	if (EntLeaf->contents == CONTENTS_SOLID)
	{
		// oops.  go for the center point instead
		originhack[2] = (mins[2] + maxs[2]) / 2.0;
		EntLeaf = Mod_PointInLeaf (originhack, cl.worldmodel);
	}
	else return EntLeaf;

	// ooops again.  try corners of the bbox
	if (EntLeaf->contents == CONTENTS_SOLID)
	{
		for (i = 0; i < 2; i++)
		{
			// decide whether to test mins or maxs first.  this condition ensures that whatever is
			// tested first will be flipped for the second iteration.
			// if the waterline is most likely to be below the viewpoint, test maxs first, otherwise test mins first
			if (r_viewleaf->contents == CONTENTS_EMPTY && !i || r_viewleaf->contents != CONTENTS_EMPTY && i)
			{
				// may the lord himself save us and preserve us from vile wee beasties that live inside our bowels and 
				// nibble at our most tender parts in the dead of night when we least expect it.
				// VectorCopy, as a #define, and despite having braces as part of the #define, still requires them here!!!
				// or else the compiler is just stupid...
				VectorCopy (maxs, originhack);
			}
			else
			{
				// may the lord himself save us and preserve us from vile wee beasties that live inside our bowels and 
				// nibble at our most tender parts in the dead of night when we least expect it.
				// VectorCopy, as a #define, and despite having braces as part of the #define, still requires them here!!!
				// or else the compiler is just stupid...
				VectorCopy (mins, originhack);
			}

			// test all corners at z min or z max as appropriate
			// the first corner is the full min or max point itself
			EntLeaf = Mod_PointInLeaf (originhack, cl.worldmodel);
			if (EntLeaf->contents != CONTENTS_SOLID) return EntLeaf;

			// now test the other points, leaving z as it is
			originhack[0] = maxs[0]; originhack[1] = mins[1];
			EntLeaf = Mod_PointInLeaf (originhack, cl.worldmodel);
			if (EntLeaf->contents != CONTENTS_SOLID) return EntLeaf;

			originhack[0] = mins[0]; originhack[1] = maxs[1];
			EntLeaf = Mod_PointInLeaf (originhack, cl.worldmodel);
			if (EntLeaf->contents != CONTENTS_SOLID) return EntLeaf;

			originhack[0] = mins[0]; originhack[1] = mins[1];
			EntLeaf = Mod_PointInLeaf (originhack, cl.worldmodel);
			if (EntLeaf->contents != CONTENTS_SOLID) return EntLeaf;
		}

		// no joy with the mins/maxs tests.
		// we really should, just to cover all bases, start testing the other midpoints of each side now, then test verts in the model,
		// and if they fail we start testing random points in the bbox volume until a runaway condition occurs.  in practice, getting 
		// here means that the model actually has all of it's corners enclosed in solid leafs, and the only way it could be visible in 
		// most cases if we get to this stage is if we're inside of it!!!  if it becomes an issue we'll take it as far as we need to.
		return r_viewleaf;
	}

	// return the leaf we found.  this part of the function should never be reached...
	return EntLeaf;
}

/*
===============
R_TextureAnimation

Returns the proper texture for a given time and base texture
===============
*/
texture_t *R_TextureAnimation (texture_t *base)
{
	// MH - spelled relative correctly
	int		relative;
	int		count;

	// store the initial base in the event of an error
	texture_t *errorbase = base;

	if (currententity->frame)
	{
		if (base->alternate_anims)
		{
			base = base->alternate_anims;
		}
	}

	if (!base->anim_total)
	{
		return base;
	}
	relative = (int)(cl.time*10) % base->anim_total;

	count = 0;

	while (base->anim_min > relative || base->anim_max <= relative)
	{
		base = base->anim_next;

		// if any error occurs, just put bask the base texture
		if (!base)
		{
			return errorbase;
		}

		if (++count > 100) 
		{
			return errorbase;
		}
	}
	return base;
}


/*
=============================================================

	BRUSH MODELS

=============================================================
*/


// setup the texture chains and animations for drawing
void R_InitTextureChains (void)
{
	gltexture_t *tex;

	for (tex = TextureList; tex; tex = tex->TextureList)
	{
		tex->texturechain = NULL;
	}
}

/*
================================================================================================================================

						WORLD RENDER PASSES

	These give a properly bumpmapped MHQuake.  I've tested by plugging the quake textures and normal maps
	generated by this engine into several of the various bumpmapping tutorials out there, and have gotten
	the same results as this gives in Quake maps, so I'm happy with it.

	MHQuake bumpmapping has nothing to do with lights.  Instead, it's using the player coords to generate
	the info used for bumpmapping.  Rotating the texture matrix for the normalisation cube map was a stroke
	of deadly genius - not documented anywhere...

	there's a lot of state change going on here, and i'd love to batch it up in a display list, but i can't
	with vertex arrays, so i don't.  opengl's outdated client/server mode strikes again!!!  at the very
	least there should be a toggle to recognise that both are on the same box so none of that nonsense.  it's
	OK if I was rendering the latest pixar masterpiece over a network, but i'm not, you know...

	I've split a lot of stuff out into seperate functions so that i only have to make changes in one place, if i
	ever need to change anything.  It means that the code could be cleaner, but at least it's still a lot cleaner
	than a single unified generic function for everything...

================================================================================================================================
*/

// flags for pickups
#define PICKUP_NOTHING		0
#define PICKUP_DETAIL		1
#define PICKUP_CAUSTICS		2
#define PICKUP_LUMAS		4
#define PICKUP_NOBLEND		8
#define PICKUP_STENCIL		16
#define PICKUP_STAIN		32

// flags for a single unified TMU shutdown routine
// this isn't serious or properly exhaustive, but only really exists so i can keep the
// various render paths cleaner and have reusable code for them so changes only have
// to be made in one place.
#define TMU_INACTIVE		0
#define TMU_ACTIVE			1
#define TMU_CUBEMAP			2
#define TMU_TEXMATRIX		4
#define TMU_RGBSCALE		8
#define TMU_PICMIP			16

int TMUFlags[4];

// single unified TMU shutdown routine
void R_KillTMUs (void)
{
	int i;
	GLenum MHGL_CURRENT_TMU;

	// do in reverse order to ensure that GL_TEXTURE0 is the last one activated.
	// this isn't robust enough to handle situations where i'm only rendering in,
	// for example, GL_TEXTURE2, but it's good for use in this instance.
	for (i = 3; i >= 0; i--)
	{
		MHGL_CURRENT_TMU = GL_TEXTURE0 + i;

		// TMU not active, don't bother
		if (!(TMUFlags[i] & TMU_ACTIVE)) continue;

		// always activate
		glClientActiveTexture (MHGL_CURRENT_TMU);

		// kill the texcoord array pointer (leave GL_TEXTURE0 up)
		if (MHGL_CURRENT_TMU > GL_TEXTURE0)
		{
			glDisableClientState (GL_TEXTURE_COORD_ARRAY);
		}

		// activate the TMU
		glActiveTexture (MHGL_CURRENT_TMU);

		// restore LOD
		if (TMUFlags[i] & TMU_PICMIP) glTexEnvf (GL_TEXTURE_FILTER_CONTROL, GL_TEXTURE_LOD_BIAS, 0);

		// take down rgb scaling
		if (TMUFlags[i] & TMU_RGBSCALE) glTexEnvi (GL_TEXTURE_ENV, GL_RGB_SCALE, 1);

		// restore the texture matrix
		if (TMUFlags[i] & TMU_TEXMATRIX)
		{
			glMatrixMode (GL_TEXTURE);
			glLoadIdentity ();
			glMatrixMode (GL_MODELVIEW);
		}

		// set replace texenv always
		glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

		// kill the texture state
		if (TMUFlags[i] & TMU_CUBEMAP)
		{
			glDisable (GL_TEXTURE_CUBE_MAP);
		}
		else 
		{
			glDisable (GL_TEXTURE_2D);
		}

		// set flags to nothing
		TMUFlags[i] = 0;
	}

	// GL_TEXTURE0 will be the last here, so enable texture2d on it
	glEnable (GL_TEXTURE_2D);
}


void R_InitSurfVArrays (qboolean TMU4, qboolean UseBumps)
{
	// enable vertex pointer - index 3 is the transformed verts
	// common to everything
	glEnableClientState (GL_VERTEX_ARRAY);
	glVertexPointer (3, GL_FLOAT, sizeof (polyverts_t), &surfarray[3]);

	if (TMU4)
	{
		// enable texcoord pointer for norm cube map
		glClientActiveTexture (GL_TEXTURE0);
		glEnableClientState (GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer (3, GL_FLOAT, sizeof (polyverts_t), &surfarray[18]);

		// enable texcoord pointer for normal maps
		glClientActiveTexture (GL_TEXTURE1);
		glEnableClientState (GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer (2, GL_FLOAT, sizeof (polyverts_t), &surfarray[6]);

		// enable texcoord pointer for textures
		glClientActiveTexture (GL_TEXTURE2);
		glEnableClientState (GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer (2, GL_FLOAT, sizeof (polyverts_t), &surfarray[6]);

		// enable texcoord pointer for lightmaps
		glClientActiveTexture (GL_TEXTURE3);
		glEnableClientState (GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer (2, GL_FLOAT, sizeof (polyverts_t), &surfarray[8]);
	}
	else if (UseBumps)
	{
		// enable texcoord pointer for norm cube map
		glClientActiveTexture (GL_TEXTURE0);
		glEnableClientState (GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer (3, GL_FLOAT, sizeof (polyverts_t), &surfarray[18]);

		// enable texcoord pointer for normal maps
		glClientActiveTexture (GL_TEXTURE1);
		glEnableClientState (GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer (2, GL_FLOAT, sizeof (polyverts_t), &surfarray[6]);
	}
	else
	{
		// enable texcoord pointer for textures
		glClientActiveTexture (GL_TEXTURE0);
		glEnableClientState (GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer (2, GL_FLOAT, sizeof (polyverts_t),  &surfarray[6]);

		// enable texcoord pointer for lightmaps
		glClientActiveTexture (GL_TEXTURE1);
		glEnableClientState (GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer (2, GL_FLOAT, sizeof (polyverts_t), &surfarray[8]);
	}

	if (GL_ExtensionBits & HAS_CVA)
	{
		glLockArrays (0, (sizeof (polyverts_t) / sizeof (float)) * MAX_MAP_VERTS * 2);
	}
}


void R_InitBumpyTMUs (GLenum MHGL_TEXTURE0, GLenum MHGL_TEXTURE1)
{
	// TMU0 writes the normalisation cubemap as the default primary colour
	glActiveTexture (MHGL_TEXTURE0);
	glDisable (GL_TEXTURE_2D);
	glEnable (GL_TEXTURE_CUBE_MAP);
	glBindTexture (GL_TEXTURE_CUBE_MAP, r_normcubemap.gltexnum);

	// now rotate the texture matrix by the same transforms as the modelview matrix to get the correct orientation
	glMatrixMode (GL_TEXTURE);
	glLoadIdentity ();

	// rotate around Y to bring blue/pink to the front
	glRotatef (90, 0, 1, 0);

	// flip axes
	glScalef (1, 1, -1);

	// alter the up/down rotation by 180 to ensure proper alignment of the cubemap
	glRotatef (r_refdef.viewangles[2], 1, 0, 0);
	glRotatef (-r_refdef.viewangles[0], 0, 1, 0);
	glRotatef (-r_refdef.viewangles[1], 0, 0, 1);

	// now scale the texture matrix linearly to ensure that we always have valid texcoords
	glScalef (666, 666, 666);

	glMatrixMode (GL_MODELVIEW);

	// basic replace texenv
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

	// set flags for this TMU
	TMUFlags[MHGL_TEXTURE0 - GL_TEXTURE0] = (TMU_ACTIVE | TMU_CUBEMAP | TMU_TEXMATRIX);

	// we need this so we can support drawing the cubemap only for testing
	if (!MHGL_TEXTURE1) return;

	// TMU1 holds the normal map
	glActiveTexture (MHGL_TEXTURE1);
	glEnable (GL_TEXTURE_2D);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
	glTexEnvi (GL_TEXTURE_ENV, GL_COMBINE_RGB, GL_DOT3_RGB);
	glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE0_RGB, GL_PREVIOUS);
	glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE1_RGB, GL_TEXTURE);
	glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND0_RGB, GL_SRC_COLOR);
	glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND1_RGB, GL_SRC_COLOR);

	// we scaled down the normal maps when we created them so there's no need to reduce LOD
	glTexEnvf (GL_TEXTURE_FILTER_CONTROL, GL_TEXTURE_LOD_BIAS, gl_picmip.value);

	// set flags for this TMU
	TMUFlags[MHGL_TEXTURE1 - GL_TEXTURE0] = (TMU_ACTIVE | TMU_PICMIP);
}

void R_InitRegularTMUs (GLenum MHGL_TEXTURE0, GLenum MHGL_TEXTURE1, qboolean worldblend)
{
	// TMU2 holds the texture
	glActiveTexture (MHGL_TEXTURE0);
	glEnable (GL_TEXTURE_2D);

	// need to boost lightmaps by 17/13 if bumpmapping is on to compensate...
	if (worldblend)
		glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	else
		glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

	glTexEnvf (GL_TEXTURE_FILTER_CONTROL, GL_TEXTURE_LOD_BIAS, gl_picmip.value);

	// set flags for this TMU
	TMUFlags[MHGL_TEXTURE0 - GL_TEXTURE0] = (TMU_ACTIVE | TMU_PICMIP);

	// TMU3 holds the lightmap
	glActiveTexture (MHGL_TEXTURE1);
	glEnable (GL_TEXTURE_2D);

	if (r_lightmap.value)
	{
		glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

		// set flags for this TMU
		TMUFlags[MHGL_TEXTURE1 - GL_TEXTURE0] = (TMU_ACTIVE);
	}
	else if (!gl_doublelight.value)
	{
		glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

		// set flags for this TMU
		TMUFlags[MHGL_TEXTURE1 - GL_TEXTURE0] = (TMU_ACTIVE);
	}
	else
	{
		glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
		glTexEnvi (GL_TEXTURE_ENV, GL_COMBINE_RGB, GL_MODULATE);
		glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE0_RGB, GL_PREVIOUS);
		glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE1_RGB, GL_TEXTURE);
		glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND0_RGB, GL_SRC_COLOR);
		glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND1_RGB, GL_SRC_COLOR);

		switch ((int) gl_doublelight.value)
		{
		case 1:
			glTexEnvi (GL_TEXTURE_ENV, GL_RGB_SCALE, 2);
			break;

		case 2:
			glTexEnvi (GL_TEXTURE_ENV, GL_RGB_SCALE, 4);
			break;

		default:
			glTexEnvi (GL_TEXTURE_ENV, GL_RGB_SCALE, 1);
			break;
		}

		// set flags for this TMU
		TMUFlags[MHGL_TEXTURE1 - GL_TEXTURE0] = (TMU_ACTIVE | TMU_RGBSCALE);
	}
}


void R_InitDetailOnlyTMU (GLenum MHGL_TEXTURE0)
{
	// bring up TMU for detail
	glActiveTexture (MHGL_TEXTURE0);
	glEnable (GL_TEXTURE_2D);

	// this combine approximately replicates the blendfunc for the pickup version.
	// it's a tiny bit brighter, but GL_RGB_SCALE only uses integers internally,
	// as per the OpenGL spec (values of 1, 2 or 4), so we're stuck with it.
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
	glTexEnvi (GL_TEXTURE_ENV, GL_COMBINE_RGB, GL_MODULATE);
	glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE0_RGB, GL_TEXTURE);
	glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND0_RGB, GL_SRC_COLOR);
	glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE1_RGB, GL_PREVIOUS);
	glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND1_RGB, GL_SRC_COLOR);
	glTexEnvi (GL_TEXTURE_ENV, GL_RGB_SCALE, 2);

	// bring up the varray and point at the 64 x 64 texcoords.  vertex arrays are great.
	// don't have to check for the 3 tmu path at every vertex when using them.  if only
	// they were documented better so more people *would* use them... sigh...
	glClientActiveTexture (MHGL_TEXTURE0);
	glEnableClientState (GL_TEXTURE_COORD_ARRAY);
	glTexCoordPointer (2, GL_FLOAT, sizeof (polyverts_t), &surfarray[10]);

	// same texture for all of these.  the texture data is stored in a seperate texture
	// object with a slightly rescaled colour balance to maintain consistency.  see above,
	// and also my rant in R_InitDetailTexture for more info on why this was needed.
	glBindTexture (GL_TEXTURE_2D, r_detailtexture.lumatexnum);

	// scale the texture matrix
	glMatrixMode (GL_TEXTURE);
	glLoadIdentity ();

	// we only need to scale the X and Y co-ords, corresponding to texture s and t.
	glScalef (gl_detailtexture.value, gl_detailtexture.value, 1);
	glMatrixMode (GL_MODELVIEW);

	// set the TMU flags
	TMUFlags[MHGL_TEXTURE0 - GL_TEXTURE0] = (TMU_ACTIVE | TMU_TEXMATRIX | TMU_RGBSCALE);
}

void R_SetupSurfaceVArrayVerts (msurface_t *surf)
{
	int i;
	float *tv;
	float sadd;
	float tadd;
	float fogdist;
	float fogalpha;

	for (i = 0; i < surf->polys[0].numverts; i++)
	{
		// keep things a bit easier to read...!
		tv = surf->polys[0].verts[i].tv;

		// get the vector from the origin to the vert and store in the norm cm texcoords
		// we don't want to normalize this as it's used for texcoords...
		// normalising also gives the weird "shiny plastic" effect, whereas this way we just get nice bumps...
		if (gl_dot3bump.value)
		{
			// leave in the -1 to 1 range, don't renormalize, as we're doing texture lookups
			// into a 3D cubemap.  rotating the cubemap with the player gives us this luxury,
			// lets us put a lot of this gruntwork on the GPU, and is cool.
			// we could also translate the cubemap to avoid the maths altogether... hmmm...
			surf->polys[0].verts[i].cm[0] = (r_refdef.vieworg[0] - tv[0]);
			surf->polys[0].verts[i].cm[1] = (r_refdef.vieworg[1] - tv[1]);
			surf->polys[0].verts[i].cm[2] = (r_refdef.vieworg[2] - tv[2]);
		}

		if (surf->volumetexture)
		{
			// you're warped, you are
			sadd = sintable[(int) ((tv[surf->warpindex[1]] * 0.025 + realtime) * STMult) & STBitMask];
			tadd = sintable[(int) ((tv[surf->warpindex[0]] * 0.025 + realtime) * STMult) & STBitMask];

			surf->polys[0].verts[i].warp[0] = tv[surf->warpindex[0]] + sadd;
			surf->polys[0].verts[i].warp[1] = tv[surf->warpindex[1]] + tadd;

			// set the correct fog colour
			surf->polys[0].verts[i].rgba[0] = surf->volumetexture->gl_texture->rgb[0];
			surf->polys[0].verts[i].rgba[1] = surf->volumetexture->gl_texture->rgb[1];
			surf->polys[0].verts[i].rgba[2] = surf->volumetexture->gl_texture->rgb[2];

			if (surf->volumetexture->gl_texture->liquidtype == LIQUIDTYPE_LAVA)
			{
				// constant fog
				fogalpha = (float)FOG_ALPHA_LAVA_CONST;
			}
			else if (surf->volumetexture->gl_texture->liquidtype == LIQUIDTYPE_SLIME)
			{
				// constant fog
				fogalpha = (float)FOG_ALPHA_SLIME_CONST;
			}
			else
			{
				// distance fog
				// use the transformed verts to get the dist
				fogdist = (tv[0] - r_origin[0]) * vpn[0] + (tv[1] - r_origin[1]) * vpn[1] + (tv[2] - r_origin[2]) * vpn[2];

				if (fogdist < 0.0f)
				{
					fogdist = 0.0f;
				}
				else if (fogdist > (float)FOG_RANGE)
				{
					fogdist = (float)FOG_RANGE;
				}
				fogalpha = 0.3f + (fogdist / (float)FOG_RANGE * (float)FOG_ALPHA_MAX) / 2;
			}
			surf->polys[0].verts[i].rgba[3] = fogalpha;
		}
	}
}

#define R_4TMUBUMPPASS		1
#define R_2TMUBUMPPASS		2
#define R_2TMUWORLDPASS		4
#define R_BMBLENDPASS		8

int R_DrawSurfaceChain (int rflags)
{
	int			PickupBits = PICKUP_NOTHING;
	msurface_t	*surf;
	gltexture_t *tex;
	int			oldlm = -1;
	GLenum		MHGL_WORLD_TMU;
	GLenum		MHGL_LIGHT_TMU;

	// oooops...
	if (!rflags)
	{
		return PICKUP_NOTHING;
	}

	if (rflags & R_BMBLENDPASS)
	{
		// source x destination + 0 blend (same as GL_MODULATE in a multitextured environment
		glEnable (GL_BLEND);
		glBlendFunc (GL_DST_COLOR, GL_ZERO);
	}

	if (rflags & R_4TMUBUMPPASS)
	{
		MHGL_WORLD_TMU = GL_TEXTURE2;
		MHGL_LIGHT_TMU = GL_TEXTURE3;
	}
	else if (rflags & R_2TMUWORLDPASS)
	{
		MHGL_WORLD_TMU = GL_TEXTURE0;
		MHGL_LIGHT_TMU = GL_TEXTURE1;
	}

	// set up vertex arrays
	R_InitSurfVArrays ((rflags & R_4TMUBUMPPASS), (rflags & R_4TMUBUMPPASS) || (rflags & R_2TMUBUMPPASS));

	// set up TMUs for bumpmaps - always GL_TEXTURE0 and GL_TEXTURE1 if we're bumpmapping
	if ((rflags & R_4TMUBUMPPASS) || (rflags & R_2TMUBUMPPASS))
	{
		R_InitBumpyTMUs (GL_TEXTURE0, GL_TEXTURE1);
	}

	// set up TMUs for world and light - these go into the TMUs we selected above
	if (!(rflags & R_2TMUBUMPPASS))
	{
		R_InitRegularTMUs (MHGL_WORLD_TMU, MHGL_LIGHT_TMU, ((rflags & R_BMBLENDPASS) || (rflags & R_4TMUBUMPPASS)));
	}

	// handle detail textures
	if (gl_detailtexture.value && (rflags & R_4TMUBUMPPASS))
	{
		PickupBits |= PICKUP_DETAIL;
	}
	else if (gl_detailtexture.value && GLCaps.GL_TextureUnits >= 3 && (rflags & R_2TMUWORLDPASS))
	{
		R_InitDetailOnlyTMU (GL_TEXTURE2);
	}
	else if (gl_detailtexture.value && (rflags & R_2TMUWORLDPASS))
	{
		PickupBits |= PICKUP_DETAIL;
	}

	// fog, caustics and lumas are not always present on each surf, 
	// so we'll run them as pickups no matter how many tmus we have
	for (tex = TextureList; tex; tex = tex->TextureList)
	{
		// no texture chain
		if (!(surf = tex->texturechain)) continue;

		// invalid draw flags for this stage
		if (surf->flags & SURF_DRAWTURB) continue;
		if (surf->flags & SURF_DRAWSKY) continue;

		if (rflags & R_2TMUBUMPPASS)
		{
			// bind the normal map in TMU1 (active coming in here)
			glBindTexture (GL_TEXTURE_2D, tex->normtexnum);
		}
		else
		{
			if (rflags & R_4TMUBUMPPASS)
			{
				// activate TMU1 for the normal map
				glActiveTexture (GL_TEXTURE1);

				// bind the normal map in TMU1
				glBindTexture (GL_TEXTURE_2D, tex->normtexnum);
			}

			// activate TMU for the world
			glActiveTexture (MHGL_WORLD_TMU);

			// bind the texture in TMU
			glBindTexture (GL_TEXTURE_2D, tex->gltexnum);

			// activate TMU for the lightmap
			glActiveTexture (MHGL_LIGHT_TMU);
		}

		for (/**/; surf; surf = surf->texturechain)
		{
			if (!(rflags & R_2TMUBUMPPASS))
			{
				// only if necessary!!!
				if (surf->lightmap->gltexturenumber != oldlm)
				{
					// bind lightmap to tmu 3
					glBindTexture (GL_TEXTURE_2D, surf->lightmap->gltexturenumber);

					// only upload lightmaps that need it
					if (surf->lightmap->modified)
					{
						R_UploadSingleLightmap (surf->lightmap);
					}

					// set the cached lightmap number
					oldlm = surf->lightmap->gltexturenumber;
				}
			}

			// draw vertex arrays
			glDrawArrays (GL_TRIANGLE_FAN, surf->polys->vertindex, surf->polys->numverts);

			if (!(rflags & R_2TMUBUMPPASS))
			{
				// accumulate in the master drawlist
				surf->drawlist = drawlist;
				drawlist = surf;

				// check for pickups per surf
				if (surf->volumetexture) 
				{
					PickupBits |= PICKUP_CAUSTICS;
				}

				if (surf->lightmap->stainactive) 
				{
					PickupBits |= PICKUP_STAIN;
				}
			}
		}

		// check for pickups per texture
		if (tex->lumatexnum && !(rflags & R_2TMUBUMPPASS))
		{
			PickupBits |= PICKUP_LUMAS;
		}
	}

	// shut down TMUs
	R_KillTMUs ();

	if ((rflags & R_4TMUBUMPPASS) || (rflags & R_2TMUBUMPPASS))
	{
		// this is crazy, but somewhere along the line doing something above implicitly modifies the blend modes
		// behind our backs, so we have to restore them now.  where *did* i put that bottle of jd...
		if (!PickupBits && !lavathisframe)
		{
			glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
			glDisable (GL_BLEND);

			PickupBits |= PICKUP_NOBLEND;
		}
	}
	else if (rflags & R_BMBLENDPASS)
	{
		if (!PickupBits && !lavathisframe)
		{
			// if blending, shut it off
			glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
			glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
			glDisable (GL_BLEND);

			PickupBits |= PICKUP_NOBLEND;
		}
	}

	// return any pickups we found
	return PickupBits;
}

void R_DetailPickup (void)
{
	msurface_t *surf;

	// change the texcoord pointer to the 64 x 64 texcoords
	glTexCoordPointer (2, GL_FLOAT, sizeof (polyverts_t), &surfarray[10]);

	// same texture for all of these
	glBindTexture (GL_TEXTURE_2D, r_detailtexture.gltexnum);

	// set the correct blending mode
	// the detail texture is rescaled so that the original colour balance is maintained
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	// src * dst + dst * src = 2 * (src * dst)
	glBlendFunc (GL_DST_COLOR, GL_SRC_COLOR);

	// here we get clever and do our texture scaling in the texture matrix.  this saves
	// us from more runtime calculations in software (moving this into software *could*
	// potentially be an option for faster CPUs)
	glMatrixMode (GL_TEXTURE);
	glLoadIdentity ();

	// we only need to scale the X and Y co-ords, corresponding to texture s and t.
	glScalef (gl_detailtexture.value, gl_detailtexture.value, 1);

	// we can safely leave the texture matrix as the active mode while we're doing the draw
	for (surf = drawlist; surf; surf = surf->drawlist)
	{
		// draw vertex arrays
		// that was easy!!!
		glDrawArrays (GL_TRIANGLE_FAN, surf->polys->vertindex, surf->polys->numverts);
	}

	// restore original texture matrix
	glLoadIdentity ();
	glMatrixMode (GL_MODELVIEW);
}

void R_CausticPickup (void)
{
	msurface_t *surf;

	// if we didn't successfully load a caustic texture, don't do caustics (yah boo sucks and pigs to you)
	// this actually never happens cos it comes from an exe resource now...
	if (r_caustictexture.gltexnum)
	{
		// change the texcoord pointer to the warp texcoords
		glTexCoordPointer (2, GL_FLOAT, sizeof (polyverts_t), &surfarray[12]);

		// same texture for all of these
		glBindTexture (GL_TEXTURE_2D, r_caustictexture.gltexnum);

		// set the correct blending mode
		glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

		// this needs nv_blend_square - part of the core since 1.4 so if a driver
		// claims >= 1.4 it has it.  but ATI doesn't...
		glBlendFunc (GL_DST_COLOR, GL_ONE_MINUS_DST_COLOR);

		// here we get clever and do our texture scaling in the texture matrix.  this saves
		// us from more runtime calculations in software (moving this into software *could*
		// potentially be an option for faster CPUs)
		glMatrixMode (GL_TEXTURE);
		glLoadIdentity ();

		// we only need to scale the X and Y co-ords, corresponding to texture s and t.
		glScalef (1.0 / (float) r_caustictexture.width, 1.0 / (float) r_caustictexture.height, 1);

		// we can safely leave the texture matrix as the active mode while we're doing the draw
		for (surf = drawlist; surf; surf = surf->drawlist)
		{
			if (!surf->volumetexture) continue;

			// draw the surf
			glDrawArrays (GL_TRIANGLE_FAN, surf->polys->vertindex, surf->polys->numverts);
		}

		// restore original texture matrix
		glLoadIdentity ();
		glMatrixMode (GL_MODELVIEW);
	}

	// do fog in the same chain cos they're the same surfs.
	// OpenGL fog is not very robust when combined with multiple blended passes, so here's a hack for ya!
	// Fog is still distance based, but the distance calcs are done in software while the actual fogging 
	// is done in hardware using a GL_COLOR_ARRAY.  Texturing is disabled for this pass.  this is very
	// similar in appearance to standard OpenGL fog, but it does break down a bit when interpolated across
	// large surfs.  i could subdivide the surfs, i suppose... even just take a midpoint...
	// first off, switch our vertex arrays
	glDisableClientState (GL_TEXTURE_COORD_ARRAY);
	glEnableClientState (GL_COLOR_ARRAY);
	glColorPointer (4, GL_FLOAT, sizeof (polyverts_t), &surfarray[14]);

	// turn off texturing and ensure we have smooth shading
	glDisable (GL_TEXTURE_2D);
	glShadeModel (GL_SMOOTH);

	// default blendfunc for this (we're already in replace texenv mode, so no need there)
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	for (surf = drawlist; surf; surf = surf->drawlist)
	{
		if (!surf->volumetexture) continue;

		// draw the surf
		glDrawArrays (GL_TRIANGLE_FAN, surf->polys->vertindex, surf->polys->numverts);
	}

	// bring texturing back
	glEnable (GL_TEXTURE_2D);

	// paranoia!!!
	glColor4f (1, 1, 1, 1);

	// now bring vertex arrays back to the way they were.
	glDisableClientState (GL_COLOR_ARRAY);
	glEnableClientState (GL_TEXTURE_COORD_ARRAY);
}

void R_LumaPickup (void)
{
	msurface_t *surf;
	gltexture_t *tex;

	// switch the texcoord pointer to the regular texture s and t
	glTexCoordPointer (2, GL_FLOAT, sizeof (polyverts_t), &surfarray[6]);

	// run lumas last of all so they shine through everything
	// set the correct blending mode for fullbrights
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

	// src + dst
	glBlendFunc (GL_ONE, GL_ONE);

	for (tex = TextureList; tex; tex = tex->TextureList)
	{
		if (!(surf = tex->texturechain)) continue;
		if (!tex->lumatexnum) continue;

		if (surf->flags & SURF_DRAWTURB) continue;
		if (surf->flags & SURF_DRAWSKY) continue;

		glBindTexture (GL_TEXTURE_2D, tex->lumatexnum);

		for (/**/; surf; surf = surf->texturechain)
		{
			// draw vertex arrays
			glDrawArrays (GL_TRIANGLE_FAN, surf->polys->vertindex, surf->polys->numverts);
		}
	}
}

void R_HeatHazePickup (void)
{
	msurface_t	*surf;
	gltexture_t *tex;
	float		args[4] = {0.05f, 0, 0, 0.04f};

	// switch the texcoord pointer to the st64 coords
	glTexCoordPointer (2, GL_FLOAT, sizeof (polyverts_t), &surfarray[10]);

	// src * dst + dst
	glBlendFunc (GL_DST_COLOR, GL_ONE);

	glBindTexture (GL_TEXTURE_2D, r_caustictexture.gltexnum);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	glMatrixMode (GL_TEXTURE);
	glLoadIdentity ();
	glTranslatef (0, realtime / 5, 0);
	glRotatef (realtime / 5, 1, 0, 0);
	glRotatef (realtime / 5, 0, 1, 0);
	glRotatef (realtime / 5, 0, 0, 1);
	glScalef (0.1f, 0.1f, 0.1f);

	for (tex = TextureList; tex; tex = tex->TextureList)
	{
		if (!(surf = tex->texturechain)) continue;
		if (surf->flags & SURF_DRAWTURB) continue;
		if (surf->flags & SURF_DRAWSKY) continue;

		for (/**/; surf; surf = surf->texturechain)
		{
			// draw vertex arrays
			glDrawArrays (GL_TRIANGLE_FAN, surf->polys->vertindex, surf->polys->numverts);
		}
	}
	glLoadIdentity ();
	glMatrixMode (GL_MODELVIEW);
}


/*
==================
R_RenderDrawList

OK hippy scum, here it is.  This is a lot more convoluted than I had originally intended as it became necessary
to do bumpmapping properly.  

Unfortunately, our bumpmapping only went and reduced our texture colour intensity by a factor of 187/255, so it's
necessary to scale it back up to compensate.  This is done using the vilest hack known to humanity...

Caustics, fog and lumas are always run as pickups because not all surfs will have them, and trying to integrate
them would only serve to over-compilate things for very very little (if any) gain.

There's a certain inefficiency and redundancy in the way I'm handling state changes here, but it's not much
extra overhead and it keeps things a whole lot cleaner than it would have been otherwise...
==================
*/
void R_RenderDrawList (void)
{
	int PickupBits = PICKUP_NOTHING;

	drawlist = NULL;

	// initial TMU flags
	TMUFlags[0] = TMUFlags[1] = TMUFlags[2] = TMUFlags[3] = 0;

	if (lavathisframe)
	{
		PickupBits |= PICKUP_STENCIL;
	}

	switch ((int) gl_dot3bump.value)
	{
	case 1:
		// standard bumpmapping with 2 or 4 passes, depending on the number of TMUs available.
		if (GLCaps.GL_TextureUnits >= 4)
		{
			PickupBits |= R_DrawSurfaceChain (R_4TMUBUMPPASS);
			break;
		}

		// fall through if there's less than 4 TMUs
	case 2:
		// standard bumpmapping with 2 passes.  also used if we have a > 3 TMU card as an override for the 4 TMU
		// path so we can compare speeds.  the 2 pass method may be faster in some cases as it involves less TMU
		// switches.  not on my GeForce 6600 GT though...
		PickupBits |= R_DrawSurfaceChain (R_2TMUBUMPPASS);

		// don't bother writing to z on the second pass
		glDepthMask (GL_FALSE);

		PickupBits |= R_DrawSurfaceChain (R_2TMUWORLDPASS | R_BMBLENDPASS);
		break;

	case 3:
		// "fun" cvar - draw bumpmaps and norm cm only.  development/testing purposes.
		// don;t bother with pickups.
		PickupBits |= R_DrawSurfaceChain (R_2TMUBUMPPASS);
		return;

	default:
		// invalid values are no bumpmaps
	case 0:
		// no bumpmapping
		PickupBits |= R_DrawSurfaceChain (R_2TMUWORLDPASS);
		break;
	}

	// no pickups needed
	if (PickupBits != PICKUP_NOTHING)
	{
		// we know that we're running pickups so now we handle them
		// don't bother writing to z cos it's been done for the regular draw already
		glDepthMask (GL_FALSE);

		// more state change management - blending came out of the original render enabled, so...
		if (!(PickupBits & PICKUP_NOBLEND)) glEnable (GL_BLEND);

		// vertex array pointers will already be up, so just go for it
		// run all the pickups we accumulated
		if (PickupBits & PICKUP_DETAIL) R_DetailPickup ();
		if (PickupBits & PICKUP_CAUSTICS) R_CausticPickup ();
		if (PickupBits & PICKUP_LUMAS) R_LumaPickup ();
	}

	// unlock the arrays now because we'll need to switch them a bit for heat haze
	if (GL_ExtensionBits & HAS_CVA)
	{
		glUnlockArrays ();
	}

	// haze particle stencil testing
	if (lavathisframe)
	{
		// stenciling is free when we're depth buffering, so hold back on that heart attack...
		glEnable (GL_STENCIL_TEST);

		// stencil test always passes
		glStencilFunc (GL_ALWAYS, 1, 1);

		// draw the haze particles into the stencil buffer
		// stencil test fails, depth can go rot: 0 stencil buffer (never happens)
		// stencil passes, depth fails: 1 stencil buffer
		// both stencil and depth pass: 0 z buffer
		glStencilOp (GL_ZERO, GL_REPLACE, GL_ZERO);
		R_StencilHazeParticles ();

		// now we only touch anything that's 0
		glStencilFunc (GL_NOTEQUAL, 1, 1);
		glStencilOp (GL_KEEP, GL_KEEP, GL_ZERO);

		// now run the heathaze pickup
		// switch the vertex pointer back
		glVertexPointer (3, GL_FLOAT, sizeof (polyverts_t), &surfarray[3]);
		glDepthMask (GL_FALSE);
		R_HeatHazePickup ();

		// done with the stencilling
		glDisable (GL_STENCIL_TEST);
	}

	if (PickupBits != PICKUP_NOTHING)
	{
		// back to replace mode
		glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

		// restore the original blend mode
		glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

		// switch off blending
		glDisable (GL_BLEND);
		glDepthMask (GL_TRUE);
	}

	// kill all remaining vertex arrays
	glDisableClientState (GL_TEXTURE_COORD_ARRAY);
	glDisableClientState (GL_VERTEX_ARRAY);
}


void R_AddTransformedBModelToList (entity_t *ent)
{
	int i;
	int x;
	float *v;
	float dotp;
	float 	matrix[16];
	msurface_t *surf;
	model_t *mod;
	qboolean rotated;
	vec3_t RelativeOrigin;
	vec3_t originhack;
	vec3_t mins, maxs;

	// store ent in currententity for R_TextureAnimation
	currententity = ent;

	mod = ent->model;

	// adjust the bounding box if the model has been rotated
	if (ent->angles[0] || ent->angles[1] || ent->angles[2])
	{
		rotated = true;

		for (i = 0; i < 3; i++)
		{
			mins[i] = ent->origin[i] - mod->radius;
			maxs[i] = ent->origin[i] + mod->radius;
		}
	}
	else
	{
		rotated = false;

		VectorAdd (ent->origin, mod->mins, mins);
		VectorAdd (ent->origin, mod->maxs, maxs);
	}

	// is the baby seal still alive???
	// (OK, very very poor taste.  i apologise)
	if (R_CullBox (mins, maxs)) return;

	// approximate the correct leaf for the entity in order to properly set fog and caustics.
	// this isn't 100% accurate (no method ever could be unless we use water surfs as clipping
	// planes) but it suffices pretty good for id1
	ent->modelleaf = R_LeafForEntity (mins, maxs);

	// get the model origin relative to the view origin for backfacing surf removal
	VectorSubtract (r_refdef.vieworg, ent->origin, RelativeOrigin);

	// adjust for rotation
	if (rotated)
	{
		vec3_t	temp;
		vec3_t	forward, right, up;

		VectorCopy (RelativeOrigin, temp);
		AngleVectors (ent->angles, forward, right, up);

		RelativeOrigin[0] = DotProduct (temp, forward);
		RelativeOrigin[1] = -DotProduct (temp, right);
		RelativeOrigin[2] = DotProduct (temp, up);
	}

	// save out the entity origin cos we're going to hack it next
	originhack[0] = ent->origin[0];
	originhack[1] = ent->origin[1];
	originhack[2] = ent->origin[2];

	// hack the entity origin to kill off Z fighting
	ent->origin[0] -= DIST_EPSILON;
	ent->origin[1] -= DIST_EPSILON;
	ent->origin[2] -= DIST_EPSILON;

	if (rotated)
	{
		// push the modelview matrix and grab an identity
		// get an entity specific matrix that's not in camera space for multiplying the verts by.
		// here we'll do it in hardware.  we could do it in software, but there's really little to
		// gain either way cos there will be so few of these
		glPushMatrix ();
		glLoadIdentity ();

		// now run the transforms on a clean identity matrix
		// (make sure we interpolate these)
		R_RotateForEntity (ent, true);

		// R_RotateForEntity only saves out the modelview if we're transforming an alias model,
		// so here we need to get it for real
		glGetFloatv (GL_MODELVIEW_MATRIX, matrix);

		// job done
		glPopMatrix ();
	}
	surf = &mod->surfaces[mod->firstmodelsurface];

	// calculate dynamic lighting for bmodel
	if (mod->firstmodelsurface != 0) 
	{
		R_PushDlights (mod->nodes + mod->hulls[0].firstclipnode);
	}

	// loop through all of the surfs
	for (i = 0; i < mod->nummodelsurfaces; i++, surf++)
	{
		// don't bother adding these
		if (surf->flags & (SURF_DRAWSKY | SURF_DRAWTURB)) continue;

		// get a dot product for backplane testing
		switch (surf->plane->type)
		{
		case PLANE_X:
			dotp = RelativeOrigin[0] - surf->plane->dist;
			break;
		case PLANE_Y:
			dotp = RelativeOrigin[1] - surf->plane->dist;
			break;
		case PLANE_Z:
			dotp = RelativeOrigin[2] - surf->plane->dist;
			break;
		default:
			dotp = DotProduct (RelativeOrigin, surf->plane->normal) - surf->plane->dist;
			break;
		}

		// what way is the surf facing?
		if (!(((surf->flags & SURF_PLANEBACK) && (dotp < -BACKFACE_EPSILON)) || (!(surf->flags & SURF_PLANEBACK) && (dotp > BACKFACE_EPSILON)))) continue;

		// now transform it
		// calculate transformed vert in software cos we want it added to the same list as
		// the worldmodel, and we also want to save the GPU muscle for the really fancy stuff.
		for (x = 0; x < surf->polys->numverts; x++)
		{
			if (rotated)
			{
				// keep things neat
				v = surf->polys->verts[x].v;

				// full matrix transformation
				surf->polys->verts[x].tv[0] = matrix[0] * v[0] + matrix[4] * v[1] + matrix[8] * v[2] + matrix[12];
				surf->polys->verts[x].tv[1] = matrix[1] * v[0] + matrix[5] * v[1] + matrix[9] * v[2] + matrix[13];
				surf->polys->verts[x].tv[2] = matrix[2] * v[0] + matrix[6] * v[1] + matrix[10] * v[2] + matrix[14];
			}
			else
			{
				// model is only translated so transform by origin only
				surf->polys->verts[x].tv[0] = surf->polys->verts[x].v[0] + ent->origin[0];
				surf->polys->verts[x].tv[1] = surf->polys->verts[x].v[1] + ent->origin[1];
				surf->polys->verts[x].tv[2] = surf->polys->verts[x].v[2] + ent->origin[2];
			}
		}

		// set the volume and volume texture
		surf->volume = ent->modelleaf->volume;
		surf->volumetexture = ent->modelleaf->volumetexture;

		// set the animation.
		surf->texinfo->texture->anim = R_TextureAnimation (surf->texinfo->texture);

		// add it to the draw list
		surf->texturechain = surf->texinfo->texture->anim->gl_texture->texturechain;
		surf->texinfo->texture->anim->gl_texture->texturechain = surf;

		// check for lightmap modification
		R_AnimateSurfaceLight (surf);

		// set up varray stuff here...
		R_SetupSurfaceVArrayVerts (surf);
	}

	// restore the saved entity origin
	ent->origin[0] = originhack[0];
	ent->origin[1] = originhack[1];
	ent->origin[2] = originhack[2];
}


void R_AddTransformedBSPBModelToList (entity_t *ent)
{
	int i;
	int j;
	int x;
	msurface_t *surf;
	msurface_t *csurf;
	vec3_t originhack;
	model_t *mod = ent->model;

	// get the centerpoint of the model and offset by the entrty origin
	originhack[0] = ((mod->maxs[0] + mod->mins[0]) / 2) + ent->origin[0];
	originhack[1] = ((mod->maxs[1] + mod->mins[1]) / 2) + ent->origin[1];
	originhack[2] = ((mod->maxs[2] + mod->mins[2]) / 2) + ent->origin[2];

	// get the containing leaf.  because these models are generally smaller we
	// can go straight for the mid point
	ent->modelleaf = Mod_PointInLeaf (originhack, cl.worldmodel);

	// store ent in currententity for R_TextureAnimation
	currententity = ent;

	// no culling cos these are so small
	// no origin hacking either
	// just go for it
	surf = &mod->surfaces[mod->firstmodelsurface];

	// loop through all of the surfs
	for (i = 0; i < mod->nummodelsurfaces; i++, surf++)
	{
		// don't bother adding these
		if (surf->flags & (SURF_DRAWSKY | SURF_DRAWTURB)) continue;

		// no backplane testing cos these are so small.
		// take a surf from the cache
		csurf = &surfcache[cachepos];

		// copy across the surf
		// avoid memcpy!!!
		for (j = 0; j < sizeof (msurface_t); j++)
		{
			((byte *) csurf)[j] = ((byte *) surf)[j];
		}

		// take a new poly from the cache
		csurf->polys = &polycache[cachepos];

		// copy across the poly
		// avoid memcpy!!!
		for (j = 0; j < sizeof (glpoly_t); j++)
		{
			((byte *) csurf->polys)[j] = ((byte *) surf->polys)[j];
		}

		// set a new vertex pointer
		csurf->polys->verts = &vertlist[tempallocverts];
		csurf->polys->vertindex = tempallocverts;

		// set the next position in the cache
		cachepos++;

		// increment the temp vertex pointer
		tempallocverts += surf->polys->numverts;

		// now transform it
		// calculate transformed vert in software cos we want it added to the same list as
		// the worldmodel, and we also want to save the GPU muscle for the really fancy stuff.
		for (x = 0; x < surf->polys->numverts; x++)
		{
			// copy across the base vertex data
			// avoid memcpy!!!
			// do a byte by byte copy to cover us in case we ever add to the polyverts_t struct
			for (j = 0; j < sizeof (polyverts_t); j++)
			{
				((byte *) &csurf->polys->verts[x])[j] = ((byte *) &surf->polys->verts[x])[j];
			}

			// model is only translated so transform by origin only
			csurf->polys->verts[x].tv[0] = surf->polys->verts[x].v[0] + ent->origin[0];
			csurf->polys->verts[x].tv[1] = surf->polys->verts[x].v[1] + ent->origin[1];
			csurf->polys->verts[x].tv[2] = surf->polys->verts[x].v[2] + ent->origin[2];
		}

		// set the animation.
		csurf->texinfo->texture->anim = R_TextureAnimation (surf->texinfo->texture);

		// add the cached surf to the draw list
		csurf->texturechain = csurf->texinfo->texture->anim->gl_texture->texturechain;
		csurf->texinfo->texture->anim->gl_texture->texturechain = csurf;

		// set the volume and volume texture
		csurf->volume = ent->modelleaf->volume;
		csurf->volumetexture = ent->modelleaf->volumetexture;

		// set up varray stuff here...
		R_SetupSurfaceVArrayVerts (csurf);
	}
}


void R_AddBModelsToDrawList (void)
{
	int		i;
	entity_t *ent;

	// don't draw any entities
	if (!r_drawentities.value) return;

	tempallocverts = allocverts;
	cachepos = 0;

	// go through all of the currently visible entities
	for (i = 0; i < cl_numvisedicts; i++)
	{
		// get the current entity
		ent = cl_visedicts[i];

		// entity has no model!!!
		if (!ent->model) continue;

		// check the type
		if (ent->model->type == mod_brush)
		{
			// and the type of bmodel
			if (ent->model->name[0] == '*')
			{
				// inline bmodels are never shared by entities so it's OK to add them to the list
				R_AddTransformedBModelToList (ent);
			}
			else
			{
				// bsp bmodels have their surfs written to a cache prior to drawing, since
				// multiple ents can share a single model.
				R_AddTransformedBSPBModelToList (ent);
			}
		}
	}
}

/*
=============================================================

	WORLD MODEL

=============================================================
*/

/*
================
R_RecursiveWorldNode
================
*/
void R_RecursiveWorldNode (mnode_t *node)
{
	int			c, side;
	mplane_t	*plane;
	msurface_t	*surf, **mark;
	mleaf_t		*pleaf;
	double		dot;

	if (node->contents == CONTENTS_SOLID) return;
	if (node->visframe != r_visframecount) return;
	if (R_CullBox (node->minmaxs, node->minmaxs + 3)) return;

	// if a leaf node, draw stuff
	if (node->contents < 0)
	{
		pleaf = (mleaf_t *) node;

		mark = pleaf->firstmarksurface;
		c = pleaf->nummarksurfaces;

		if (c)
		{
			do
			{
				(*mark)->visframe = r_framecount;
				mark++;
			} while (--c);
		}

		// deal with model fragments in this leaf
		if (pleaf->efrags) 
		{
			R_StoreEfrags (&pleaf->efrags);
		}
		return;
	}

	// node is just a decision point, so go down the apropriate sides
	// find which side of the node we are on
	plane = node->plane;

	switch (plane->type)
	{
	case PLANE_X:
		dot = modelorg[0] - plane->dist;
		break;
	case PLANE_Y:
		dot = modelorg[1] - plane->dist;
		break;
	case PLANE_Z:
		dot = modelorg[2] - plane->dist;
		break;
	default:
		dot = DotProduct (modelorg, plane->normal) - plane->dist;
		break;
	}

	// this is why i love C...
	side = !(dot >= 0);

	// recurse down the children, front side first
	R_RecursiveWorldNode (node->children[side]);

	if (node->numsurfaces)
	{
		surf = cl.worldmodel->surfaces + node->firstsurface;

		if (dot < 0 - BACKFACE_EPSILON)
		{
			side = SURF_PLANEBACK;
		}
		else if (dot > BACKFACE_EPSILON)
		{
			side = 0;
		}

		for (c = node->numsurfaces; c; c--, surf++)
		{
			// not marked
			if (surf->visframe != r_framecount) continue;

			// which side are we on?
			if (((dot < 0) ^ !!(surf->flags & SURF_PLANEBACK))) continue;

			if (surf->flags & SURF_DRAWTURB)
			{
				// store it out
				surf->texturechain = surf->texinfo->texture->gl_texture->texturechain;
				surf->texinfo->texture->gl_texture->texturechain = surf;

				// mark if water is present
				waterthisframe = true;

				// if only lava or teleports are present we can skip mod_pointinleaf checks for models
				if ((surf->flags & SURF_DRAWWATER) || (surf->flags & SURF_DRAWSLIME))
				{
					onlynontransthisframe = false;
				}
				continue;
			}

			// mark if sky is present for drawing later
			if (surf->flags & SURF_DRAWSKY)
			{
				skythisframe = true;
				continue;
			}

			// because we now have a mins and maxs for each surf we can do this!
			// this avoids some VERY substantial surface overdraw - never less than
			// 30 or so, and sometimes up to 200 or more surfs can be skipped.
			// this is more of my "saving the GPU for the really fancy stuff" thing.
			if (R_CullBox (surf->mins, surf->maxs)) continue;

			// check for light animations
			R_AnimateSurfaceLight (surf);

			// set the correct texture animation
			surf->texinfo->texture->anim = R_TextureAnimation (surf->texinfo->texture);

			// we're always sorting by texture, so just store it out
			// store it in the chain for the animation
			surf->texturechain = surf->texinfo->texture->anim->gl_texture->texturechain;
			surf->texinfo->texture->anim->gl_texture->texturechain = surf;

			// set up varray stuff here...
			R_SetupSurfaceVArrayVerts (surf);
		}
	}

	// recurse down the back side
	R_RecursiveWorldNode (node->children[!side]);
}

/*
=============
R_SetAnisotropicFilter
=============
*/
void R_SetAnisotropicFilter (int gltexnum)
{
	if (!gltexnum) return;

	glBindTexture (GL_TEXTURE_2D, gltexnum);
	glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY, gl_anisotropicfilter.value);
}

/*
=============
R_DrawWorld
=============
*/
void R_DrawWorld (void)
{
	entity_t	ent;
	int			i;
	static float oldbumpvalue = -1;
	static float oldrfbvalue = -1;
	static float oldaniso = -1;

	if (gl_dot3bump.value != oldbumpvalue)
	{
		// store the old value
		oldbumpvalue = gl_dot3bump.value;

		// reinit the blocklights table
		InitBLTable ();

		// reload all lightmaps
		GL_RemakeSurfaceLightmaps ();
	}

	if (r_fullbright.value != oldrfbvalue)
	{
		// store the old value
		oldrfbvalue = r_fullbright.value;

		// reinit the blocklights table
		InitBLTable ();

		// reload all lightmaps
		GL_RemakeSurfaceLightmaps ();
	}

	// is it possible that a card might report support for this but have a max filter of 1?
	// i wouldn't put anything past some of these folks...
	if (gl_anisotropicfilter.value != oldaniso && (GL_ExtensionBits & HAS_ANISOTROPIC) && GLCaps.GL_MaxAnisotropy > 1)
	{
		gltexture_t *tex;

		// lower bound
		if (gl_anisotropicfilter.value < 1)
		{
			gl_anisotropicfilter.value = 1;
		}

		// anisotropic filter only seems to go up in powers of 2 so set to the next highest power of 2
		for (i = 1; i < gl_anisotropicfilter.value; i *= 2);

		// force to a power of 2
		if (i != gl_anisotropicfilter.value)
		{
			gl_anisotropicfilter.value = i;
		}

		// upper bound here
		if (gl_anisotropicfilter.value > GLCaps.GL_MaxAnisotropy)
		{
			gl_anisotropicfilter.value = GLCaps.GL_MaxAnisotropy;
		}

		// now set it directly to avoid a load of string comparisons through the cvars table
		Cvar_SetValueDirect (&gl_anisotropicfilter, gl_anisotropicfilter.value);

		// only run on world texes - it doesn't really make sense on the others...!
		for (tex = TextureList; tex; tex = tex->TextureList)
		{
			// this shouldnt generally happen unless the map is crap
			// it does sometimes happen on id1 maps because of the incremental compilation they used
			// my sky texture replacement might also be throwing it out
			if (!tex) continue;

			// make sure we set it on all 3 textures!!!
			R_SetAnisotropicFilter (tex->gltexnum);
			R_SetAnisotropicFilter (tex->lumatexnum);
			R_SetAnisotropicFilter (tex->normtexnum);
			R_SetAnisotropicFilter (tex->spectexnum);
		}

		// make sure that this change doesn't fire every frame!!!
		oldaniso = gl_anisotropicfilter.value;
	}
	memset (&ent, 0, sizeof(ent));
	ent.model = cl.worldmodel;

	VectorCopy (r_refdef.vieworg, modelorg);

	currententity = &ent;
	currenttexture = -1;

	glColor3f (1,1,1);

	// push the dlights here
	R_PushDlights (cl.worldmodel->nodes);

	// init textures
	R_InitTextureChains ();

	// none initially
	skythisframe = false;
	waterthisframe = false;

	// assume that only non-trans surfs are visible unless we discover otherwise
	onlynontransthisframe = true;

	// mark visible surfs in the world and build texturechains for rendering
	R_RecursiveWorldNode (cl.worldmodel->nodes);

	// add in the brush models
	R_AddBModelsToDrawList ();

	// render the world
	R_RenderDrawList ();

	// draw sky and increment rotations
	R_DrawSkyPickup (skythisframe);
}


/*
===============
R_MarkLeaves
===============
*/
// mh - auto water trans begin
extern cvar_t gl_lockpvs;
// mh - auto water trans end

void R_MarkLeaves (void)
{
	byte		*vis;
	mnode_t		*node;
	int			i;
	byte		solid[4096];
	int			c;
	msurface_t	**mark;

	if (r_oldviewleaf == r_viewleaf && !r_novis.value) return;

	// mh - auto water trans begin
	if (gl_lockpvs.value) return;
	// mh - auto water trans end

	r_visframecount++;
	r_oldviewleaf = r_viewleaf;

	if (r_novis.value)
	{
		vis = solid;
		memset (solid, 0xff, (cl.worldmodel->numleafs + 7) >> 3);
	}
	else 
	{
		vis = Mod_LeafPVS (r_viewleaf, cl.worldmodel);
	}

	// unmark surfs
	for (i = 0; i < cl.worldmodel->numsurfaces; i++) 
	{
		cl.worldmodel->surfaces[i].lavavishack = false;
	}
	lavathisframe = false;

	for (i = 0; i < cl.worldmodel->numleafs; i++)
	{
		if (vis[i >> 3] & (1 << (i & 7)))
		{
			// mark surfs
			// vile hack necessitated by view frustum culling where a lava surf may be culled but it's
			// smoke should still be visible
			mark = cl.worldmodel->leafs[i + 1].firstmarksurface;
			c = cl.worldmodel->leafs[i + 1].nummarksurfaces;

			if (c)
			{
				do
				{
					(*mark)->lavavishack = true;

					if ((*mark)->flags & SURF_DRAWLAVA)
					{
						lavathisframe = true;
					}
					mark++;
				} while (--c);
			}
			node = (mnode_t *) &cl.worldmodel->leafs[i + 1];

			do
			{
				if (node->visframe == r_visframecount) break;

				node->visframe = r_visframecount;
				node = node->parent;
			} while (node);
		}
	}

	// don't draw the haze effect if in lava
	if (r_viewleaf->contents == CONTENTS_LAVA) 
	{
		lavathisframe = false;
	}
}

/*
================
GL_BuildPolygonFromSurface
================
*/
void GL_BuildPolygonFromSurface (model_t *mod, msurface_t *surf)
{
	int			i, lindex, lnumverts;
	medge_t		*pedges, *r_pedge;
	int			vertpage;
	float		   *vec;
	float		   s, t;
	glpoly_t	   *poly;
	vec3_t		surfmaxs = {-99999999.0f, -99999999.0f, -99999999.0f};
	vec3_t		surfmins = {99999999.0f, 99999999.0f, 99999999.0f};

	// reconstruct the polygon
	pedges = mod->edges;
	lnumverts = surf->numedges;
	vertpage = 0;

	// draw texture
	poly = Hunk_Alloc (sizeof (glpoly_t));
	poly->next = surf->polys;
	poly->flags = surf->flags;
	surf->polys = poly;
	poly->numverts = lnumverts;

	// set the vertex pointer
	poly->verts = &vertlist[allocverts];
	poly->vertindex = allocverts;
	allocverts += lnumverts;

	for (i=0 ; i<lnumverts ; i++)
	{
		lindex = mod->surfedges[surf->firstedge + i];

		if (lindex > 0)
		{
			r_pedge = &pedges[lindex];
			vec = mod->vertexes[r_pedge->v[0]].position;
		}
		else
		{
			r_pedge = &pedges[-lindex];
			vec = mod->vertexes[r_pedge->v[1]].position;
		}

		s = DotProduct (vec, surf->texinfo->vecs[0]) + surf->texinfo->vecs[0][3];
		s /= surf->texinfo->texture->width;

		t = DotProduct (vec, surf->texinfo->vecs[1]) + surf->texinfo->vecs[1][3];
		t /= surf->texinfo->texture->height;

		// copy into both v and tv to have valid initial values
		VectorCopy (vec, poly->verts[i].v);
		VectorCopy (vec, poly->verts[i].tv);

		// set bbox for the surf
		// useful to have for testing and stuff...
		if (vec[0] > surfmaxs[0]) surfmaxs[0] = vec[0];
		if (vec[1] > surfmaxs[1]) surfmaxs[1] = vec[1];
		if (vec[2] > surfmaxs[2]) surfmaxs[2] = vec[2];

		if (vec[0] < surfmins[0]) surfmins[0] = vec[0];
		if (vec[1] < surfmins[1]) surfmins[1] = vec[1];
		if (vec[2] < surfmins[2]) surfmins[2] = vec[2];

		// world verts
		poly->verts[i].st[0] = s;
		poly->verts[i].st[1] = t;

		// lightmap texture coordinates
		s = DotProduct (vec, surf->texinfo->vecs[0]) + surf->texinfo->vecs[0][3];
		s -= surf->texturemins[0];
		s += surf->light_s*16;
		s += 8;
		s /= BLOCK_WIDTH*16; //surf->texinfo->texture->width;

		t = DotProduct (vec, surf->texinfo->vecs[1]) + surf->texinfo->vecs[1][3];
		t -= surf->texturemins[1];
		t += surf->light_t*16;
		t += 8;
		t /= BLOCK_HEIGHT*16; //surf->texinfo->texture->height;

		poly->verts[i].lm[0] = s;
		poly->verts[i].lm[1] = t;

		// 64 x 64 version
		s = (DotProduct (vec, surf->texinfo->vecs[0]) + surf->texinfo->vecs[0][3]) / 64.0f;
		t = (DotProduct (vec, surf->texinfo->vecs[1]) + surf->texinfo->vecs[1][3]) / 64.0f;

		poly->verts[i].st64[0] = s;
		poly->verts[i].st64[1] = t;
	}

	// store out the completed bbox
	VectorCopy (surfmins, surf->mins);
	VectorCopy (surfmaxs, surf->maxs);
}


/*
==================
GL_BuildLightmaps

Builds the lightmap textures
with all the surfaces from all brush models
==================
*/
void GL_AllocateLightmap (int texnum, int baseintensity);


void GL_BuildLightmaps (void)
{
	int		i, j;
	model_t	*m;
	msurface_t *surf;
	msurface_t  *waterchain = NULL;

	R_ClearLightmaps ();

	r_framecount = 1;		// no dlightcache

	allocverts = 0;

	// release any memory used by added water verts
	if (waterverts) free (waterverts);

	// set to a valid pointer
	waterverts = NULL;

	// none to begin with
	watervertcount = 0;

	// go through all of the models
	for (j = 1; j < MAX_MODELS; j++)
	{
		// get the current model
		m = cl.model_precache[j];

		// no more models
		if (!m) break;

		// don;t do * models (their data comes from the world)
		if (m->name[0] == '*') continue;

		// make sure we only get brush models
		if (m->type != mod_brush) continue;

		currentmodel = m;

		// now go through all of the surfs in the model
		for (i = 0; i < m->numsurfaces; i++)
		{
			// moved this check to first so that the redundant call to
			// GL_CreateSurfaceLightmap is avoided
			if (m->surfaces[i].flags & SURF_DRAWTURB)
			{
				// add to the water surf chain
				m->surfaces[i].texturechain = waterchain;
				waterchain = &m->surfaces[i];

				// subdivide the surf here
				GL_SubdivideSurface (m, m->surfaces + i);
				continue;
			}

			if (m->surfaces[i].flags & SURF_DRAWSKY) continue;

			// create the lightmap
			GL_CreateSurfaceLightmap (m->surfaces + i);

			// build the verts
			GL_BuildPolygonFromSurface (m, m->surfaces + i);
		}
	}

	// now link in the water verts
	// first up, allocate memory to hold them
	waterverts = malloc (watervertcount * sizeof (waterverts_t));

	for (surf = waterchain; surf; surf = surf->texturechain)
	{
		glpoly_t *p;

		// do all the polys in the surf
		for (p = surf->polys; p; p = p->next)
		{
			// link in the new verts
			p->water = &waterverts[p->waterindex];

			// compute a random offset to prevent the "tunnel effect" when viewed at right angles
			p->fxofs = (rand () % 64) - 64;

			// set v here to avoid doing it at runtime
			for (i = 0; i < p->numverts; i++)
			{
				p->water[i].v[0] = p->verts[i].v[0];
				p->water[i].v[1] = p->verts[i].v[1];
				p->water[i].v[2] = p->verts[i].v[2];
			}
		}
	}
}


// normalization cubemap code from "A Practical and Robust Bump-mapping Technique for Todays GPUs"
void WriteTGA (char *name, int width, int height, int bpp, byte *data);

static void getCubeVector (int i, int cubesize, int x, int y, float *vector)
{
	float s, t, sc, tc, mag;

	s = ((float) x + 0.5) / (float) cubesize;
	t = ((float) y + 0.5) / (float) cubesize;
	sc = s * 2.0 - 1.0;
	tc = t * 2.0 - 1.0;

	// cleaned manky braces again here...
	switch (i)
	{
		case 0: vector[0] = 1.0; vector[1] = -tc; vector[2] = -sc; break;
		case 1: vector[0] = -1.0; vector[1] = -tc; vector[2] = sc; break;
		case 2: vector[0] = sc; vector[1] = 1.0; vector[2] = tc; break;
		case 3: vector[0] = sc; vector[1] = -1.0; vector[2] = -tc; break;
		case 4: vector[0] = sc; vector[1] = -tc; vector[2] = 1.0; break;
		case 5: vector[0] = -sc; vector[1] = -tc; vector[2] = -1.0; break;
	}

	mag = 1.0 / sqrt (vector[0] * vector[0] + vector[1] * vector[1] + vector[2] * vector[2]);

	vector[0] *= mag;
	vector[1] *= mag;
	vector[2] *= mag;
}

void R_InitNormCubeMap (void)
{
	// this texture doesn't need to be huge.  I've seen examples where it goes up to 128 or higher, but 32
	// is perfectly adequate.  Who am I to argue with Mr. Kilgard?
#define NORMCMSIZE 64
	float vector[3];
	int i, x, y;
	GLubyte pixels[NORMCMSIZE * NORMCMSIZE * 4];

	// set up a gltexture_t and a texture object for it
	memset (&r_normcubemap, 0, sizeof (gltexture_t));
	GL_MakeTextureObjects (1, &r_normcubemap.gltexnum);

	// bind
	glBindTexture (GL_TEXTURE_CUBE_MAP, r_normcubemap.gltexnum);

	// set up the texture parameters - did a clamp on R as well...
	glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

	// this is one area where i'd gladly argue with anyone.
	// corrected the vile braces and added better whitespace...
	for (i = 0; i < 6; i++)
	{
		for (y = 0; y < NORMCMSIZE; y++)
		{
			for (x = 0; x < NORMCMSIZE; x++)
			{
				getCubeVector (i, NORMCMSIZE, x, y, vector);

				pixels[4 * (y * NORMCMSIZE + x) + 0] = 128 + 127 * vector[0];
				pixels[4 * (y * NORMCMSIZE + x) + 1] = 128 + 127 * vector[1];
				pixels[4 * (y * NORMCMSIZE + x) + 2] = 128 + 127 * vector[2];
			}
		}

		// cube map me baybeee
		glTexImage2D (GL_TEXTURE_CUBE_MAP_POSITIVE_X + i, 0, GL_RGBA, NORMCMSIZE, NORMCMSIZE, 0, GL_RGBA, GL_UNSIGNED_BYTE, pixels);
	}
#undef NORMCMSIZE
}


/*
===============
R_InitDSTTex

Create the texture which warps texture shaders
===============
*/
void R_InitDSTTex (void)
{
#define DST_SIZE 16
	signed char data[DST_SIZE][DST_SIZE][2];
	int x, y;

	if (!(GL_ExtensionBits & HAS_NVTEXSHADER)) return;

	for (x = 0; x < DST_SIZE; x++)
	{
		for (y = 0; y < DST_SIZE; y++)
		{
			data[x][y][0] = rand () % 255 - 128;
			data[x][y][1] = rand () % 255 - 128;
		}
	}

	// set up a gltexture_t and a texture object for it
	memset (&r_dst_texture, 0, sizeof (gltexture_t));
	GL_MakeTextureObjects (1, &r_dst_texture.gltexnum);

	glBindTexture (GL_TEXTURE_2D, r_dst_texture.gltexnum);
	glTexImage2D (GL_TEXTURE_2D, 0, GL_DSDT8_NV, DST_SIZE, DST_SIZE, 0, GL_DSDT_NV, GL_BYTE, data);

	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
#undef DST_SIZE
}

void R_InitSurfStuff (void)
{
	int   i;
	float j;
	float inc = (2.0f * (float)M_PI) / (float)SINTABLE_SIZE;

	// set all lightmaps to a valid pointer
	for (i = 0; i < MAX_LIGHTMAPS; i++)
	{
		// set the lightmapinfo structs to NULL on first run
		lightmapinfo[i] = NULL;
	}

	// allocate the caches
	// we have no way of knowing how many entities use these types of models on the client
	// side, so lets just allocate enough space for all of them, and assume 6 surfs is an average
	surfcache = malloc (sizeof (msurface_t) * MAX_MAP_ENTITIES * 6);
	polycache = malloc (sizeof (glpoly_t) * MAX_MAP_ENTITIES * 6);

	// init the blocklights lookup table
	InitBLTable ();

	// make a display list for the sky sphere
	R_InitSkySphere ();

	// set up the sin table for warps
	for (i = 0, j = 0; i < SINTABLE_SIZE; i++, j += inc)
	{
		sintable[i] = sin (j) * 8.0;
	}
	STBitMask = SINTABLE_SIZE - 1;

	STMult = (float)SINTABLE_SIZE / (2.0f * (float)M_PI);

	// register cvars
	Cvar_RegisterVariable (&gl_detailtexture);
	Cvar_RegisterVariable (&gl_dot3bump);
	Cvar_RegisterVariable (&gl_doublelight);
	Cvar_RegisterVariable (&r_fullbright);
	Cvar_RegisterVariable (&gl_picmip);
	Cvar_RegisterVariable (&gl_anisotropicfilter);

	R_InitNormCubeMap ();
	R_InitDSTTex ();
}

