/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_main.c

#include "quakedef.h"
#include "gl_mtex.h"
#include "gl_mirror.h"

entity_t	r_worldentity;

qboolean	r_cache_thrash;		// compatability

vec3_t		modelorg, r_entorigin;
entity_t	*currententity;

int			r_visframecount;	// bumped when going to a new PVS
int			r_framecount;		// used for dlight push checking

mplane_t	frustum[4];

int			c_brush_polys, c_alias_polys;

qboolean	envmap;				// true during envmap command capture 

int			cnttextures[2] = {-1, -1};     // cached

int			particletexture;	// little dot for particles
int			playertextures;		// up to 16 color translated skins

// view origin
vec3_t	vup;
vec3_t	vpn;
vec3_t	vright;
vec3_t	r_origin;

float	r_world_matrix[16];

float PowerupGlowRadius = 0.0;		// variable with frametime for throbbing effect
float TorchGlowFlicker;		// variable for throbbing effect
float BoltGlowFlicker;

// screen size info
refdef_t	r_refdef;

mleaf_t		*r_viewleaf, *r_oldviewleaf;

gltexture_t	*r_notexture_mip;

int		d_lightstylevalue[256];	// 8.8 fraction of base light value

void R_MarkLeaves (void);

cvar_t	r_norefresh = {"r_norefresh","0"};
cvar_t	r_drawentities = {"r_drawentities","1"};
cvar_t	r_drawviewmodel = {"r_drawviewmodel","1"};
cvar_t	r_speeds = {"r_speeds","0"};
cvar_t	r_lightmap = {"r_lightmap","0"};
cvar_t	r_shadows = {"r_shadows","0", true};
cvar_t	r_wateralpha = {"r_wateralpha","1"};
cvar_t	r_dynamic = {"r_dynamic","1", true};
cvar_t	r_novis = {"r_novis","0"};

cvar_t	gl_finish = {"gl_finish","0"};
cvar_t	gl_clear = {"gl_clear","0"};
cvar_t	gl_cull = {"gl_cull","1"};
cvar_t	gl_smoothmodels = {"gl_smoothmodels","1"};
cvar_t	gl_affinemodels = {"gl_affinemodels","0"};
cvar_t	gl_polyblend = {"gl_polyblend","1"};
cvar_t	gl_playermip = {"gl_playermip","0"};
cvar_t	gl_nocolors = {"gl_nocolors","0"};
cvar_t	gl_keeptjunctions = {"gl_keeptjunctions","0", true};
cvar_t	gl_reporttjunctions = {"gl_reporttjunctions","0"};
cvar_t	gl_doubleeyes = {"gl_doubleeys", "1"};

// fenix@io.com: model interpolation
cvar_t  r_interpolate_alias = {"r_interpolate_alias", "1", true};
cvar_t  r_interpolate_brush = {"r_interpolate_brush", "1", true};


extern	cvar_t	gl_ztrick;
extern int gl_fogenabled;

extern cvar_t r_particles;

// muff quake jello effects
void R_DrawAliasModelShell (entity_t *ent);

// glow effects
#define MAX_GLOWS 2048
int num_glows;

typedef struct glows_s
{
	float red;
	float green;
	float blue;
	float radius;
	vec3_t origin;
} glows_t;

glows_t glow_effects[MAX_GLOWS];

int num_sprites;

extern qboolean water_present;

void R_RenderGlowEffects (void)
{
	int i, j, k;
	vec3_t v;
	float rad;

	if (gl_fogenabled) glDisable (GL_FOG);

	glDepthMask (0);
	glDisable (GL_TEXTURE_2D);
	glEnable (GL_BLEND);
	glBlendFunc (GL_ONE, GL_ONE);

	for (k = 0; k < num_glows; k++)
	{
		rad = glow_effects[k].radius * 0.35;
		VectorSubtract (glow_effects[k].origin, r_origin, v);

		if (VLength (v) < rad) continue;

		glBegin (GL_TRIANGLE_FAN);
		glColor3f (glow_effects[k].red, glow_effects[k].green, glow_effects[k].blue);

		for (i = 0; i < 3; i++)
			v[i] = glow_effects[k].origin[i] - vpn[i] * glow_effects[k].radius;

		glVertex3fv (v);
		glColor3f (0,0,0);

		for (i = 16; i >= 0; i--)
		{
			for (j = 0; j < 3; j++)
				v[j] = glow_effects[k].origin[j] + vright[j] * glowcos[i] * 
					glow_effects[k].radius + vup[j] * glowsin[i] * glow_effects[k].radius;

			glVertex3fv (v);
		}

		glEnd ();
	}

	glColor3f (1,1,1);
	glDisable (GL_BLEND);
	glEnable (GL_TEXTURE_2D);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDepthMask (1);

	if (gl_fogenabled) glEnable (GL_FOG);
}


void R_AddGlowEffect (float red, float green, float blue, float radius, vec3_t origin)
{
	if (num_glows >= MAX_GLOWS)
	{
		Con_DPrintf ("Too many glows\n");
		return;
	}

	glow_effects[num_glows].red = red;
	glow_effects[num_glows].green = green;
	glow_effects[num_glows].blue = blue;

	glow_effects[num_glows].radius = radius;

	glow_effects[num_glows].origin[0] = origin[0];
	glow_effects[num_glows].origin[1] = origin[1];
	glow_effects[num_glows].origin[2] = origin[2];

	num_glows++;
}


/*
=============================================================

  SPRITE MODELS
  No longer uses the Hexen II Sprite Model Code!
  (Serious problems with water surfaces)

=============================================================
*/

/*
================
R_GetSpriteFrame
================
*/
mspriteframe_t *R_GetSpriteFrame (entity_t *currententity)
{
	msprite_t		*psprite;
	mspritegroup_t	*pspritegroup;
	mspriteframe_t	*pspriteframe;
	int				i, numframes, frame;
	float			*pintervals, fullinterval, targettime, time;

	psprite = currententity->model->cache.data;
	frame = currententity->frame;

	if ((frame >= psprite->numframes) || (frame < 0))
	{
		Con_DPrintf ("R_DrawSprite: no such frame %d\n", frame);
		frame = 0;
	}

	if (psprite->frames[frame].type == SPR_SINGLE)
		pspriteframe = psprite->frames[frame].frameptr;
	else
	{
		pspritegroup = (mspritegroup_t *)psprite->frames[frame].frameptr;
		pintervals = pspritegroup->intervals;
		numframes = pspritegroup->numframes;
		fullinterval = pintervals[numframes-1];

		time = cl.time + currententity->syncbase;

		// when loading in Mod_LoadSpriteGroup, we guaranteed all interval values
		// are positive, so we don't have to worry about division by 0
		targettime = time - ((int)(time / fullinterval)) * fullinterval;

		for (i=0 ; i<(numframes-1) ; i++)
		{
			if (pintervals[i] > targettime)
				break;
		}

		pspriteframe = pspritegroup->frames[i];
	}

	return pspriteframe;
}


/*
=================
R_DrawSpriteModel

=================
*/
void R_ZerstorerRain (vec3_t org);

void R_DrawSpriteModel (entity_t *e)
{
	vec3_t	point;
	mspriteframe_t	*frame;
	float		*up, *right;
	vec3_t		v_forward, v_right, v_up;
	msprite_t		*psprite;

	// standard quake sprites are handled differently.  other types use the normal sprite code
	// the new rocket explosion effect means that we no longer need the sprite for it.
	// update - draw the rocket explosion sprite if we're using the classic particle system.
	if (!strcmp (currententity->model->name, "progs/s_explod.spr") && r_particles.value)
		return;
	else if (!strcmp (currententity->model->name, "progs/s_light.spr"))
	{
		R_AddGlowEffect (RTable[236], GTable[236], BTable[236], 15, e->origin);
		return;
	}
	else if (!strcmp (currententity->model->name, "progs/s_bubble.spr"))
	{
		R_AddGlowEffect (RTable[244], GTable[244], BTable[244], 5, e->origin);
		return;
	}
	else if (!strcmp (currententity->model->name, "progs/s_rain.spr"))
	{
		// zerstorer rain
		R_ZerstorerRain (currententity->origin);
		return;
	}

	// don't even bother culling, because it's just a single
	// polygon without a surface cache
	frame = R_GetSpriteFrame (e);
	psprite = currententity->model->cache.data;

	if (psprite->type == SPR_ORIENTED)
	{
		// bullet marks on walls
		AngleVectors (currententity->angles, v_forward, v_right, v_up);
		up = v_up;
		right = v_right;
	}
	else
	{
		// normal sprite
		up = vup;
		right = vright;
	}

	glColor3f (1,1,1);

	// these are always changing, so we can use GL_Bind
	GL_Bind (frame->gl_texturenum);

	glEnable (GL_ALPHA_TEST);
	glBegin (GL_QUADS);

	glTexCoord2f (0, 1);
	VectorMA (e->origin, frame->down, up, point);
	VectorMA (point, frame->left, right, point);
	glVertex3fv (point);

	glTexCoord2f (0, 0);
	VectorMA (e->origin, frame->up, up, point);
	VectorMA (point, frame->left, right, point);
	glVertex3fv (point);

	glTexCoord2f (1, 0);
	VectorMA (e->origin, frame->up, up, point);
	VectorMA (point, frame->right, right, point);
	glVertex3fv (point);

	glTexCoord2f (1, 1);
	VectorMA (e->origin, frame->down, up, point);
	VectorMA (point, frame->right, right, point);
	glVertex3fv (point);
	
	glEnd ();

	glDisable (GL_ALPHA_TEST);
}


/*
=============================================================

  ALIAS MODELS

=============================================================
*/

// moved to gl_aliasmod.c

//==================================================================================

int num_amodel_glows;
int num_amodel_shadows;

/*
=============
R_DrawEntitiesOnList
=============
*/
void R_SetupDraw (qboolean findfunc);
void R_ShutdownDraw (void);
int	num_amodels;

void R_DrawEntitiesOnList (void)
{
	int	i;
	aliashdr_t *model_header;

	num_sprites = 0;
	num_amodels = 0;
	num_amodel_glows = 0;
	num_amodel_shadows = 0;

	if (!r_drawentities.value)
		return;

	// check for the presence of the various model types - only brush models are
	// actually drawn at this stage
	// the draw has already been set up for the world, and brush models are the first
	// thing drawn after it, so...
	//R_SetupDraw (false);

	for (i=0 ; i<cl_numvisedicts ; i++)
	{
		switch (cl_visedicts[i]->model->type)
		{
		case mod_brush:
			currententity = cl_visedicts[i];
			R_DrawBrushModel (currententity);
			break;

		case mod_alias:
			num_amodels++;
			break;

		case mod_sprite:
			num_sprites++;
			break;

		default:
			break;
		}
	}

	R_ShutdownDraw ();

	if (num_amodels)
	{
		// draw the model
		for (i=0 ; i<cl_numvisedicts ; i++)
		{
			switch (cl_visedicts[i]->model->type)
			{
			case mod_alias:
				currententity = cl_visedicts[i];
				model_header = (aliashdr_t *) Mod_Extradata (currententity->model);

				// fixed chasecam - frika-c
				if (currententity == &cl_entities[cl.viewentity])
					currententity->angles[0] *= 0.3;

				R_DrawAliasModel (currententity, model_header);
				break;

			default:
				break;
			}
		}
	}
}


// MH - moved this outside so we can get the drawing order correct.
void R_DrawEntityFX (void)
{
	int	i;

	if (!r_drawentities.value)
		return;

	if (num_amodels)
	{
		// draw glow effects
		if (num_amodel_glows)
		{
			// disable fogging for the glow
			if (gl_fogenabled) glDisable (GL_FOG);
			glDepthMask (0);
	        glDisable (GL_TEXTURE_2D);
		    glEnable (GL_BLEND);
			glBlendFunc (GL_ONE, GL_ONE);

			for (i=0 ; i<cl_numvisedicts ; i++)
			{
				switch (cl_visedicts[i]->model->type)
				{
				case mod_alias:
					currententity = cl_visedicts[i];
					R_DrawAliasGlowFX (currententity, (aliashdr_t *)Mod_Extradata (currententity->model));
					break;

				default:
					break;
				}
			}

			glColor3f (1,1,1);
	        glDisable (GL_BLEND);
		    glEnable (GL_TEXTURE_2D);
			glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	        glDepthMask (1);
			if (gl_fogenabled) glEnable (GL_FOG);
		}

		// draw shadow effects
		if (num_amodel_shadows)
		{
			for (i=0 ; i<cl_numvisedicts ; i++)
			{
				switch (cl_visedicts[i]->model->type)
				{
				case mod_alias:
					currententity = cl_visedicts[i];
					R_DrawAliasShadowFX (currententity, (aliashdr_t *)Mod_Extradata (currententity->model));
					break;

				default:
					break;
				}
			}
		}
	}
}


void R_DrawSpriteModels (void)
{
	int		i;

	if (!r_drawentities.value)
		return;

	// most sprites are drawn at the same time as particles
	for (i=0 ; i<cl_numvisedicts ; i++)
	{
		currententity = cl_visedicts[i];

		switch (currententity->model->type)
		{
		case mod_sprite:
			R_DrawSpriteModel (currententity);
			break;

		default:
			break;
		}
	}
}

/*
============
R_PolyBlend
============
*/
void R_PolyBlend (void)
{
	if (!gl_polyblend.value) return;
	if (!v_blend[3]) return;

	glDisable (GL_ALPHA_TEST);
	glEnable (GL_BLEND);
	glDisable (GL_DEPTH_TEST);
	glDisable (GL_TEXTURE_2D);

    glLoadIdentity ();

    glRotatef (-90, 1, 0, 0);	    // put Z going up
    glRotatef (90, 0, 0, 1);	    // put Z going up

	glColor4fv (v_blend);

	glBegin (GL_QUADS);

	glVertex3f (10, 100, 100);
	glVertex3f (10, -100, 100);
	glVertex3f (10, -100, -100);
	glVertex3f (10, 100, -100);
	glEnd ();

	glDisable (GL_BLEND);
	glEnable (GL_TEXTURE_2D);
	glEnable (GL_ALPHA_TEST);
}


int SignbitsForPlane (mplane_t *out)
{
	int	bits, j;

	// for fast box on planeside test

	bits = 0;
	for (j=0 ; j<3 ; j++)
	{
		if (out->normal[j] < 0)
			bits |= 1<<j;
	}
	return bits;
}


// -------------------------------------------------------------------------------------------
// This stuff was all in mathlib.c but it's actually only ever used in this file

//#define DEG2RAD (a) (a * M_PI) / 180.0F

// no need to make this a macro since it's rarely used (4 times per frame)
float DEG2RAD (float a)
{
	return (a * M_PI) / 180.0F;
}


void ProjectPointOnPlane (vec3_t dst, const vec3_t p, const vec3_t normal)
{
	float d;
	vec3_t n;
	float inv_denom;

	inv_denom = 1.0F / DotProduct (normal, normal);

	d = DotProduct (normal, p) * inv_denom;

	n[0] = normal[0] * inv_denom;
	n[1] = normal[1] * inv_denom;
	n[2] = normal[2] * inv_denom;

	dst[0] = p[0] - d * n[0];
	dst[1] = p[1] - d * n[1];
	dst[2] = p[2] - d * n[2];
}

// assumes src is normalized
void PerpendicularVector (vec3_t dst, const vec3_t src)
{
	int	pos;
	int i;
	float minelem = 1.0F;
	vec3_t tempvec;

	// find the smallest magnitude axially aligned vector
	for (pos = 0, i = 0; i < 3; i++)
	{
		if (fabs (src[i]) < minelem)
		{
			pos = i;
			minelem = fabs (src[i]);
		}
	}

	tempvec[0] = tempvec[1] = tempvec[2] = 0.0F;
	tempvec[pos] = 1.0F;

	// project the point onto the plane defined by src
	ProjectPointOnPlane (dst, tempvec, src);

	// normalize the result
	VectorNormalize (dst);
}

void R_ConcatRotations (float in1[3][3], float in2[3][3], float out[3][3])
{
	out[0][0] = in1[0][0] * in2[0][0] + in1[0][1] * in2[1][0] + in1[0][2] * in2[2][0];
	out[0][1] = in1[0][0] * in2[0][1] + in1[0][1] * in2[1][1] + in1[0][2] * in2[2][1];
	out[0][2] = in1[0][0] * in2[0][2] + in1[0][1] * in2[1][2] + in1[0][2] * in2[2][2];
	out[1][0] = in1[1][0] * in2[0][0] + in1[1][1] * in2[1][0] + in1[1][2] * in2[2][0];
	out[1][1] = in1[1][0] * in2[0][1] + in1[1][1] * in2[1][1] + in1[1][2] * in2[2][1];
	out[1][2] = in1[1][0] * in2[0][2] + in1[1][1] * in2[1][2] + in1[1][2] * in2[2][2];
	out[2][0] = in1[2][0] * in2[0][0] + in1[2][1] * in2[1][0] + in1[2][2] * in2[2][0];
	out[2][1] = in1[2][0] * in2[0][1] + in1[2][1] * in2[1][1] + in1[2][2] * in2[2][1];
	out[2][2] = in1[2][0] * in2[0][2] + in1[2][1] * in2[1][2] + in1[2][2] * in2[2][2];
}


#ifdef _WIN32
#pragma optimize( "", off )
#endif


void RotatePointAroundVector (vec3_t dst, const vec3_t dir, const vec3_t point, float degrees)
{
	float m[3][3];
	float im[3][3];
	float zrot[3][3];
	float tmpmat[3][3];
	float rot[3][3];
	int	i;
	vec3_t vr, vup, vf;

	vf[0] = dir[0];
	vf[1] = dir[1];
	vf[2] = dir[2];

	PerpendicularVector (vr, dir);
	CrossProduct (vr, vf, vup);

	m[0][0] = vr[0];
	m[1][0] = vr[1];
	m[2][0] = vr[2];

	m[0][1] = vup[0];
	m[1][1] = vup[1];
	m[2][1] = vup[2];

	m[0][2] = vf[0];
	m[1][2] = vf[1];
	m[2][2] = vf[2];

	memcpy (im, m, sizeof (im));

	im[0][1] = m[1][0];
	im[0][2] = m[2][0];
	im[1][0] = m[0][1];
	im[1][2] = m[2][1];
	im[2][0] = m[0][2];
	im[2][1] = m[1][2];

	memset (zrot, 0, sizeof (zrot));
	zrot[0][0] = zrot[1][1] = zrot[2][2] = 1.0F;

	zrot[0][0] = cos (DEG2RAD (degrees));
	zrot[0][1] = sin (DEG2RAD (degrees));
	zrot[1][0] = -sin (DEG2RAD (degrees));
	zrot[1][1] = cos (DEG2RAD (degrees));

	R_ConcatRotations (m, zrot, tmpmat);
	R_ConcatRotations (tmpmat, im, rot);

	for (i = 0; i < 3; i++)
	{
		dst[i] = rot[i][0] * point[0] + rot[i][1] * point[1] + rot[i][2] * point[2];
	}
}

// -------------------------------------------------------------------------------------------


void R_SetFrustum (void)
{
	int	i;

	if (r_refdef.fov_x == 90) 
	{
		// front side is visible
		VectorAdd (vpn, vright, frustum[0].normal);
		VectorSubtract (vpn, vright, frustum[1].normal);
		VectorAdd (vpn, vup, frustum[2].normal);
		VectorSubtract (vpn, vup, frustum[3].normal);
	}
	else
	{
		RotatePointAroundVector (frustum[0].normal, vup, vpn, -(90 - r_refdef.fov_x / 2 ));
		RotatePointAroundVector (frustum[1].normal, vup, vpn, 90 - r_refdef.fov_x / 2);
		RotatePointAroundVector (frustum[2].normal, vright, vpn, 90-r_refdef.fov_y / 2 );
		RotatePointAroundVector (frustum[3].normal, vright, vpn, -(90 - r_refdef.fov_y / 2 ));
	}

	for (i=0 ; i<4 ; i++)
	{
		frustum[i].type = PLANE_ANYZ;
		frustum[i].dist = DotProduct (r_origin, frustum[i].normal);
		frustum[i].signbits = SignbitsForPlane (&frustum[i]);
	}
}



/*
===============
R_SetupFrame
===============
*/
extern qboolean inwater;
qboolean mirror_particles;

void R_SetupFrame (void)
{
	if (!mirror_particles)
	{
		R_AnimateLight ();
		r_framecount++;
	}

	// build the transformation matrix for the given view angles
	VectorCopy (r_refdef.vieworg, r_origin);

	AngleVectors (r_refdef.viewangles, vpn, vright, vup);

	// current viewleaf
	r_oldviewleaf = r_viewleaf;
	r_viewleaf = Mod_PointInLeaf (r_origin, cl.worldmodel);

	inwater = false;

	V_SetContentsColor (r_viewleaf->contents);
	V_CalcBlend ();

	r_cache_thrash = false;
}


void MYgluPerspective (GLdouble fovy, GLdouble aspect, GLdouble zNear, GLdouble zFar)
{
   GLdouble xmin, xmax, ymin, ymax;

   ymax = zNear * tan( fovy * M_PI / 360.0 );
   ymin = -ymax;

   xmin = ymin * aspect;
   xmax = ymax * aspect;

   glFrustum( xmin, xmax, ymin, ymax, zNear, zFar );
}


/*
=============
R_SetupGL
=============
*/
extern cvar_t r_drawdist;
float sky_mult;

void R_SetupGL (void)
{
	float	screenaspect;
	extern	int glwidth, glheight;

	// set up viewpoint
	glMatrixMode(GL_PROJECTION);
    glLoadIdentity ();
	glViewport (0, 0, glwidth, glheight);
    screenaspect = (float)r_refdef.vrect.width/r_refdef.vrect.height;
    MYgluPerspective (r_refdef.fov_y,  screenaspect,  4,  r_drawdist.value);  //4096);

	sky_mult = r_drawdist.value / 2;

	if (mirror)
		Mirror_Scale ();
	else glCullFace(GL_FRONT);

	glMatrixMode(GL_MODELVIEW);
    glLoadIdentity ();

    glRotatef (-90,  1, 0, 0);	    // put Z going up
    glRotatef (90,  0, 0, 1);	    // put Z going up

    glRotatef (-r_refdef.viewangles[2],  1, 0, 0);
    glRotatef (-r_refdef.viewangles[0],  0, 1, 0);
    glRotatef (-r_refdef.viewangles[1],  0, 0, 1);

    glTranslatef (-r_refdef.vieworg[0],  -r_refdef.vieworg[1],  -r_refdef.vieworg[2]);

	glGetFloatv (GL_MODELVIEW_MATRIX, r_world_matrix);

	// set drawing parms
	if (gl_cull.value)
		glEnable(GL_CULL_FACE);
	else glDisable(GL_CULL_FACE);

	glDisable(GL_BLEND);
	glDisable(GL_ALPHA_TEST);
	glEnable(GL_DEPTH_TEST);
}


/*
================
R_RenderScene

r_refdef must be set before the first call
MH - rearranged a lot of the drawing orders in here
================
*/
void R_CheckLightBlends (void);
extern cvar_t gl_fogstart;
extern cvar_t gl_fogend;

extern float r_waterfog_colour[4];
extern float r_waterfog_density;
extern cvar_t r_autowatertrans;

// this is a global cos we need it in gl_rsurf too...
GLfloat fog_colours[4];

// fog code was moved in here to enable fog to be rendered in reflections!
void R_RenderScene (void)
{
	// automatically set water alpha for automatic water translucency if it
	// hasn't already been set
	if (r_autowatertrans.value && r_wateralpha.value >= 1.0)
		Cvar_SetValue ("r_wateralpha", 0.65);

	if (gl_fogenable.value)
	{
		// prevent any silliness with fogstart and fogend - this also ensures that fogstart
		// is always less than fogend
		if (gl_fogstart.value < 5)
			Cvar_SetValue ("gl_fogstart", 5);

#if 0
		// removed cos drawdist could be > 4096
		if (gl_fogstart.value > 4095)
			Cvar_SetValue ("gl_fogstart", 4095);
#endif

		if (gl_fogend.value < 6)
			Cvar_SetValue ("gl_fogend", 6);

#if 0
		// removed cos drawdist could be > 4096
		if (gl_fogend.value > 4096)
			Cvar_SetValue ("gl_fogend", 4096);
#endif

		// we have to set end to greater than start, cos setting start to less than end
		// may result in an invalid start value - any value of end will always be valid.
		if (gl_fogstart.value >= gl_fogend.value)
			Cvar_SetValue ("gl_fogend", gl_fogstart.value + 1);
		
		fog_colours[0] = (GLfloat) gl_fogred.value;
		fog_colours[1] = (GLfloat) gl_foggreen.value;
		fog_colours[2] = (GLfloat) gl_fogblue.value;
		fog_colours[3] = 0.2;

		glFogi (GL_FOG_MODE, GL_LINEAR);
		glFogfv (GL_FOG_COLOR, fog_colours);
		glFogf (GL_FOG_START, gl_fogstart.value);
		glFogf (GL_FOG_END, gl_fogend.value);
		glFogf (GL_FOG_DENSITY, 0.2);

		glEnable (GL_FOG);

		// it's faster to get at an int than at a structure member, so we'll use this to
		// check instead of the cvar!  we can also use it for checking when underwater fog
		// should not be drawn (underwater fog doesn't use the cvar)
		gl_fogenabled = 1;
	}
	else
	{
		// we had previously enabled it and now wanna disable it
		gl_fogenabled = 0;
		glDisable (GL_FOG);
	}

	num_glows = 0;

	R_CheckLightBlends ();
	R_SetupFrame ();
	R_SetFrustum ();
	R_SetupGL ();
	R_MarkLeaves ();
	water_present = false;
	R_DrawWorld ();
	S_ExtraUpdate ();
	R_DrawEntitiesOnList ();
	if (num_sprites) R_DrawSpriteModels ();
	R_DrawTranslucentSurfaces ();

#ifdef MH_MIRRORS
	if (mirror_render)
	{
		// draw reflected particles
		R_DrawParticles ();
		mirror_particles = true;	// origin, etc has been modified, so don't do it again
		R_DrawEntityFX ();
		if (num_glows) R_RenderGlowEffects ();
	}
	else if (!mirror)
#endif
	{
		// draw world particles if there's no mirror
		R_DrawParticles ();
		R_DrawEntityFX ();
		if (num_glows) R_RenderGlowEffects ();
	}

	if (gl_fogenabled) glDisable(GL_FOG);
	if (!mirror_render) R_DrawViewModel ();
}


/*
=============
R_Clear
=============
*/
void R_Clear (void)
{
	if (r_mirroralpha.value < 1.0)
		Mirror_Clear ();
	else if (gl_ztrick.value)
	{
		static int trickframe;

		if (gl_clear.value)
			glClear (GL_COLOR_BUFFER_BIT);

		trickframe++;
		if (trickframe & 1)
		{
			gldepthmin = 0;
			gldepthmax = 0.49999;
			glDepthFunc (GL_LEQUAL);
		}
		else
		{
			gldepthmin = 1;
			gldepthmax = 0.5;
			glDepthFunc (GL_GEQUAL);
		}
	}
	else
	{
		if (gl_clear.value)
			glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		else
			glClear (GL_DEPTH_BUFFER_BIT);
		gldepthmin = 0;
		gldepthmax = 1;
		glDepthFunc (GL_LEQUAL);
	}

	glDepthRange (gldepthmin, gldepthmax);
}


/*
================
R_RenderView

r_refdef must be set before the first call
================
*/
float frametime;
float TGF_Tick = 0.0;
qboolean PGR_Dir = true;
float watermove;

void R_RenderView (void)
{
	// fixme - crash to the console?
	if (!r_worldentity.model || !cl.worldmodel)
		Sys_Error ("R_RenderView: NULL worldmodel");

#ifdef NOT_THE_MOST_STABLE
	// HA!
	if (!rand () % 10) Sys_Error ("Not the most stable!");
#endif

#ifdef MH_MIRRORS
	// lock mirroralpha at sensible values
	if (r_mirroralpha.value >= 1.0)
		r_mirroralpha.value = 1.0;
	else if (r_mirroralpha.value <= 0)
		r_mirroralpha.value = 0;
#endif

	frametime = cl.time - cl.oldtime;

	// make powerups throb
	if (PowerupGlowRadius > 50) {PGR_Dir = false; PowerupGlowRadius = 50;}
	if (PowerupGlowRadius < 30) {PGR_Dir = true; PowerupGlowRadius = 30;}

	if (PGR_Dir)
		PowerupGlowRadius = PowerupGlowRadius + (frametime * 20);
	else PowerupGlowRadius = PowerupGlowRadius - (frametime * 20);

	// flicker for torches & bolts
	if (TGF_Tick == 0)
	{
		TorchGlowFlicker = 19.0 + (float) (rand () & 3);
		BoltGlowFlicker = 27 + (float) (rand () & 7);
	}

	// ticks of 0.05 second (20 flickers per second)
	TGF_Tick = TGF_Tick + frametime;
	if (TGF_Tick > 0.05) TGF_Tick = 0;

	mirror = false;
	mirror_render = false;
	mirror_particles = false;

	if (gl_finish.value) glFinish ();
	R_Clear ();
	R_RenderScene ();

#ifdef MH_MIRRORS
	if (mirror)
	{
		R_Mirror ();
		mirror = false;
		mirror_particles = true;

		// set everything up all over again just for particles!  bet you're glad you wanted
		// bloody mirrors now, eh?
		R_SetupFrame ();
		R_SetFrustum ();
		R_SetupGL ();
		R_DrawParticles ();
		R_DrawEntityFX ();
		if (num_glows) R_RenderGlowEffects ();
	}
#endif

	R_PolyBlend ();
}

