/*
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

// halflife bsp and wad loading routines - halflife uses bsp30 - however, mhquake also
// supports the fake bsp30 popularised by the original qsg coloured light tutorials.  i've
// gotten round this with a cheap and nasty hack - see gl_edicts for details.

// note that mhquake only supports the halflife bsp and wad formats - you will have to use a 
// quake entity set and progs.dat with custom halflife bsps in order to run them on mhquake 
// properly...

// mhquake allows the halflife bsp format to be used with both the world and with ammo boxes,
// etc.

#include "quakedef.h"
#include "hlwad.h"

extern model_t	*loadmodel;
extern byte	*mod_base;
extern char	loadname[32];	// for hunk tags

extern char wadstring[1024];

int MHGL_LoadTexture (char *identifier, int width, int height, byte *data, qboolean mipmap, qboolean alpha, int bitsperpixel);

int notexture_num;

// these have to be globals cos we need them outside this function too.  ugh.
int fog_colour[3];
int fog_density;

// the palette immediately follows the texture.  load it and convert the texture to 32 bit,
// then upload it.
int LoadHLTexture (char *texname, int width, int height, byte *texpix)
{
#ifdef HALFLIFEBSP
	// each halflife texture has it's own palette
	byte HL_Red[256];
	byte HL_Green[256];
	byte HL_Blue[256];

	int palette_offset;
	int i, j;
	byte *hl_rgba;
	int hltexnum;
	qboolean alpha;
	qboolean mipmap;

	// add 2 cos the first 2 bytes after the pixels are a short giving the number of palette
	// entries.  while this can, in theory, change, in reality it's always 256.  i have yet
	// to see a halflife bsp or wad where it wasn't 256, anyway.
	palette_offset = (int) (((float) height * (float) width * 1.328125) + 2.0);

	for (i = 0; i < 256; i++)
	{
		HL_Red[i] = texpix[palette_offset++];
		HL_Green[i] = texpix[palette_offset++];
		HL_Blue[i] = texpix[palette_offset++];

		// load the underwater fog stuff if we have a fog texture
		if (texname[0] == '!')
		{ 
			if (i == 4) fog_density = HL_Red[i];

			if (i == 3)
			{
				fog_colour[0] = HL_Red[i];
				fog_colour[1] = HL_Green[i];
				fog_colour[2] = HL_Blue[i];
			}
		}
	}

	hl_rgba = (byte *) malloc (height * width * 4);
	alpha = false;

	// convert to rgba
	for (i = 0, j = 0; i < height * width; i++)
	{
		if (texpix[i] == 255 && texname[0] == '{')
		{
			// yeah, i know it looks amateurish, just go fuck off.
			hl_rgba[j++] = 0;
			hl_rgba[j++] = 0;
			hl_rgba[j++] = 0;
			hl_rgba[j++] = 0;
			alpha = true;
		}
		else 
		{
			hl_rgba[j++] = HL_Red[texpix[i]];
			hl_rgba[j++] = HL_Green[texpix[i]];
			hl_rgba[j++] = HL_Blue[texpix[i]];
			hl_rgba[j++] = 255;
		}
	}

	// don't mipmap grates cos it's ugly
	if (texname[0] == '{')
		mipmap = false;
	else mipmap = true;

	hltexnum = MHGL_LoadTexture (texname, width, height, &hl_rgba[0], mipmap, alpha, 32);
	free (hl_rgba);
	return hltexnum;
#else
	return -1;
#endif
}


// most of the code in here comes from a wad3 to tga utility i wrote a year ago - so please don't
// expect me to be able to explain what's going on in it.
int HL_SearchWAD3File (char *wad3name, miptex_t *mt, gltexture_t *tx)
{
#ifdef HALFLIFEBSP
	FILE *wadfile;
	char wad3path[1024];
	int i, j;

	hlwadinfo_t wadheader;
	hllumpinfo_t waddirectory;
	hlmiptex_t wadtexture;

	byte *loadtexture;

	int pixels;

	sprintf (wad3path, "textures/%s.wad", wad3name);

	i = COM_FOpenFile (wad3path, &wadfile);

	if (!wadfile) return -1;

	// search for the texture
	// read and verify the header
	fread (&wadheader, sizeof (wadheader), 1, wadfile);

	if (strncmp (wadheader.identification, "WAD3", 4))
	{
		// try for WAD2 - this is dangerous.  Half-Life supports WAD2 as a valid ID which means
		// that this program WILL TRY TO DO SOMETHING with a standard Quake WAD2.  So watch your
		// step here. this returns -1 rather than crashing out.
		if (strncmp (wadheader.identification, "WAD2", 4))
			return -1;
	}

	for (i = 0; i < wadheader.numlumps; i++)
	{
		// read the current entry in  the wad directory
		fseek (wadfile, (wadheader.infotableofs + (sizeof (waddirectory) * i)), SEEK_SET);
		fread (&waddirectory, sizeof (waddirectory), 1, wadfile);

		// fucking uppercase texture names
		if (!strcmp (strlwr (waddirectory.name), strlwr (tx->name)))
		{
			// gotcha, ya little bastard!
			// seek to the texture data
			fseek (wadfile, waddirectory.filepos, SEEK_SET);
			fread (&wadtexture, sizeof (wadtexture), 1, wadfile);

			// don't really need this in glquake, but what the hell...
			for (j = 0; j < MIPLEVELS; j++)
				tx->offsets[j] = wadtexture.offsets[j] + sizeof(gltexture_t) - sizeof(hlmiptex_t);

			// at this stage, wadtexture contains the same data as mt in the main texture loading
			// function, so we can do similar stuff with it - firstly, make the contents of mt
			// the same as those of wadtexture (just call me paranoid)
			mt->height = wadtexture.height;
			strcpy (mt->name, wadtexture.name);
			mt->width = wadtexture.width;

			for (j = 0; j < MIPLEVELS; j++)
				mt->offsets[j] = wadtexture.offsets[j];

			// i had hoped that i wouldn't have had to do this...
			pixels = (mt->width * mt->height / 64 * 85) + 770;

			// our file pointer is in the right position for this, so just read away
			loadtexture = (byte *) malloc (pixels);
			fread (loadtexture, pixels, 1, wadfile);

			// then load the texture
			tx->gl_texturenum = LoadHLTexture (tx->name, tx->width, tx->height, loadtexture);

			free (loadtexture);
			fclose (wadfile);
			return tx->gl_texturenum;
		}
	}

	// didn't find it
	fclose (wadfile);
#endif
	return -1;
}


int LoadHLWAD3Texture (miptex_t *mt, gltexture_t *tx)
{
#ifdef HALFLIFEBSP
	int i, j, k;
	char wadname[64];
	int texnum;

	// no WAD3 specified and Quake is expecting one!  Oh shit!!
	if (wadstring[0] == '\0') return notexture_num;

	// parse the wadstring
	for (i = 0; i < strlen (wadstring);)
	{
		// hunt for a '.'
		for (; wadstring[i] != '.'; i++)
			if (i >= strlen (wadstring)) return notexture_num;

		// save the current i (+ 1)
		j = i + 1;

		// read backwards for a '/' or a '\' or a ' ' or until i is 0
		for (; wadstring[i] != ' ' && wadstring[i] != '/' && wadstring[i] != '\\'; i--)
			if (i <= 0) {i = 0; break;}		// just in case it ever does go < 0...

		// read forwards to the '.' again, this time loading the wadname - we increment i here
		// cos we want to start at the first character after the '/' or whatever (only if i != 0)
		if (i) i++;

		for (k = 0; wadstring[i] != '.';) 
			wadname[k++] = wadstring[i++];

		wadname[k] = '\0';

		// find the texture
		texnum = HL_SearchWAD3File (wadname, mt, tx);

		// got it!
		if (texnum != -1) return texnum;

		// restore the saved i
		i = j;
	}
#endif

	// if we've got this far, the engine was unable to find a texture, so return the 
	// notexture texture.
	return notexture_num;
}


void Mod_LoadHLTextures (miptex_t *mt, gltexture_t *tx, int pixels)
{
#ifdef HALFLIFEBSP
	int j;

	fog_colour[0] = fog_colour[1] = fog_colour[2] = fog_density = 255;

	// the pixels immediately follow the structures - halflife doesn't do sky the same
	// as quake, so we're cool...
	texture_mode = GL_LINEAR_MIPMAP_LINEAR;

	if (mt->offsets[0])
	{
		for (j=0 ; j<MIPLEVELS ; j++)
			tx->offsets[j] = mt->offsets[j] + sizeof(gltexture_t) - sizeof(miptex_t);

		// texture is stored in the BSP
		memcpy (tx+1, mt+1, pixels);

		// external texture loading
		tx->gl_texturenum = GL_LoadTexture (mt->name, &tx->width, &tx->height, NULL, true, true);

		if (tx->gl_texturenum == -1)
			tx->gl_texturenum = LoadHLTexture (mt->name, tx->width, tx->height, (byte *)(tx+1));
	}
	else
	{
		// external texture loading (TGA/JPG/PCX, not WAD3)
		tx->gl_texturenum = GL_LoadTexture (mt->name, &tx->width, &tx->height, NULL, true, true);

		// grab the texture from an external WAD3
		if (tx->gl_texturenum == -1)
			tx->gl_texturenum = LoadHLWAD3Texture (mt, tx);
	}

	// halflife has no fullbrights
	tx->fullbright = -1;	// because 0 is a potentially valid texture number

	texture_mode = GL_LINEAR;

	// set underwater fog colour and density
	tx->water_fog_density = (float) fog_density / 255.0;
	tx->water_fog_colour[0] = (float) fog_colour[0] / 255.0;
	tx->water_fog_colour[1] = (float) fog_colour[1] / 255.0;
	tx->water_fog_colour[2] = (float) fog_colour[2] / 255.0;
#endif
}


extern int gl_useclight;

// doesn't support LIT files
void Mod_LoadHLLighting (lump_t *l)
{
#ifdef HALFLIFEBSP
	int i, j;
	byte *explit;
	int r, g, b;

	loadmodel->lightdata = NULL;

	if (!l->filelen)
		return;

	// halflife uses 3 component lightmaps in the bsp, so we expand them to 4
	Con_DPrintf ("Using Light Data from %s - HalfLife BSP format\n", loadmodel->name);

	if (gl_useclight)
	{
		// load from the bsp and expand to rgba
		loadmodel->lightdata = Hunk_AllocName ((l->filelen / 3) * 4, loadname);
		explit = (byte *) malloc (l->filelen);
		memcpy (explit, mod_base + l->fileofs, l->filelen);

		for (i = 0, j = 0; i < l->filelen; i += 3, j += 4)
		{
			r = loadmodel->lightdata[j] = explit[i];
			g = loadmodel->lightdata[j + 1] = explit[i + 1];
			b = loadmodel->lightdata[j + 2] = explit[i + 2];
			loadmodel->lightdata[j + 3] = (r + g + b) / 3;
		}

		free (explit);
	}
	else
	{
		// load from the bsp and condense - this means we can explicitly force white light
		// in halflife bsps if the performance hit caused by coloured light is too much to
		// cope with!
		loadmodel->lightdata = Hunk_AllocName (l->filelen / 3, loadname);
		explit = (byte *) malloc (l->filelen);
		memcpy (explit, mod_base + l->fileofs, l->filelen);

		for (i = 0, j = 0; i < l->filelen; i += 3, j++)
		{
			r = explit[i];
			g = explit[i+1];
			b = explit[i+2];
			loadmodel->lightdata[j] = (r + g + b) / 3;
		}

		free (explit);
	}
#endif
}


