/* MOD_MD3.c
 * Based on code from the Aftershock 3D rendering engine
 * Copyright (C) 1999 Stephen C. Taylor
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#include "quakedef.h"
#include "mod_md3.h"
#include "mod_common.h"

/*
=============
R_DrawQ3Model
=============
*/
// Sucks Need New rendering function eg:for now no lightning/shadows :(
void R_DrawQ3Model(entity_t *e)
{
	md3header_t		*md3header;
	int				i, j;
	unsigned int	*elem;
	md3mesh_t		*mesh;

	md3header = (md3header_t *)Mod_Extradata (e->model);

	glPushMatrix();

	if (r_interpolate_alias.value)
	{
		if (e == &cl.viewent)
		{
			R_RotateForEntity (e);
		}
		else
		{
			R_BlendedRotateForEntity (e, 0);
		}
	}

	mesh = (md3mesh_t *)((byte *)md3header + md3header->mesh_offs);

	for (i = 0; i < md3header->nummeshes; i++)
	{
		int numelems, frameoffset;

		frameoffset = e->frame;

		glBindTexture (GL_TEXTURE_2D, (((int *)((byte *)md3header + mesh->skin_offs))[e->skinnum%mesh->numskins]));

		glBegin(GL_TRIANGLES);

		elem = (unsigned int *)((byte *)mesh + mesh->elem_offs);

		numelems = mesh->numtris*3;

		if(e->frame > md3header->numboneframes)
		{
			frameoffset = md3header->numboneframes - 1;
		}
		frameoffset *= mesh->numverts;

		for (j = 0; j < numelems; j++)
		{
			glTexCoord2fv(((vec2_t *)((byte *)mesh + mesh->tc_offs))[*elem]);
			glVertex3fv(((vec3_t *)((byte *)md3header + mesh->vert_offs))[*elem + frameoffset]);
			*elem++;
		}
		glEnd();

		mesh = (md3mesh_t *)((char *)mesh + mesh->meshsize);
	}
	glPopMatrix();
}

/*
=============
Mod_LoadQ3Model
=============
*/
// fuxored i dunno what the hell does go wrong with large md3 models :S
void Mod_LoadQ3Model(model_t *mod, void *buffer)
{
	md3header_t *header;
	md3mesh_t	*mesh;
	byte		fake[16];	
	int			base;
	int			texhack = 4;
	int			i, j, size;
	int			defaultskin;
	int			numskins, numskinsfound;
	int			*skins;
	char		*skin;
	char		*skinsfound[16];
	char		skinname[MAX_QPATH];

	header = (md3header_t *)buffer;
	mesh = (md3mesh_t *)((byte *)buffer + header->mesh_offs);
	
	for (size = 0, i = 0; i < header->nummeshes; i++)
	{
		size += mesh->numverts * mesh->numframes * sizeof(vec3_t);
		mesh = (md3mesh_t *)((char *)mesh + mesh->meshsize);
		fake[i] = 254;
	}

	i = Q_strrchr(mod->name, '/') - mod->name;
	Q_strncpy(skinname, mod->name, i+1);
	Q_strcat(skinname, loadname);
	
	numskinsfound = COM_MultipleSearch(va("%s*.skin", skinname), skinsfound, 16, 0);
	
	numskins = numskinsfound;

	if(!numskins)
	{
		numskins = 1;
	}
	size += sizeof(int) * numskins * header->nummeshes;

	header = Cache_Alloc (&mod->cache, com_filesize + size, loadname);

	Q_memcpy(header, buffer, com_filesize);

	base = com_filesize;
	
    mod->flags = 0;mod->type = mod_md3;
	mod->version = header->version;
	mod->numframes = header->numboneframes;

	if (!Q_strcmp (mod->name, "progs/g_shot.mdl") || //Hack to give .md3 files renamed to .mdl rotate effects - Eradicator
 		!Q_strcmp (mod->name, "progs/g_nail.mdl") ||
 		!Q_strcmp (mod->name, "progs/g_nail2.mdl") ||
 		!Q_strcmp (mod->name, "progs/g_rock.mdl") ||
 		!Q_strcmp (mod->name, "progs/g_rock2.mdl") || 
 		!Q_strcmp (mod->name, "progs/g_light.mdl") ||
		!Q_strcmp (mod->name, "progs/armor.mdl") ||
		!Q_strcmp (mod->name, "progs/backpack.mdl") ||
 		!Q_strcmp (mod->name, "progs/w_g_key.mdl") ||
 		!Q_strcmp (mod->name, "progs/w_s_key.mdl") ||
 		!Q_strcmp (mod->name, "progs/m_g_key.mdl") ||
 		!Q_strcmp (mod->name, "progs/m_s_key.mdl") ||
 		!Q_strcmp (mod->name, "progs/b_g_key.mdl") ||
 		!Q_strcmp (mod->name, "progs/b_s_key.mdl") ||
 		!Q_strcmp (mod->name, "progs/quaddama.mdl") ||
 		!Q_strcmp (mod->name, "progs/invisibl.mdl") ||
 		!Q_strcmp (mod->name, "progs/invulner.mdl") ||
 		!Q_strcmp (mod->name, "progs/jetpack.mdl") || 
 		!Q_strcmp (mod->name, "progs/cube.mdl") ||
 		!Q_strcmp (mod->name, "progs/suit.mdl") ||
 		!Q_strcmp (mod->name, "progs/boots.mdl") ||
 		!Q_strcmp (mod->name, "progs/end1.mdl") ||
 		!Q_strcmp (mod->name, "progs/end2.mdl") ||
 		!Q_strcmp (mod->name, "progs/end3.mdl") ||
		!Q_strcmp (mod->name, "progs/end4.mdl")) {
		mod->flags |= EF_ROTATE; 
	}
	
	VectorCopy(((md3boneframe_t *)((byte *)header + header->bone_offs))->mins, mod->mins);
	VectorCopy(((md3boneframe_t *)((byte *)header + header->bone_offs))->maxs, mod->maxs);
	
	mesh = (md3mesh_t *)((byte *)header + header->mesh_offs);
	
	Con_DPrintf("Procesing  %d meshes...\n", header->nummeshes);

	for (i = 0; i < header->nummeshes; i++)
	{
		md3vert_t *vert;

		if (mesh->numskins != 1)
		{
			Sys_Error("Unexpected number of skins (%i) in %s", mesh->numskins, mesh->name);
		}
		mesh->numskins = numskins;

		skins = (int *)((byte *)header + base);
		
		if(mesh->skin_offs)
		{
			skin = (char *)mesh + mesh->skin_offs;
			
			j = Q_strlen(skin);

			if(j)
			{
				int texnum;

				Con_DPrintf("Loading %s skin...\n", skin);

				texnum = GL_LoadTexture (skin, &texhack, &texhack, (byte *)fake[0], true, false);

				if (texnum == 0)
				{
					texnum = GL_LoadTexture ("notexture", &texhack, &texhack, (byte *)fake[0], true, false);
				}

				for(j = 0; j < numskins; j++)
				{
					skins[j] = texnum;
					Con_DPrintf("Load texture OK skindata!\n");
				}
			}
		}
		else
		{
			for(j = 0; j < numskins; j++)
			{
				skins[j] = 0;
				Con_DPrintf("Load texture failed skindata!\n");
			}
		}
		
		mesh->skin_offs = base;
		base += sizeof(int *)numskins;
		vert = (md3vert_t *)((char *)mesh + mesh->vert_offs);
		mesh->vert_offs = base;
		size = mesh->numverts * mesh->numframes;
		base += size * sizeof(vec3_t);

		for (j=0; j < size; j++)
		{
			vec3_t *vertexes = (vec3_t *)((byte *)header + mesh->vert_offs);
			vertexes[j][0] = (float)vert->vec[0] / 64.0f;
			vertexes[j][1] = (float)vert->vec[1] / 64.0f;
			vertexes[j][2] = (float)vert->vec[2] / 64.0f;
			vert++;

		}
		mesh = (md3mesh_t *)((char *)mesh + mesh->meshsize);
	}

	for(i = 0; i < numskinsfound; i++)
	{
		if(Q_strlcmp(skinsfound[i], va("%s_default.skin", skinname))==0)
		{
			defaultskin = i;
			break;
		}
	}
	for(i = 0; i < numskinsfound; i++)
	{
		char *token, *oldtoken;
		int	 pos;
		char *skindata;

		if(!i)
		{
			skindata = (byte *)COM_LoadFile(skinsfound[defaultskin], 0); //To prevent the base model being thrown out of temporary hunk use Z_Malloc
		}
		else if(i < defaultskin)
		{
			skindata = (byte *)COM_LoadFile(skinsfound[i+1], 0); //To prevent the base model being thrown out of temporary hunk use Z_Malloc
		}
		else
		{
			skindata = (byte *)COM_LoadFile(skinsfound[i], 0); //To prevent the base model being thrown out of temporary hunk use Z_Malloc
		}
		if(skindata)
		{
			pos = 0;
			
			while(skindata[pos])
			{
				oldtoken = &skindata[pos];

				token = &skindata[pos];

				while(skindata[pos] != ',' && skindata[pos]) pos++;
				
				skindata[pos++] = '\0';

				j = 0;

				mesh = (md3mesh_t *)((byte *)header + header->mesh_offs);

				for(j = 0; j < header->nummeshes && Q_strlcmp(mesh->name ,token)!=0; j++)
				{
					mesh = (md3mesh_t *)((char *)mesh + mesh->meshsize);
				}
				token = &skindata[pos];

				while(skindata[pos] != '\n' && skindata[pos]) pos++;

				skindata[pos++-1] = '\0';

				if(token[0] && j < header->nummeshes)
				{
					skins = (int *)((byte *)header + mesh->skin_offs);

					skins[i] = GL_LoadTexture (token, &texhack, &texhack, (byte *)fake[0], true, false);

					if (skins[i] == 0)
					{
						skins[i] = GL_LoadTexture ("notexture", &texhack, &texhack, (byte *)fake[0], true, false);
					}
				}
			}
			Z_Free(mainzone, skindata);
		}
		Z_Free(mainzone, skinsfound[i]);
	}
}

