/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// map script parsing

#include "quakedef.h"

#ifdef USE_SCRIPTS
#include "script_common.h"

void Host_LoadSkybox (char *sb_name, qboolean report);

void Map_Script_Execute (char *scriptname, FILE *scriptfile)
{
	int i;
	char skey[64], value[64];
	char mapname[64];
	float water_alpha;
	int global_fog;
	float fog_start, fog_end, fog_red, fog_green, fog_blue;
	char skybox_name[64];
	float x_rotate, y_rotate, z_rotate;
	char mdl_script_name[64];
	char tex_script_name[64];
	char hud_script_name[64];

	while (1)
	{
		// init the variables
		water_alpha = 1.0;
		global_fog = 0;
		fog_start = 5;
		fog_end = 4096;
		fog_red = 0.25;
		fog_green = 0.25;
		fog_blue = 0.25;
		x_rotate = 0.0;
		y_rotate = 0.0;
		z_rotate = 0.0;
		sprintf (skybox_name, "none");
		sprintf (mdl_script_name, "goaway");
		sprintf (tex_script_name, "goaway");
		sprintf (hud_script_name, "goaway");

		if (!Script_ParseOpeningBrace (scriptfile)) goto finish_script;

		while (1)
		{
			if (!Script_Read (scriptfile, &skey[0])) goto bad_script;
			if (!Script_Read (scriptfile, &value[0])) goto bad_script;

			if (!strcmp (skey, "map"))
				sprintf (mapname, "%s", value);
			else if (!strcmp (skey, "wateralpha"))
				water_alpha = atof (value);
			else if (!strcmp (skey, "fog"))
			{
				// initially followed by a 1 or a zero - if it's 1, then we need to read 5 more
				// values for (respectively) start, end, red, green, blue
				global_fog = atoi (value);

				if (global_fog)
				{
					if (!Script_Read (scriptfile, &value[0])) goto bad_script;
					fog_start = atof (value);

					if (!Script_Read (scriptfile, &value[0])) goto bad_script;
					fog_end = atof (value);

					if (!Script_Read (scriptfile, &value[0])) goto bad_script;
					fog_red = atof (value);

					if (!Script_Read (scriptfile, &value[0])) goto bad_script;
					fog_green = atof (value);

					if (!Script_Read (scriptfile, &value[0])) goto bad_script;
					fog_blue = atof (value);
				}
			}
			else if (!strcmp (skey, "mdlscript"))
				sprintf (mdl_script_name, "%s", value);
			else if (!strcmp (skey, "texscript"))
				sprintf (tex_script_name, "%s", value);
			else if (!strcmp (skey, "hudscript"))
				sprintf (hud_script_name, "%s", value);
			else if (!strcmp (skey, "skyname"))
				sprintf (skybox_name, "%s", value);
			else if (!strcmp (skey, "skyrotate"))
			{
				// if present there will be 3 values - we already have the first, so
				// assign it and then read the other 2
				x_rotate = atof (value);

				if (!Script_Read (scriptfile, &value[0])) goto bad_script;
				y_rotate = atof (value);

				if (!Script_Read (scriptfile, &value[0])) goto bad_script;
				z_rotate = atof (value);
			}

			i = Script_ParseClosingBrace (scriptfile);

			if (i == -1)
				goto bad_script;
			else if (i == 1)
			{
				if (!strcmp (sv.name, mapname))
				{
					// set the stuff
					Cvar_SetValue ("r_wateralpha", water_alpha);

					Cvar_SetValue ("gl_fogenable", global_fog);
					Cvar_SetValue ("gl_fogstart", fog_start);
					Cvar_SetValue ("gl_fogend", fog_end);

					Cvar_SetValue ("gl_fogred", fog_red);
					Cvar_SetValue ("gl_foggreen", fog_green);
					Cvar_SetValue ("gl_fogblue", fog_blue);

					Host_LoadSkybox (skybox_name, false);

					Cvar_SetValue ("sky_rotate_x", x_rotate);
					Cvar_SetValue ("sky_rotate_y", y_rotate);
					Cvar_SetValue ("sky_rotate_z", z_rotate);

					Cbuf_AddText (va ("script %s false", mdl_script_name));
					Cbuf_AddText (va ("script %s false", tex_script_name));
					Cbuf_AddText (va ("script %s false", hud_script_name));

					// since only one map can be loaded at a time, we can finish up now
					goto finish_script;
				}

				goto next_script_item;
			}
		}
next_script_item:;
	}

bad_script:
	Con_Printf ("Error parsing scripts/%s.mhs\n", scriptname);

finish_script:
	return;
}
#endif
