#include "quakedef.h"
#include "gl_frontend_shared.h"
#include "glext.h"


// count how many alias polys we've drawn
int c_alias_polys;

// vertex array for drawing alias models
GLfloat VArray[MAXALIASVERTS * 9];



/*
=====================
GL_DrawAliasFrameNOLERPFLAT

Flat coloured non-interpolated alias model drawing - it don't get much simpler!
=====================
*/
void GL_DrawAliasFrameNOLERPFLAT (aliashdr_t *paliashdr, int posenum, float Light)
{
	trivertx_t	*v, *verts;
	int		*order;
	int		count;
	float	*aliasVerts;
	GLenum	TRI_TYPE;
	int		VCount;

	verts = (trivertx_t *)((byte *)paliashdr + paliashdr->posedata);
	verts += posenum * paliashdr->poseverts;
	order = (int *)((byte *)paliashdr + paliashdr->commands);

	while (1)
	{
		count = *order++;

		if (!count)
			break;

		if (count < 0)
		{
			count = -count;
			TRI_TYPE = GL_TRIANGLE_FAN;
		}
		else TRI_TYPE = GL_TRIANGLE_STRIP;

		aliasVerts = VArray;
		VCount = 0;

		do
		{
			aliasVerts[0] = verts->v[0];
			aliasVerts[1] = verts->v[1];
			aliasVerts[2] = verts->v[2];

			aliasVerts[3] = ((float *)order)[0];
			aliasVerts[4] = ((float *)order)[1];

			aliasVerts[5] = Light;
			aliasVerts[6] = Light;
			aliasVerts[7] = Light;
			aliasVerts[8] = 1.0; // paliashdr->translevel;

			order += 2;
			verts++;

			aliasVerts += 9;
			VCount++;
		} while (--count);

		glDrawArrays (TRI_TYPE, 0, VCount);
	}
}


/*
==================
R_SetupAliasFrameBase

Basic setting up of an Alias model frame with nothing fancy like interpolation
or lighting - only use for models that don't need them!!!
==================
*/
void R_SetupAliasFrameBase (int frame, aliashdr_t *paliashdr, float Light)
{
	int pose;
	int numposes;

	float interval;

	if ((frame >= paliashdr->numframes) || (frame < 0)) frame = 0;

	pose = paliashdr->frames[frame].firstpose;
	numposes = paliashdr->frames[frame].numposes;

	if (numposes > 1)
	{
		interval = paliashdr->frames[frame].interval;
		pose += (int)(realtime / interval) % numposes;
	}

	// never interpolate these cos we're using temporary entities
	GL_DrawAliasFrameNOLERPFLAT (paliashdr, pose, Light);
}


/*
==================
R_DrawAliasModelLOCK

Used for drawing alias models whose position is generally locked at a certain place
on the screen, normally status bar, menu and console widgets
==================
*/
void R_DrawAliasModelLOCK (entity_t *e, float Light, int skin, float trans)
{
	struct aliashdr_s *paliashdr = (struct aliashdr_s *) Mod_Extradata (e->model);
	float oldTrans;

    glPushMatrix ();

	// apply modelview transformations
	glTranslatef (e->origin[0], e->origin[1], e->origin[2]);
	glRotatef (e->angles[1],  0, 0, 1);
	glRotatef (-e->angles[0],  0, 1, 0);
	glRotatef (e->angles[2],  1, 0, 0);

	// scale the wee buggah
	glTranslatef (paliashdr->scale_origin[0], paliashdr->scale_origin[1], paliashdr->scale_origin[2]);
	glScalef (paliashdr->scale[0], paliashdr->scale[1], paliashdr->scale[2]);

    glBindTexture (GL_TEXTURE_2D, paliashdr->gl_texturenum[e->skinnum + skin][0]);
	R_SetupAliasFrameBase (e->frame, paliashdr, Light);

	glPopMatrix ();
}



/*
=================================================================

ALIAS MODEL DISPLAY LIST GENERATION

These are no longer cached on disk as the standard "fall back on id1" method
can cause massive trouble with mods, etc.  On the fly generation is not that
bad anyway...

=================================================================
*/

model_t		*aliasmodel;
aliashdr_t	*paliashdr;

qboolean	used[8192];


// the command list holds counts and s/t values that are valid for
// every frame
int		commands[8192];
int		numcommands;


// all frames will have their vertexes rearranged and expanded
// so they are in the order expected by the command list
int		vertexorder[8192];
int		numorder;


int		allverts, alltris;


int		stripverts[128];
int		striptris[128];
int		stripcount;


/*
================
StripLength
================
*/
int	StripLength (int starttri, int startv)
{
	int			m1, m2;
	int			j;
	mtriangle_t	*last, *check;
	int			k;

	used[starttri] = 2;

	last = &triangles[starttri];

	stripverts[0] = last->vertindex[(startv) % 3];
	stripverts[1] = last->vertindex[(startv + 1) % 3];
	stripverts[2] = last->vertindex[(startv + 2) % 3];

	striptris[0] = starttri;
	stripcount = 1;

	m1 = last->vertindex[(startv + 2) % 3];
	m2 = last->vertindex[(startv + 1) % 3];

	// look for a matching triangle
nexttri:;
	for (j = starttri + 1, check = &triangles[starttri + 1]; j < pheader->numtris; j++, check++)
	{
		if (check->facesfront != last->facesfront) continue;

		for (k = 0; k < 3; k++)
		{
			if (check->vertindex[k] != m1) continue;
			if (check->vertindex[(k + 1) % 3] != m2) continue;

			// this is the next part of the fan
			// if we can't use this triangle, this tristrip is done
			if (used[j]) goto done;

			// the new edge
			if (stripcount & 1)
				m2 = check->vertindex[(k + 2) % 3];
			else m1 = check->vertindex[(k + 2) % 3];

			stripverts[stripcount + 2] = check->vertindex[(k + 2) % 3];
			striptris[stripcount] = j;
			stripcount++;

			used[j] = 2;
			goto nexttri;
		}
	}
done:;

	// clear the temp used flags
	for (j = starttri + 1; j < pheader->numtris; j++)
	{
		if (used[j] == 2)
		{
			used[j] = 0;
		}
	}

	return stripcount;
}

/*
===========
FanLength
===========
*/
int	FanLength (int starttri, int startv)
{
	int		m1, m2;
	int		j;
	mtriangle_t	*last, *check;
	int		k;

	used[starttri] = 2;

	last = &triangles[starttri];

	stripverts[0] = last->vertindex[(startv)%3];
	stripverts[1] = last->vertindex[(startv+1)%3];
	stripverts[2] = last->vertindex[(startv+2)%3];

	striptris[0] = starttri;
	stripcount = 1;

	m1 = last->vertindex[(startv+0)%3];
	m2 = last->vertindex[(startv+2)%3];


	// look for a matching triangle
nexttri:;
	for (j = starttri + 1, check = &triangles[starttri + 1]; j < pheader->numtris; j++, check++)
	{
		if (check->facesfront != last->facesfront) continue;

		for (k = 0; k < 3; k++)
		{
			if (check->vertindex[k] != m1) continue;
			if (check->vertindex[(k + 1) % 3] != m2) continue;

			// this is the next part of the fan
			// if we can't use this triangle, this tristrip is done
			if (used[j]) goto done;

			// the new edge
			m2 = check->vertindex[(k + 2) % 3];

			stripverts[stripcount + 2] = m2;
			striptris[stripcount] = j;
			stripcount++;

			used[j] = 2;
			goto nexttri;
		}
	}
done:;

	// clear the temp used flags
	for (j = starttri + 1; j < pheader->numtris; j++)
	{
		if (used[j] == 2)
		{
			used[j] = 0;
		}
	}

	return stripcount;
}


/*
================
BuildTris

Generate a list of trifans or strips
for the model, which holds for all frames
================
*/
void BuildTris (void)
{
	int		i, j, k;
	int		startv;
	mtriangle_t	*last, *check;
	int		m1, m2;
	int		striplength;
	trivertx_t	*v;
	mtriangle_t *tv;
	float	s, t;
	int		index;
	int		len, bestlen, besttype;
	int		bestverts[1024];
	int		besttris[1024];
	int		type;

	// build tristrips
	numorder = 0;
	numcommands = 0;
	memset (used, 0, sizeof (used));

	for (i = 0; i < pheader->numtris; i++)
	{
		// pick an unused triangle and start the trifan
		if (used[i]) continue;

		bestlen = 0;

		for (type = 0; type < 2; type++)
		{
			//	type = 1;
			for (startv = 0; startv < 3; startv++)
			{
				if (type == 1)
					len = StripLength (i, startv);
				else len = FanLength (i, startv);

				if (len > bestlen)
				{
					besttype = type;
					bestlen = len;

					for (j = 0; j < bestlen + 2; j++) bestverts[j] = stripverts[j];
					for (j = 0; j < bestlen; j++) besttris[j] = striptris[j];
				}
			}
		}

		// mark the tris on the best strip as used
		for (j = 0; j < bestlen; j++) used[besttris[j]] = 1;

		if (besttype == 1)
			commands[numcommands++] = (bestlen + 2);
		else commands[numcommands++] = -(bestlen + 2);

		for (j = 0; j < bestlen + 2; j++)
		{
			// emit a vertex into the reorder buffer
			k = bestverts[j];
			vertexorder[numorder++] = k;

			// emit s/t coords into the commands stream
			s = stverts[k].s;
			t = stverts[k].t;

			// on back side
			if (!triangles[besttris[0]].facesfront && stverts[k].onseam)
				s += pheader->skinwidth / 2;

			s = (s + 0.5) / pheader->skinwidth;
			t = (t + 0.5) / pheader->skinheight;

			*(float *) &commands[numcommands++] = s;
			*(float *) &commands[numcommands++] = t;
		}
	}

	// end of list marker
	commands[numcommands++] = 0;

	Con_DPrintf ("%3i tri %3i vert %3i cmd\n", pheader->numtris, numorder, numcommands);

	allverts += numorder;
	alltris += pheader->numtris;
}


/*
================
GL_MakeAliasModelDisplayLists
================
*/
void GL_MakeAliasModelDisplayLists (model_t *m, aliashdr_t *hdr)
{
	int			i, j;
	int			*cmds;
	trivertx_t	*verts;

	aliasmodel = m;
	paliashdr = hdr;

	BuildTris ();

	// save the data out
	paliashdr->poseverts = numorder;

	cmds = Hunk_Alloc (numcommands * 4);
	paliashdr->commands = (byte *) cmds - (byte *) paliashdr;
	memcpy (cmds, commands, numcommands * 4);

	verts = Hunk_Alloc (paliashdr->numposes * paliashdr->poseverts * sizeof (trivertx_t));
	paliashdr->posedata = (byte *) verts - (byte *) paliashdr;

	for (i = 0; i < paliashdr->numposes; i++)
	{
		for (j = 0; j < numorder; j++)
		{
			*verts++ = poseverts[i][vertexorder[j]];
		}
	}
}



