// 3D surface rendering

#include "quakedef.h"
#include "gl_frontend_shared.h"
#include "glext.h"


qboolean liquidRender;
qboolean skyRender;


/*
===============
R_TextureAnimation

Returns the proper texture for a given time and base texture
===============
*/
texture_t *R_TextureAnimation (texture_t *base)
{
	int		reletive;
	int		count;

	// save out the base texture in case we have a bad cycle
	texture_t *bad = base;

	if (currententity->frame)
	{
		if (base->alternate_anims)
			base = base->alternate_anims;
	}
	
	if (!base->anim_total)
		return base;

	reletive = (int)(realtime*10) % base->anim_total;

	count = 0;

	while (base->anim_min > reletive || base->anim_max <= reletive)
	{
		base = base->anim_next;

		if (!base) return bad;
		if (++count > 100) return bad;
	}

	return base;
}


__inline void SelectDrawMode (numverts)
{
	// select the least expensive drawing mode
	switch (numverts)
	{
	case 3:
		glBegin (GL_TRIANGLES);
		break;

	case 4:
		glBegin (GL_QUADS);
		break;

	default:
		glBegin (GL_POLYGON);
	}
}


void R_RenderSurf (msurface_t *surf)
{
	int i;
	float *v = surf->polys->verts[0];
	int numverts = surf->polys->numverts;

	SelectDrawMode (numverts);

	for (i = 0; i < numverts; i++, v += VERTEXSIZE)
	{
		// and draw it
		glTexCoord2fv (&v[3]);
		glVertex3fv (v);
	}

	glEnd ();
}


/*
================
Draw_TextureChains
================
*/
void Draw_TextureChains (model_t *mod)
{
	int i;
	msurface_t *s;
	texture_t *t;

	for (i = 0; i < mod->numtextures; i++)
	{
		if (!(t = mod->textures[i])) continue;
		if (!(s = t->texturechain)) continue;
		if (i == skytexturenum) continue;
		if (s->flags & SURF_DRAWTURB) continue;

		// get the correct animation and bind it
		glBindTexture (GL_TEXTURE_2D, (R_TextureAnimation (s->texinfo->texture))->gl_texturenum);

		for (; s; s = s->texturechain)
		{
			R_UpdateSurfaceLightmap (s);
			R_RenderSurf (s);
		}

		t->texturechain = NULL;
	}
}


/*
=============
R_DrawWorld
=============
*/
void R_DrawWorld (void)
{
	entity_t	ent;
	int			i;

	// sky rotation
	rotateBack = anglemod (realtime * 25);
	rotateFore = anglemod (realtime * 40);

	memset (&ent, 0, sizeof (ent));
	ent.model = cl.worldmodel;

	VectorCopy (r_refdef.vieworg, modelorg);

	currententity = &ent;
	currenttexture = -1;

	glColor3f (1, 1, 1);

	memset (lightmap_polys, 0, sizeof (lightmap_polys));

	liquidRender = false;
	skyRender = false;

	R_RecursiveWorldNode (cl.worldmodel->nodes);

	R_PushDlights ();

	Draw_TextureChains (cl.worldmodel);

	R_BlendLightmaps ();

	if (skyRender) R_DrawSkyChain ();
}
