#include "quakedef.h"
#include "gl_frontend_shared.h"
#include "glext.h"


int		solidskytexture;
int		alphaskytexture;
float	speedscale;


/*
=================
R_DrawSkyChain
=================
*/
// these are incremented in R_RenderView cos we want them always updated even if
// the sky is not being currently drawn
float rotateBack = 0;
float rotateFore = 0;


void R_DrawSkyChain (void)
{
	GLUquadricObj *pObj;

	// background
	// ----------
	glPushMatrix ();

	// center it on the players position
	glTranslatef (r_origin[0], r_origin[1], r_origin[2]);

	// flatten the sphere
	glScalef (1, 1, 0.5);

	// orient it so that the poles are unobtrusive
	glRotatef (-90, 1, 0, 0);

	// make it not always at right angles to the player
	glRotatef (-22, 0 ,1, 0);

	// rotate it around the poles
	glRotatef (-rotateBack, 0, 0, 1);

	// gimme the sky texture
	GL_Bind (solidskytexture);

	// allocate a new quadric object for the sphere
	pObj = gluNewQuadric();

	// tell it we want it textured and drawn filled
	gluQuadricTexture (pObj, true);
	gluQuadricDrawStyle (pObj, GLU_FILL);

	// radius, definition x, definition y
	// higher definitions are slower to draw
	// 10 is perfectly adequate
	gluSphere (pObj, 4000, 10, 10);

	glPopMatrix ();

	// foreground
	// ----------
	glPushMatrix ();

	// center it on the players position
	glTranslatef (r_origin[0], r_origin[1], r_origin[2]);

	// flatten the sphere
	glScalef (1, 1, 0.5);

	// orient it so that the poles are unobtrusive
	glRotatef (-90, 1, 0, 0);

	// make it not always at right angles to the player
	glRotatef (-22, 0 ,1, 0);

	// rotate it around the poles
	glRotatef (-rotateFore, 0, 0, 1);

	// blend mode
	glEnable (GL_BLEND);

	// foreground texture
	GL_Bind (alphaskytexture);

	// draw the sphere - reduce the radius to prevent "poke-through"
	gluSphere (pObj, 3500, 10, 10);

	// back to normal mode
	glDisable (GL_BLEND);

	// restore the world matrix
	glPopMatrix ();

	// free the object
	gluDeleteQuadric (pObj);
}


/*
=============
R_InitSky

A sky texture is 256*128, with the right side being a masked overlay
==============
*/
void GL_TileTexture (unsigned int *in, unsigned int *out, int inWidth, int inHeight, int tileX, int tileY)
{
	int outPos = 0;
	int i, j, k, l;
	int Row;

	for (l = 0; l < tileY; l++)
	{
		for (k = 0, Row = 0; k < inHeight; k++, Row += inHeight)
		{
			for (j = 0; j < tileX; j++)
			{
				for (i = 0; i < inWidth; i++)
				{
					out[outPos++] = in[i + Row];
				}
			}
		}
	}
}


void R_InitSky (texture_t *mt)
{
	int			i, j, p;
	byte		*src;
	unsigned	transpix;
	int			r, g, b;
	unsigned	*rgba;
	extern int 	skytexturenum;
	unsigned	topalpha;
	int			div;

	unsigned	*trans;
	unsigned	*newTrans;

	// were #defines but made dependent on hardware max texture size here
	int SKY_TILE_X, SKY_TILE_Y;

	// this of course will look better on hardware capable of displaying larger
	// textures, and look not so good on hardware that is not so capable (i.e. 3dfx)
	// such is the price of crap hardware
	if (maxGLSize >= 1024)
		SKY_TILE_X = 2;		// 8 = more detail, 4 = faster render
	else if (maxGLSize >= 512)
		SKY_TILE_X = 2;
	else SKY_TILE_X = 2;

	SKY_TILE_Y = SKY_TILE_X / 2;

	// initialize our pointers
	trans = malloc (128 * 128 * 4);
	newTrans = malloc ((SKY_TILE_X * 128) * (SKY_TILE_Y * 128) * 4);

	// -----------------
	// solid sky texture
	// -----------------
	src = (byte *)mt + mt->offsets[0];

	// make an average value for the back to avoid
	// a fringe on the top level
	r = g = b = 0;

	for (i=0 ; i<128 ; i++)
	{
		for (j=0 ; j<128 ; j++)
		{
			p = src[i*256 + j + 128];
			rgba = &d_8to24table[p];
			trans[(i*128) + j] = *rgba;

			r += ((byte *)rgba)[0];
			g += ((byte *)rgba)[1];
			b += ((byte *)rgba)[2];
		}
	}

	((byte *)&transpix)[0] = r/(128*128);
	((byte *)&transpix)[1] = g/(128*128);
	((byte *)&transpix)[2] = b/(128*128);
	((byte *)&transpix)[3] = 0;

	if (!solidskytexture) solidskytexture = texture_extension_number++;

	GL_TileTexture (trans, newTrans, 128, 128, SKY_TILE_X, SKY_TILE_Y);

	GL_Bind (solidskytexture);

	glTexImage2D (GL_TEXTURE_2D, 0, gl_solid_format, (SKY_TILE_X * 128), (SKY_TILE_Y * 128), 0, GL_RGBA, GL_UNSIGNED_BYTE, newTrans);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	// -----------------
	// alpha sky texture
	// -----------------
	// get another average cos the bottom layer average can be too dark
	div = r = g = b = 0;

	for (i=0 ; i<128 ; i++)
	{
		for (j=0 ; j<128 ; j++)
		{
			p = src[i*256 + j];

			if (p == 0)
				topalpha = transpix;
			else
			{
				rgba = &d_8to24table[p];

				g += ((byte *)rgba)[1];

				r += ((byte *)rgba)[0];
				b += ((byte *)rgba)[2];

				div++;

				topalpha = d_8to24table[p];
			}

			((byte *)&topalpha)[3] = ((byte *)&topalpha)[3] / 2;

			trans[(i*128) + j] = topalpha;
		}
	}

	if (!alphaskytexture) alphaskytexture = texture_extension_number++;

	GL_TileTexture (trans, newTrans, 128, 128, SKY_TILE_X, SKY_TILE_Y);

	GL_Bind (alphaskytexture);

	glTexImage2D (GL_TEXTURE_2D, 0, gl_alpha_format, (SKY_TILE_X * 128), (SKY_TILE_Y * 128), 0, GL_RGBA, GL_UNSIGNED_BYTE, newTrans);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	// free the used memory
	free (trans);
	free (newTrans);
}

